#Include-once

#include <File.au3>
#Include "FASM.au3"
#Include "BinaryCall.au3"


Func GetProcAddress( $sDllFile, $sFunction )
	Local $aRet = DllCall( "kernel32.dll", "ptr", "GetProcAddress", "ptr", _WinAPI_GetModuleHandle( $sDllFile ), "str", $sFunction  )
	If Not @error And $aRet[0] Then Return $aRet[0]
	Return 0
EndFunc


Func ReadFasmCode( $sAsmFile, ByRef $oFasm, ByRef $cBinCode, $iOrgDirective = 0 )

	; x86 mode
	If @AutoItX64 Then
		MsgBox( 16, @ScriptName, "This script can not be run as a 64 bit program." )
		Return 0
	EndIf

	; File type
	Local $sSuffix = StringRight( $sAsmFile, StringLen( $sAsmFile ) - StringInStr( $sAsmFile, ".", 0, -1 ) )
	If $sSuffix <> "asm" Then
		MsgBox( 16, @ScriptName, "The assembler source file """ & $sAsmFile & """ must be an ASM-file." )
		Return 0
	EndIf

	; File exists
	If Not FileExists( $sAsmFile ) Then
		MsgBox( 16, @ScriptName, "The assembler source file """ & $sAsmFile & """ does not exists." )
		Return 0
	EndIf

	Local $aCode, $aTmp[1], $sStr, $iCnt = 0

	; Read source
	_FileReadToArray( $sAsmFile, $aCode )

	ReDim $aTmp[$aCode[0]+1]

	; Get assembler code
	For $i = 1 To $aCode[0]
		$sStr = StringStripWS( $aCode[$i], 7 )							; Strip redundant blanks
		If StringLen( $sStr ) = 0 Then ContinueLoop					; Skip empty lines
		If StringLeft( $sStr, 1 ) = ";" Then ContinueLoop		; Remove comment lines
		While StringInStr( $sStr, ";" )											; Remove comments to the right of a code line
			$sStr = StringLeft( $sStr, StringInStr( $sStr, ";", 0, -1 ) - 1 )
			$sStr = StringStripWS( $sStr, 2 )
		WEnd
		$aTmp[$iCnt+1] = $sStr
		$iCnt += 1
	Next

	; Code lines
	If Not $iCnt Then
		MsgBox( 16, @ScriptName, "The file """ & $sAsmFile & """ seems not to contain any assembler source code." )
		Return 0
	EndIf

	$aTmp[0] = $iCnt
	ReDim $aTmp[$iCnt+1]
	$aCode = $aTmp

	; ORG directive
	If $iOrgDirective Then
		if $iOrgDirective < 1 Or $iOrgDirective > $aCode[0] Then
			MsgBox( 16, @ScriptName, "$iOrgDirective must be between 1 and " & $aCode[0] & "." )
			Return 0
		EndIf
		If $aCode[$iOrgDirective] <> "org" Then
			MsgBox( 16, @ScriptName, "$aCode[" & $iOrgDirective & "] = " & $aCode[$iOrgDirective] & " is not the ORG directive (org or ORG)." )
			Return 0
		EndIf
	EndIf

	; ORG directive
	If $iOrgDirective Then _
		$aCode[$iOrgDirective] &= " " & FasmGetBasePtr( $oFasm )

	ConsoleWrite( @CRLF & "Source file """ & $sAsmFile & """" & @CRLF )

	FasmReset( $oFasm )
	ConsoleWrite( @CRLF & "FASM code:" & @CRLF )
	For $i = 1 To $aCode[0]
		ConsoleWrite( StringFormat( "%4i  ", $i ) & $aCode[$i] & @CRLF )
		FasmAdd( $oFasm, $aCode[$i] )
	Next

	FasmCompile( $oFasm )
	If @Error Then
		Local $Error = FasmGetLastError()
		ConsoleWrite( @CRLF )
		ConsoleWrite( "Error Code:       " & $Error[$FASMERRINDEX_CODE] & @CRLF )
		ConsoleWrite( "Error Message:    " & $Error[$FASMERRINDEX_MESSAGE] & @CRLF )
		ConsoleWrite( "Error LineNumber: " & $Error[$FASMERRINDEX_LINENUMBER] & @CRLF )
		ConsoleWrite( "Error Line:       " & $Error[$FASMERRINDEX_LINE] & @CRLF & @CRLF )
		_FasmError()
	Else
		$cBinCode = Binary( FasmGetBinary( $oFasm ) )
		ConsoleWrite( @CRLF & "Binary code:" & @CRLF )
		ConsoleWrite( String( $cBinCode ) & @CRLF & @CRLF )
		Return 1
	EndIf

EndFunc


Func MakeX64code( $sAsmFile, $iOrgDirective = 0 )

	Local $oFasm = FasmInit(), $cBinCode, $sBinFile, $sTxtFile, $hFile, $iRet

	; Read FASM code
	If Not ReadFasmCode( $sAsmFile, $oFasm, $cBinCode, $iOrgDirective ) Then Return 0

	$sBinFile = StringLeft( $sAsmFile, StringInStr( $sAsmFile, ".", 0, -1 ) - 1 ) & ".bin"
	$sTxtFile = StringLeft( $sAsmFile, StringInStr( $sAsmFile, ".", 0, -1 ) - 1 ) & ".txt"

	If StringLen( String( $cBinCode ) ) = 0 Then
		MsgBox( 16, @ScriptName, "An error occurred. No binary code generated." )
		$iRet = 0
	Else
		ConsoleWrite( "Write code to file """ & $sBinFile & """" & @CRLF )
		ConsoleWrite( "Write code to file """ & $sTxtFile & """" & @CRLF & @CRLF )
		$iRet = 1
	EndIf

	; Save binary code
	$hFile = FileOpen( $sBinFile, 18 )
	FileWrite( $hFile, $cBinCode )
	FileClose( $hFile )

	; Save as string
	$hFile = FileOpen( $sTxtFile, 2 )
	FileWrite( $hFile, String( $cBinCode ) )
	FileClose( $hFile )

	FasmExit( $oFasm )

	Return $iRet

EndFunc


Func ReadX64code( $sBinFile, $sMakeScript, ByRef $cBinCode )

	If Not @AutoItX64 Then
		MsgBox( 16, @ScriptName, "This script must be run as a 64 bit program." )
		Return
	EndIf

	Local $hFile

	If FileExists( $sBinFile ) And FileGetSize( $sBinFile ) Then

		$hFile = FileOpen( $sBinFile, 16 )
		$cBinCode = String( FileRead( $hFile ) )
		FileClose( $hFile )

		ConsoleWrite( @CRLF & "Binary code:" & @CRLF )
		ConsoleWrite( $cBinCode & @CRLF & @CRLF )

		Return 1

	Else

		MsgBox( 16, @ScriptName, _
						"" & $sBinFile & """ does not exists or is empty." & @CRLF & _
						"Run """ & $sMakeScript & """ to create the binary file." )
		Return 0

	EndIf

EndFunc
