; Functions
;		ShowFavoritesMenu()		Opens the Favorites menu
;		FavMenuMsgUnHook()		Unhooks the FavMenuMsg() hook procedure
;													Must be executed on the $idFavMenuMsgUnHook event in the main message loop

#include-once

#include <GuiMenu.au3>
#include <WinAPI.au3>
#include <WinAPICom.au3>
#include <WindowsConstants.au3>
#include <APIShellExConstants.au3>
#include <WinAPIvkeysConstants.au3>

; Dynamic Link Libraries
Global Const $dllOle32    = DllOpen( "ole32.dll" )
Global Const $dllShell32  = DllOpen( "shell32.dll" )
;Global Const $dllUser32   = DllOpen( "user32.dll" )

; Initialize COM
OleInitialize()

; Register cleanup function
OnAutoItExitRegister( "ShellFavoritesMenuExit" )

; --- Global variables ---

Global $hFavMenu

Global $oIMenuBand, $pIMenuBand = 0x0000

; Favorites menu messages
Global $hFavMenuMsg = 0x0000, $tFavMenuMsg
Global $hFavMenuMsgHook, $idFavMenuMsgUnHook = 0

; --- Global constants ---

Global Const $NULL = 0x0000

; Flag used with CoCreateInstance
Global Const $CLSCTX_INPROC_SERVER = 0x1

; Flags for the IShellMenu Interface
; Flags that control how the menu operates
Global Const $SMINIT_DEFAULT            = 0x00000000  ; No Options
Global Const $SMINIT_RESTRICT_DRAGDROP  = 0x00000002  ; Don't allow Drag and Drop
Global Const $SMINIT_TOPLEVEL           = 0x00000004  ; This is the top band.
Global Const $SMINIT_CACHED             = 0x00000010
Global Const $SMINIT_VERTICAL           = 0x10000000  ; This is a vertical menu
Global Const $SMINIT_HORIZONTAL         = 0x20000000  ; This is a horizontal menu (does not inherit)
Global Const $ANCESTORDEFAULT           = 0xFFFFFFFF  ; dword(-1);
Global Const $SMSET_TOP                 = 0x10000000  ; Bias this namespace to the top of the menu
Global Const $SMSET_BOTTOM              = 0x20000000  ; Bias this namespace to the bottom of the menu
Global Const $SMSET_DONTOWN             = 0x00000001  ; The Menuband doesn't own the non-ref counted object
Global Const $SMSET_USEBKICONEXTRACTION = 0x00000008

; --- Enumeration constants ---

; dwFlags constants for the IMenuPopup::Popup method
Global Enum _
	$MPPF_SETFOCUS						= 0x1, _				; The menu should have focus when it appears.
	$MPPF_INITIALSELECT				= 0x2, _				; The first item in the menu should be selected.
	$MPPF_NOANIMATE						= 0x4, _				; Do not animate this show.
	$MPPF_KEYBOARD						= 0x10, _				; The menu is activated by the keyboard.
	$MPPF_REPOSITION					=	0x20, _				; Reposition the displayed bar.
	$MPPF_FORCEZORDER					= 0x40, _				; The menu bar should ignore submenu positions.
	$MPPF_FINALSELECT					= 0x80, _				; The last item in the menu should be selected.
	$MPPF_TOP									= 0x20000000, _	; Display the pop-up menu above the point specified in ppt.
	$MPPF_LEFT								= 0x40000000, _	; Display the pop-up menu to the left of the point specified in ppt.
	$MPPF_RIGHT								= 0x60000000, _	; Display the pop-up menu to the right of the point specified in ppt.
	$MPPF_BOTTOM							= 0x80000000, _	; Display the pop-up menu below the point specified in ppt.
	$MPPF_POS_MASK						= 0xE0000000, _	; Mask for position values MPPF_TOP, MPPF_LEFT, and MPPF_RIGHT.
	$MPPF_ALIGN_LEFT					= 0x2000000, _	; Default alignment.
	$MPPF_ALIGN_RIGHT					= 0x4000000			; The pop-up menu should be aligned to the right of the excluded rectangle specified by prcExclude.

; --- Structure constants ---

Global $tagMSG = "hwnd hwnd;uint message;wparam wParam;lparam lParam;dword time;int X;int Y"

; --- Class Identifier constants ---

; CLSID_MenuBand
Global Const $sCLSID_MenuBand = "{5B4DAE26-B807-11D0-9815-00C04FD91972}"
Global Const $tCLSID_MenuBand = CLSIDFromString( $sCLSID_MenuBand )

; CLSID_MenuDeskBar
Global Const $sCLSID_MenuDeskBar = "{ECD4FC4F-521C-11D0-B792-00A0C90312E1}"
Global Const $tCLSID_MenuDeskBar = CLSIDFromString( $sCLSID_MenuDeskBar )

; CLSID_MenuBandSite
Global Const $sCLSID_MenuBandSite = "{E13EF4E4-D2F2-11D0-9816-00C04FD91972}"
Global Const $tCLSID_MenuBandSite = CLSIDFromString( $sCLSID_MenuBandSite )

; --- Interface definitions ---

; IShellMenu Interface
Global Const $sIID_IShellMenu = "{EE1F7637-E138-11d1-8379-00C04FD918D0}"
Global Const $tRIID_IShellMenu = CLSIDFromString( $sIID_IShellMenu )
Global Const $dtag_IShellMenu = _
	"Initialize hresult(ptr;uint;uint;dword);" & _
	"GetMenuInfo hresult(ptr*;uint*;uint*;dword*);" & _
	"SetShellFolder hresult(ptr;ptr;handle;dword);" & _
	"GetShellFolder hresult(ptr*;ptr*;struct*;ptr*);" & _
	"Setmenu hresult(handle;hwnd;dword);" & _
	"GetMenu hresult(handle*;hwnd*;dword*);" & _
	"InvalidateItem hresult(ptr;dword);" & _
	"GetState hresult(ptr);" & _
	"SetMenuToolbar hresult(ptr*;dword);"

; IShellFolder Interface
Global Const $sIID_IShellFolder = "{000214E6-0000-0000-C000-000000000046}"
Global Const $tRIID_IShellFolder = CLSIDFromString( $sIID_IShellFolder )
Global Const $dtag_IShellFolder = _
	"ParseDisplayName hresult(hwnd;ptr;wstr;dword*;ptr*;dword*);" & _
	"EnumObjects hresult(hwnd;dword;ptr*);" & _
	"BindToObject hresult(ptr;ptr;struct*;ptr*);" & _
	"BindToStorage hresult(ptr;ptr;ptr;ptr*);" & _
	"CompareIDs hresult(lparam;ptr;ptr);" & _
	"CreateViewObject hresult(hwnd;struct*;ptr*);" & _
	"GetAttributesOf hresult(uint;struct*;ulong*);" & _
	"GetUIObjectOf hresult(hwnd;uint;struct*;struct*;uint*;ptr*);" & _
	"GetDisplayNameOf hresult(ptr;uint;struct*);" & _
	"SetNameOf hresult(hwnd;ptr;wstr;dword;ptr*);"

; IMenuPopup Interface
Global Const $sIID_IMenuPopup = "{D1E7AFEB-6A2E-11d0-8C78-00C04FD918B4}"
Global Const $tRIID_IMenuPopup = CLSIDFromString( $sIID_IMenuPopup )
Global Const $dtag_IMenuPopup = _
	"GetWindow hresult(hwnd);" & _								; HWND* phwnd
	"ContextSensitiveHelp hresult(bool);" & _			; BOOL fEnterMode
	"SetClient hresult(ptr);" & _									; IUnknown* punkClient
	"GetClient hresult(ptr*);" & _								; IUnknown** ppunkClient
	"OnPosRectChangeDB hresult(struct*);" & _			; LPRECT prc
	"Popup hresult(struct*;struct*;dword);" & _		; POINTL* ppt, RECTL* prcExclude, MP_POPUPFLAGS dwFlags
	"OnSelect hresult(dword);" & _								; DWORD dwSelectType
	"SetSubMenu hresult(ptr;bool);"								; IMenuPopup* pmp, BOOL fSet

; IDeskBar Interface
Global Const $sIID_IDeskBar = "{EB0FE173-1A3A-11D0-89B3-00A0C90A90AC}"
Global Const $tRIID_IDeskBar = CLSIDFromString( $sIID_IDeskBar )
Global Const $dtag_IDeskBar = _
	"SetClient hresult(ptr);" & _									; IUnknown* punkClient
	"GetClient hresult(ptr*);" & _								; IUnknown** ppunkClient
	"OnPosRectChangeDB hresult(struct*);"					; LPRECT prc

; IDeskBand Interface
Global Const $sIID_IDeskBand = "{EB0FE172-1A3A-11D0-89B3-00A0C90A90AC}"
Global Const $tRIID_IDeskBand = CLSIDFromString( $sIID_IDeskBand )
Global Const $dtag_IDeskBand = _
	"GetBandInfo hresult(dword;dword;struct*);"		; DWORD dwBandID, DWORD dwViewMode, DESKBANDINFO* pdbi

; IUnknown Interface
Global Const $sIID_IUnknown = "{00000000-0000-0000-C000-000000000046}"
Global Const $tRIID_IUnknown = CLSIDFromString( $sIID_IUnknown )
Global Const $dtag_IUnknown = _
	"QueryInterface hresult(struct*;ptr*);" & _	; REFIID iid, void** ppvObject		; Returns pointers to supported interfaces.
	"AddRef ulong();" & _												; void														; Increments reference count.
	"Release ulong();"													; void														; Decrements reference count.

; IBandSite Interface
Global Const $sIID_IBandSite = "{4CF504B0-DE96-11D0-8B3F-00A0C911E8E5}"
Global Const $tRIID_IBandSite = CLSIDFromString( $sIID_IBandSite )
Global Const $dtag_IBandSite = _
	"AddBand hresult(ptr);" & _														; IUnknown* punk
	"EnumBands hresult(uint;dword*);" & _									; UINT uBand, DWORD* pdwBandID
	"QueryBand hresult(dword;ptr*;dword*;wstr;int);" & _	; DWORD dwBandID, IDeskBand** ppstb, DWORD* pdwState, LPWSTR pszName, int cchName
	"SetBandState hresult(dword;dword;dword);" & _				; DWORD dwBandID, DWORD dwMask, DWORD dwState
	"RemoveBand hresult(dword);" & _											; DWORD dwBandID
	"GetBandObject hresult(dword;struct*;ptr*);" & _			; DWORD dwBandID, REFIID riid, VOID** ppv
	"SetBandSiteInfo hresult(ptr);" & _										; BANDSITEINFO* pbsinfo
	"GetBandSiteInfo hresult(ptr*);"											; BANDSITEINFO* pbsinfo

; IMenuBand Interface
Global Const $sIID_IMenuBand = "{568804CD-CBD7-11d0-9816-00C04FD91972}"
Global Const $tRIID_IMenuBand = CLSIDFromString( $sIID_IMenuBand )
Global Const $dtag_IMenuBand = _
	"IsMenuMessage hresult(struct*);" & _								; MSG* pmsg
	"TranslateMenuMessage hresult(struct*;lresult*);"		; MSG* pmsg, LRESULT* plRet

; --- Functions ---

; Copied from AutoItObject.au3 by
; the AutoItObject team / Prog@ndy.
Func CLSIDFromString($sString)
	Local $tCLSID = DllStructCreate("dword;word;word;byte[8]")
	Local $aRet = DllCall($dllOle32, "long", "CLSIDFromString", "wstr", $sString, "ptr", DllStructGetPtr($tCLSID))
	If @error Then Return SetError(1, @error, 0)
	If $aRet[0] <> 0 Then Return SetError(2, $aRet[0], 0)
	Return $tCLSID
EndFunc

Func OleInitialize()
	Local $aRet = DllCall($dllOle32, "long", "OleInitialize", "ptr", 0)
	If @error Then Return SetError(1, 0, 0)
	Return $aRet[0]
EndFunc

Func OleUninitialize()
	DllCall($dllOle32, "none", "OleUninitialize")
	If @error Then Return SetError(1, 0, 0)
	Return 0
EndFunc

Func CoCreateInstance( $clsid, $pUnkOuter, $ClsContext, $iid, ByRef $pOutObj )
	Local $aRet = DllCall( $dllOle32, "long_ptr", "CoCreateInstance", "ptr", DllStructGetPtr($clsid), "ptr", $pUnkOuter, "dword", $ClsContext, "ptr", DllStructGetPtr($iid), "ptr*", 0 )
	If @error Then Return SetError(1,0,0)
	$pOutObj = $aRet[5]
	Return $aRet[0]
EndFunc

Func SHGetDesktopFolder( ByRef $pIShellFolder )
	Local $aRet = DllCall( $dllShell32, "uint", "SHGetDesktopFolder", "ptr*", 0 )
	If @error Then Return SetError(1,0,0)
	$pIShellFolder = $aRet[1]
	Return $aRet[0]
EndFunc

Func SHGetSpecialFolderLocation( $hwndOwner, $nFolder, ByRef $pidl )
	Local $aRet = DllCall( $dllShell32, "uint", "SHGetSpecialFolderLocation", "hwnd", $hwndOwner, "int", $nFolder, "ptr*", 0 )
	If @error Then Return SetError(1,0,0)
	$pidl = $aRet[3]
	Return $aRet[0]
EndFunc

Func SHILCreateFromPath( $pszPath, ByRef $pidlAbs, ByRef $rgflnOut )
	Local $aRet = DllCall( $dllShell32, "uint", "SHILCreateFromPath", "wstr", $pszPath, "ptr*", 0, "dword*", $rgflnOut )
	If @error Then Return SetError(1,0,0)
	$pidlAbs = $aRet[2]
	$rgflnOut = $aRet[3]
	Return $aRet[0]
EndFunc

Func ShowFavoritesMenu( $hWnd, $hMenu, $xCoord, $yCoord, $sFavFolder = "" )

	; --- Create a menu band object and the IShellMenu interface ---

	Local $oIShellMenu, $pIShellMenu
	; Create a menu band object and get a pointer to the IShellMenu interface
	CoCreateInstance( $tCLSID_MenuBand, $NULL, $CLSCTX_INPROC_SERVER, $tRIID_IShellMenu, $pIShellMenu )
	$oIShellMenu = ObjCreateInterface( $pIShellMenu, $sIID_IShellMenu, $dtag_IShellMenu )
	If Not IsObj( $oIShellMenu ) Then Return SetError(1,0,0)

	; --- Initialize the interface ---

	If $oIShellMenu.Initialize( $NULL, -1, $ANCESTORDEFAULT, BitOR( $SMINIT_TOPLEVEL, $SMINIT_VERTICAL ) ) <> $S_OK Then Return SetError(2,0,0)

	; --- Specify the folder for the menu band to browse ---

	; Get a pointer to the IShellFolder interface for the favorites folder relative to the desktop
	Local $pIShellFolder = $NULL, $oIShellFolderDesktop, $pIShellFolderDesktop, $pidlRel, $rgflnOut = 0
	SHGetDesktopFolder( $pIShellFolderDesktop )
	If $sFavFolder <> "" And FileExists( $sFavFolder ) And StringInStr( FileGetAttrib( $sFavFolder ), "D" ) > 0 Then
		SHILCreateFromPath( $sFavFolder, $pidlRel, $rgflnOut )
	Else
		SHGetSpecialFolderLocation( $NULL, $CSIDL_FAVORITES, $pidlRel )
	EndIf
	$oIShellFolderDesktop = ObjCreateInterface( $pIShellFolderDesktop, $sIID_IShellFolder, $dtag_IShellFolder )
	$oIShellFolderDesktop.BindToObject( $pidlRel, $NULL, $tRIID_IShellFolder, $pIShellFolder )
	If $pIShellFolder = $NULL Then Return SetError(3,0,0)

	; Specify the folder for the menu band to browse
	If $oIShellMenu.SetShellFolder( $pIShellFolder, $pidlRel, $NULL, BitOR( $SMSET_BOTTOM, $SMSET_USEBKICONEXTRACTION ) ) <> $S_OK Then Return SetError(5,0,0)
	_WinAPI_CoTaskMemFree( $pidlRel )

	; --- Append a static menu to the menu band ---

	If $oIShellMenu.SetMenu( $hMenu, $hWnd, $SMSET_TOP ) <> $S_OK Then Return SetError(6,0,0)

	; --- Get access to IMenuPopup, IDeskBar and IDeskBand interfaces ---

	Local $pIMenuPopup, $pIDeskBar, $pIDeskBand

	; The IMenuPopup interface
	If $pIShellMenu Then
		If $oIShellMenu.QueryInterface( $tRIID_IMenuPopup, $pIMenuPopup ) <> $S_OK Then Return SetError(7,0,0)
	EndIf

	; The IDeskBar interface
	If $pIMenuPopup Then
		Local $oIMenuPopup = ObjCreateInterface( $pIMenuPopup, $sIID_IMenuPopup, $dtag_IMenuPopup )
		If $oIMenuPopup.QueryInterface( $tRIID_IDeskBar, $pIDeskBar ) <> $S_OK Then Return SetError(8,0,0)
	EndIf

	; The IDeskBand interface
	If $pIDeskBar Then
		Local $oIDeskBar = ObjCreateInterface( $pIDeskBar, $sIID_IDeskBar, $dtag_IDeskBar )
		If $oIDeskBar.QueryInterface( $tRIID_IDeskBand, $pIDeskBand ) <> $S_OK Then Return SetError(9,0,0)
	EndIf

	; --- Create a menu desk bar object and the IMenuPopup interface ---

	Local $oIUnknown, $pIUnknown, $oIMenuPopup, $pIMenuPopup

	; Create a menu desk bar object and get a pointer to the IUnknown interface
	CoCreateInstance( $tCLSID_MenuDeskBar, $NULL, $CLSCTX_INPROC_SERVER, $tRIID_IUnknown, $pIUnknown )
	$oIUnknown = ObjCreateInterface( $pIUnknown, $sIID_IUnknown, $dtag_IUnknown )

	; Create the IMenuPopup interface
	$oIUnknown.QueryInterface( $tRIID_IMenuPopup, $pIMenuPopup )
	$oIMenuPopup = ObjCreateInterface( $pIMenuPopup, $sIID_IMenuPopup, $dtag_IMenuPopup )
	If Not IsObj( $oIMenuPopup ) Then Return SetError(10,0,0)

	; --- Create a menu band site object and the IBandSite interface ---

	Local $oIBandSite, $pIBandSite
	; Create a menu band site object and get a pointer to the IBandSite interface
	CoCreateInstance( $tCLSID_MenuBandSite, $NULL, $CLSCTX_INPROC_SERVER, $tRIID_IBandSite, $pIBandSite )
	$oIBandSite = ObjCreateInterface( $pIBandSite, $sIID_IBandSite, $dtag_IBandSite )
	If Not IsObj( $oIBandSite ) Then Return SetError(11,0,0)

	; --- Set the band site object as client for the desk bar object ---

	If $oIMenuPopup.SetClient( $pIBandSite ) <> $S_OK Then Return SetError(12,0,0)

	; --- Add the desk band object to the band site object ---

	If $oIBandSite.AddBand( $pIDeskBand ) <> $S_OK Then Return SetError(13,0,0)

	; --- Show the Favorites menu ---

	Local $tPOINT = DllStructCreate( $tagPOINT )
	DllStructSetData( $tPOINT, "X", $xCoord )
	DllStructSetData( $tPOINT, "Y", $yCoord )

	Local $tRECT = DllStructCreate( $tagRECT )
	DllStructSetData( $tRECT, "Left", $xCoord )
	DllStructSetData( $tRECT, "Top", $yCoord+1 )
	DllStructSetData( $tRECT, "Right", $xCoord )
	DllStructSetData( $tRECT, "Bottom", $yCoord+1 )

	$oIMenuPopup.Popup( $tPOINT, $tRECT, $MPPF_ALIGN_RIGHT )

	; --- Create the IMenuBand interface ---

	Local $oIDeskBand
	$oIDeskBand = ObjCreateInterface( $pIDeskBand, $sIID_IDeskBand, $dtag_IDeskBand )
	$oIDeskBand.QueryInterface( $tRIID_IMenuBand, $pIMenuBand )
	$oIMenuBand = ObjCreateInterface( $pIMenuBand, $sIID_IMenuBand, $dtag_IMenuBand )
	If Not IsObj( $oIMenuBand ) Then Return SetError(14,0,0)

	; $oIMenuBand and $pIMenuBand are global variables
	; $oIMenuBand is used to handle the messages from the Favorites menu in the function below
	; $pIMenuBand is used to test if the Favorites menu is open ($pIMenuBand <> $NULL)

	; --- Favorites menu messages ---

	If $idFavMenuMsgUnHook = 0 Then $idFavMenuMsgUnHook = GUICtrlCreateDummy()
	$hFavMenuMsg = DllCallbackRegister( "FavMenuMsg", "long", "int;wparam;lparam" )
	$hFavMenuMsgHook = _WinAPI_SetWindowsHookEx( $WH_MSGFILTER, DllCallbackGetPtr( $hFavMenuMsg ), 0, _WinAPI_GetCurrentThreadId() )
	If $hFavMenuMsgHook = 0 Then Return SetError(15,0,0)
	; $WH_MSGFILTER installs a hook procedure that monitors messages generated as a
	; result of an input event in a dialog box, message box, menu, or scroll bar.

	Return 0

EndFunc

Func TranslateFavMenuMessage()
	Local $i = $oIMenuBand.IsMenuMessage( $tFavMenuMsg )
	If $i >= 0 Then
		Local $tLRESULT = DllStructCreate( "lresult" )
		If $oIMenuBand.TranslateMenuMessage( $tFavMenuMsg, $tLRESULT ) = $S_OK Then Return True
	ElseIf $i = $E_FAIL Then
		; The menu has exited the menu mode and can be destroyed
		$pIMenuBand = $NULL
		Return False
	EndIf
	Return False
EndFunc

Func FavMenuMsg( $nCode, $wParam, $lParam )
	Local $msg, $vKey
	Local Static $fVK_RETURN = False
	If $pIMenuBand Then
		If $nCode >= 0 Then
			$tFavMenuMsg = DllStructCreate( $tagMSG, $lParam )
			$msg = DllStructGetData( $tFavMenuMsg, "message" )
			Switch $msg
				Case $WM_NULL, $WM_PAINT, $WM_TIMER, 0x0118, _	; These messages are not relevant for the Favorites menu
						 $WM_MOUSEFIRST To $WM_XBUTTONDBLCLK				; 0x0118 ~ WM_SYSTIMER, WM_XBUTTONDBLCLK ~ WM_MOUSELAST
					Return False
					;Return _WinAPI_CallNextHookEx( 0, $nCode, $wParam, $lParam )

				Case $WM_KEYDOWN
					$vKey = DllStructGetData( $tFavMenuMsg, "wParam" )
					Switch $vKey
						Case $VK_RETURN, $VK_LEFT, $VK_UP, $VK_RIGHT, $VK_DOWN
							If $vKey = $VK_RETURN Then $fVK_RETURN = True
							TranslateFavMenuMessage()
							Return True
					EndSwitch

				Case Else
					If $fVK_RETURN Then
						$fVK_RETURN = False
						Return False
						;Return _WinAPI_CallNextHookEx( 0, $nCode, $wParam, $lParam )
					EndIf
					If TranslateFavMenuMessage() Then
						Return True			; If the message is handled by TranslateFavMenuMessage() then return True
					Else							; In all other cases the message should be send to the next hook
						Return False		; But you can save some time by returning immediately
						;Return _WinAPI_CallNextHookEx( 0, $nCode, $wParam, $lParam )
					EndIf
			EndSwitch
		Else
			Return False
			;Return _WinAPI_CallNextHookEx( 0, $nCode, $wParam, $lParam )
		EndIf
	Else
		If $hFavMenuMsg <> $NULL Then
			$msg = DllStructGetData( $tFavMenuMsg, "message" )
			GUICtrlSendToDummy( $idFavMenuMsgUnHook )
			; At this point the Favorites menu has exited
			; The $idFavMenuMsgUnHook event must execute the FavMenuMsgUnHook() function below
			; This will disable the FavMenuMsg() hook procedure
		EndIf
		Return False
		;Return _WinAPI_CallNextHookEx( 0, $nCode, $wParam, $lParam )
	EndIf
EndFunc

Func FavMenuMsgUnHook()
	If $hFavMenuMsg <> $NULL Then
		_GUICtrlMenu_DestroyMenu( $hFavMenu )
		_WinAPI_UnhookWindowsHookEx( $hFavMenuMsgHook )
		DllCallbackFree( $hFavMenuMsg )
		$hFavMenuMsg = $NULL
	EndIf
EndFunc

Func ShellFavoritesMenuExit()
	OleUninitialize()
	DllClose( $dllOle32 )
	DllClose( $dllShell32 )
	DllClose( $dllUser32 )
EndFunc
