; Utility functions

#include-once

Global Const $dllGDI32    = DllOpen( "gdi32.dll" )
Global Const $dllGLU32    = DllOpen( "glu32.dll" )
Global Const $dllKernel32 = DllOpen( "kernel32.dll" )
Global Const $dllOpenGL32 = DllOpen( "opengl32.dll" )
Global Const $dllUser32   = DllOpen( "user32.dll" )

Global $iHeightXP = 0
If @OSVersion = "WIN_XP" Then $iHeightXP = 3

#include "..\oglIncludes\oglConstants.au3"
#include "..\oglIncludes\oglFunctions.au3"
#include "..\oglIncludes\extConstants.au3"
#include "..\oglIncludes\extFunctions.au3"
#include "..\oglIncludes\oglUtilities.au3"
#include "..\oglIncludes\oglContext.au3"

Func CloseDlls( $bUser32 = True )
	DllClose( $dllGDI32 )
	DllClose( $dllGLU32 )
	DllClose( $dllKernel32 )
	DllClose( $dllOpenGL32 )
	If $bUser32 Then _
		DllClose( $dllUser32 )
EndFunc

Func GetProjectPath()
	Local $p = ""
	While Not FileExists( $p & "Includes\Utilities.au3" )
		$p &= "..\"
	WEnd
	Return $p
EndFunc

; The purpose of this function is to read the parameters for creation of
; an OpenGL rendering context in an inifile. By changing the parameters
; in the inifile, the examples can be tested under different contexts.				; $fExLauncher = 0, 1, 2
Func GetOpenGLContext( $hWnd, ByRef $hDC, ByRef $hRC, $fExLauncher = 0 )			; 0   - An example is executed without the launcher
	Local $sVerPath = GetProjectPath() & "Version.ini", $iStatus								;       E.i. run directly by double clicking the example file
	Local $sContextType = "Simple", $iMajorVersion = -1, $iMinorVersion = -1		; 1,2 - The launcher is opening or executing an example

	; Read Version.ini
	If FileExists( $sVerPath ) And FileGetSize( $sVerPath ) Then
		$sContextType = IniRead( $sVerPath, "OpenGL Context", "ContextType", "None" )						; Setting illegal default values will
		$iMajorVersion = Number( IniRead( $sVerPath, "OpenGL Context", "MajorVersion", -2 ) )		; reveal missing keys in the inifile.
		$iMinorVersion = Number( IniRead( $sVerPath, "OpenGL Context", "MinorVersion", -2 ) )

		; The inifile can be edited manually. Setting illegal values for the parameters can
		; make creation of a context to fail. To avoid this the parameters in the inifile are
		; checked carefully.

		; Even if the parameters in the inifile have legal values the context creation can fail.
		; But the context creation should not fail because of illegal values.

		; The interesting parameters and values are:
		If $sContextType = "Proper" And ( $iMajorVersion <> -1 Or $iMinorVersion <> -1 ) Then
			; This is an inifile where the rendering context is set to a specific version.
			; Check that the major and minor version numbers are within the limits of your
			; video card.

			If $iMajorVersion <= 0 Or $iMinorVersion < 0 Then _
				Return HandleMajorMinorErrors( $fExLauncher )

			Local $hGuiTst, $hDCTst, $hRCTst, $sVersion, $iMajor = -1, $iMinor = -1

			$hGuiTst = GUICreate( "OpenGL Test" )
			If EnableOpenGL( $hGuiTst, $hDCTst, $hRCTst ) Then

				$sVersion = StringLeft( glGetString( $GL_VERSION ), 3 )
				$iMajor = Number( StringLeft( $sVersion, 1 ) )   ; Major version supported by video card
				$iMinor = Number( StringRight( $sVersion, 1 ) )  ; Minor version supported by video card
				DisableOpenGL( $hGuiTst, $hDCTst, $hRCTst )
				GUIDelete( $hGuiTst )

				; A "Proper" context is only supported by 3.0+ versions of OpenGL
				If $iMajor < 3 Then Return HandleMajorMinorErrors( $fExLauncher )

				; You can't get a higher version than your video card supports
				If $iMajor = 3 And $iMajorVersion = 3 And $iMinor < $iMinorVersion Then _
					Return HandleMajorMinorErrors( $fExLauncher )

				; You can't get a higher version than your video card supports
				If $iMajor < $iMajorVersion Then Return HandleMajorMinorErrors( $fExLauncher )

				; Now the legal values for $iMajorVersion and $iMinorVersion are:
				Select
					Case $iMajorVersion = 1 And $iMinorVersion = 1
					Case $iMajorVersion = 2 And $iMinorVersion = 1
					Case $iMajorVersion = 3 And ( $iMinorVersion >= 0 And $iMinorVersion <= 3 )
					Case Else
						Return HandleMajorMinorErrors( $fExLauncher )
				EndSelect

			Else
				DisableOpenGL( $hGuiTst, $hDCTst, $hRCTst )
				GUIDelete( $hGuiTst )
			EndIf
		EndIf
	EndIf

	If $sContextType = "Simple" Then
		If EnableOpenGLEx( $hWnd, $hDC, $hRC, -1, -1, "MSAAbest" ) = 0 Then
			;MsgBox( 48, "GetOpenGLContext", "MSAA context is not supported" )
			Return OpenGLSimpleContext( $hWnd, $hDC, $hRC )
		Else
			Return 1
		EndIf

	ElseIf $sContextType = "Proper" Then
		$iStatus = OpenGLProperContext( $hWnd, $hDC, $hRC, $iMajorVersion, $iMinorVersion )
		If $iStatus = 1 And $fExLauncher = 1 Then
			; $fExLauncher = 1 means that the launcher is opening
			; Return info to the launcher to set (click) the proper radio button
			; The values are already verified above when checking the inifile
			Select
				Case $iMajorVersion = -1 And $iMinorVersion = -1
					Return 2
				Case $iMajorVersion =  3 And $iMinorVersion <= 3
					Return 6 - $iMinorVersion
				Case $iMajorVersion =  2 And $iMinorVersion =  1
					Return 7
				Case $iMajorVersion =  1 And $iMinorVersion =  1
					Return 8
				Case Else
					; Fejl
			EndSelect
		Else
			Return $iStatus
		EndIf

	; Illegal context type in inifile
	; Or missing key "ContextType" in section "OpenGL Context" indicated by the value "None"

	ElseIf $fExLauncher = 0 Then
		; Example is executed without the launcher
		; E.i. run directly by double clicking the example file
		PrintError( "Function GetOpenGLContext()" & @CRLF & _
			"ERROR: Illegal or missing context type """ & $sContextType & """ in Version.ini." & @CRLF & _
			"Please test and save a new context with ExLauncher.au3 in the top folder." )
		Return 0

	ElseIf $fExLauncher = 1 Then
		; The launcher is opening
		PrintError( "Function GetOpenGLContext()" & @CRLF & _
			"ERROR: Illegal or missing context type """ & $sContextType & """ in Version.ini." & @CRLF & _
			"Please test and save a new context." )
		Return 0

	ElseIf $fExLauncher = 2 Then
		; Example is executed from the launcher
		; The example file is executed with a parameter $fExLauncher = 2
		; In the example file this parameter is testet, if it has the value 2 we know it's startet from the launcher
		; In this case we don't have to check for the context type in the inifile, this has already been done when the launcher startet up

	EndIf
EndFunc

Func HandleMajorMinorErrors( $fExLauncher )
	If $fExLauncher = 0 Then
		; Example is executed without the launcher
		; E.i. run directly by double clicking the example file
		PrintError( "Function GetOpenGLContext()" & @CRLF & _
			"ERROR: Illegal or missing major or minor OpenGL version numbers in Version.ini." & @CRLF & _
			"Please test and save a new context with ExLauncher.au3 in the top folder." )
	Else
		; The launcher is opening or executing an example file
		PrintError( "Function GetOpenGLContext()" & @CRLF & _
			"ERROR: Illegal or missing major or minor OpenGL version numbers in Version.ini." & @CRLF & _
			"Please test and save a new context." )
	EndIf
	Return 0
EndFunc

Func ScriptName()
	Return StringLeft( @ScriptName, StringInStr( @ScriptName, ".", 0, -1 ) - 1 )
EndFunc

Func ExLauncherWin()
	If WinExists( "[TITLE:ExLauncher; CLASS:AutoIt v3 GUI]", "" ) Then _
		WinActivate( "[TITLE:ExLauncher; CLASS:AutoIt v3 GUI]", "" )
EndFunc
