#include <GUIConstantsEx.au3>
#include <StaticConstants.au3>
#include <WindowsConstants.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

Global $bOpenGL21 = False
Global $nNormals, $pNormals, $tNormals
Global $sObjModel = "Small\Art\Greek vase\greek_vase.obj"

; --- OpenGL globals ---

Global $aModel[14]
Global $aMaterialData[1][9]
Global $aTransforms[7] ; [ $fScale, $fRotX, $fRotY, $fRotZ, $fTransX, $fTransY, $fTransZ ]

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 800, $iHeight = 600

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30
Global Const $iMenuBarHeight   = 20
Global Const $iStatusBarHeight = 28

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight + $iMenuBarHeight + $iStatusBarHeight

Global $hGui, $idInput, $bResize = True, $bClose = False, $bFocus = True

; --- Project includes ---

#include "utilities\models.au3"
#include "utilities\includes.au3"
#include "utilities\utilities.au3"

; --- Program code ---

MainFunc()

While 1
	; Load a new model
	If $sObjModel <> $sNewModel Then
		If $bOpen Then
			; A model is opened through the "Open File Dialog"
			; Let the OpenGL window update once before the model is loaded
			; (To refresh the part of the OpenGL window where the "Open File Dialog" was displayed)
			$bOpen = 0
		Else
			Local $aNewModel[14]
			If $sNewModel = "" Then $sNewModel = $sObjModel
			If CreateResources( $sNewModel, $aNewModel ) Then
				; If the new model is loaded without errors then make it the current model
				DeletePreviousBuffers()
				$sObjModel = $sNewModel
				$aModel = $aNewModel
			Else
				; Continue with the old model in case of errors
				$sNewModel = $sObjModel
			EndIf
		EndIf
	EndIf

	; Menu commands that affect OpenGL should not
	; be performed in the middle of drawing cycle.
	If $bMenuCmd Then
		$bMenuCmd = False
		Switch $iMenuId

			; --- Models menu ---

			Case $idModelsMtls To $idModelsMtlsMax
				If $sMtlFileNew <> $sMtlFile Then
					DeletePreviousMaterials()
					Local $s = $sMtlFileNew, $aMtlInfo
					Local $j = StringInStr( $s, "\", 0, -1 )
					ReadMaterials( StringLeft( $s, $j ), StringRight( $s, StringLen( $s ) - $j ), $aMtlInfo )
					$aModel[6] = $aMtlInfo
					LoadMaterials( $aModel, $aMaterialData )
					LoadTextures( $aModel, $aMaterialData )
					$sMtlFile = $sMtlFileNew
					$aModel[6] = StringRight( $sMtlFile, StringLen( $sMtlFile ) - StringInStr( $sMtlFile, "\", 0, -1 ) )
				EndIf

			; --- View menu ---

			Case $idViewWire
				If $bWire Then
					glPolygonMode( $GL_FRONT_AND_BACK, $GL_LINE )
				Else
					glPolygonMode( $GL_FRONT_AND_BACK, $GL_FILL )
				EndIf
			Case $idViewNormals, $idMaxVectors3000, $idMaxVectorsNoLimit
				If BitAND( $bVectors, 4 ) Then
					CalcNormals( $aModel )
				Else
					$nNormals = 0
					$pNormals = 0
					$tNormals = 0
				EndIf

			; --- Options menu ---

			; Light
			Case $idOptionsLight
				SetLight()
				$bDrawLight = $bModelLight
				If $bDrawLight Then           ; Light is turned on
					If $bModelNormals = -1 Then ; Normal vectors isn't loaded
						; If light is turned on and normal vectors isn't loaded, then reload the
						; model WITH normal vectors. When light is turned on we need normal vectors.
						If ReloadModel( $aModel, 0 ) Then
							SetNormals( 0 )
							$bDrawNormals = $bModelNormals
							glEnable( $GL_LIGHTING )
						Else
							; ReloadModel cancelled
							; Undo light change
							SetLight()
							$bDrawLight = $bModelLight
						EndIf
					Else ; $bModelNormals <> -1
						glEnable( $GL_LIGHTING )
					EndIf
				Else
					glDisable( $GL_LIGHTING )
				EndIf

			; Textures
			Case $idOptionsTextures
				SetTextures()
				$bDrawTextures = $bModelTextures
				If Not $bDrawTextures And _ ; Textures are turned off
					 Not $bDrawLight Then     ; Light is turned off
						 ; When textures are turned off it is assumed that you want to see the model with light instead of
						 ; textures. To see the model with ligth we need normal vectors and the light must be turned on.
						 If $bModelNormals = -1 Then
							 If ReloadModel( $aModel, 0 ) Then
								 SetNormals( 0 )
								 $bDrawNormals = $bModelNormals
								 SetLight()
								 glEnable( $GL_LIGHTING )
								 $bDrawLight = $bModelLight
							 Else
								 ; ReloadModel cancelled
								 ; Undo texture change
								 SetTextures()
								 $bDrawTextures = $bModelTextures
							 EndIf
						 EndIf
				EndIf

			; Normals
			Case $idCalculateNormals, $idAverageNormals
				; Calculate normal vectors with or without average
				; When settings are switched normal vectors must be recalculated
				SetNormals( $bModelNormals )
				If ReloadModel( $aModel, $bModelNormals ) Then
					If Not $bModelLight Then ; Light is turned off
						; When you change normal vector settings you expect to
						; see a change. To see a change light must be turned on.
						SetLight()
						glEnable( $GL_LIGHTING )
						$bDrawLight = $bModelLight
					EndIf
					$bDrawNormals = $bModelNormals
				Else
					SetNormals( $bModelNormals )
				EndIf
			Case $idNoNormals
				; This decides whether to load normal vectors or not
				; When settings are switched normal vectors must be recalculated
				SetNormals( -$bModelNormals - 1 )
				If ReloadModel( $aModel, $bModelNormals ) Then
					If $bModelNormals = -1 _ ; Normal vectors isn't loaded
						 And $bModelLight Then ; Light is turned on
							 ; If normal vectors isn't loaded
							 ; then turn light off.
							 SetLight()
							 $bDrawLight = $bModelLight
							 glDisable( $GL_LIGHTING )
					EndIf
					$bDrawNormals = $bModelNormals
				Else
					; ReloadModel cancelled
					; Undo normal vector change
					SetNormals( -$bModelNormals - 1 )
				EndIf

			; Quads
			Case $idOptionsQuads
				SetQuads()
				If ReloadModel( $aModel, $bModelNormals ) Then
					If $bModelQuads Then
						$bModelDrawMode = $GL_QUADS
					Else
						$bModelDrawMode = $GL_TRIANGLES
					EndIf
				Else
					; ReloadModel cancelled
					; Undo draw mode change
					SetQuads()
				EndIf

		EndSwitch
	EndIf

	If $bResize Then
		glViewport( 0, 0, $iWidth, $iHeight )
		InitializeView()
		$bResize = False
	EndIf

	If $bClose Then
		CloseProgram()
	EndIf

	; OpenGL drawing cycle
	DisplayFunction()

	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then CloseProgram()
	If Not InitOpenGLwindow() Then CloseProgram()
	;If Not CheckOpenGLversion( 2, 1 ) Then CloseProgram()
	If Not CreateResources( $sObjModel, $aModel ) Then CloseProgram()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 50, $WS_OVERLAPPEDWINDOW )

	; Menu system
	CreateMenuSystem()

	; Statusbar
	Local $idStatus, $idLabel
	$idInput = GUICtrlCreateInput( "", 0, 0, 1, 1, 2048 ) ; 2048 = $ES_READONLY
	$idStatus = GUICtrlCreateLabel( "", 0, $iHeight, $iGuiWidth + 2, $iStatusBarHeight+2, $SS_ETCHEDFRAME )
	$idLabel = GUICtrlCreateLabel( "", 3, $iHeight + 6, $iGuiWidth - 4, $iStatusBarHeight - 10 )
	GUICtrlSetData( $idLabel, "Rotate: Q/A, W/S, E/D  |  Move: U/J, I/K, O/L  |  Zoom: Page up/down  |  Reset: Home   (Close model list to use shortcuts)" )
	GUICtrlSetResizing( $idStatus, $GUI_DOCKSTATEBAR )
	GUICtrlSetResizing( $idLabel, $GUI_DOCKSTATEBAR )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	GUIRegisterMsg( $WM_ACTIVATE, "WM_ACTIVATE" )
	GUIRegisterMsg( $WM_COMMAND,  "WM_COMMAND" )

	GUISetState()

	Return 1

EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	;If EnableOpenGLEx( $hOpenGL, $hDC, $hRC, -1, -1, "MSAAbest" ) = 0 Then
	;	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	;EndIf
	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	#cs
	If CheckOpenGLextension( "GL_EXT_texture_filter_anisotropic" ) Then
		Local $fMaxAnisotrophy
		glGetFloatv( $GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, $fMaxAnisotrophy )
		If $fMaxAnisotrophy > 1.0 Then _
			glTexParameterf( $GL_TEXTURE_2D, $GL_TEXTURE_MAX_ANISOTROPY_EXT, $fMaxAnisotrophy )
	EndIf
	#ce

	glClearColor( 0.75, 0.75, 0.75, 1.0 )

	glEnable( $GL_TEXTURE_2D )
	glEnable( $GL_DEPTH_TEST )
	;glEnable( $GL_CULL_FACE )
	glEnable( $GL_LIGHTING )
	glEnable( $GL_LIGHT0 )

	glEnable( $GL_BLEND )
	glBlendFunc( $GL_SRC_ALPHA, $GL_ONE_MINUS_SRC_ALPHA )

	glPolygonMode( $GL_FRONT_AND_BACK, $GL_FILL )
	;glPolygonMode( $GL_FRONT_AND_BACK, $GL_LINE )

	;glShadeModel( $GL_SMOOTH )

	GUISwitch( $hGui )

	Return 1

EndFunc


Func CreateResources( $sObjModel, ByRef $aModel )

	; Check file format and file existence of the OBJ-model
	; Set $aModel[11] = full path to the OBJ-file WITHOUT the .obj suffix
	If Not CheckModelFile( $sObjModel, $aModel ) Then Return 0

	; Save previous settings
	; This makes it possible to restore settings if load of a new model is cancelled
	$bModelLightPrev    = $bModelLight
	$bModelTexturesPrev = $bModelTextures
	$bModelNormalsPrev  = $bModelNormals
	$bModelQuadsPrev    = $bModelQuads

	; Get model settings from inifile
	; If no inifile set defaults
	; Store settings in $aModel[10]
	GetModelSettings( $aModel, False )
	Local $aSettings = $aModel[10]
	$bModelLight    = $aSettings[0]
	$bModelTextures = $aSettings[2]
	$bModelNormals  = $aSettings[3]
	$bModelQuads    = $aSettings[4]

	; Load model
	; Parameters for model loading functions:
	; LoadObjModel, LoadObjModelBins, LoadObjQuads, LoadObjQuadsBins
	; LoadObjModel( ByRef $aModel, $bOpenGL21 = True, $bNormals = 0 )
	; $bOpenGL21 is used to handle the OpenGL version
	;   Set $bOpenGL21 = True to use the model for OpenGL version 2.1 or higher
	;   Set $bOpenGL21 = False to use the model for version 2.0 or lower
	; $bNormals is used to handle normal vectors. Values: 0, 1, -1
	; $bNormals = 0/1 is for models which does not contain any normal vectors
	;   Set $bNormals = 0 to calculate normal vectors (no average)
	;   Set $bNormals = 1 to calculate average normal vectors
	; $bNormals = -1 is for models which does contain normal vectors
	;   Set $bNormals = -1 to exclude normal vectors when they are not needed

	Local $bLoadBinaries = False

	If Not $bModelQuads Then

		If FileExists( $aModel[11] & "V.bin" ) _
		Or FileExists( $aModel[11] & "VCal.bin" ) _
		Or FileExists( $aModel[11] & "VAvg.bin" ) Then
			If LoadObjModelBins( $aModel, $bOpenGL21, $bModelNormals ) Then _
				$bLoadBinaries = True
		EndIf

		If Not $bLoadBinaries Then
			If Not LoadObjModel( $aModel, $bOpenGL21, $bModelNormals ) Then
				$bModelLight    = $bModelLightPrev
				$bModelTextures = $bModelTexturesPrev
				$bModelNormals  = $bModelNormalsPrev
				$bModelQuads    = $bModelQuadsPrev
				Return 0
			EndIf
		EndIf

	Else

		; Use LoadObjQuads/LoadObjQuadsBins to force a model to be loaded as a quad mesh and rendered with $GL_QUADS draw mode

		If FileExists( $aModel[11] & "V4.bin" ) _
		Or FileExists( $aModel[11] & "VCal4.bin" ) _
		Or FileExists( $aModel[11] & "VAvg4.bin" ) Then
			If LoadObjQuadsBins( $aModel, $bOpenGL21, $bModelNormals ) Then _
				$bLoadBinaries = True
		EndIf

		If Not $bLoadBinaries Then
			If Not LoadObjQuads( $aModel, $bOpenGL21, $bModelNormals ) Then
				$bModelLight    = $bModelLightPrev
				$bModelTextures = $bModelTexturesPrev
				$bModelNormals  = $bModelNormalsPrev
				$bModelQuads    = $bModelQuadsPrev
				Return 0
			EndIf
		EndIf

	EndIf

	DeletePreviousMaterials()

	Local $iObjects = $aModel[13]
	Redim $aMaterialData[$iObjects][9]

	; Load materials
	; Always load materials, if no material information then load default materials
	LoadMaterials( $aModel, $aMaterialData )

	; Load texture maps
	; Always load texture maps, if no texture maps then load default texture map ids
	LoadTextures( $aModel, $aMaterialData )

	; Some models are provided with several different material files which can be changed
	; dynamically without the need to reload the entire model. Get the names of these files
	; and add the names to the Model menu.
	$sMtlFile = ""
	Local $aMtlInfo
	If IsArray( $aModel[6] ) Then
		$aMtlInfo = $aModel[6]
		$sMtlFile = $aMtlInfo[0][0]
	EndIf
	SetModelsMenuMaterials( $aModel[11], $sMtlFile )

	; $aModel[6] = $aMaterialInfo is used to load materials and texture maps, not needed any more
	; Instead of $aModel[6] will be used to store the name of the material file
	If IsArray( $aModel[6] ) Then $aModel[6] = $sMtlFile

	; Light
	If $bModelLight Then
		glEnable( $GL_LIGHTING )
	Else
		glDisable( $GL_LIGHTING )
	EndIf

	; Draw mode
	If $bModelQuads Then
		$bModelDrawMode = $GL_QUADS
	Else
		$bModelDrawMode = $GL_TRIANGLES
	EndIf

	; Set state op Option menu items
	SetOptionsMenuStates( $aModel )

	; Read transformations
	Local $aTransforms0[7] = _
		[ 1, 0, 0, 0, 0, 0, 0 ]
	$aTransforms = $aTransforms0
	ReadTransforms( $aModel[11], $aTransforms )

	; Apply transformations
	ResetModel( $aModel )

	; Normal vectors
	$nNormals = 0
	$pNormals = 0
	$tNormals = 0
	glColor3f( 1, 1, 1 )
	If BitAND( $bVectors, 1 ) Then
		If BitAND( $bVectors, 4 ) Then _
			CalcNormals( $aModel )
	EndIf

	Return 1

EndFunc


Func ReloadModel( ByRef $aModel, $bNormals )
	Local $aNewModel[14], $aSettings[5], $bOK
	$aNewModel[10] = $aSettings
	$aNewModel[11] = $aModel[11]
	If Not $bModelQuads Then
		If FileExists( $aNewModel[11] & "V.bin" ) _
		Or FileExists( $aNewModel[11] & "VCal.bin" ) _
		Or FileExists( $aNewModel[11] & "VAvg.bin" ) Then _
			$bOK = LoadObjModelBins( $aNewModel, $bOpenGL21, $bNormals )
		If Not $bOK Then _
			$bOK = LoadObjModel( $aNewModel, $bOpenGL21, $bNormals, True )
	Else
		If FileExists( $aNewModel[11] & "V4.bin" ) _
		Or FileExists( $aNewModel[11] & "VCal4.bin" ) _
		Or FileExists( $aNewModel[11] & "VAvg4.bin" ) Then _
			$bOK = LoadObjQuadsBins( $aNewModel, $bOpenGL21, $bNormals )
		If Not $bOK Then _
			$bOK = LoadObjQuads( $aNewModel, $bOpenGL21, $bNormals, True )
	EndIf
	If $bOK Then
		Local $aBuffer
		For $i = 0 To 3                  ; Vertex coordinates
			; Delete old data              ; Texture coordinates
			If IsArray( $aModel[$i] ) Then ; Normal vectors
				$aBuffer = $aModel[$i]       ; Face triangles
				$aBuffer[1] = 0 ; Delete pointer
				$aBuffer[0] = 0 ; Delete structure
				$aModel[$i] = 0 ; Delete array
			EndIf
			; Add new data
			$aModel[$i] = $aNewModel[$i]
		Next
		$aModel[4] = $aNewModel[4]       ; $aVertInfo
	EndIf
	; Normal vectors
	$nNormals = 0
	$pNormals = 0
	$tNormals = 0
	glColor3f( 1, 1, 1 )
	If BitAND( $bVectors, 1 ) Then
		If BitAND( $bVectors, 4 ) Then _
			CalcNormals( $aModel )
	EndIf
	Return $bOK
EndFunc


Func CalcNormals( ByRef $aModel )

	If Not IsArray( $aModel[2] ) Then Return

	Local $aNormals = $aModel[2]
	Local $aVertices = $aModel[0], $j = 0
	$nNormals = DllStructGetSize( $aVertices[0] ) / 12

	If BitAND( $bVectors, 2 ) And $nNormals > 3000 Then
		$nNormals = 0
		Return
	EndIf

	$tNormals = DllStructCreate( "float[" & 3 * $nNormals & "]" )

	For $i = 0 To $nNormals - 1
		DllStructSetData( $tNormals, 1, DllStructGetData( $aVertices[0], 1, $j+1 ) + 0.1 * DllStructGetData( $aNormals[0], 1, $j+1 ), $j+1 )
		DllStructSetData( $tNormals, 1, DllStructGetData( $aVertices[0], 1, $j+2 ) + 0.1 * DllStructGetData( $aNormals[0], 1, $j+2 ), $j+2 )
		DllStructSetData( $tNormals, 1, DllStructGetData( $aVertices[0], 1, $j+3 ) + 0.1 * DllStructGetData( $aNormals[0], 1, $j+3 ), $j+3 )
		$j += 3
	Next

	$pNormals = DllStructGetPtr( $tNormals )

EndFunc


Func ResetModel( ByRef $aModel )

	$aModel[12] = $aTransforms

EndFunc


Func InitializeView()

	glMatrixMode( $GL_PROJECTION )
	glLoadIdentity()
	gluPerspective( 45, $iWidth / $iHeight, 0.5, 100 )

	glMatrixMode( $GL_MODELVIEW )
	glLoadIdentity( )
	gluLookAt( 0, 0, 2, 0, 0, 0, 0.0, 1.0, 0.0 )

EndFunc


Func DrawModel( ByRef $aModel )

	Local $aFaces      = $aModel[3]
	Local $aVertInfo   = $aModel[4]
	Local $aTransforms = $aModel[12]
	Local $iObjects    = $aModel[13]
	Local $aVertStart  = $aVertInfo[0]
	Local $aVertCnts   = $aVertInfo[1]

	; Model transformations
	glRotatef( $aTransforms[1], 1.0, 0.0, 0.0 )
	glRotatef( $aTransforms[2], 0.0, 1.0, 0.0 )
	glRotatef( $aTransforms[3], 0.0, 0.0, 1.0 )

	glTranslatef( $aTransforms[4], $aTransforms[5], $aTransforms[6] )

	glScalef( $aTransforms[0], $aTransforms[0], $aTransforms[0] )

	; Draw model
	For $i = 0 To $iObjects - 1

		; Materials
		If $bDrawLight Then
			glMaterialfv( $GL_FRONT_AND_BACK, $GL_AMBIENT,   $aMaterialData[$i][1] ) ; Ka
			glMaterialfv( $GL_FRONT_AND_BACK, $GL_DIFFUSE,   $aMaterialData[$i][2] ) ; Kd
			glMaterialfv( $GL_FRONT_AND_BACK, $GL_SPECULAR,  $aMaterialData[$i][3] ) ; Ks
			glMaterialf(  $GL_FRONT_AND_BACK, $GL_SHININESS, $aMaterialData[$i][4] ) ; Ns
		EndIf

		; Textures
		If BitAND( $aModel[7], 3 ) = 3 Then
			If $bDrawTextures Then
				glBindTexture( $GL_TEXTURE_2D, $aMaterialData[$i][7] ) ; map_Kd
			Else
				glBindTexture( $GL_TEXTURE_2D, 0 )
			EndIf
		EndIf

		; Draw model
		If Not $aModel[7] And $bDrawNormals Then
			glDrawElements( $bModelDrawMode, $aVertCnts[$i], $GL_UNSIGNED_SHORT, $aFaces[1] + $aVertStart[$i] * 2 )
		Else
			glDrawArrays( $bModelDrawMode, $aVertStart[$i], $aVertCnts[$i] )
		EndIf

	Next

EndFunc


Func DisplayFunction()

	If $bFocus And Not $bModelWin Then
		HandleKeystrokes( $aModel, False )
		If $bTransforms Then CalcTransforms()
	EndIf

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )

	glPushMatrix()

	DrawModel( $aModel )

	; Draw normal vectors
	If $nNormals Then
		If $bDrawLight Then _
			glDisable( $GL_LIGHTING )
		Local $aVertices = $aModel[0]
		Local $p1 = $aVertices[1], $p2 = $pNormals
		glColor3f( 0, 0, 1 ) ; Blue
		glBegin( $GL_LINES )
			For $i = 0 To $nNormals - 1
				DllCall( $dllOpenGL32, "none", "glVertex3fv", "ptr", $p1 )
				DllCall( $dllOpenGL32, "none", "glVertex3fv", "ptr", $p2 )
				$p1 += 12
				$p2 += 12
			Next
		glEnd()
		If $bDrawLight Then _
			glEnable( $GL_LIGHTING )
	EndIf

	glPopMatrix()

	SwapBuffers( $hDC )

EndFunc


Func CalcTransforms()

	Local $iSpeedFactor = 15
	Local $aTransforms = $aModel[12]

	If $iScaleFactor <> 0 Then
		$aTransforms[0] += $iScaleFactor * $iSpeedFactor * 0.00125
		$iScaleFactor = 0
	EndIf

	If $iRotXdirection <> 0 Then
		$aTransforms[1] += $iRotXdirection * $iSpeedFactor * 0.125
		$iRotXdirection = 0
	EndIf

	If $iRotYdirection <> 0 Then
		$aTransforms[2] += $iRotYdirection * $iSpeedFactor * 0.125
		$iRotYdirection = 0
	EndIf

	If $iRotZdirection <> 0 Then
		$aTransforms[3] += $iRotZdirection * $iSpeedFactor * 0.125
		$iRotZdirection = 0
	EndIf

	If $iTransXdirection <> 0 Then
		$aTransforms[4] += $iTransXdirection * $iSpeedFactor * 0.0025
		$iTransXdirection = 0
	EndIf

	If $iTransYdirection <> 0 Then
		$aTransforms[5] += $iTransYdirection * $iSpeedFactor * 0.0025
		$iTransYdirection = 0
	EndIf

	If $iTransZdirection <> 0 Then
		$aTransforms[6] += $iTransZdirection * $iSpeedFactor * 0.0025
		$iTransZdirection = 0
	EndIf

	$aModel[12] = $aTransforms
	$bTransforms = 0

EndFunc


Func DeletePreviousBuffers()
	Local $aBuffer                     ; Vertex coordinates
	For $i = 0 To 3                    ; Texture coordinates
		If IsArray( $aModel[$i] ) Then   ; Normal vectors
			$aBuffer = $aModel[$i]         ; Face triangles
			$aBuffer[1] = 0 ; Delete pointer
			$aBuffer[0] = 0 ; Delete structure
			$aModel[$i] = 0 ; Delete array
		EndIf
	Next
	$aModel[4] = 0 ; Delete $aVertInfo
	$aModel[5] = 0 ; Delete $aMaterials
	$aModel[6] = 0 ; Delete $aMaterialInfo
EndFunc


Func DeletePreviousMaterials()
	; Delete textures
	If $aModel[13] > 0 Then
		Local $iTexIdLast = 0
		For $obj = 0 To $aModel[13] - 1
			If $aMaterialData[$obj][6] > $iTexIdLast Then
				$iTexIdLast = $aMaterialData[$obj][6]
				glDeleteTextures( 1, $iTexIdLast )
			EndIf
			If $aMaterialData[$obj][7] > $iTexIdLast Then
				$iTexIdLast = $aMaterialData[$obj][7]
				glDeleteTextures( 1, $iTexIdLast )
			EndIf
			If $aMaterialData[$obj][8] > $iTexIdLast Then
				$iTexIdLast = $aMaterialData[$obj][8]
				glDeleteTextures( 1, $iTexIdLast )
			EndIf
		Next
	EndIf
EndFunc


Func WM_ACTIVATE( $hWnd, $Msg, $wParam, $lParam )
	Switch $hWnd
		Case $hGui
			If $wParam Then
				$bFocus = True
			Else
				$bFocus = False
			EndIf
	EndSwitch
	Return $GUI_RUNDEFMSG
EndFunc


Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth - $iMenuBarHeight - $iStatusBarHeight
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	$bResize = True
EndFunc


Func GUI_EVENT_CLOSE()
	$bClose = True
EndFunc


Func CloseProgram()
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
