#include <GUIConstantsEx.au3>
#include <StaticConstants.au3>
#include <WindowsConstants.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

Global $bOpenGL21 = True
Global $sObjModel = "Small\Art\Greek vase\greek_vase.obj"

; --- OpenGL globals ---

Global $aModel[14]
Global $aMaterialData[1][9]
Global $iProgram, $aPrograms[3], $iAttributeVertexXYZ[3], $iAttributeTextureUV[3], $iAttributeNormalXYZ[3]
Global $iUniformM[3], $iUniformV[3], $iUniformP[3], $iUniformM3x3InvTransp[3], $iUniformTexture[3], $iTexWhite
Global $mModelMatrix, $mViewMatrix, $mProjectionMatrix

Global $axisX[3] = [ 1.0, 0.0, 0.0 ], $axisY[3] = [ 0.0, 1.0, 0.0 ], $axisZ[3] = [ 0.0, 0.0, 1.0 ]

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 800, $iHeight = 600

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30
Global Const $iMenuBarHeight   = 20
Global Const $iStatusBarHeight = 28

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight + $iMenuBarHeight + $iStatusBarHeight

Global $hGui, $idInput, $bResize = True, $bClose = False, $bFocus = True

; --- Project includes ---

#include "utilities\models.au3"
#include "utilities\includes.au3"
#include "utilities\utilities.au3"

; --- Program code ---

MainFunc()

While 1
	; Load a new model
	If $sObjModel <> $sNewModel Then
		If $bOpen Then
			; A model is opened through the "Open File Dialog"
			; Let the OpenGL window update once before the model is loaded
			; (To refresh the part of the OpenGL window where the "Open File Dialog" was displayed)
			$bOpen = 0
		Else
			Local $aNewModel[14]
			If $sNewModel = "" Then $sNewModel = $sObjModel
			If CreateResources( $sNewModel, $aNewModel ) Then
				; If the new model is loaded without errors then make it the current model
				DeletePreviousBuffers()
				$sObjModel = $sNewModel
				$aModel = $aNewModel
			Else
				; Continue with the old model in case of errors
				$sNewModel = $sObjModel
			EndIf
		EndIf
	EndIf

	; Menu commands that affect OpenGL should not
	; be performed in the middle of drawing cycle.
	If $bMenuCmd Then
		$bMenuCmd = False
		Switch $iMenuId

			; --- Models menu ---

			Case $idModelsMtls To $idModelsMtlsMax
				If $sMtlFileNew <> $sMtlFile Then
					DeletePreviousMaterials()
					Local $s = $sMtlFileNew, $aMtlInfo
					Local $j = StringInStr( $s, "\", 0, -1 )
					ReadMaterials( StringLeft( $s, $j ), StringRight( $s, StringLen( $s ) - $j ), $aMtlInfo )
					$aModel[6] = $aMtlInfo
					LoadMaterials( $aModel, $aMaterialData )
					LoadTextures( $aModel, $aMaterialData )
					$sMtlFile = $sMtlFileNew
					$aModel[6] = StringRight( $sMtlFile, StringLen( $sMtlFile ) - StringInStr( $sMtlFile, "\", 0, -1 ) )
				EndIf

			; --- View menu ---

			Case $idViewWire
				If $bWire Then
					glPolygonMode( $GL_FRONT_AND_BACK, $GL_LINE )
				Else
					glPolygonMode( $GL_FRONT_AND_BACK, $GL_FILL )
				EndIf

			; --- Options menu ---

			; Textures
			Case $idOptionsTextures
				SetTextures()
				$bDrawTextures = $bModelTextures
				If Not $bDrawTextures And _                                 ; Textures are turned off
					 ( BitAND( $aModel[7], 4 ) Or $bModelNormals = -1 ) And _ ; Normal vectors are loaded or are not loaded
					 $bCurrentShader = 3 Then                                 ; Texture-Coords shaders can't handle light (normal vectors)
						 ; When textures are turned off it is assumed that you want to see the model with light instead of
						 ; textures. To see the model with ligth we need normal vectors and a set of shaders to handle light.
						 If $bModelNormals = -1 Then
							 ; Normal vectors are not loaded
							 If ReloadModel( $aModel, 0 ) Then
								 SetNormals( 0 )
								 $bDrawNormals = $bModelNormals
								 SetShader( 1 )
								 $bCurrentShader = $bModelShader
								 $iProgram = $bCurrentShader - 1
								 glUseProgram( $aPrograms[$iProgram] )
							 Else
								 ; ReloadModel cancelled
								 ; Undo texture change
								 SetTextures()
								 $bDrawTextures = $bModelTextures
							 EndIf
						 Else
							 ; Normal vectors are loaded
							 SetShader( 1 )
							 $bCurrentShader = $bModelShader
							 $iProgram = $bCurrentShader - 1
							 glUseProgram( $aPrograms[$iProgram] )
						 EndIf
				EndIf

			; Shaders
			Case $idBlinnPhong, $idPerpixelLight
				If $bModelNormals = -1 Then ; Normal vectors are not loaded
					; For Blinn-Phong and Perpixel-Light shaders normal vectors are needed
					Local $iMenuId2 = $iMenuId
					If ReloadModel( $aModel, 0 ) Then
						SetNormals( 0 )
						$bDrawNormals = $bModelNormals
						SetShader( $iMenuId2 - $idBlinnPhong + 1 )
						$bCurrentShader = $bModelShader
						$iProgram = $bCurrentShader - 1
						glUseProgram( $aPrograms[$iProgram] )
					EndIf
				Else ; Normal vectors are loaded
					SetShader( $iMenuId - $idBlinnPhong + 1 )
					$bCurrentShader = $bModelShader
					$iProgram = $bCurrentShader - 1
					glUseProgram( $aPrograms[$iProgram] )
				EndIf
			Case $idTextureCoords
				SetShader( $iMenuId - $idBlinnPhong + 1 )
				$bCurrentShader = $bModelShader
				$iProgram = $bCurrentShader - 1
				glUseProgram( $aPrograms[$iProgram] )

			; Normals
			Case $idCalculateNormals, $idAverageNormals
				; Calculate normal vectors with or without average
				; When settings are switched normal vectors must be recalculated
				SetNormals( $bModelNormals )
				If ReloadModel( $aModel, $bModelNormals ) Then
					$bDrawNormals = $bModelNormals
				Else
					SetNormals( $bModelNormals )
				EndIf
			Case $idNoNormals
				; This decides whether to load normal vectors or not
				; When settings are switched normal vectors must be recalculated
				SetNormals( -$bModelNormals - 1 )
				If ReloadModel( $aModel, $bModelNormals ) Then
					$bDrawNormals = $bModelNormals
					If $bModelNormals = -1 And $bCurrentShader <> 3 Then
						SetShader( 3 )
						$bCurrentShader = $bModelShader
						$iProgram = $bCurrentShader - 1
						glUseProgram( $aPrograms[$iProgram] )
					EndIf
				Else
					SetNormals( -$bModelNormals - 1 )
				EndIf

			; Quads
			Case $idOptionsQuads
				SetQuads()
				If ReloadModel( $aModel, $bModelNormals ) Then
					If $bModelQuads Then
						$bModelDrawMode = $GL_QUADS
					Else
						$bModelDrawMode = $GL_TRIANGLES
					EndIf
				Else
					; ReloadModel cancelled
					; Undo draw mode change
					SetQuads()
				EndIf

		EndSwitch
	EndIf

	If $bResize Then
		glViewport( 0, 0, $iWidth, $iHeight )
		$mProjectionMatrix = glmPerspective( 45, $iWidth / $iHeight, 0.5, 100.0 )
		$bResize = False
	EndIf

	If $bClose Then
		CloseProgram()
	EndIf

	; OpenGL drawing cycle
	DisplayFunction()

	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then CloseProgram()
	If Not InitOpenGLwindow() Then CloseProgram()
	If Not CheckOpenGLversion( 2, 1 ) Then CloseProgram()
	If Not CreateResources( $sObjModel, $aModel ) Then CloseProgram()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 50, $WS_OVERLAPPEDWINDOW )

	; Menu system
	CreateMenuSystem()

	; Statusbar
	Local $idStatus, $idLabel
	$idInput = GUICtrlCreateInput( "", 0, 0, 1, 1, 2048 ) ; 2048 = $ES_READONLY
	$idStatus = GUICtrlCreateLabel( "", 0, $iHeight, $iGuiWidth + 2, $iStatusBarHeight+2, $SS_ETCHEDFRAME )
	$idLabel = GUICtrlCreateLabel( "", 3, $iHeight + 6, $iGuiWidth - 4, $iStatusBarHeight - 10 )
	GUICtrlSetData( $idLabel, "Rotate: Q/A, W/S, E/D  |  Move: U/J, I/K, O/L  |  Zoom: Page up/down  |  Reset: Home   (Close model list to use shortcuts)" )
	GUICtrlSetResizing( $idStatus, $GUI_DOCKSTATEBAR )
	GUICtrlSetResizing( $idLabel, $GUI_DOCKSTATEBAR )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	GUIRegisterMsg( $WM_ACTIVATE, "WM_ACTIVATE" )
	GUIRegisterMsg( $WM_COMMAND,  "WM_COMMAND" )

	GUISetState()

	Return 1

EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 0.75, 0.75, 0.75, 1.0 )

	glEnable( $GL_TEXTURE_2D )
	glEnable( $GL_DEPTH_TEST )
	;glEnable( $GL_CULL_FACE )
	glEnable( $GL_LIGHTING )
	glEnable( $GL_LIGHT0 )

	glEnable( $GL_BLEND )
	glBlendFunc( $GL_SRC_ALPHA, $GL_ONE_MINUS_SRC_ALPHA )

	glPolygonMode( $GL_FRONT_AND_BACK, $GL_FILL )
	;glPolygonMode( $GL_FRONT_AND_BACK, $GL_LINE )

	;glShadeModel( $GL_SMOOTH )

	GUISwitch( $hGui )

	Return 1

EndFunc


Func CreateResources( $sObjModel, ByRef $aModel )

	; Create programs and shaders
	; Only necessary to do this once
	If Not $aPrograms[0] Then
		If Not CreateShaders() Then Return 0
	EndIf

	; Check file format and file existence of the OBJ-model
	; Set $aModel[11] = full path to the OBJ-file WITHOUT the .obj suffix
	If Not CheckModelFile( $sObjModel, $aModel ) Then Return 0

	; Save previous settings
	; This makes it possible to restore settings if load of a new model is cancelled
	$bModelShaderPrev   = $bModelShader
	$bModelTexturesPrev = $bModelTextures
	$bModelNormalsPrev  = $bModelNormals
	$bModelQuadsPrev    = $bModelQuads

	; Get model settings from inifile
	; If no inifile set defaults
	; Store settings in $aModel[10]
	GetModelSettings( $aModel, True )
	Local $aSettings = $aModel[10]
	$bModelShader   = $aSettings[1]
	$bModelTextures = $aSettings[2]
	$bModelNormals  = $aSettings[3]
	$bModelQuads    = $aSettings[4]

	; Load model
	; Parameters for model loading functions:
	; LoadObjModel, LoadObjModelBins, LoadObjQuads, LoadObjQuadsBins
	; LoadObjModel( ByRef $aModel, $bOpenGL21 = True, $bNormals = 0 )
	; $bOpenGL21 is used to handle the OpenGL version
	;   Set $bOpenGL21 = True to use the model for OpenGL version 2.1 or higher
	;   Set $bOpenGL21 = False to use the model for version 2.0 or lower
	; $bNormals is used to handle normal vectors. Values: 0, 1, -1
	; $bNormals = 0/1 is for models which does not contain any normal vectors
	;   Set $bNormals = 0 to calculate normal vectors (no average)
	;   Set $bNormals = 1 to calculate average normal vectors
	; $bNormals = -1 is for models which does contain normal vectors
	;   Set $bNormals = -1 to exclude normal vectors when they are not needed

	Local $bLoadBinaries = False

	If Not $bModelQuads Then

		If FileExists( $aModel[11] & "V.bin" ) _
		Or FileExists( $aModel[11] & "VCal.bin" ) _
		Or FileExists( $aModel[11] & "VAvg.bin" ) Then
			If LoadObjModelBins( $aModel, $bOpenGL21, $bModelNormals ) Then _
				$bLoadBinaries = True
		EndIf

		If Not $bLoadBinaries Then
			If Not LoadObjModel( $aModel, $bOpenGL21, $bModelNormals ) Then
				$bModelShader   = $bModelShaderPrev
				$bModelTextures = $bModelTexturesPrev
				$bModelNormals  = $bModelNormalsPrev
				$bModelQuads    = $bModelQuadsPrev
				Return 0
			EndIf
		EndIf

	Else

		; Use LoadObjQuads/LoadObjQuadsBins to force a model to be loaded as a quad mesh and rendered with $GL_QUADS draw mode

		If FileExists( $aModel[11] & "V4.bin" ) _
		Or FileExists( $aModel[11] & "VCal4.bin" ) _
		Or FileExists( $aModel[11] & "VAvg4.bin" ) Then
			If LoadObjQuadsBins( $aModel, $bOpenGL21, $bModelNormals ) Then _
				$bLoadBinaries = True
		EndIf

		If Not $bLoadBinaries Then
			If Not LoadObjQuads( $aModel, $bOpenGL21, $bModelNormals ) Then
				$bModelShader   = $bModelShaderPrev
				$bModelTextures = $bModelTexturesPrev
				$bModelNormals  = $bModelNormalsPrev
				$bModelQuads    = $bModelQuadsPrev
				Return 0
			EndIf
		EndIf

	EndIf

	DeletePreviousMaterials()

	Local $iObjects = $aModel[13], $w, $h
	Redim $aMaterialData[$iObjects][9]

	; Load materials
	; Always load materials, if no material information then load default materials
	LoadMaterials( $aModel, $aMaterialData )

	; Load texture maps
	; Always load texture maps, if no texture maps then load default texture map ids
	LoadTextures( $aModel, $aMaterialData )

	; Load default white texture
	; (To prevent errors in shaders for models that have no textures)
	LoadTextureImage( $iTexWhite, $models & "\White.jpg", $w, $h )

	; Some models are provided with several different material files which can be changed
	; dynamically without the need to reload the entire model. Get the names of these files
	; and add the names to the Model menu.
	$sMtlFile = ""
	Local $aMtlInfo
	If IsArray( $aModel[6] ) Then
		$aMtlInfo = $aModel[6]
		$sMtlFile = $aMtlInfo[0][0]
	EndIf
	SetModelsMenuMaterials( $aModel[11], $sMtlFile )

	; $aModel[6] = $aMaterialInfo is used to load materials and texture maps, not needed any more
	; Instead of $aModel[6] will be used to store the name of the material file
	If IsArray( $aModel[6] ) Then $aModel[6] = $sMtlFile

	; Program
	$iProgram = 0                          ; Blinn-Phong shaders (default)
	If $bModelShader Then
		$iProgram = $bModelShader - 1
	Else
		If $aModel[7] = 0 Then $iProgram = 1 ; Perpixel-Light shaders
		If $aModel[7] = 3 Then $iProgram = 2 ; Texture-Coords shaders
	EndIf
	$bModelShader = $iProgram + 1
	glUseProgram( $aPrograms[$iProgram] )

	; Draw mode
	If $bModelQuads Then
		$bModelDrawMode = $GL_QUADS
	Else
		$bModelDrawMode = $GL_TRIANGLES
	EndIf

	; Set state op Option menu items
	SetOptionsMenuStates( $aModel )

	; Read transformations
	$mModelMatrix = $IDENTITY_MATRIX
	ReadTransforms( $aModel[11], $mModelMatrix )

	; Apply transformations
	ResetModel( $aModel )

	; Initialize view
	InitializeView()

	Return 1

EndFunc


Func ReloadModel( ByRef $aModel, $bNormals )
	Local $aNewModel[14], $aSettings[5], $bOK
	$aNewModel[10] = $aSettings
	$aNewModel[11] = $aModel[11]
	Local $bLoadBinaries = False
	If Not $bModelQuads Then
		If FileExists( $aNewModel[11] & "V.bin" ) _
		Or FileExists( $aNewModel[11] & "VCal.bin" ) _
		Or FileExists( $aNewModel[11] & "VAvg.bin" ) Then _
			$bOK = LoadObjModelBins( $aNewModel, $bOpenGL21, $bNormals )
		If Not $bOK Then _
			$bOK = LoadObjModel( $aNewModel, $bOpenGL21, $bNormals, True )
	Else
		If FileExists( $aNewModel[11] & "V4.bin" ) _
		Or FileExists( $aNewModel[11] & "VCal4.bin" ) _
		Or FileExists( $aNewModel[11] & "VAvg4.bin" ) Then _
			$bOK = LoadObjQuadsBins( $aNewModel, $bOpenGL21, $bNormals )
		If Not $bOK Then _
			$bOK = LoadObjQuads( $aNewModel, $bOpenGL21, $bNormals, True )
	EndIf
	If $bOK Then
		For $i = 0 To 3                         ; Vertex coordinates
			; Delete old data                     ; Texture coordinates
			If $aModel[$i] Then _                 ; Normal vectors
				glDeleteBuffers( 1, $aModel[$i] )   ; Face triangles
			; Add new data
			$aModel[$i] = $aNewModel[$i]
		Next
		$aModel[4] = $aNewModel[4]              ; $aVertInfo
	EndIf
	Return $bOK
EndFunc


Func CreateShaders()

	; Three programs $aPrograms[0], $aPrograms[1] and $aPrograms[2] are used for three set of shaders.


	; The default shaders are the Blinn-Phong shaders for a single directional light source.

	; $aPrograms[0]
	Local $iVs, $iFs
	If Not LoadGLSLshader( $iVs, "shaders\Blinn-Phong.v.glsl", $GL_VERTEX_SHADER ) Then Return 0
	If Not LoadGLSLshader( $iFs, "shaders\Blinn-Phong.f.glsl", $GL_FRAGMENT_SHADER ) Then Return 0
	If Not CreateProgram( $aPrograms[0], $iVs, $iFs ) Then Return 0

	; Get handles for attributes
	If Not GetAttribLocation( $aPrograms[0], $iAttributeVertexXYZ[0], "vertexXYZ" ) Then Return 0
	If Not GetAttribLocation( $aPrograms[0], $iAttributeTextureUV[0], "textureUV" ) Then Return 0
	If Not GetAttribLocation( $aPrograms[0], $iAttributeNormalXYZ[0], "normalXYZ" ) Then Return 0

	; Get handles for uniforms
	If Not GetUniformLocation( $aPrograms[0], $iUniformM[0], "m" ) Then Return 0
	If Not GetUniformLocation( $aPrograms[0], $iUniformV[0], "v" ) Then Return 0
	If Not GetUniformLocation( $aPrograms[0], $iUniformP[0], "p" ) Then Return 0
	If Not GetUniformLocation( $aPrograms[0], $iUniformM3x3InvTransp[0], "m_3x3_inv_transp" ) Then Return 0
	If Not GetUniformLocation( $aPrograms[0], $iUniformTexture[0], "colorMap" ) Then Return 0


	; If the model type is "f v1 v2 v3 ..." ($aModel[7] = 0) then use the Perpixel-Light shaders.
	; The model has no texture coordinates and no normal vectors. In this case normal vectors are
	; calculated by LoadObjModel.

	; $aPrograms[1]
	Local $iVs, $iFs
	If Not LoadGLSLshader( $iVs, "shaders\Perpixel-Light.v.glsl", $GL_VERTEX_SHADER ) Then Return 0
	If Not LoadGLSLshader( $iFs, "shaders\Perpixel-Light.f.glsl", $GL_FRAGMENT_SHADER ) Then Return 0
	If Not CreateProgram( $aPrograms[1], $iVs, $iFs ) Then Return 0

	; Get handles for attributes
	If Not GetAttribLocation( $aPrograms[1], $iAttributeVertexXYZ[1], "vertexXYZ" ) Then Return 0
	If Not GetAttribLocation( $aPrograms[1], $iAttributeTextureUV[1], "textureUV" ) Then Return 0
	If Not GetAttribLocation( $aPrograms[1], $iAttributeNormalXYZ[1], "normalXYZ" ) Then Return 0

	; Get handles for uniforms
	If Not GetUniformLocation( $aPrograms[1], $iUniformM[1], "m" ) Then Return 0
	If Not GetUniformLocation( $aPrograms[1], $iUniformV[1], "v" ) Then Return 0
	If Not GetUniformLocation( $aPrograms[1], $iUniformP[1], "p" ) Then Return 0
	If Not GetUniformLocation( $aPrograms[1], $iUniformM3x3InvTransp[1], "m_3x3_inv_transp" ) Then Return 0
	If Not GetUniformLocation( $aPrograms[1], $iUniformTexture[1], "colorMap" ) Then Return 0


	; If the model type is "f v1/vt1 v2/vt2 v3/vt3 ..." ($aModel[7] = 3) then use the Texture-Coords shaders.
	; There are no normal vectors, and light can't be used if there are no normal vectors.

	; $aPrograms[2]
	Local $iVs, $iFs
	If Not LoadGLSLshader( $iVs, "shaders\Texture-Coords.v.glsl", $GL_VERTEX_SHADER ) Then Return 0
	If Not LoadGLSLshader( $iFs, "shaders\Texture-Coords.f.glsl", $GL_FRAGMENT_SHADER ) Then Return 0
	If Not CreateProgram( $aPrograms[2], $iVs, $iFs ) Then Return 0

	; Get handles for attributes
	If Not GetAttribLocation( $aPrograms[2], $iAttributeVertexXYZ[2], "vertexXYZ" ) Then Return 0
	If Not GetAttribLocation( $aPrograms[2], $iAttributeTextureUV[2], "textureUV" ) Then Return 0
	;If Not GetAttribLocation( $aPrograms[2], $iAttributeNormalXYZ[2], "normalXYZ" ) Then Return 0

	; Get handles for uniforms
	If Not GetUniformLocation( $aPrograms[2], $iUniformM[2], "m" ) Then Return 0
	If Not GetUniformLocation( $aPrograms[2], $iUniformV[2], "v" ) Then Return 0
	If Not GetUniformLocation( $aPrograms[2], $iUniformP[2], "p" ) Then Return 0
	;If Not GetUniformLocation( $aPrograms[2], $iUniformM3x3InvTransp[2], "m_3x3_inv_transp" ) Then Return 0
	If Not GetUniformLocation( $aPrograms[2], $iUniformTexture[2], "colorMap" ) Then Return 0

	Return 1

EndFunc


Func ResetModel( ByRef $aModel )

	; $mModelMatrix
	$aModel[12] = $mModelMatrix

EndFunc


Func InitializeView()

	; Look at view matrix
	Local $eye[3] = [ 0.0, 0.0, 2.0 ], $center[3] = [ 0.0, 0.0, 0.0 ], $up[3] = [ 0.0, 1.0, 0.0 ]
	$mViewMatrix = glmLookAt( $eye, $center, $up )

EndFunc


Func DrawModel( ByRef $aModel )

	Local $aVertInfo  = $aModel[4]
	Local $mMatrix    = $aModel[12]
	Local $iObjects   = $aModel[13]
	Local $aVertStart = $aVertInfo[0]
	Local $aVertCnts  = $aVertInfo[1]

	; Vertices
	glEnableVertexAttribArray( $iAttributeVertexXYZ[$iProgram] )
	glBindBuffer( $GL_ARRAY_BUFFER, $aModel[0] )
	glVertexAttribPointer(  _
		$iAttributeVertexXYZ[$iProgram], _ ; attribute
		3,                               _ ; number of elements per vertex, here (x,y,z)
		$GL_FLOAT,                       _ ; the type of each element
		$GL_FALSE,                       _ ; take our values as-is
		0,                               _ ; no extra data between each position
		0 )                                ; offset of first element

	; Texture coordinates
	If BitAND( $aModel[7], 2 ) Then
		glEnableVertexAttribArray( $iAttributeTextureUV[$iProgram] )
		glBindBuffer( $GL_ARRAY_BUFFER, $aModel[1] )
		glVertexAttribPointer(  _
			$iAttributeTextureUV[$iProgram], _ ; attribute
			2,                               _ ; number of elements per vertex, here (u,v)
			$GL_FLOAT,                       _ ; the type of each element
			$GL_FALSE,                       _ ; take our values as-is
			0,                               _ ; no extra data between each position
			0 )                                ; offset of first element
	EndIf

	; Normal vectors
	If $iProgram <> 2 Then
		glEnableVertexAttribArray( $iAttributeNormalXYZ[$iProgram] )
		glBindBuffer( $GL_ARRAY_BUFFER, $aModel[2] )
		glVertexAttribPointer(  _
			$iAttributeNormalXYZ[$iProgram], _ ; attribute
			3,                               _ ; number of elements per vertex, here (x,y,z)
			$GL_FLOAT,                       _ ; the type of each element
			$GL_FALSE,                       _ ; take our values as-is
			0,                               _ ; no extra data between each position
			0 )                                ; offset of first element
	EndIf

	; Apply transformation matrix
	glUniformMatrix4fv( $iUniformM[$iProgram], 1, $GL_FALSE, $mMatrix )

	If $iProgram <> 2 Then
		; Transform normal vectors with transpose of inverse of upper left 3x3 model matrix
		Local $mMat3[9] = [ $mMatrix[0], $mMatrix[1], $mMatrix[2], _
		                    $mMatrix[4], $mMatrix[5], $mMatrix[6], _
		                    $mMatrix[8], $mMatrix[9], $mMatrix[10] ]
		Local $mInv = glmInverseMat3( $mMat3 )
		Local $mTransp = glmTransposeMat3( $mInv )
		glUniformMatrix3fv( $iUniformM3x3InvTransp[$iProgram], 1, $GL_FALSE, $mTransp )
	EndIf

	; Texture
	glActiveTexture( $GL_TEXTURE0 )
	glUniform1i( $iUniformTexture[$iProgram], 0 ) 

	; Draw model
	For $i = 0 To $iObjects - 1

		; Materials
		glMaterialfv( $GL_FRONT_AND_BACK, $GL_AMBIENT,   $aMaterialData[$i][1] ) ; Ka
		glMaterialfv( $GL_FRONT_AND_BACK, $GL_DIFFUSE,   $aMaterialData[$i][2] ) ; Kd
		glMaterialfv( $GL_FRONT_AND_BACK, $GL_SPECULAR,  $aMaterialData[$i][3] ) ; Ks
		glMaterialf(  $GL_FRONT_AND_BACK, $GL_SHININESS, $aMaterialData[$i][4] ) ; Ns

		; Textures
		If BitAND( $aModel[7], 3 ) = 3 Then
			If $bDrawTextures Then
				glBindTexture( $GL_TEXTURE_2D, $aMaterialData[$i][7] ) ; map_Kd
			Else
				glBindTexture( $GL_TEXTURE_2D, $iTexWhite )
				;glBindTexture( $GL_TEXTURE_2D, 0 )
			EndIf
		EndIf

		; Draw model
		If Not $aModel[7] And $bDrawNormals Then
			glBindBuffer( $GL_ELEMENT_ARRAY_BUFFER, $aModel[3] )
			glDrawElements( $bModelDrawMode, $aVertCnts[$i], $GL_UNSIGNED_SHORT, $aVertStart[$i] * 2 )
		Else
			glDrawArrays( $bModelDrawMode, $aVertStart[$i], $aVertCnts[$i] )
		EndIf

	Next

	glDisableVertexAttribArray( $iAttributeVertexXYZ[$iProgram] )
	glDisableVertexAttribArray( $iAttributeTextureUV[$iProgram] )
	If $iProgram <> 2 Then _
		glDisableVertexAttribArray( $iAttributeNormalXYZ[$iProgram] )

EndFunc


Func DisplayFunction()

	If $bFocus And Not $bModelWin Then
		HandleKeystrokes( $aModel, True )
		If $bTransforms Then CalcTransforms()
	EndIf

	glUniformMatrix4fv( $iUniformV[$iProgram], 1, $GL_FALSE, $mViewMatrix )
	glUniformMatrix4fv( $iUniformP[$iProgram], 1, $GL_FALSE, $mProjectionMatrix )

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )

	DrawModel( $aModel )

	SwapBuffers( $hDC )

EndFunc


Func CalcTransforms()

	Local $iSpeedFactor = 15

	If $iScaleFactor <> 0 Then
		Local $fDeltaScale = 1.0 + $iScaleFactor * $iSpeedFactor * 0.00125
		Local $scale[3] = [ $fDeltaScale, $fDeltaScale, $fDeltaScale ]
		$aModel[12] = glmScale( $aModel[12], $scale )
		$iScaleFactor = 0
	EndIf

	If $iRotXdirection <> 0 Then
		Local $fDeltaRotX = $iRotXdirection * $iSpeedFactor * 0.125
		$aModel[12] = glmRotate( $aModel[12], $fDeltaRotX, $axisX )
		$iRotXdirection = 0
	EndIf

	If $iRotYdirection <> 0 Then
		Local $fDeltaRotY = $iRotYdirection * $iSpeedFactor * 0.125
		$aModel[12] = glmRotate( $aModel[12], $fDeltaRotY, $axisY )
		$iRotYdirection = 0
	EndIf

	If $iRotZdirection <> 0 Then
		Local $fDeltaRotZ = $iRotZdirection * $iSpeedFactor * 0.125
		$aModel[12] = glmRotate( $aModel[12], $fDeltaRotZ, $axisZ )
		$iRotZdirection = 0
	EndIf

	If $iTransXdirection <> 0 Then
		Local $fDeltaTransX = $iTransXdirection * $iSpeedFactor * 0.0025
		Local $trans[3] = [ $fDeltaTransX, 0, 0 ]
		$aModel[12] = glmTranslate( $aModel[12], $trans )
		$iTransXdirection = 0
	EndIf

	If $iTransYdirection <> 0 Then
		Local $fDeltaTransY = $iTransYdirection * $iSpeedFactor * 0.0025
		Local $trans[3] = [ 0, $fDeltaTransY, 0 ]
		$aModel[12] = glmTranslate( $aModel[12], $trans )
		$iTransYdirection = 0
	EndIf

	If $iTransZdirection <> 0 Then
		Local $fDeltaTransZ = $iTransZdirection * $iSpeedFactor * 0.0025
		Local $trans[3] = [ 0, 0, $fDeltaTransZ ]
		$aModel[12] = glmTranslate( $aModel[12], $trans )
		$iTransZdirection = 0
	EndIf

	$bTransforms = 0

EndFunc


Func DeletePreviousBuffers()
	For $i = 0 To 3                         ; Vertex coordinates
		If $aModel[$i] Then _                 ; Texture coordinates
			glDeleteBuffers( 1, $aModel[$i] )   ; Normal vectors
	Next                                    ; Face triangles
	$aModel[4] = 0 ; Delete $aVertCnts
	$aModel[5] = 0 ; Delete $aMaterials
	$aModel[6] = 0 ; Delete $aMaterialInfo
EndFunc


Func DeletePreviousMaterials()
	; Delete textures
	If $aModel[13] > 0 Then
		Local $iTexIdLast = 0
		For $obj = 0 To $aModel[13] - 1
			If $aMaterialData[$obj][6] > $iTexIdLast Then
				$iTexIdLast = $aMaterialData[$obj][6]
				glDeleteTextures( 1, $iTexIdLast )
			EndIf
			If $aMaterialData[$obj][7] > $iTexIdLast Then
				$iTexIdLast = $aMaterialData[$obj][7]
				glDeleteTextures( 1, $iTexIdLast )
			EndIf
			If $aMaterialData[$obj][8] > $iTexIdLast Then
				$iTexIdLast = $aMaterialData[$obj][8]
				glDeleteTextures( 1, $iTexIdLast )
			EndIf
		Next
	EndIf
EndFunc


Func WM_ACTIVATE( $hWnd, $Msg, $wParam, $lParam )
	Switch $hWnd
		Case $hGui
			If $wParam Then
				$bFocus = True
			Else
				$bFocus = False
			EndIf
	EndSwitch
	Return $GUI_RUNDEFMSG
EndFunc


Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth - $iMenuBarHeight - $iStatusBarHeight
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	$bResize = True
EndFunc


Func GUI_EVENT_CLOSE()
	$bClose = True
EndFunc


Func CloseProgram()
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
