#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

Global $sObjModel = "Small\Aircrafts\Mitsubishi A6M Zero\A6M_ZERO.obj"

; --- OpenGL globals ---

Global $aModel[14]
Global $aMaterialData[1][9]
Global $iTexture, $iCallList

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 800, $iHeight = 600

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight

Global $hGui, $bResize = True, $bClose = False

; --- Project includes ---

#include "utilities\models.au3"
#include "utilities\includes.au3"
#include "utilities\utilities.au3"

; --- Program code ---

MainFunc()

While 1
	If $bResize Then
		glViewport( 0, 0, $iWidth, $iHeight )
		InitializeView()
		$bResize = False
	EndIf
	If $bClose Then
		CloseProgram()
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then CloseProgram()
	If Not InitOpenGLwindow() Then CloseProgram()
	;If Not CheckOpenGLversion( 2, 1 ) Then CloseProgram()
	If Not CreateResources() Then CloseProgram()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 50, $WS_OVERLAPPEDWINDOW )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	Return 1

EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 0.75, 0.75, 0.75, 1.0 )

	glEnable( $GL_TEXTURE_2D )
	glEnable( $GL_DEPTH_TEST )

	glEnable( $GL_BLEND )
	glBlendFunc( $GL_SRC_ALPHA, $GL_ONE_MINUS_SRC_ALPHA )

	GUISwitch( $hGui )
	GUISetState()

	Return 1

EndFunc


Func CreateResources()

	Local $w, $h
	If Not LoadTextureImage( $iTexture, $models & "\Small\Aircrafts\Mitsubishi A6M Zero\clouds_large.jpg", $w, $h ) Then Return 0

	$iCallList = glGenLists( 1 )
	glNewList( $iCallList, $GL_COMPILE )
		glBindTexture( $GL_TEXTURE_2D, $iTexture )
		glBegin( $GL_QUADS )
			glTexCoord2f(  0.0,  0.0 )
			glVertex3f  ( -5.0, -5.0, -5.0 )
			glTexCoord2f(  1.0,  0.0 )
			glVertex3f  ( +5.0, -5.0, -5.0 )
			glTexCoord2f(  1.0,  1.0 )
			glVertex3f  ( +5.0, +5.0, -5.0 )
			glTexCoord2f(  0.0,  1.0 )
			glVertex3f  ( -5.0, +5.0, -5.0 )
		glEnd()
	glEndList()

	If Not CheckModelFile( $sObjModel, $aModel ) Then Return 0

	If Not LoadObjModelBins( $aModel, False, -1 ) Then Return 0

	Local $iObjects = $aModel[13]
	Redim $aMaterialData[$iObjects][9]

	LoadTextures( $aModel, $aMaterialData )

	Return 1

EndFunc


Func DrawModel()

	Local $aVertInfo  = $aModel[4]
	Local $aVertStart = $aVertInfo[0]
	Local $aVertCnts  = $aVertInfo[1]

	; Draw aircraft
	glBindTexture( $GL_TEXTURE_2D, $aMaterialData[0][7] )
	glDrawArrays( $GL_TRIANGLES, $aVertStart[0], $aVertCnts[0] )

	; Draw propeller
	glBindTexture( $GL_TEXTURE_2D, $aMaterialData[1][7] )
	glDrawArrays( $GL_TRIANGLES, $aVertStart[1], $aVertCnts[1] )

EndFunc


Func DisplayFunction()

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )

	; Draw clouds
	glCallList( $iCallList )

	glPushMatrix()

	glRotatef( 30, 1.0, 0.0, 0.0 )
	glRotatef( 45, 0.0, 1.0, 0.0 )
	glScalef( 0.75, 0.75, 0.75 )

	; First aircraft
	glTranslatef( -0.75, 0, 0 )
	DrawModel()

	; Second aircraft
	glTranslatef( +1.25, 0, 0 )
	DrawModel()

	; Third aircraft
	glTranslatef( +1.25, 0, 0 )
	DrawModel()

	glPopMatrix()

	SwapBuffers( $hDC )

EndFunc


Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	$bResize = True
EndFunc


Func GUI_EVENT_CLOSE()
	$bClose = True
EndFunc


Func CloseProgram()
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
