#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

Global $sObjPath = "Samples\Blender\"

; --- OpenGL globals ---

Global $aModel[17]

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 800, $iHeight = 600

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight

Global $hGui, $bResize = True, $bClose = False

; --- Project includes ---

#include "utilities\models.au3"
#include "utilities\includes.au3"
#include "utilities\utilities.au3"

; --- Program code ---

MainFunc()

While 1
	If $bResize Then
		glViewport( 0, 0, $iWidth, $iHeight )
		InitializeView()
		$bResize = False
	EndIf
	If $bClose Then
		CloseProgram()
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then CloseProgram()
	If Not InitOpenGLwindow() Then CloseProgram()
	;If Not CheckOpenGLversion( 2, 1 ) Then CloseProgram()
	If Not CreateResources() Then CloseProgram()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 50, $WS_OVERLAPPEDWINDOW )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	Return 1

EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 0.75, 0.75, 0.75, 1.0 )

	glEnable( $GL_DEPTH_TEST )
	glEnable( $GL_LIGHTING )
	glEnable( $GL_LIGHT0 )

	GUISwitch( $hGui )

	Return 1

EndFunc


Func CreateResources()

	Local $sObjModel = $sObjPath & "tree2.obj"
	If Not CheckModelFile( $sObjModel, $aModel ) Then Return 0

	If Not LoadObjModel( $aModel, False ) Then Return 0

	Local $iObjects = $aModel[13]
	Local $aMtlData[$iObjects][9]
	$sObjPath = $models & "\" & $sObjPath

	Local $iInstances = 3
	Local $aMaterialData[$iInstances]
	Local $aTransforms[$iInstances]

	; Instance #0
	Local $aTrans[7] = [ 0.8, 28.1250, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000 ]
	LoadMaterials( $aModel, $aMtlData )
	$aMaterialData[0] = $aMtlData
	$aTransforms[0] = $aTrans

	; Instance #1
	Local $aMtlInfo, $aTrans[7] = [ 0.8, 28.1250, 0.0000, 0.0000, -0.75, 0.0000, 0.0000 ]
	ReadMaterials( $sObjPath, "tree2-3.mtl", $aMtlInfo )
	$aModel[6] = $aMtlInfo
	LoadMaterials( $aModel, $aMtlData )
	$aMaterialData[1] = $aMtlData
	$aTransforms[1] = $aTrans

	; Instance #2
	Local $aMtlInfo, $aTrans[7] = [ 0.8, 28.1250, 0.0000, 0.0000, +0.75, 0.0000, 0.0000 ]
	ReadMaterials( $sObjPath, "tree2-5.mtl", $aMtlInfo )
	$aModel[6] = $aMtlInfo
	LoadMaterials( $aModel, $aMtlData )
	$aMaterialData[2] = $aMtlData
	$aTransforms[2] = $aTrans

	$aModel[14] = $iInstances
	$aModel[15] = $aMaterialData
	$aModel[16] = $aTransforms

	GUISetState()

	Return 1

EndFunc


Func DrawModel()

	Local $aVertInfo     = $aModel[4]
	Local $iInstances    = $aModel[14]
	Local $aMaterialData = $aModel[15]
	Local $aTransforms   = $aModel[16]
	Local $aVertStart    = $aVertInfo[0]
	Local $aVertCnts     = $aVertInfo[1]
	Local $aMtlData, $aTrans

	For $i = 0 To $iInstances - 1

		; Instance data
		$aMtlData = $aMaterialData[$i]
		$aTrans = $aTransforms[$i]

		glPushMatrix()

		; Model transformations
		glRotatef( $aTrans[1], 1.0, 0.0, 0.0 )
		glScalef( $aTrans[0], $aTrans[0], $aTrans[0] )
		glTranslatef( $aTrans[4], $aTrans[5], $aTrans[6] )

		; Tree top
		glMaterialfv( $GL_FRONT_AND_BACK, $GL_AMBIENT,   $aMtlData[0][1] ) ; Ka
		glMaterialfv( $GL_FRONT_AND_BACK, $GL_DIFFUSE,   $aMtlData[0][2] ) ; Kd
		glMaterialfv( $GL_FRONT_AND_BACK, $GL_SPECULAR,  $aMtlData[0][3] ) ; Ks
		glMaterialf(  $GL_FRONT_AND_BACK, $GL_SHININESS, $aMtlData[0][4] ) ; Ns
		glDrawArrays( $GL_TRIANGLES, $aVertStart[0], $aVertCnts[0] )

		; Tree trunk
		glMaterialfv( $GL_FRONT_AND_BACK, $GL_AMBIENT,   $aMtlData[1][1] ) ; Ka
		glMaterialfv( $GL_FRONT_AND_BACK, $GL_DIFFUSE,   $aMtlData[1][2] ) ; Kd
		glMaterialfv( $GL_FRONT_AND_BACK, $GL_SPECULAR,  $aMtlData[1][3] ) ; Ks
		glMaterialf(  $GL_FRONT_AND_BACK, $GL_SHININESS, $aMtlData[1][4] ) ; Ns
		glDrawArrays( $GL_TRIANGLES, $aVertStart[1], $aVertCnts[1] )

		glPopMatrix()

	Next

EndFunc


Func DisplayFunction()

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )

	DrawModel()

	SwapBuffers( $hDC )

EndFunc


Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	$bResize = True
EndFunc


Func GUI_EVENT_CLOSE()
	$bClose = True
EndFunc


Func CloseProgram()
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
