#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

Global $sObjModel = "Samples\Blender\tree.obj"

; --- OpenGL globals ---

Global $aModel[16]
Global $aMaterialData[1][9]

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 800, $iHeight = 600

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight

Global $hGui, $bResize = True, $bClose = False

; --- Project includes ---

#include "utilities\models.au3"
#include "utilities\includes.au3"
#include "utilities\utilities.au3"

; --- Program code ---

MainFunc()

While 1
	If $bResize Then
		glViewport( 0, 0, $iWidth, $iHeight )
		InitializeView()
		$bResize = False
	EndIf
	If $bClose Then
		CloseProgram()
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then CloseProgram()
	If Not InitOpenGLwindow() Then CloseProgram()
	;If Not CheckOpenGLversion( 2, 1 ) Then CloseProgram()
	If Not CreateResources() Then CloseProgram()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 50, $WS_OVERLAPPEDWINDOW )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	Return 1

EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 0.75, 0.75, 0.75, 1.0 )

	glEnable( $GL_DEPTH_TEST )
	glEnable( $GL_LIGHTING )
	glEnable( $GL_LIGHT0 )

	GUISwitch( $hGui )

	Return 1

EndFunc


Func CreateResources()

	If Not CheckModelFile( $sObjModel, $aModel ) Then Return 0

	If Not LoadObjModel( $aModel, False ) Then Return 0

	Local $iObjects = $aModel[13]
	Redim $aMaterialData[$iObjects][9]

	LoadMaterials( $aModel, $aMaterialData )

	Local $oInstances = 0
	Local $iInstances = 20
	Local $aTransforms[$iInstances]

	For $i = 0 To $iInstances - 1
		Local $aTrans[7] = [ 1.0000, 0.0000, 0.0000, 0.0000, -0.75 + ( 0.05 * $i ) * 0.25, 0.1 * $i, -1 - 0.5 * $i ]
		$aTransforms[$i] = $aTrans
	Next

	$oInstances = $iInstances
	$iInstances = $iInstances + 20
	ReDim $aTransforms[$iInstances]

	For $i = $oInstances To $iInstances - 1
		Local $aTrans[7] = [ 1.0000, 0.0000, 0.0000, 0.0000, 0.75 - ( 0.05 * ( $i - $oInstances ) ) * 0.25, 0.1 * ( $i - $oInstances ), -1 - 0.5 * ( $i - $oInstances ) ]
		$aTransforms[$i] = $aTrans
	Next

	$aModel[14] = $iInstances
	$aModel[15] = $aTransforms

	GUISetState()

	Return 1

EndFunc


Func DrawModel()

	Local $aVertInfo   = $aModel[4]
	Local $iInstances  = $aModel[14]
	Local $aTransforms = $aModel[15]
	Local $aVertStart  = $aVertInfo[0]
	Local $aVertCnts   = $aVertInfo[1]
	Local $aTrans

	For $i = 0 To $iInstances - 1

		; Instance data
		$aTrans = $aTransforms[$i]

		glPushMatrix()

		glTranslatef( $aTrans[4], $aTrans[5], $aTrans[6] )

		; Tree top
		glMaterialfv( $GL_FRONT_AND_BACK, $GL_DIFFUSE, $aMaterialData[0][2] ) ; Kd
		glDrawArrays( $GL_TRIANGLES, $aVertStart[0], $aVertCnts[0] )

		; Tree trunk
		glMaterialfv( $GL_FRONT_AND_BACK, $GL_DIFFUSE, $aMaterialData[1][2] ) ; Kd
		glDrawArrays( $GL_TRIANGLES, $aVertStart[1], $aVertCnts[1] )

		glPopMatrix()

	Next

EndFunc


Func DisplayFunction()

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )

	glDisable( $GL_LIGHTING )
	glBegin( $GL_QUADS )
		glColor3f( 0.3, 0.3, 0.3 )
		glVertex3f( -0.5, -0.5, -0.9 )
		glVertex3f( -0.35, 1.4, -10.5 )
		glVertex3f(  0.35, 1.4, -10.5 )
		glVertex3f(  0.5, -0.5, -0.9 )
	glEnd()
	glEnable( $GL_LIGHTING )

	DrawModel()

	SwapBuffers( $hDC )

EndFunc


Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	$bResize = True
EndFunc


Func GUI_EVENT_CLOSE()
	$bClose = True
EndFunc


Func CloseProgram()
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
