#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

Global $sObjModel = "Medium\Plants\Trees\Tree 01\tree.obj"

; --- OpenGL globals ---

Global $aModel[14]
Global $aMaterialData[1][9]

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 800, $iHeight = 600

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight

Global $hGui, $bResize = True, $bClose = False

; --- Project includes ---

#include "utilities\models.au3"
#include "utilities\includes.au3"
#include "utilities\utilities.au3"

; --- Program code ---

MainFunc()

While 1
	If $bResize Then
		glViewport( 0, 0, $iWidth, $iHeight )
		InitializeView()
		$bResize = False
	EndIf
	If $bClose Then
		CloseProgram()
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then CloseProgram()
	If Not InitOpenGLwindow() Then CloseProgram()
	;If Not CheckOpenGLversion( 2, 1 ) Then CloseProgram()
	If Not CreateResources() Then CloseProgram()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 50, $WS_OVERLAPPEDWINDOW )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	Return 1

EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 0.75, 0.75, 0.75, 1.0 )

	glEnable( $GL_TEXTURE_2D )
	glEnable( $GL_DEPTH_TEST )

	glEnable( $GL_BLEND )
	glBlendFunc( $GL_SRC_ALPHA, $GL_ONE_MINUS_SRC_ALPHA )

	GUISwitch( $hGui )
	GUISetState()

	Return 1

EndFunc


Func CreateResources()

	If Not CheckModelFile( $sObjModel, $aModel ) Then Return 0

	If Not LoadObjModelBins( $aModel, False ) Then Return 0

	Local $iObjects = $aModel[13]
	Redim $aMaterialData[$iObjects][9]

	LoadTextures( $aModel, $aMaterialData )

	Return 1

EndFunc


Func DrawModel()

	Local $aVertInfo  = $aModel[4]
	Local $iObjects   = $aModel[13]
	Local $aVertStart = $aVertInfo[0]
	Local $aVertCnts  = $aVertInfo[1]

	For $i = 0 To $iObjects - 1
		glBindTexture( $GL_TEXTURE_2D, $aMaterialData[$i][7] )
		glDrawArrays( $GL_TRIANGLES, $aVertStart[$i], $aVertCnts[$i] )
	Next

EndFunc


Func DisplayFunction()

	Local Static $i = 0

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )

	glPushMatrix()

	Select
		Case $i <= 360
			glScalef( 1.5, 1.5, 1.5 )
			glRotatef( $i, 0.0, 1.0, 0.0 )
		Case $i <= 1080
			glRotatef( -1.875, 0.0, 1.0, 0.0 )
			glTranslatef( 0.0, 0.3 - ( $i - 361 ) / 1000, 1.5 )
		Case $i <= 1800
			glRotatef( -1.875, 0.0, 1.0, 0.0 )
			glTranslatef( 0.0, -0.42 + ( $i - 1081 ) / 1000, 1.5 )
		Case $i <= 2650
			glRotatef( 84.375, 1.0, 0.0, 0.0 )
			glTranslatef( -0.375, -11.00 + ( $i - 1801 ) / 50, 0.15 )
			glScalef( 11.443, 11.443, 11.443 )
		Case $i <= 2700
			glRotatef( 84.375, 1.0, 0.0, 0.0 )
			glTranslatef( -0.375, 6.00, 0.15 )
			glScalef( 11.443, 11.443, 11.443 )
		Case $i <= 3550
			glRotatef( 84.375, 1.0, 0.0, 0.0 )
			glTranslatef( -0.375, 6.00 - ( $i - 2701 ) / 50, 0.15 )
			glScalef( 11.443, 11.443, 11.443 )
		Case Else
			$i = 0
	EndSelect

	DrawModel()

	glPopMatrix()

	SwapBuffers( $hDC )

	$i += 1

EndFunc


Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	$bResize = True
EndFunc


Func GUI_EVENT_CLOSE()
	$bClose = True
EndFunc


Func CloseProgram()
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
