#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

Global $sObjPath = "Medium\Plants\Trees\Tree 01\"
Global $sObjModel = $sObjPath & "tree.obj"

; --- OpenGL globals ---

Global $aModel[14]
Global $aMaterialData[1][9]

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 800, $iHeight = 600

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30
Global Const $iStatusBarHeight = 28

Global $hStatusbar, $idRadio1, $idRadio2, $idRadio3
Global $bMtlFile = 2, $bNewMtlFile = 2

Global $iGuiWidth	= $iWidth
Global $iGuiHeight = $iHeight + $iStatusBarHeight

Global $hGui, $bResize = True, $bClose = False

; --- Project includes ---

#include "utilities\models.au3"
#include "utilities\includes.au3"
#include "utilities\utilities.au3"

; --- Program code ---

MainFunc()

While 1
	If $bNewMtlFile <> $bMtlFile Then
		Local $aMtlInfo
		DeletePreviousMaterials()
		Switch $bNewMtlFile
			Case 1
				ReadMaterials( $models & "\" & $sObjPath, "tree.mtl", $aMtlInfo )
			Case 2
				ReadMaterials( $models & "\" & $sObjPath, "tree-png.mtl", $aMtlInfo )
			Case 3
				ReadMaterials( $models & "\" & $sObjPath, "tree-brown.mtl", $aMtlInfo )
		EndSwitch
		$aModel[6] = $aMtlInfo
		LoadTextures( $aModel, $aMaterialData )
		$bMtlFile = $bNewMtlFile
	EndIf
	If $bResize Then
		glViewport( 0, 0, $iWidth, $iHeight )
		InitializeView()
		$bResize = False
	EndIf
	If $bClose Then
		CloseProgram()
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then CloseProgram()
	If Not InitOpenGLwindow() Then CloseProgram()
	;If Not CheckOpenGLversion( 2, 1 ) Then CloseProgram()
	If Not CreateResources() Then CloseProgram()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 50, $WS_OVERLAPPEDWINDOW )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	GUIRegisterMsg( $WM_SIZING, "WM_SIZING" )

	CreateStatusbar()

	Return 1

EndFunc


Func CreateStatusbar()

	; Use a child window as a container for controls in the Statusbar
	$hStatusbar = GUICreate( "", $iGuiWidth - 4 , 24, 2, $iHeight + 3 , $WS_CHILD, -1, $hGui )

	; Label
	GUICtrlCreateLabel( "Material file:", 3, 5, 57, 14 )

	; Radio buttons
	GUIStartGroup()
	$idRadio1 = GUICtrlCreateRadio( "tree.mtl",        75, 3,  60, 18 )
	$idRadio2 = GUICtrlCreateRadio( "tree-png.mtl",   160, 3,  75, 18 )
	$idRadio3 = GUICtrlCreateRadio( "tree-brown.mtl", 270, 3, 100, 18 )
	GUICtrlSetOnEvent( $idRadio1, "RadioClick" )
	GUICtrlSetOnEvent( $idRadio2, "RadioClick" )
	GUICtrlSetOnEvent( $idRadio3, "RadioClick" )
	GUICtrlSetState( $idRadio2, $GUI_CHECKED )

	GUISetState( @SW_SHOW, $hStatusbar )
	GUISwitch( $hGui )

	; Label for $hStatusbar to make it look nicer
	GUICtrlCreateLabel( "", 0, $iHeight + 1, $iGuiWidth + 4, $iStatusBarHeight, $SS_ETCHEDFRAME )
	GUICtrlSetResizing( -1, $GUI_DOCKSTATEBAR )

EndFunc


Func RadioClick()
	Switch @GUI_CTRLID
		Case $idRadio1
			$bNewMtlFile = 1
		Case $idRadio2
			$bNewMtlFile = 2
		Case $idRadio3
			$bNewMtlFile = 3
	EndSwitch
EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 0.75, 0.75, 0.75, 1.0 )

	glEnable( $GL_TEXTURE_2D )
	glEnable( $GL_DEPTH_TEST )

	glEnable( $GL_BLEND )
	glBlendFunc( $GL_SRC_ALPHA, $GL_ONE_MINUS_SRC_ALPHA )

	GUISwitch( $hGui )
	GUISetState()

	Return 1

EndFunc


Func CreateResources()

	If Not CheckModelFile( $sObjModel, $aModel ) Then Return 0

	If Not LoadObjModelBins( $aModel, False ) Then Return 0

	Local $iObjects = $aModel[13]
	Redim $aMaterialData[$iObjects][9]

	LoadTextures( $aModel, $aMaterialData )

	Return 1

EndFunc


Func DeletePreviousMaterials()
	; Delete textures
	Local $iTexIdLast = 0
	Local $iObjects = $aModel[13]
	For $obj = 0 To $iObjects - 1
		If $aMaterialData[$obj][6] > $iTexIdLast Then
			$iTexIdLast = $aMaterialData[$obj][6]
			glDeleteTextures( 1, $iTexIdLast )
		EndIf
		If $aMaterialData[$obj][7] > $iTexIdLast Then
			$iTexIdLast = $aMaterialData[$obj][7]
			glDeleteTextures( 1, $iTexIdLast )
		EndIf
		If $aMaterialData[$obj][8] > $iTexIdLast Then
			$iTexIdLast = $aMaterialData[$obj][8]
			glDeleteTextures( 1, $iTexIdLast )
		EndIf
	Next
EndFunc


Func DrawModel()

	Local $aVertInfo  = $aModel[4]
	Local $iObjects   = $aModel[13]
	Local $aVertStart = $aVertInfo[0]
	Local $aVertCnts  = $aVertInfo[1]

	For $i = 0 To $iObjects - 1
		glBindTexture( $GL_TEXTURE_2D, $aMaterialData[$i][7] )
		glDrawArrays( $GL_TRIANGLES, $aVertStart[$i], $aVertCnts[$i] )
	Next

EndFunc


Func DisplayFunction()

	Local Static $i = 0

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )

	glPushMatrix()

	Select
		Case $i <= 360
			glScalef( 1.5, 1.5, 1.5 )
			glRotatef( $i, 0.0, 1.0, 0.0 )
		Case $i <= 1080
			glRotatef( -1.875, 0.0, 1.0, 0.0 )
			glTranslatef( 0.0, 0.3 - ( $i - 361 ) / 1000, 1.5 )
		Case $i <= 1800
			glRotatef( -1.875, 0.0, 1.0, 0.0 )
			glTranslatef( 0.0, -0.42 + ( $i - 1081 ) / 1000, 1.5 )
		Case $i <= 2650
			glRotatef( 84.375, 1.0, 0.0, 0.0 )
			glTranslatef( -0.375, -11.00 + ( $i - 1801 ) / 50, 0.15 )
			glScalef( 11.443, 11.443, 11.443 )
		Case $i <= 2700
			glRotatef( 84.375, 1.0, 0.0, 0.0 )
			glTranslatef( -0.375, 6.00, 0.15 )
			glScalef( 11.443, 11.443, 11.443 )
		Case $i <= 3550
			glRotatef( 84.375, 1.0, 0.0, 0.0 )
			glTranslatef( -0.375, 6.00 - ( $i - 2701 ) / 50, 0.15 )
			glScalef( 11.443, 11.443, 11.443 )
		Case Else
			$i = 0
	EndSelect

	DrawModel()

	glPopMatrix()

	SwapBuffers( $hDC )

	$i += 1

EndFunc


Func WM_SIZING( $hWnd, $iMsg, $iwParam, $ilParam )
	#forceref $iMsg, $iwParam
	Switch $hWnd
		Case $hGui
			Local $tRECT = DllStructCreate( $tagRECT, $ilParam )
			Local $y = DllStructGetData( $tRECT, "Top" )
			Local $iHeight = DllStructGetData( $tRECT, "Bottom" ) - $y - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight
			WinMove( $hStatusbar, "", 2, $iHeight + $iHeightXP )
	EndSwitch
	Return $GUI_RUNDEFMSG
EndFunc


Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	WinMove( $hStatusbar, "", 2, $iHeight + $iHeightXP )
	$bResize = True
EndFunc


Func GUI_EVENT_CLOSE()
	$bClose = True
EndFunc


Func CloseProgram()
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
