#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

Global $sObjPath = "Small\Aircrafts\Mitsubishi A6M Zero\"

; --- OpenGL globals ---

Global $aModel[17]

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 800, $iHeight = 600

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight

Global $hGui, $bResize = True, $bClose = False

; --- Project includes ---

#include "utilities\models.au3"
#include "utilities\includes.au3"
#include "utilities\utilities.au3"

; --- Program code ---

MainFunc()

While 1
	If $bResize Then
		glViewport( 0, 0, $iWidth, $iHeight )
		$mProjectionMatrix = glmPerspective( 45, $iWidth / $iHeight, 0.5, 100.0 )
		$bResize = False
	EndIf
	If $bClose Then
		CloseProgram()
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then CloseProgram()
	If Not InitOpenGLwindow() Then CloseProgram()
	If Not CheckOpenGLversion( 2, 1 ) Then CloseProgram()
	If Not CreateResources() Then CloseProgram()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 50, $WS_OVERLAPPEDWINDOW )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	Return 1

EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 0.75, 0.75, 0.75, 1.0 )

	glEnable( $GL_TEXTURE_2D )
	glEnable( $GL_DEPTH_TEST )

	glEnable( $GL_BLEND )
	glBlendFunc( $GL_SRC_ALPHA, $GL_ONE_MINUS_SRC_ALPHA )

	GUISwitch( $hGui )
	GUISetState()

	Return 1

EndFunc


Func CreateResources()

	If Not CreateTextureCoordsShaders() Then Return 0

	Local $sObjModel = $sObjPath & "A6M_ZERO.obj"
	If Not CheckModelFile( $sObjModel, $aModel ) Then Return 0

	If Not LoadObjModelBins( $aModel, True, -1 ) Then Return 0

	Local $iObjects = $aModel[13]
	Local $aMtlData[$iObjects][9]
	$sObjPath = $models & "\" & $sObjPath

	Local $iInstances = 3
	Local $aMaterialData[$iInstances]
	Local $aTransforms[$iInstances]
	Local $aMtlInfo

	; Instance #0
	$aTransforms[0] = Transforms( 0.85, 30, 45, 30, -0.75, 0, -0.50 )
	ReadMaterials( $sObjPath, "A6M_ZERO-blue.mtl", $aMtlInfo )
	$aModel[6] = $aMtlInfo
	LoadTextures( $aModel, $aMtlData )
	$aMaterialData[0] = $aMtlData

	; Instance #1
	$aTransforms[1] = Transforms( 0.85, 30, 45, 0, -0.10, 0.25, 0.25 )
	ReadMaterials( $sObjPath, "A6M_ZERO-green.mtl", $aMtlInfo )
	$aModel[6] = $aMtlInfo
	LoadTextures( $aModel, $aMtlData )
	$aMaterialData[1] = $aMtlData

	; Instance #2
	$aTransforms[2] = Transforms( 0.80, 30, 45, -56.25, 0, 0.9, 0.2625 )
	ReadMaterials( $sObjPath, "A6M_ZERO-yellow.mtl", $aMtlInfo )
	$aModel[6] = $aMtlInfo
	LoadTextures( $aModel, $aMtlData )
	$aMaterialData[2] = $aMtlData

	$aModel[14] = $iInstances
	$aModel[15] = $aMaterialData
	$aModel[16] = $aTransforms

	InitializeView()

	Return 1

EndFunc


Func DrawModel()

	Local $aVertInfo     = $aModel[4]
	Local $iInstances    = $aModel[14]
	Local $aMaterialData = $aModel[15]
	Local $aTransforms   = $aModel[16]
	Local $aVertStart    = $aVertInfo[0]
	Local $aVertCnts     = $aVertInfo[1]
	Local $aMtlData, $mMatrix

	; Vertices
	glEnableVertexAttribArray( $iAttributeVertexXYZ )
	glBindBuffer( $GL_ARRAY_BUFFER, $aModel[0] )
	glVertexAttribPointer(  _
		$iAttributeVertexXYZ, _ ; attribute
		3,                    _ ; number of elements per vertex, here (x,y,z)
		$GL_FLOAT,            _ ; the type of each element
		$GL_FALSE,            _ ; take our values as-is
		0,                    _ ; no extra data between each position
		0 )                     ; offset of first element

	; Texture coordinates
	glEnableVertexAttribArray( $iAttributeTextureUV )
	glBindBuffer( $GL_ARRAY_BUFFER, $aModel[1] )
	glVertexAttribPointer(  _
		$iAttributeTextureUV, _ ; attribute
		2,                    _ ; number of elements per vertex, here (u,v)
		$GL_FLOAT,            _ ; the type of each element
		$GL_FALSE,            _ ; take our values as-is
		0,                    _ ; no extra data between each position
		0 )                     ; offset of first element

	For $i = 0 To $iInstances - 1

		; Instance data
		$aMtlData = $aMaterialData[$i]
		$mMatrix = $aTransforms[$i]

		; Apply transformation matrix
		glUniformMatrix4fv( $iUniformM, 1, $GL_FALSE, $mMatrix )

		; Textures
		glActiveTexture( $GL_TEXTURE0 )
		glUniform1i( $iUniformTexture, 0 ) 

		; Draw aircraft
		glBindTexture( $GL_TEXTURE_2D, $aMtlData[0][7] ) ; map_Kd
		glDrawArrays( $GL_TRIANGLES, $aVertStart[0], $aVertCnts[0] )

		; Draw propeller
		glBindTexture( $GL_TEXTURE_2D, $aMtlData[1][7] ) ; map_Kd
		glDrawArrays( $GL_TRIANGLES, $aVertStart[1], $aVertCnts[1] )

	Next

	glDisableVertexAttribArray( $iAttributeVertexXYZ )
	glDisableVertexAttribArray( $iAttributeTextureUV )

EndFunc


Func DisplayFunction()

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )

	glUniformMatrix4fv( $iUniformV, 1, $GL_FALSE, $mViewMatrix )
	glUniformMatrix4fv( $iUniformP, 1, $GL_FALSE, $mProjectionMatrix )

	DrawModel()

	SwapBuffers( $hDC )

EndFunc


Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	$bResize = True
EndFunc


Func GUI_EVENT_CLOSE()
	$bClose = True
EndFunc


Func CloseProgram()
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
