#include-once

Global $iProgram, $iAttributeVertexXYZ, $iAttributeTextureUV, $iAttributeNormalXYZ
Global $iUniformM, $iUniformV, $iUniformP, $iUniformM3x3InvTransp, $iUniformTexture, $iTexWhite
Global $mModelMatrix, $mViewMatrix, $mProjectionMatrix

Global $axisX[3] = [ 1.0, 0.0, 0.0 ], $axisY[3] = [ 0.0, 1.0, 0.0 ], $axisZ[3] = [ 0.0, 0.0, 1.0 ]

Func CreateBlinnPhongShaders()

	; The default shaders are the Blinn-Phong shaders for a single directional light source.

	; Program
	Local $iVs, $iFs
	If Not LoadGLSLshader( $iVs, "shaders\Blinn-Phong.v.glsl", $GL_VERTEX_SHADER ) Then Return 0
	If Not LoadGLSLshader( $iFs, "shaders\Blinn-Phong.f.glsl", $GL_FRAGMENT_SHADER ) Then Return 0
	If Not CreateProgram( $iProgram, $iVs, $iFs ) Then Return 0

	; Get handles for attributes
	If Not GetAttribLocation( $iProgram, $iAttributeVertexXYZ, "vertexXYZ" ) Then Return 0
	If Not GetAttribLocation( $iProgram, $iAttributeTextureUV, "textureUV" ) Then Return 0
	If Not GetAttribLocation( $iProgram, $iAttributeNormalXYZ, "normalXYZ" ) Then Return 0

	; Get handles for uniforms
	If Not GetUniformLocation( $iProgram, $iUniformM, "m" ) Then Return 0
	If Not GetUniformLocation( $iProgram, $iUniformV, "v" ) Then Return 0
	If Not GetUniformLocation( $iProgram, $iUniformP, "p" ) Then Return 0
	If Not GetUniformLocation( $iProgram, $iUniformM3x3InvTransp, "m_3x3_inv_transp" ) Then Return 0
	If Not GetUniformLocation( $iProgram, $iUniformTexture, "colorMap" ) Then Return 0

	glUseProgram( $iProgram )

	Return 1

EndFunc

Func CreatePerpixelLightShaders()

	; If the model type is "f v1 v2 v3 ..." ($aModel = 0) then use the Perpixel-Light shaders.
	; The model has no texture coordinates and no normal vectors. In this case normal vectors are
	; calculated by LoadObjModel

	; Program
	Local $iVs, $iFs
	If Not LoadGLSLshader( $iVs, "shaders\Perpixel-Light.v.glsl", $GL_VERTEX_SHADER ) Then Return 0
	If Not LoadGLSLshader( $iFs, "shaders\Perpixel-Light.f.glsl", $GL_FRAGMENT_SHADER ) Then Return 0
	If Not CreateProgram( $iProgram, $iVs, $iFs ) Then Return 0

	; Get handles for attributes
	If Not GetAttribLocation( $iProgram, $iAttributeVertexXYZ, "vertexXYZ" ) Then Return 0
	If Not GetAttribLocation( $iProgram, $iAttributeTextureUV, "textureUV" ) Then Return 0
	If Not GetAttribLocation( $iProgram, $iAttributeNormalXYZ, "normalXYZ" ) Then Return 0

	; Get handles for uniforms
	If Not GetUniformLocation( $iProgram, $iUniformM, "m" ) Then Return 0
	If Not GetUniformLocation( $iProgram, $iUniformV, "v" ) Then Return 0
	If Not GetUniformLocation( $iProgram, $iUniformP, "p" ) Then Return 0
	If Not GetUniformLocation( $iProgram, $iUniformM3x3InvTransp, "m_3x3_inv_transp" ) Then Return 0
	If Not GetUniformLocation( $iProgram, $iUniformTexture, "colorMap" ) Then Return 0

	glUseProgram( $iProgram )

	Return 1

EndFunc

Func CreateTextureCoordsShaders()

	; If the model type is "f v1/vt1 v2/vt2 v3/vt3 ..." ($aModel = 3) then use the Texture-Coords shaders.
	; There are no normal vectors, and light can't be used if there are no normal vectors.

	; Program
	Local $iVs, $iFs
	If Not LoadGLSLshader( $iVs, "shaders\Texture-Coords.v.glsl", $GL_VERTEX_SHADER ) Then Return 0
	If Not LoadGLSLshader( $iFs, "shaders\Texture-Coords.f.glsl", $GL_FRAGMENT_SHADER ) Then Return 0
	If Not CreateProgram( $iProgram, $iVs, $iFs ) Then Return 0

	; Get handles for attributes
	If Not GetAttribLocation( $iProgram, $iAttributeVertexXYZ, "vertexXYZ" ) Then Return 0
	If Not GetAttribLocation( $iProgram, $iAttributeTextureUV, "textureUV" ) Then Return 0
	;If Not GetAttribLocation( $iProgram, $iAttributeNormalXYZ, "normalXYZ" ) Then Return 0

	; Get handles for uniforms
	If Not GetUniformLocation( $iProgram, $iUniformM, "m" ) Then Return 0
	If Not GetUniformLocation( $iProgram, $iUniformV, "v" ) Then Return 0
	If Not GetUniformLocation( $iProgram, $iUniformP, "p" ) Then Return 0
	;If Not GetUniformLocation( $iProgram, $iUniformM3x3InvTransp, "m_3x3_inv_transp" ) Then Return 0
	If Not GetUniformLocation( $iProgram, $iUniformTexture, "colorMap" ) Then Return 0

	glUseProgram( $iProgram )

	Return 1

EndFunc

Func Transforms( $s, $rx, $ry, $rz, $tx, $ty, $tz )
	$mModelMatrix = $IDENTITY_MATRIX
	$mModelMatrix = glmRotate( $mModelMatrix, $rx, $axisX )
	$mModelMatrix = glmRotate( $mModelMatrix, $ry, $axisY )
	$mModelMatrix = glmRotate( $mModelMatrix, $rz, $axisZ )
	Local $scale[3] = [ $s, $s, $s ]
	$mModelMatrix = glmScale( $mModelMatrix, $scale )
	Local $trans[3] = [ $tx, $ty, $tz ]
	Return glmTranslate( $mModelMatrix, $trans )
EndFunc

Func InitializeView()

	; Look at view matrix
	Local $eye[3] = [ 0.0, 0.0, 2.0 ], $center[3] = [ 0.0, 0.0, 0.0 ], $up[3] = [ 0.0, 1.0, 0.0 ]
	$mViewMatrix = glmLookAt( $eye, $center, $up )

EndFunc
