; A star is a short line segment. The segment starts with the RGB color (0,0,0) (black) and ends
; with the color (z/5,z/5,z/5). z is a number with an initial value between 0.0 and 0.1. To make
; an impression of movement the z-coordinate is slowly increased up to a value of 10. It looks as
; if the star is coming closer. Because of the perspective view this will make the line strip look
; longer. And because of the color definition this also makes the star look brighter.

; Make these assumptions about a line segment with start point p0 and end point p1:
; p1.x = p0.x and p1.y = p0.y. Only the z-coordinates are different.

; Then a line segment can be described with 5 values: ( p0.x, p0.y, p0.z, p1.z, dz ).
; dz is used to increase the z-coordinate.

; dz is also used to control the speed of the stars. The larger dz the greater the velocity. This
; means that the stars (the line segments) are bigger for high speeds and smaller for low speeds.
; (A fast moving object must be large to be able to see it, while a slow moving object can be seen
; even if it's small.)

#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; Stars globals
Global $iStars = 250
Global $aStars[1][5]
Global $iSpeed = 1 ; 1 - 5, 1 => fastest
Global $bShowStars = True

; GUI globals
Global $hStatusbar = 0
Global $iStarsXpos = 4

; Project includes
#include "utilities\globals.au3"
#include "utilities\includes.au3"
#include "utilities\stars 1.au3"
#include "utilities\utils.au3"

; --- Program code ---

MainFunc()

While 1
	If $bStars Then
		CreateResources()
		$bStars = False
	EndIf
	If $bResize Then
		ResizeOpenGLwindow( 0, 0, $iWidth, $iHeight, 45.0, 1.0, 100.0 )
		$bResize = False
	EndIf
	If $bClose Then
		; Because most program running time is spent in the loop in DisplayFunction(), it must be taken
		; into account that the program is not shut down in the middle of an OpenGL drawing cycle.
		CloseProgram()
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then CloseProgram()
	If Not InitOpenGLwindow() Then CloseProgram()
	If Not CreateResources() Then CloseProgram()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 50, $WS_OVERLAPPEDWINDOW )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	GUIRegisterMsg( $WM_SIZING, "WM_SIZING" )

	; Statusbar
	StarsControls( $hGui, $iStarsXpos, $iHeight + 3, True )
	GUICtrlCreateLabel( "", 0, $iHeight + 1, $iGuiWidth + 2, $iStatusBarHeight, $SS_ETCHEDFRAME )
	GUICtrlSetResizing( -1, $GUI_DOCKSTATEBAR )

	Return 1

EndFunc


Func CreateResources()

	$iStars = $newStars
	$iSpeed = $newSpeed
	ReDim $aStars[$iStars+1][5]

	For $i = 0 To $iStars
		$aStars[$i][0] = ( Random( 200 ) - 100 ) / 100	; p0.x
		$aStars[$i][1] = ( Random( 200 ) - 100 ) / 100	; p0.y
		$aStars[$i][2] = 0															; p0.z
		$aStars[$i][3] = 0															; p1.z
		$aStars[$i][4] = Random( 100 ) / ( ( $iSpeed / 2 ) * 1000	) ; dz
	Next

	Return 1

EndFunc


Func DisplayFunction()

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )
	glLoadIdentity()

	glTranslatef( 0, 0, -7 )

	For $i = 0 To $iStars

		glBegin( $GL_LINE_STRIP )

			; Start point
			glColor3f( 0, 0, 0 )
			glVertex3f( $aStars[$i][0], $aStars[$i][1], $aStars[$i][2] ) ; p0: x,y,z

			; End point
			Local $rgbColor = $aStars[$i][3] / 5
			glColor3f( $rgbColor, $rgbColor, $rgbColor )
			glVertex3f( $aStars[$i][0], $aStars[$i][1], $aStars[$i][3] ) ; p1: x,y,z

			; Start point = end point
			$aStars[$i][2] = $aStars[$i][3] ; z

			; Increase z-coordinate of end point
			$aStars[$i][3] += $aStars[$i][4] ; p1.z += dz

			If $aStars[$i][2] >= 10 Then
				$aStars[$i][2] = 0 ; p0.z
				$aStars[$i][3] = 0 ; p1.z
			EndIf

		glEnd()

	Next

	SwapBuffers( $hDC )

EndFunc


Func GUI_EVENT_CLOSE()
	$bClose = True
EndFunc


Func CloseProgram()
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
