#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>
#include <EditConstants.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

Global $iThreadCount  = 15		; 10 - 20
Global $iThreadLength = 20		; 15 - 25, $iThreadCount + $iThreadLength = 35
Global $fThreadWidth  =  0.07	; 0.03, 0.07, 0.15 (space between threads)
Global $iThreadColor  =  1		; 1 = red, 2 = green, 3 = blue
Global $aThreads[1][1][3]

; Make a small difference between each thread
Global $aThreadOffsets[1][3]

; --- Project includes ---

#include "utilities\globals.au3"
#include "utilities\includes.au3"
#include "utilities\status 1.au3"
#include "utilities\utils.au3"

; --- Program code ---

MainFunc()

While 1
	If $bThreads Then
		CreateResources()
		$bThreads = False
	EndIf
	If $bResize Then
		ResizeOpenGLwindow( 0, 0, $iWidth, $iHeight, 45.0, 1.0, 100.0 )
		$fWidthHeightRatio = $iWidth / $iHeight / $fWidthHeightRatioInit
		$bResize = False
	EndIf
	If $bClose Then
		; Because most program running time is spent in the loop in DisplayFunction(), it must be taken
		; into account that the program is not shut down in the middle of an OpenGL drawing cycle.
		CloseProgram()
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then CloseProgram()
	If Not InitOpenGLwindow() Then CloseProgram()
	If Not CreateResources() Then CloseProgram()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 50, $WS_OVERLAPPEDWINDOW )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	GUIRegisterMsg( $WM_MOUSEMOVE, "WM_MOUSEMOVE" )
	GUIRegisterMsg( $WM_SIZING,    "WM_SIZING"    )

	CreateStatusbar()

	Return 1

EndFunc


Func CreateResources()

	$iThreadCount  = $newThreadCount
	$iThreadLength = $newThreadLength
	$fThreadWidth  = $newThreadWidth
	$iThreadColor  = $newThreadColor

	ReDim $aThreads[$iThreadCount+1][$iThreadLength+1][3]

	ReDim $aThreadOffsets[$iThreadCount+1][3]

	For $i = 0 To $iThreadCount
		$aThreadOffsets[$i][0] = ( Random( 200 ) - 100 ) / 1000 * $fThreadWidth
		$aThreadOffsets[$i][1] = ( Random( 200 ) - 100 ) / 1000 * $fThreadWidth
		$aThreadOffsets[$i][2] = ( Random( 200 ) - 100 ) / 1000 * $fThreadWidth
	Next

	Return 1

EndFunc


Func DisplayFunction()

	Local $x0 = 4.95 * $fWidthHeightRatio, $y0 = 3.7

	Local Static $iStartTime = TimerInit(), $fLastTime = 0
	Local $fElapsedTime = TimerDiff( $iStartTime )
	$fElapsedTime = ( $fLastTime + $fElapsedTime ) / 2
	$fLastTime = $fElapsedTime

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )
	glLoadIdentity()

	glTranslatef( 0, 0, -7 )

	For $i = 0 To $iThreadCount
		$aThreads[$i][0][0] = -$x0 + 2 * $x0 * ( $xMouse / $iWidth )
		$aThreads[$i][0][1] = +$y0 - 2 * $y0 * ( $yMouse / $iHeight )
		$aThreads[$i][0][2] = -2

		For $j = $iThreadLength To 1 Step -1
			$aThreads[$i][$j][0] = ( 2 * $aThreads[$i][$j-1][0] + $aThreads[$i][$j][0] ) / 2.9 + $aThreadOffsets[$i][0]
			$aThreads[$i][$j][1] = ( 2 * $aThreads[$i][$j-1][1] + $aThreads[$i][$j][1] ) / 2.9 + $aThreadOffsets[$i][1]
			$aThreads[$i][$j][2] = ( 2 * $aThreads[$i][$j-1][2] + $aThreads[$i][$j][2] ) / 2.9 + $aThreadOffsets[$i][2]
		Next
	Next

	Local $c
	For $i = 0 To $iThreadCount
		glBegin( $GL_LINE_STRIP )
			$c = Abs( Sin( $fElapsedTime / 3000 ) )
			Switch $iThreadColor
				Case 1 ; Red
					glColor3f( $c, 1 - $c, 1 - $c )
				Case 2 ; Green
					glColor3f( 1 - $c, $c, 1 - $c )
				Case 3 ; Blue
					glColor3f( 1 - $c, 1 - $c, $c )
			EndSwitch
			For $j = 0 To $iThreadLength - 1
				glVertex3f( $aThreads[$i][$j][0], $aThreads[$i][$j][1], $aThreads[$i][$j][2] )
			Next
		glEnd()
	Next

	SwapBuffers( $hDC )

EndFunc


Func GUI_EVENT_CLOSE()
	$bClose = True
EndFunc


Func CloseProgram()
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
