#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>
#include <EditConstants.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

; Threads
Global $iThreadCount  =  50
Global $iThreadLength = 100
Global $fThreadWidth  =   0.07	; 0.03, 0.07, 0.15 (space between threads)
Global $aThreads, $pThreads

; Make a small difference between each thread
Global $aThreadOffsets, $pThreadOffsets

; --- FASM globals ---

; Thread points
Global $pFasm1Data, $pCode1

; Thread colors, displaying threads
Global $pFasm2Data, $pCode2

; --- Project includes ---

#include "utilities\globals.au3"
#include "utilities\includes.au3"
#include "..\utilities\Stars 2.au3"
#include "utilities\status 2.au3"
#include "utilities\utils.au3"

; --- Program code ---

MainFunc()

While 1
	If $bThreads Or $bStars Then
		CreateResources()
		$bThreads = False
		$bStars = False
	EndIf
	If $bResize Then
		ResizeOpenGLwindow( 0, 0, $iWidth, $iHeight, 45.0, 1.0, 100.0 )
		$fWidthHeightRatio = $iWidth / $iHeight / $fWidthHeightRatioInit
		$bResize = False
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then GUI_EVENT_CLOSE()
	If Not InitOpenGLwindow() Then GUI_EVENT_CLOSE()
	If Not CreateResources() Then GUI_EVENT_CLOSE()
	If Not GetFASMcode() Then GUI_EVENT_CLOSE()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 50, $WS_OVERLAPPEDWINDOW )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	GUIRegisterMsg( $WM_MOUSEMOVE, "WM_MOUSEMOVE" )
	GUIRegisterMsg( $WM_SIZING,    "WM_SIZING"    )

	CreateStatusbar()

	Return 1

EndFunc


Func CreateResources()

	; --- Threads ---

	$iThreadCount  = $newThreadCount
	$iThreadLength = $newThreadLength
	$fThreadWidth  = $newThreadWidth

	$aThreads = DllStructCreate( "float[" & ( $iThreadCount + 1 ) * ( $iThreadLength + 1 ) * 3 & "]" )
	$pThreads = DllStructGetPtr( $aThreads )

	$aThreadOffsets = DllStructCreate( "float[" & ( $iThreadCount + 1 ) * 3 & "]" )
	$pThreadOffsets = DllStructGetPtr( $aThreadOffsets )

	Local $j = 1
	For $i = 0 To $iThreadCount
		DllStructSetData( $aThreadOffsets, 1, ( Random( 200 ) - 100 ) / 1000 * $fThreadWidth, $j   ) ; x
		DllStructSetData( $aThreadOffsets, 1, ( Random( 200 ) - 100 ) / 1000 * $fThreadWidth, $j+1 ) ; y
		DllStructSetData( $aThreadOffsets, 1, ( Random( 200 ) - 100 ) / 1000 * $fThreadWidth, $j+2 ) ; z
		$j += 3
	Next

	; --- Stars ---

	If Not CreateStars() Then Return 0

	Return 1

EndFunc


Func GetFASMcode()

	; --- Fasm1 ---

	Local Static $oFasm1, $aFasm1Data, $sCode1, $tCode1

	; Data structure
	$aFasm1Data = DllStructCreate( "ushort;ushort;uint;float;ptr" )
	; Element 1, offset  0 : Ushort $iThreadCount
	; Element 2, offset  2 : Ushort $iThreadLength
	; Element 3, offset  4 : Uint ( $iThreadLength + 1 ) * 3 * 4
	; Element 4, offset  8 : Constant float 2.9
	; Element 5, offset 12 : Pointer in $aThreadOffsets
	$pFasm1Data = DllStructGetPtr( $aFasm1Data )

	; Fasm1 data
	DllStructSetData( $aFasm1Data, 4, 2.9 ) ; Element 4 : Constant float 2.9

	; Fasm1 code
	If @AutoItX64 Then
		If Not ReadX64code( "utilities\fasm\Fasm1-x64.bin", "utilities\fasm\Fasm1-mk-x64-bin.au3", $sCode1 ) Then Return 0
	Else
		$oFasm1 = FasmInit()
		If Not ReadFasmCode( "utilities\fasm\Fasm1.asm", $oFasm1, $sCode1 ) Then Return 0
		FasmExit( $oFasm1 )
	EndIf
	$pCode1 = _MemVirtualAlloc( 0, 1024, $MEM_COMMIT, $PAGE_EXECUTE_READWRITE )
	$tCode1 = DllStructCreate( "byte[1024]", $pCode1 )
	DllStructSetData( $tCode1, 1, $sCode1 )

	; --- Fasm2 ---

	Local Static $oFasm2, $aFasm2Data, $sCode2, $tCode2

	; Data structure
	$aFasm2Data = DllStructCreate( "float[11];uint;ptr[4];ushort[2];ushort[2];float" )
	; Element 1, index 1,      offset  0      : Constant float 0.3
	; Element 1, index 2 -  4, offset  4      : Three constant floats: 1300, 700, 1000
	; Element 1, index 5 -  7, offset 16      : Three constant floats: 50, 45, 60
	; Element 1, index 8,      offset 28      : Float $m = 1 - $j / $iThreadLength
	; Element 1, index 9 - 11, offset 32      : Three floats: Color $r, $g, $b
	; Element 2, index 1,      offset 44      : Uint $GL_LINE_STRIP
	; Element 3, index 1 -  4, offset 48      : Pointers: glBegin, glColor3f, glVertex3f, glEnd  ; 8-bytes pointers in x64
	; Element 4, index 1,      offset 64 (80) : Ushort $iThreadCount                             ; not used in x86
	; Element 4, index 2,      offset 66 (82) : $i in "For $i = 0 To $iThreadCount"
	; Element 5, index 1,      offset 68 (84) : Ushort $iThreadLength                            ; not used in x86
	; Element 5, index 2,      offset 70 (86) : $j in "For $j = 0 To $iThreadLength"
	; Element 6, index 1,      offset 72 (88) : Float $fElapsedTime                              ; not used in x86
	; (The numbers in brackets are x64 offsets)
	$pFasm2Data = DllStructGetPtr( $aFasm2Data )

	; Fasm2 data
	DllStructSetData( $aFasm2Data, 1, 0.3, 1 )	; Element 1, index 1,     : Constant float 0.3
	DllStructSetData( $aFasm2Data, 1, 1300, 2 )	; Element 1, index 2 -  4 : Three constant floats: 1300, 700, 1000
	DllStructSetData( $aFasm2Data, 1,  700, 3 )
	DllStructSetData( $aFasm2Data, 1, 1000, 4 )
	DllStructSetData( $aFasm2Data, 1, 50, 5 )		; Element 1, index 5 -  7 : Three constant floats: 50, 45, 60
	DllStructSetData( $aFasm2Data, 1, 45, 6 )
	DllStructSetData( $aFasm2Data, 1, 60, 7 )
	DllStructSetData( $aFasm2Data, 2, $GL_LINE_STRIP )	; Element 2, index 1,     : Uint $GL_LINE_STRIP
	DllStructSetData( $aFasm2Data, 3, GetProcAddress( "opengl32.dll", "glBegin" ), 1 ) ; Element 3, index 1 -  4 : Pointers: glBegin, glColor3f, glVertex3f, glEnd
	DllStructSetData( $aFasm2Data, 3, GetProcAddress( "opengl32.dll", "glColor3f" ), 2 )
	DllStructSetData( $aFasm2Data, 3, GetProcAddress( "opengl32.dll", "glVertex3f" ), 3 )
	DllStructSetData( $aFasm2Data, 3, GetProcAddress( "opengl32.dll", "glEnd" ), 4 )

	; Fasm2 code
	If @AutoItX64 Then
		If Not ReadX64code( "utilities\fasm\Fasm2-x64.bin", "utilities\fasm\Fasm2-mk-x64-bin.au3", $sCode2 ) Then Return 0
	Else
		$oFasm2 = FasmInit()
		If Not ReadFasmCode( "utilities\fasm\Fasm2.asm", $oFasm2, $sCode2 ) Then Return 0
		FasmExit( $oFasm2 )
	EndIf
	$pCode2 = _MemVirtualAlloc( 0, 1024, $MEM_COMMIT, $PAGE_EXECUTE_READWRITE )
	$tCode2 = DllStructCreate( "byte[1024]", $pCode2 )
	DllStructSetData( $tCode2, 1, $sCode2 )

	; --- Stars ---

	If Not StarsFASMcode() Then Return 0

	Return 1

EndFunc


Func DisplayFunction()

	Local $x0 = 4.95 * $fWidthHeightRatio, $y0 = 3.7

	Local Static $iStartTime = TimerInit(), $fLastTime = 0
	Local $fElapsedTime = TimerDiff( $iStartTime )
	$fElapsedTime = ( $fLastTime + $fElapsedTime ) / 2
	$fLastTime = $fElapsedTime

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )
	glLoadIdentity()

	glTranslatef( 0, 0, -7 )

	DllCallAddress( "none", $pCode1, _
		"byte", $iThreadCount, "byte", $iThreadLength, "ptr", $pThreads, "ptr", $pThreadOffsets, _
		"float", -$x0 + 2 * $x0 * ( $xMouse / $iWidth ), "float", +$y0 - 2 * $y0 * ( $yMouse / $iHeight ), "float", -2, "ptr", $pFasm1Data )

	DllCallAddress( "none", $pCode2, _
		"byte", $iThreadCount, "byte", $iThreadLength, "ptr", $pThreads, "float", $fElapsedTime, "ptr", $pFasm2Data )

	If $bShowStars Then DllCallAddress( "none", $pCodeStars, "byte", $iStars, "ptr", $pStars, "ptr", $pDataStars )

	SwapBuffers( $hDC )

EndFunc


Func GUI_EVENT_CLOSE()
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
