        ; FASM assembler code

        ; $aFasmData = DllStructCreate( "ushort;ushort;uint;float;ptr" )
        ; Element 1, offset  0 : Ushort $iThreadCount                    ; x64: r11 is used in stead of $aFasmData element 1
        ; Element 2, offset  2 : Ushort $iThreadLength
        ; Element 3, offset  4 : Uint ( $iThreadLength + 1 ) * 3 * 4
        ; Element 4, offset  8 : Constant float 2.9
        ; Element 5, offset 12 : Pointer in $aThreadOffsets              ; x64: r9 is used in stead of $aFasmData element 5

        ; $aThreads = DllStructCreate( "float[" & ( $iThreadCount + 1 ) * ( $iThreadLength + 1 ) * 3 & "]" )

        ; $aThreadOffsets = DllStructCreate( "float[" & ( $iThreadCount + 1 ) * 3 & "]" )

        ; For $i = 0 To $iThreadCount
        ;   $aThreads[$i][0][0] = $xMouse
        ;   $aThreads[$i][0][1] = $yMouse
        ;   $aThreads[$i][0][2] = $zMouse
        ;
        ;   For $j = $iThreadLength To 1 Step -1
        ;     $aThreads[$i][$j][0] = ( 2 * $aThreads[$i][$j-1][0] + $aThreads[$i][$j][0] ) / 2.9 + $aThreadOffsets[$i][0]
        ;     $aThreads[$i][$j][1] = ( 2 * $aThreads[$i][$j-1][1] + $aThreads[$i][$j][1] ) / 2.9 + $aThreadOffsets[$i][1]
        ;     $aThreads[$i][$j][2] = ( 2 * $aThreads[$i][$j-1][2] + $aThreads[$i][$j][2] ) / 2.9 + $aThreadOffsets[$i][2]
        ;   Next
        ; Next

        ; Parameters:
        ; rcx        : $iThreadCount
        ; rdx        : $iThreadLength
        ; r8         : $pThreads
        ; r9         : $pThreadOffsets
        ; [rsp + 40] : $xMouse
        ; [rsp + 48] : $yMouse
        ; [rsp + 56] : $zMouse
        ; [rsp + 64] : $pFasmData


        use64


        ; --- Initializations ---

        ; Handle parameters and fill in the $aFasmData table

        ; r10 is pointer in $aFasmData table
        mov    r10,                [rsp + 64]         ; $aFasmData start

        ; r11 is $i in "For $i = 0 To $iThreadCount"
        ; Here is used "For $i = $iThreadCount To 0 Step -1"
        mov    r11,                rcx                ; $iThreadCount
        inc    r11                                    ; $iThreadCount += 1

        ; r12w is $j in "For $j = $iThreadLength To 1 Step -1"
        mov    r12,                rdx                ; $iThreadLength
        mov    [r10 + 02],         r12w               ; Store value in $aFasmData element 2

        ; Calculate the number of bytes per thread
        ; Bytes per thread = ( $iThreadLength + 1 ) * 3 * 4
        mov    [r10 + 04],         word 12            ; $aFasmData element 3
        mov    ax,                 r12w               ; $iThreadLength
        inc    ax                                     ; $iThreadLength + 1
        mul    word [r10 + 04]                        ; * ( 3 * 4 )
        mov    [r10 + 04],         ax                 ; Store least significant bytes (little-endian)
        mov    [r10 + 06],         dx                 ; Store most significant bytes
        ; $iThreadLength = 100 => 1212 bytes per thread
        ; $iThreadCount = 50 => 61812 bytes in $aThreads

        ; r8 is pointer in $aThreads table
        sub    r8d,                [r10 + 04]         ; Subtract ( $iThreadLength + 1 ) * 3 * 4 bytes from pointer

        ; r9 is pointer in $aThreadOffsets table
        add    r9,                 12                 ; Add 12 bytes to the pointer, now it's positioned at the first point right after the z-coordinate


        ; --- Calculations ---

iLoop:  ; For $i = 0 To $iThreadCount

        ; Calculate pointer (r8) for the next thread in $aThreads
        add    r8d,                [r10 + 04]         ; Add ( $iThreadLength + 1 ) * 3 * 4 bytes to pointer

        finit                                         ; Init FPU

        ; $aThreads[$i][0][0] = $xMouse
        fld    dword [rsp + 40]                       ; $xMouse
        fst    dword [r8  + 00]                       ; $xMouse -> $aThreads[$i][0][0]

        ; $aThreads[$i][0][1] = $yMouse
        fld    dword [rsp + 48]                       ; $yMouse
        fst    dword [r8  + 04]                       ; $yMouse -> $aThreads[$i][0][1]

        ; $aThreads[$i][0][2] = $zMouse
        fld    dword [rsp + 56]                       ; $zMouse
        fst    dword [r8  + 08]                       ; $zMouse -> $aThreads[$i][0][2]

        ; Initialize $j in the jLoop, start with $j = $iThreadLength and count backwards to 1
        ; Including the first point above this gives a total of $iThreadLength + 1 points in a thread

        mov    r12w,               [r10 + 02]         ; r12w is $j in "For $j = $iThreadLength To 1 Step -1"

        ; The jLoop is moving backwards down the thread, set the pointer to the end of the thread
        add    r8d,                [r10 + 04]         ; Add the number of bytes in a thread ( ( $iThreadLength + 1 ) * 3 * 4 ) to get to the end

        ; In the jLoop rcx is a pointer in $aThreadOffsets
        mov    rcx,                r9                 ; r9 is a pointer in $aThreadOffsets to row $i

jLoop:  ; For $j = $iThreadLength To 1 Step -1 (going backwards)

        mov    dl,                 3                  ; dl = 3 is used as counter in the zyxLoop, 3 coordinates

zyxLoop:; Do the calculations for each of the coordinates z,y,x in that order (going backwards)

        sub    r8,                 4                  ; Subtract 4 bytes from the pointer in $aThreads, the first coordinate for a point is z
        sub    rcx,                4                  ; Subtract 4 bytes from the pointer in $aThreadOffsets, the first coordinate for a point is z

        finit                                         ; Init FPU

        ; z ~ $aThreads[$i][$j][2], y ~ $aThreads[$i][$j][1], x ~ $aThreads[$i][$j][0]
        ; $aThreads[$i][$j][2] = ( 2 * $aThreads[$i][$j-1][2] + $aThreads[$i][$j][2] ) / 2.9 + $aThreadOffsets[$i][2]
        ; It's exactly the same formula for each of the coordinates

        fld    dword [r8  - 12]                       ; $aThreads[$i][$j-1][2]
        fadd   st0,                st0                ; * 2
        fadd   dword [r8]                             ; + $aThreads[$i][$j][2]
        fdiv   dword [r10 + 08]                       ; / 2.9 ( 2.9 is stored in $aFasmData element 4 )
        fadd   dword [rcx]                            ; + $aThreadOffsets[$i][2]
        fst    dword [r8]                             ; Result -> $aThreads[$i][$j][2]

        dec    dl                                     ; Decrement coordinate counter
        jnz    zyxLoop                                ; If > 0 jump to zyxLoop

jNext:  ; For $j = $iThreadLength To 1 Step -1

        add    rcx,                12                 ; Add 12 bytes to the $aThreadOffsets pointer

        dec    r12w                                   ; This is $j in "For $j = $iThreadLength To 1 Step -1", $j -= 1
        jnz    jLoop                                  ; If > 0 jump to previous (going backwards) point in the thread

iNext:  ; For $i = 0 To $iThreadCount

        sub    r8,                 12                 ; The jLoop terminates at second point, subtract 12 bytes to get to first point

        add    r9,                 12                 ; Add 12 bytes to the $aThreadOffsets pointer to make it point to the next row right after the z-coordinate

        dec    r11                                    ; This is $i in "For $i = $iThreadCount To 0 Step -1", $i -= 1
        jnz    iLoop                                  ; Jump to next thread if > 0

        ret
