        ; FASM assembler code

        ; $aFasmData = DllStructCreate( "ushort;ushort;uint;float;ptr" )
        ; Element 1, offset  0 : Ushort $iThreadCount
        ; Element 2, offset  2 : Ushort $iThreadLength
        ; Element 3, offset  4 : Uint ( $iThreadLength + 1 ) * 3 * 4
        ; Element 4, offset  8 : Constant float 2.9
        ; Element 5, offset 12 : Pointer in $aThreadOffsets

        ; $aThreads = DllStructCreate( "float[" & ( $iThreadCount + 1 ) * ( $iThreadLength + 1 ) * 3 & "]" )

        ; $aThreadOffsets = DllStructCreate( "float[" & ( $iThreadCount + 1 ) * 3 & "]" )

        ; For $i = 0 To $iThreadCount
        ;   $aThreads[$i][0][0] = $xMouse
        ;   $aThreads[$i][0][1] = $yMouse
        ;   $aThreads[$i][0][2] = $zMouse
        ;
        ;   For $j = $iThreadLength To 1 Step -1
        ;     $aThreads[$i][$j][0] = ( 2 * $aThreads[$i][$j-1][0] + $aThreads[$i][$j][0] ) / 2.9 + $aThreadOffsets[$i][0]
        ;     $aThreads[$i][$j][1] = ( 2 * $aThreads[$i][$j-1][1] + $aThreads[$i][$j][1] ) / 2.9 + $aThreadOffsets[$i][1]
        ;     $aThreads[$i][$j][2] = ( 2 * $aThreads[$i][$j-1][2] + $aThreads[$i][$j][2] ) / 2.9 + $aThreadOffsets[$i][2]
        ;   Next
        ; Next

        ; Parameters:
        ; [ebp + 08] : $iThreadCount
        ; [ebp + 12] : $iThreadLength
        ; [ebp + 16] : $pThreads
        ; [ebp + 20] : $pThreadOffsets
        ; [ebp + 24] : $xMouse
        ; [ebp + 28] : $yMouse
        ; [ebp + 32] : $zMouse
        ; [ebp + 36] : $pFasmData


        use32
        push   ebp
        mov    ebp,                esp


        ; --- Initializations ---

        ; Handle parameters and fill in the $aFasmData table

        ; ebx is pointer in $aFasmData table
        mov    ebx,                [ebp + 36]         ; $aFasmData start

        ; Initialize $i in the iLoop
        mov    [ebx],              word 0             ; This is $i in "For $i = 0 To $iThreadCount"

        ; Calculate the number of bytes per thread
        ; Bytes per thread = ( $iThreadLength + 1 ) * 3 * 4
        mov    [ebx + 04],         word 12            ; $aFasmData element 3
        mov    ax,                 [ebp + 12]         ; $iThreadLength
        inc    ax                                     ; $iThreadLength + 1
        mul    word [ebx + 04]                        ; * ( 3 * 4 )
        mov    [ebx + 04],         ax                 ; Store least significant bytes (little-endian)
        mov    [ebx + 06],         dx                 ; Store most significant bytes
        ; $iThreadLength = 100 => 1212 bytes per thread
        ; $iThreadCount = 50 => 61812 bytes in $aThreads

        ; eax is pointer in $aThreads table
        mov    eax,                [ebp + 16]         ; $aThreads start
        sub    eax,                [ebx + 04]         ; Subtract ( $iThreadLength + 1 ) * 3 * 4 bytes from pointer

        ; Pointer in $aThreadOffsets table
        mov    ecx,                [ebp + 20]         ; $pThreadOffsets is pointer in $aThreadOffsets
        add    ecx,                12                 ; Add 12 bytes to the pointer, now it's positioned at the first point right after the z-coordinate
        mov    [ebx + 12],         ecx                ; Store pointer in $aFasmData element 5


        ; --- Calculations ---

iLoop:  ; For $i = 0 To $iThreadCount

        ; Calculate pointer (eax) for the next thread in $aThreads
        add    eax,                [ebx + 04]         ; Add ( $iThreadLength + 1 ) * 3 * 4 bytes to pointer

        finit                                         ; Init FPU

        ; $aThreads[$i][0][0] = $xMouse
        fld    dword [ebp + 24]                       ; $xMouse
        fst    dword [eax + 00]                       ; $xMouse -> $aThreads[$i][0][0]

        ; $aThreads[$i][0][1] = $yMouse
        fld    dword [ebp + 28]                       ; $yMouse
        fst    dword [eax + 04]                       ; $yMouse -> $aThreads[$i][0][1]

        ; $aThreads[$i][0][2] = $zMouse
        fld    dword [ebp + 32]                       ; $zMouse
        fst    dword [eax + 08]                       ; $zMouse -> $aThreads[$i][0][2]

        ; Initialize $j in the jLoop, start with $j = $iThreadLength and count backwards to 1
        ; Including the first point above this gives a total of $iThreadLength + 1 points in a thread

        mov    cx,                 [ebp + 12]         ; $iThreadLength
        mov    [ebx + 02],         cx                 ; This is $j in "For $j = $iThreadLength To 1 Step -1"

        ; The jLoop is moving backwards down the thread, set the pointer to the end of the thread
        add    eax,                [ebx + 04]         ; Add the number of bytes in a thread ( ( $iThreadLength + 1 ) * 3 * 4 ) to get to the end

        ; In the jLoop ecx is a pointer in $aThreadOffsets
        mov    ecx,                [ebx + 12]         ; $aFasmData element 5 contains a pointer in $aThreadOffsets to row $i

jLoop:  ; For $j = $iThreadLength To 1 Step -1 (going backwards)

        mov    dl,                 3                  ; dl = 3 is used as counter in the zyxLoop, 3 coordinates

zyxLoop:; Do the calculations for each of the coordinates z,y,x in that order (going backwards)

        sub    eax,                4                  ; Subtract 4 bytes from the pointer in $aThreads, the first coordinate for a point is z
        sub    ecx,                4                  ; Subtract 4 bytes from the pointer in $aThreadOffsets, the first coordinate for a point is z

        finit                                         ; Init FPU

        ; z ~ $aThreads[$i][$j][2], y ~ $aThreads[$i][$j][1], x ~ $aThreads[$i][$j][0]
        ; $aThreads[$i][$j][2] = ( 2 * $aThreads[$i][$j-1][2] + $aThreads[$i][$j][2] ) / 2.9 + $aThreadOffsets[$i][2]
        ; It's exactly the same formula for each of the coordinates

        fld    dword [eax - 12]                       ; $aThreads[$i][$j-1][2]
        fadd   st0,                st0                ; * 2
        fadd   dword [eax]                            ; + $aThreads[$i][$j][2]
        fdiv   dword [ebx + 08]                       ; / 2.9 ( 2.9 is stored in $aFasmData element 4 )
        fadd   dword [ecx]                            ; + $aThreadOffsets[$i][2]
        fst    dword [eax]                            ; Result -> $aThreads[$i][$j][2]

        dec    dl                                     ; Decrement coordinate counter
        jnz    zyxLoop                                ; If > 0 jump to zyxLoop

jNext:  ; For $j = $iThreadLength To 1 Step -1

        add    ecx,                12                 ; Add 12 bytes to the $aThreadOffsets pointer

        dec    word [ebx + 02]                        ; This is $j in "For $j = $iThreadLength To 1 Step -1", $j -= 1
        jnz    jLoop                                  ; If > 0 jump to previous (going backwards) point in the thread

iNext:  ; For $i = 0 To $iThreadCount

        sub    eax,                12                 ; The jLoop terminates at second point, subtract 12 bytes to get to first point

        add    [ebx + 12],         dword 12           ; Add 12 bytes to the $aThreadOffsets pointer to make it point to the next row right after the z-coordinate

        inc    word [ebx]                             ; This is $i in "For $i = 0 To $iThreadCount", $i += 1
        mov    cx,                 [ebp + 08]         ; $iThreadCount
        inc    cx                                     ; $iThreadCount += 1
        cmp    cx,                 word [ebx]         ; Compare $i with $iThreadCount
        jne    iLoop                                  ; Jump to next thread if not equal

        pop    ebp
        ret    32
