        ; FASM assembler code

        ; $aFasmData = DllStructCreate( "float[11];uint;ptr[4];ushort[2];ushort[2];float" )
        ; Element 1, index 1,      offset  0 : Constant float 0.3
        ; Element 1, index 2 -  4, offset  4 : Three constant floats: 1300, 700, 1000
        ; Element 1, index 5 -  7, offset 16 : Three constant floats: 50, 45, 60
        ; Element 1, index 8,      offset 28 : Float $m = 1 - $j / $iThreadLength
        ; Element 1, index 9 - 11, offset 32 : Three floats: Color $r, $g, $b
        ; Element 2, index 1,      offset 44 : Uint $GL_LINE_STRIP
        ; Element 3, index 1 -  4, offset 48 : Pointers: glBegin, glColor3f, glVertex3f, glEnd ; x64: 8-bytes pointers
        ; Element 4, index 1,      offset 80 : Ushort $iThreadCount
        ; Element 4, index 2,      offset 82 : $i in "For $i = 0 To $iThreadCount"
        ; Element 5, index 1,      offset 84 : Ushort $iThreadLength
        ; Element 5, index 2,      offset 86 : $j in "For $j = 0 To $iThreadLength"
        ; Element 6, index 1,      offset 88 : Float $fElapsedTime

        ; $aThreads = DllStructCreate( "float[" & ( $iThreadCount + 1 ) * ( $iThreadLength + 1 ) * 3 & "]" )

        ; For $i = 0 To $iThreadCount
        ;   glBegin( $GL_LINE_STRIP )
        ;     For $j = 0 To $iThreadLength
        ;       $m = 1 - $j / $iThreadLength
        ;       $r = 0.3 * ( 0.3 + Sin( $fElapsedTime / 1300 - $i / 50 ) ) * $m
        ;       $g = 0.3 * ( 0.3 + Sin( $fElapsedTime /  700 - $i / 45 ) ) * $m
        ;       $b = 0.3 * ( 0.3 + Sin( $fElapsedTime / 1000 - $i / 60 ) ) * $m
        ;       glColor3f( $r, $g, $b )
        ;       glVertex3f( $x, $y, $z )
        ;     Next
        ;   glEnd()
        ; Next

        ; Parameters:
        ; rcx        : $iThreadCount
        ; rdx        : $iThreadLength
        ; r8         : $pThreads
        ; xmm3       : $fElapsedTime
        ; [rsp + 40] : $pFasmData


        use64


        ; --- Initializations ---

        ; Handle parameters and fill in the $aFasmData table

        ; r12 is pointer in $aFasmData table
        mov    r12,                [rsp + 40]         ; $aFasmData start

        ; Store $iThreadCount in $aFasmData element 4
        mov    [r12 + 80],         cx                 ; $iThreadCount -> $aFasmData element 4

        ; Store $iThreadLength in $aFasmData element 5
        mov    [r12 + 84],         dx                 ; $iThreadLength -> $aFasmData element 5

        ; r14 is pointer in $aThreads table
        mov    r14,                r8                 ; $aThreads start

        ; Store $fElapsedTime in $aFasmData element 6
        movss  [r12 + 88],         xmm3               ; $fElapsedTime -> $aFasmData element 6

        ; Initialize $i in iLoop
        mov    [r12 + 82],         word 0             ; This is $i in "For $i = 0 To $iThreadCount"


        ; --- Calculations ---

iLoop:  ; For $i = 0 To $iThreadCount

        ; glBegin( $GL_LINE_STRIP )
        sub    rsp,                40
        mov    ecx,                dword [r12 + 44]   ; $GL_LINE_STRIP
        call   qword [r12 + 48]                       ; glBegin
        add    rsp,                40

        ; Initialize $j in jLoop
        mov    [r12 + 86],         word 0             ; This is $j in "For $j = 0 To $iThreadLength"

jLoop:  ; For $j = 0 To $iThreadLength

        finit                                         ; Init FPU

        ; $m = 1 - $j / $iThreadLength
        fild   word [r12 + 86]                        ; $j
        fidiv  word [r12 + 84]                        ; / $iThreadLength
        fld1                                          ; Load +1.0 into st0 (and push prev st0 to st1)
        fsub   st0,                st1                ; 1 - $j / $iThreadLength
        fst    dword [r12 + 28]                       ; Store result, $m

        mov    rcx,                r12                ; Temporary store r12 in rcx

        mov    dl,                 3                  ; dl = 3 is used as counter in the rgbLoop, 3 colors

rgbLoop:; Do the calculations for each of the colors r,g,b

        finit                                         ; Init FPU

        ; $r = 0.3 * ( 0.3 + Sin( $fElapsedTime / 1300 - $i / 50 ) ) * $m,  $r = [r12 + 0 + 32], 1300 = [r12 + 0 + 04], 50 = [r12 + 0 + 16]
        ; $g = 0.3 * ( 0.3 + Sin( $fElapsedTime /  700 - $i / 45 ) ) * $m,  $g = [r12 + 4 + 32],  700 = [r12 + 4 + 04], 45 = [r12 + 4 + 16]
        ; $b = 0.3 * ( 0.3 + Sin( $fElapsedTime / 1000 - $i / 60 ) ) * $m,  $b = [r12 + 8 + 32], 1000 = [r12 + 8 + 04], 60 = [r12 + 8 + 16]

        fild   word  [rcx + 82]                       ; $i
        fdiv   dword [r12 + 16]                       ; / 50
        fld    dword [rcx + 88]                       ; $fElapsedTime -> st0, st0 -> st1
        fdiv   dword [r12 + 04]                       ; / 1300
        fsub   st0,                st1                ; $fElapsedTime / 1300 - $i / 50
        fsin                                          ; Sin
        fadd   dword [rcx + 00]                       ; + 0.3
        fmul   dword [rcx + 00]                       ; * 0.3
        fmul   dword [rcx + 28]                       ; * $m
        fst    dword [r12 + 32]                       ; Store result, $r

        add    r12,                4                  ; Next color

        dec    dl                                     ; Decrement color counter
        jnz    rgbLoop                                ; If > 0 jump to rgbLoop

        mov    r12,                rcx                ; Restore r12

        ; glColor3f( $r, $g, $b )
        sub    rsp,                40
        movss  xmm0,               dword [r12 + 32]   ; RGBcolor r
        movss  xmm1,               dword [r12 + 36]   ; RGBcolor g
        movss  xmm2,               dword [r12 + 40]   ; RGBcolor b
        call   qword [r12 + 56]                       ; glColor3f
        add    rsp,                40

        ; glVertex3f( $x, $y, $z )
        sub    rsp,                40
        movss  xmm0,               dword [r14 + 00]   ; x-coordinate
        movss  xmm1,               dword [r14 + 04]   ; y-coordinate
        movss  xmm2,               dword [r14 + 08]   ; z-coordinate
        call   qword [r12 + 64]                       ; glVertex3f
        add    rsp,                40

        add    r14,                12                 ; Next thread point

jNext:  ; For $j = 0 To $iThreadLength

        inc    word [r12 + 86]                        ; This is $j in "For $j = 0 To $iThreadLength", $j += 1
        mov    cx,                 [r12 + 84]         ; $iThreadLength
        inc    cx                                     ; $iThreadLength += 1
        cmp    cx,                 [r12 + 86]         ; Compare $j with $iThreadLength
        jne    jLoop                                  ; Jump to next point if not equal

        ; glEnd()
        sub    rsp,                40
        call   qword [r12 + 72]                       ; glEnd
        add    rsp,                40

iNext:  ; For $i = 0 To $iThreadCount

        inc    word [r12 + 82]                        ; This is $i in "For $i = 0 To $iThreadCount", $i += 1
        mov    cx,                 [r12 + 80]         ; $iThreadCount
        inc    cx                                     ; $iThreadCount += 1
        cmp    cx,                 [r12 + 82]         ; Compare $i with $iThreadCount
        jne    iLoop                                  ; Jump to next thread if not equal


        ret
