        ; FASM assembler code

        ; $aFasmData = DllStructCreate( "float[11];uint;ptr[4];ushort[2];ushort[2];float" )
        ; Element 1, index 1,      offset  0 : Constant float 0.3
        ; Element 1, index 2 -  4, offset  4 : Three constant floats: 1300, 700, 1000
        ; Element 1, index 5 -  7, offset 16 : Three constant floats: 50, 45, 60
        ; Element 1, index 8,      offset 28 : Float $m = 1 - $j / $iThreadLength
        ; Element 1, index 9 - 11, offset 32 : Three floats: Color $r, $g, $b
        ; Element 2, index 1,      offset 44 : Uint $GL_LINE_STRIP
        ; Element 3, index 1 -  4, offset 48 : Pointers: glBegin, glColor3f, glVertex3f, glEnd
        ; Element 4, index 1,      offset 64 : Ushort $iThreadCount                 ; not used in x86
        ; Element 4, index 2,      offset 66 : $i in "For $i = 0 To $iThreadCount"
        ; Element 5, index 1,      offset 68 : Ushort $iThreadLength                ; not used in x86
        ; Element 5, index 2,      offset 70 : $j in "For $j = 0 To $iThreadLength"
        ; Element 6, index 1,      offset 72 : Float $fElapsedTime                  ; not used in x86

        ; $aThreads = DllStructCreate( "float[" & ( $iThreadCount + 1 ) * ( $iThreadLength + 1 ) * 3 & "]" )

        ; For $i = 0 To $iThreadCount
        ;   glBegin( $GL_LINE_STRIP )
        ;     For $j = 0 To $iThreadLength
        ;       $m = 1 - $j / $iThreadLength
        ;       $r = 0.3 * ( 0.3 + Sin( $fElapsedTime / 1300 - $i / 50 ) ) * $m
        ;       $g = 0.3 * ( 0.3 + Sin( $fElapsedTime /  700 - $i / 45 ) ) * $m
        ;       $b = 0.3 * ( 0.3 + Sin( $fElapsedTime / 1000 - $i / 60 ) ) * $m
        ;       glColor3f( $r, $g, $b )
        ;       glVertex3f( $x, $y, $z )
        ;     Next
        ;   glEnd()
        ; Next

        ; Parameters:
        ; [ebp + 08] : $iThreadCount
        ; [ebp + 12] : $iThreadLength
        ; [ebp + 16] : $pThreads
        ; [ebp + 20] : $fElapsedTime
        ; [ebp + 24] : $pFasmData


        use32
        push   ebp
        mov    ebp,                esp


        ; --- Initializations ---

        ; Handle parameters and fill in the $aFasmData table

        ; eax is pointer in $aFasmData table
        mov    eax,                [ebp + 24]         ; $aFasmData start

        ; ebx is pointer in $aThreads table
        mov    ebx,                [ebp + 16]         ; $aThreads start

        ; Initialize $i in iLoop
        mov    [eax + 66],         word 0             ; This is $i in "For $i = 0 To $iThreadCount"


        ; --- Calculations ---

iLoop:  ; For $i = 0 To $iThreadCount

        ; glBegin( $GL_LINE_STRIP )
        pusha
        push   dword [eax + 44]                       ; $GL_LINE_STRIP
        call   dword [eax + 48]                       ; glBegin
        popa

        ; Initialize $j in jLoop
        mov    [eax + 70],         word 0             ; This is $j in "For $j = 0 To $iThreadLength"

jLoop:  ; For $j = 0 To $iThreadLength

        finit                                         ; Init FPU

        ; $m = 1 - $j / $iThreadLength
        fild   word [eax + 70]                        ; $j
        fidiv  word [ebp + 12]                        ; / $iThreadLength
        fld1                                          ; Load +1.0 into st0 (and push prev st0 to st1)
        fsub   st0,                st1                ; 1 - $j / $iThreadLength
        fst    dword [eax + 28]                       ; Store result, $m

        mov    ecx,                eax                ; Temporary store eax in ecx

        mov    dl,                 3                  ; dl = 3 is used as counter in the rgbLoop, 3 colors

rgbLoop:; Do the calculations for each of the colors r,g,b

        finit                                         ; Init FPU

        ; $r = 0.3 * ( 0.3 + Sin( $fElapsedTime / 1300 - $i / 50 ) ) * $m,  $r = [eax + 0 + 32], 1300 = [eax + 0 + 04], 50 = [eax + 0 + 16]
        ; $g = 0.3 * ( 0.3 + Sin( $fElapsedTime /  700 - $i / 45 ) ) * $m,  $g = [eax + 4 + 32],  700 = [eax + 4 + 04], 45 = [eax + 4 + 16]
        ; $b = 0.3 * ( 0.3 + Sin( $fElapsedTime / 1000 - $i / 60 ) ) * $m,  $b = [eax + 8 + 32], 1000 = [eax + 8 + 04], 60 = [eax + 8 + 16]

        fild   word  [ecx + 66]                       ; $i
        fdiv   dword [eax + 16]                       ; / 50
        fld    dword [ebp + 20]                       ; $fElapsedTime -> st0, st0 -> st1
        fdiv   dword [eax + 04]                       ; / 1300
        fsub   st0,                st1                ; $fElapsedTime / 1300 - $i / 50
        fsin                                          ; Sin
        fadd   dword [ecx + 00]                       ; + 0.3
        fmul   dword [ecx + 00]                       ; * 0.3
        fmul   dword [ecx + 28]                       ; * $m
        fst    dword [eax + 32]                       ; Store result, $r

        add    eax,                4                  ; Next color

        dec    dl                                     ; Decrement color counter
        jnz    rgbLoop                                ; If > 0 jump to rgbLoop

        mov    eax,                ecx                ; Restore eax

        ; glColor3f( $r, $g, $b )
        pusha
        push   dword [eax + 40]                       ; RGBcolor b
        push   dword [eax + 36]                       ; RGBcolor g
        push   dword [eax + 32]                       ; RGBcolor r
        call   dword [eax + 52]                       ; glColor3f
        popa

        ; glVertex3f( $x, $y, $z )
        pusha
        push   dword [ebx + 08]                       ; z-coordinate
        push   dword [ebx + 04]                       ; y-coordinate
        push   dword [ebx + 00]                       ; x-coordinate
        call   dword [eax + 56]                       ; glVertex3f
        popa

        add    ebx,                12                 ; Next thread point

jNext:  ; For $j = 0 To $iThreadLength

        inc    word [eax + 70]                        ; This is $j in "For $j = 0 To $iThreadLength", $j += 1
        mov    cx,                 [ebp + 12]         ; $iThreadLength
        inc    cx                                     ; $iThreadLength += 1
        cmp    cx,                 [eax + 70]         ; Compare $j with $iThreadLength
        jne    jLoop                                  ; Jump to next point if not equal

        ; glEnd()
        pusha
        call   dword [eax + 60]                       ; glEnd
        popa

iNext:  ; For $i = 0 To $iThreadCount

        inc    word [eax + 66]                        ; This is $i in "For $i = 0 To $iThreadCount", $i += 1
        mov    cx,                 [ebp + 08]         ; $iThreadCount
        inc    cx                                     ; $iThreadCount += 1
        cmp    cx,                 [eax + 66]         ; Compare $i with $iThreadCount
        jne    iLoop                                  ; Jump to next thread if not equal


        pop    ebp
        ret    20
