; A star is a short line segment. The segment starts with the RGB color (0,0,0) (black) and ends
; with the color (z/5,z/5,z/5). z is a number with an initial value between 0.0 and 0.1. To make
; an impression of movement the z-coordinate is slowly increased up to a value of 10. It looks as
; if the star is coming closer. Because of the perspective view this will make the line strip look
; longer. And because of the color definition this also makes the star look brighter.

; Make these assumptions about a line segment with start point p0 and end point p1:
; p1.x = p0.x and p1.y = p0.y. Only the z-coordinates are different.

; Then a line segment can be described with 5 values: ( p0.x, p0.y, p0.z, p1.z, dz ).
; dz is used to increase the z-coordinate.

; dz is also used to control the speed of the stars. The larger dz the greater the velocity. This
; means that the stars (the line segments) are bigger for high speeds and smaller for low speeds.
; (A fast moving object must be large to be able to see it, while a slow moving object can be seen
; even if it's small.)

#include-once

#include <EditConstants.au3>

; Stars
Global $iStars = 250
Global $aStars, $pStars
Global $iSpeed = 1 ; 1 - 5, 1 => fastest
Global $bShowStars = True

; FASM
Global $pDataStars, $pCodeStars

; Window and controls
Global $hStars = 0, $iStarsXpos
Global $idShowStars, $idStars, $idSpeed, $idUpdStars
Global $newShow = $bShowStars, $newStars = $iStars, $newSpeed = $iSpeed, $bStars = False

Func StarsControls( $hWnd, $xPos, $yPos, $bDisableCheckbox = False, $iniStars = 0, $iniSpeed = 0 )

	; Use a child window as a container for star controls
	$hStars = GUICreate( "", 320, 24, $xPos, $yPos, $WS_CHILD, -1, $hWnd )

	; Turn stars on/off
	$idShowStars = GUICtrlCreateCheckbox( "Stars", 2, 3, 50, 18 )
	GUICtrlSetState( $idShowStars, $GUI_CHECKED )
	If $bDisableCheckbox Then _
		GUICtrlSetState( $idShowStars, $GUI_DISABLE )
	GUICtrlSetOnEvent( $idShowStars, "ShowStars" )

	; Number of stars, 50 - 250
	GUICtrlCreateLabel( "Stars:", 62, 5, 32, 14 )
	$idStars = GUICtrlCreateInput( "25", 94, 2, 36, 20 )
	If $iniStars Then $newStars = $iniStars
	GUICtrlSetData( $idStars, $newStars / 10 )
	GUICtrlSetStyle( $idStars, $ES_READONLY )
	GUICtrlSetBkColor( $idStars, 0xFFFFFF )
	GUICtrlCreateUpdown( $idStars )
	GUICtrlSetLimit( -1, 25, 5 )
	GUICtrlCreateLabel( "x 10", 134, 5, 20, 14 )
	GUICtrlSetOnEvent( $idStars, "iStars" )

	; Velocity of stars, 1 - 5
	GUICtrlCreateLabel( "Speed:", 170, 5, 40, 14 )
	$idSpeed = GUICtrlCreateInput( "5", 210, 2, 36, 20 )
	If $iniSpeed Then $newSpeed = 6 - $iniSpeed
	GUICtrlSetData( $idSpeed, 6 - $newSpeed )
	GUICtrlSetStyle( $idSpeed, $ES_READONLY )
	GUICtrlSetBkColor( $idSpeed, 0xFFFFFF )
	GUICtrlCreateUpdown( $idSpeed )
	GUICtrlSetLimit( -1, 5, 1 )
	GUICtrlSetOnEvent( $idSpeed, "Speed" )

	; Implement settings
	$idUpdStars = GUICtrlCreateButton( "Stars", 266, 2, 50, 20 )
	GUICtrlSetState( $idUpdStars, $GUI_DISABLE )
	GUICtrlSetOnEvent( $idUpdStars, "Stars" )

	GUISetState( @SW_SHOW, $hStars )
	GUISwitch( $hWnd )

EndFunc


Func ShowStars()
	If GUICtrlRead( $idShowStars ) = $GUI_CHECKED Then
		$newShow = True
	Else
		$newShow = False
	EndIf
	SetStarsState()
EndFunc


Func iStars()
	$newStars = GUICtrlRead( $idStars ) * 10
	SetStarsState()
EndFunc


Func Speed()
	$newSpeed = 6 - GUICtrlRead( $idSpeed )
	SetStarsState()
EndFunc


Func SetStarsState()
	If $newShow  <> $bShowStars Or _
		 $newStars <> $iStars  Or _
		 $newSpeed <> $iSpeed  Then
		 GUICtrlSetState( $idUpdStars, $GUI_ENABLE )
	Else
		 GUICtrlSetState( $idUpdStars, $GUI_DISABLE )
	EndIf
EndFunc


Func Stars()
	$bStars = True
	GUICtrlSetState( $idUpdStars, $GUI_DISABLE )
EndFunc


Func CreateStars()

	$bShowStars = $newShow
	$iStars = $newStars
	$iSpeed = $newSpeed

	$aStars = DllStructCreate( "float[" & ( $iStars + 1 ) * 5 & "]" )
	$pStars = DllStructGetPtr( $aStars )

	Local $j = 1
	For $i = 0 To $iStars
		DllStructSetData( $aStars, 1, ( Random( 200 ) - 100 ) / 100, $j   ) ; p0.x
		DllStructSetData( $aStars, 1, ( Random( 200 ) - 100 ) / 100, $j+1 ) ; p0.y
		DllStructSetData( $aStars, 1, 0,                             $j+2 ) ; p0.z
		DllStructSetData( $aStars, 1, 0,                             $j+3 ) ; p1.z
		DllStructSetData( $aStars, 1, Random( 100 ) / ( ( $iSpeed / 2 ) * 1000	), $j+4 ) ; dz
		$j += 5
	Next

	Return 1

EndFunc


Func StarsFASMcode()

	Local Static $oFasmStars, $aDataStars, $sCodeStars, $tCodeStars

	; Data structure
	$aDataStars = DllStructCreate( "float;float;uint[2];ptr[4];ushort" )
	; Element 1, index 1,     offset  0      : Constant float 5
	; Element 2, index 1      offset  4      : Float RGB color
	; Element 3, index 1,     offset  8      : Uint $GL_LINE_STRIP
	; Element 3, index 2,     offset 12      : Unused, added to secure 8 bytes memory alignment in x64
	; Element 4, index 1 - 4, offset 16      : Pointers: glBegin, glColor3f, glVertex3f, glEnd  ; 8-bytes pointers in x64
	; Element 5, index 1,     offset 32 (48) : Ushort $iStars
	; (The numbers in brackets are x64 offsets)
	$pDataStars = DllStructGetPtr( $aDataStars )

	; Fasm data
	DllStructSetData( $aDataStars, 1, 5 )									; Element 1 : Constant float 5
	DllStructSetData( $aDataStars, 3, $GL_LINE_STRIP, 1 )	; Element 3 : Uint $GL_LINE_STRIP
	DllStructSetData( $aDataStars, 4, GetProcAddress( "opengl32.dll", "glBegin" ), 1 ) ; Element 4, index 1 -  4 : Pointers: glBegin, glColor3f, glVertex3f, glEnd
	DllStructSetData( $aDataStars, 4, GetProcAddress( "opengl32.dll", "glColor3f" ), 2 )
	DllStructSetData( $aDataStars, 4, GetProcAddress( "opengl32.dll", "glVertex3f" ), 3 )
	DllStructSetData( $aDataStars, 4, GetProcAddress( "opengl32.dll", "glEnd" ), 4 )

	; Fasm code
	If @AutoItX64 Then
		If Not ReadX64code( "..\utilities\Stars-x64.bin", "..\utilities\Stars-mk-x64-bin.au3", $sCodeStars ) Then Return 0
	Else
		$oFasmStars = FasmInit()
		If Not ReadFasmCode( "..\utilities\Stars.asm", $oFasmStars, $sCodeStars ) Then Return 0
		FasmExit( $oFasmStars )
	EndIf
	$pCodeStars = _MemVirtualAlloc( 0, 1024, $MEM_COMMIT, $PAGE_EXECUTE_READWRITE )
	$tCodeStars = DllStructCreate( "byte[1024]", $pCodeStars )
	DllStructSetData( $tCodeStars, 1, $sCodeStars )

	Return 1

EndFunc
