#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>
#include <EditConstants.au3>
#include <Date.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

Global $fWidthRatio

; --- OpenGL globals ---

Global $iListBase

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 640, $iHeight = 480

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30
Global Const $iStatusBarHeight = 28

Global $hStatusbar, $idRadio1, $idRadio2, $idRadio3, $idRadio4, $idCheckbox
Global $sFont = "Courier New", $sNewFont = $sFont
Global $bItalic = False, $bNewItalic = $bItalic

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight + $iStatusBarHeight

Global $hGui, $bResize = True

; --- Project includes ---

#include "utilities\includes.au3"

; --- Program code ---

MainFunc()

While 1
	If $sFont <> $sNewFont Or $bItalic <> $bNewItalic Then
		$sFont = $sNewFont
		$bItalic = $bNewItalic
		glDeleteLists( $iListBase, 96 )
		CreateResources()
	EndIf
	If $bResize Then
		ResizeOpenGLwindow( 0, 0, $iWidth, $iHeight, 45.0, 0.1, 100.0 )
		$fWidthRatio = $iWidth / $iGuiWidth
		$bResize = False
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then GUI_EVENT_CLOSE()
	If Not InitOpenGLwindow() Then GUI_EVENT_CLOSE()
	If Not CreateResources() Then GUI_EVENT_CLOSE()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 50, $WS_OVERLAPPEDWINDOW )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	GUIRegisterMsg( $WM_GETMINMAXINFO, "WM_GETMINMAXINFO" )
	GUIRegisterMsg( $WM_SIZING,        "WM_SIZING" )

	CreateStatusbar()

	Return 1

EndFunc


Func CreateStatusbar()

	; Use a child window as a container for controls in the Statusbar
	$hStatusbar = GUICreate( "", $iGuiWidth - 4 , 24, 2, $iHeight + 3 , $WS_CHILD, -1, $hGui )

	; Radio buttons
	GUIStartGroup()
	$idRadio1 = GUICtrlCreateRadio( "Arial",             5, 3,  45, 18 )
	$idRadio2 = GUICtrlCreateRadio( "Courier New",      60, 3,  75, 18 )
	$idRadio3 = GUICtrlCreateRadio( "Tahoma",          155, 3,  60, 18 )
	$idRadio4 = GUICtrlCreateRadio( "Times New Roman", 230, 3, 110, 18 )
	GUICtrlSetOnEvent( $idRadio1, "RadioClick" )
	GUICtrlSetOnEvent( $idRadio2, "RadioClick" )
	GUICtrlSetOnEvent( $idRadio3, "RadioClick" )
	GUICtrlSetOnEvent( $idRadio4, "RadioClick" )
	GUICtrlSetState( $idRadio2, $GUI_CHECKED )

	; Vertical splitterbar
	GUICtrlCreateLabel( "", 355, 2, 3, 20, $SS_ETCHEDFRAME )

	; Turn italic on/off
	$idCheckbox = GUICtrlCreateCheckbox( "Italic", 375, 3, 50, 18 )
	GUICtrlSetOnEvent( $idCheckbox, "Checkbox" )

	GUISetState( @SW_SHOW, $hStatusbar )
	GUISwitch( $hGui )

	; Label for $hStatusbar to make it look nicer
	GUICtrlCreateLabel( "", 0, $iHeight + 1, $iGuiWidth + 4, $iStatusBarHeight, $SS_ETCHEDFRAME )
	GUICtrlSetResizing( -1, $GUI_DOCKSTATEBAR )

EndFunc


Func RadioClick()
	Switch @GUI_CTRLID
		Case $idRadio1
			$sNewFont = "Arial"
		Case $idRadio2
			$sNewFont = "Courier New"
		Case $idRadio3
			$sNewFont = "Tahoma"
		Case $idRadio4
			$sNewFont = "Times New Roman"
	EndSwitch
EndFunc


Func Checkbox()
	If GUICtrlRead( $idCheckbox ) = $GUI_CHECKED Then
		$bNewItalic = True
	Else
		$bNewItalic = False
	EndIf
EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 0.0, 0.0, 0.0, 0.5 )

	glShadeModel( $GL_SMOOTH )

	glClearDepth( 1.0 )
	glEnable( $GL_DEPTH_TEST )
	glDepthFunc( $GL_LEQUAL )

	glHint( $GL_PERSPECTIVE_CORRECTION_HINT, $GL_NICEST )

	GUISwitch( $hGui )

	Return 1

EndFunc


Func CreateResources()

	If Not CreateOpenGLBitmapFont( _
		$iListBase, _  ; display list start
		32, _          ; first character
		96, _          ; number of characters
	  $sFont, _      ; typeface (font) name e.g. Arial
	  24, _          ; font size, default 8.5
	  $FW_BOLD, _    ; font weight, defined values in FontConstants.au3
	  $bItalic ) _   ; italic attribute option, True/False
			Then Return 0

	GUISetState()

	Return 1

EndFunc


Func DisplayFunction()

	Local Static $cnt1 = 0, $cnt2 = 0
	Local $text = "The time is: " & _NowTime()

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )
	glLoadIdentity()

	glTranslatef( 0.0, 0.0, -1.0 )

	glRasterPos2f( -0.35 / $fWidthRatio + 0.15 * Cos( $cnt1 ) * $fWidthRatio, 0.32 * Sin( $cnt2 ) )
	glColor3f( Cos( $cnt1 ), Sin( $cnt2 ), 1.0 - 0.5 * Cos( $cnt1 + $cnt2 ) )

	glListBase( $iListBase - 32 )
	glCallLists( StringLen( $text ), $GL_UNSIGNED_BYTE, $text )

	SwapBuffers( $hDC )

	$cnt1 += 0.025
	$cnt2 += 0.005

EndFunc


Func WM_GETMINMAXINFO( $hWnd, $iMsg, $iwParam, $ilParam )
	#forceref $iMsg, $iwParam
	Switch $hWnd
		Case $hGui
			Local $tMINMAXINFO = DllStructCreate( "int;int;int;int;int;int;int;int;int;int", $ilParam )
			DllStructSetData( $tMINMAXINFO, 7, 640 + 2 * $iBorderWidth )																		; Min width
			DllStructSetData( $tMINMAXINFO, 8, 480 + $iTitleBarHeight + $iBorderWidth + $iStatusBarHeight )	; Min height
	EndSwitch
	Return $GUI_RUNDEFMSG
EndFunc


Func WM_SIZING( $hWnd, $iMsg, $iwParam, $ilParam )
	#forceref $iMsg, $iwParam
	Switch $hWnd
		Case $hGui
			Local $tRECT = DllStructCreate( $tagRECT, $ilParam )
			Local $y = DllStructGetData( $tRECT, "Top" )
			Local $iHeight = DllStructGetData( $tRECT, "Bottom" ) - $y - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight
			WinMove( $hStatusbar, "", 2, $iHeight + $iHeightXP )
	EndSwitch
	Return $GUI_RUNDEFMSG
EndFunc


Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	WinMove( $hStatusbar, "", 2, $iHeight + $iHeightXP )
	$bResize = True
EndFunc


Func GUI_EVENT_CLOSE()
	glDeleteLists( $iListBase, 96 )
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
