#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>
#include <EditConstants.au3>
#include <Date.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

Global $fRot = 0.0

; --- OpenGL globals ---

Global $iListBase, $GMF

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 640, $iHeight = 480

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30
Global Const $iStatusBarHeight = 28

Global $hStatusbar, $idRotateX, $idRotateY, $idRotateZ, $idLight
Global $bRotateX = True, $bRotateY = True, $bRotateZ = True, $bLight = True, $bNewLight = $bLight

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight + $iStatusBarHeight

Global $hGui, $bResize = True

; --- Project includes ---

#include "utilities\includes.au3"

; --- Program code ---

MainFunc()

While 1
	If $bLight <> $bNewLight Then
		$bLight = $bNewLight
		If $bLight Then
			glEnable( $GL_LIGHT0 )
			glEnable( $GL_LIGHTING )
			glEnable( $GL_COLOR_MATERIAL )
		Else
			glDisable( $GL_LIGHT0 )
			glDisable( $GL_LIGHTING )
			glDisable( $GL_COLOR_MATERIAL )
		EndIf
	EndIf
	If $bResize Then
		ResizeOpenGLwindow( 0, 0, $iWidth, $iHeight, 45.0, 0.1, 100.0 )
		$bResize = False
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then GUI_EVENT_CLOSE()
	If Not InitOpenGLwindow() Then GUI_EVENT_CLOSE()
	If Not CreateResources() Then GUI_EVENT_CLOSE()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 50, $WS_OVERLAPPEDWINDOW )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	GUIRegisterMsg( $WM_SIZING,        "WM_SIZING" )

	CreateStatusbar()

	Return 1

EndFunc


Func CreateStatusbar()

	; Use a child window as a container for controls in the Statusbar
	$hStatusbar = GUICreate( "", $iGuiWidth - 4 , 24, 2, $iHeight + 3 , $WS_CHILD, -1, $hGui )

	; Turn rotation on/off
	GUICtrlCreateLabel( "Rotate about:", 2, 4, 70, 17 )
	$idRotateX = GUICtrlCreateCheckbox( "X-axis",  80, 3, 50, 18 )
	$idRotateY = GUICtrlCreateCheckbox( "Y-axis", 140, 3, 50, 18 )
	$idRotateZ = GUICtrlCreateCheckbox( "Z-axis", 200, 3, 50, 18 )
	GUICtrlSetState( $idRotateX, $GUI_CHECKED )
	GUICtrlSetState( $idRotateY, $GUI_CHECKED )
	GUICtrlSetState( $idRotateZ, $GUI_CHECKED )
	GUICtrlSetOnEvent( $idRotateX, "RotateX" )
	GUICtrlSetOnEvent( $idRotateY, "RotateY" )
	GUICtrlSetOnEvent( $idRotateZ, "RotateZ" )

	; Vertical splitterbar
	GUICtrlCreateLabel( "", 260, 2, 3, 20, $SS_ETCHEDFRAME )

	; Turn light and material on/off
	$idLight = GUICtrlCreateCheckbox( "Light and material", 278, 3, 105, 18 )
	GUICtrlSetState( $idLight, $GUI_CHECKED )
	GUICtrlSetOnEvent( $idLight, "Light" )

	GUISetState( @SW_SHOW, $hStatusbar )
	GUISwitch( $hGui )

	; Label for $hStatusbar to make it look nicer
	GUICtrlCreateLabel( "", 0, $iHeight + 1, $iGuiWidth + 4, $iStatusBarHeight, $SS_ETCHEDFRAME )
	GUICtrlSetResizing( -1, $GUI_DOCKSTATEBAR )

EndFunc


Func RotateX()
	$bRotateX = Not $bRotateX
	$fRot = 0.0
EndFunc


Func RotateY()
	$bRotateY = Not $bRotateY
	$fRot = 0.0
EndFunc


Func RotateZ()
	$bRotateZ = Not $bRotateZ
	$fRot = 0.0
EndFunc


Func Light()
	$bNewLight = Not $bNewLight
EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 0.0, 0.0, 0.0, 0.5 )

	glShadeModel( $GL_SMOOTH )

	glClearDepth( 1.0 )
	glEnable( $GL_DEPTH_TEST )
	glDepthFunc( $GL_LEQUAL )

	glEnable( $GL_LIGHT0 )
	glEnable( $GL_LIGHTING )
	glEnable( $GL_COLOR_MATERIAL )

	glHint( $GL_PERSPECTIVE_CORRECTION_HINT, $GL_NICEST )

	GUISwitch( $hGui )

	Return 1

EndFunc


Func CreateResources()

	$GMF = DllStructCreate( "float[" & 6 * 256 & "]" )
	Local $pGMF = DllStructGetPtr( $GMF )

	If Not CreateOpenGLOutlineFont( _
		$pGMF, _              ; address of buffer to receive glyph metric data, see $tagGLYPHMETRICSFLOAT above
		$iListBase, _         ; display list start
		0, _                  ; first character
		256, _                ; number of characters
		0.0, _                ; specifies the maximum chordal deviation from the true outlines
		0.2, _                ; extrusion value (font thickness) in the negative z direction
		$WGL_FONT_POLYGONS, _ ; line segments or polygons in display lists
	  "Comic Sans MS", _    ; typeface (font) name e.g. Arial
	  12, _                 ; font size, default 8.5
	  $FW_BOLD, _           ; font weight, defined values in FontConstants.au3
	  False ) _             ; italic attribute option, True/False
			Then Return 0

	GUISetState()

	Return 1

EndFunc


Func DisplayFunction()

	Local Static $fLength = 0.0
	Local $text = "The time is: " & _NowTime()

	If $fLength = 0.0 Then
		Local $aTmp = StringSplit( $text, "" ), $j
		For $i = 1 To $aTmp[0]
			$j = 6 * Asc( $aTmp[$i] ) + 5
			$fLength += DllStructGetData( $GMF, 1, $j )
		Next
	EndIf

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )
	glLoadIdentity()

	glTranslatef( 0.0, 0.0, -10.0 )

	If $bRotateX Then glRotatef( $fRot, 1.0, 0.0, 0.0 )
	If $bRotateY Then glRotatef( $fRot * 1.5, 0.0, 1.0, 0.0 )
	If $bRotateZ Then glRotatef( $fRot * 1.4, 0.0, 0.0, 1.0 )
	glColor3f( Cos( $fRot / 20.0 ), sin( $fRot / 25.0 ), 1.0 -0.5 * Cos( $fRot / 17.0 ) )

	glTranslatef( -$fLength / 2, 0.0, 0.0 )
	glListBase( $iListBase )
	glCallLists( StringLen( $text ), $GL_UNSIGNED_BYTE, $text )

	SwapBuffers( $hDC )

	$fRot += 0.25

EndFunc


Func WM_SIZING( $hWnd, $iMsg, $iwParam, $ilParam )
	#forceref $iMsg, $iwParam
	Switch $hWnd
		Case $hGui
			Local $tRECT = DllStructCreate( $tagRECT, $ilParam )
			Local $y = DllStructGetData( $tRECT, "Top" )
			Local $iHeight = DllStructGetData( $tRECT, "Bottom" ) - $y - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight
			WinMove( $hStatusbar, "", 2, $iHeight + $iHeightXP )
	EndSwitch
	Return $GUI_RUNDEFMSG
EndFunc


Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	WinMove( $hStatusbar, "", 2, $iHeight + $iHeightXP )
	$bResize = True
EndFunc


Func GUI_EVENT_CLOSE()
	glDeleteLists( $iListBase, 256 )
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
