#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>
#include <EditConstants.au3>
#include <Misc.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

Global $iScroll = 0
Global $aExtensions, $iExtensions
Global $iScissorWidth, $iScissorHeight

; --- OpenGL globals ---

Global $iTexture1, $iTexture2, $iTexture3, $iListBase

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 640, $iHeight = 480

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30
Global Const $iStatusBarHeight = 28

Global $hStatusbar, $idRadio1, $idRadio2, $idRadio3, $idLabel
Global $bImage = 1, $bNewImage = 1

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight + $iStatusBarHeight

Global $hGui, $bResize = True

; --- Project includes ---

#include "utilities\includes.au3"
#include "utilities\images.au3"

; --- Program code ---

MainFunc()

While 1
	If $bNewImage <> $bImage Then
		Switch $bNewImage
			Case 1
				glBindTexture( $GL_TEXTURE_2D, $iTexture1 )
				glDeleteLists( $iListBase, 256 )
				CreateOpenGLTextureFont()
			Case 2
				glBindTexture( $GL_TEXTURE_2D, $iTexture2 )
				glDeleteLists( $iListBase, 256 )
				CreateOpenGLTextureFont()
			Case 3
				glBindTexture( $GL_TEXTURE_2D, $iTexture3 )
				glDeleteLists( $iListBase, 256 )
				CreateOpenGLTextureFont()
		EndSwitch
		$bImage = $bNewImage
	EndIf
	If $bResize Then
		$iScissorWidth = $iWidth
		$iScissorHeight = $iHeight
		ResizeOpenGLwindowOrtho( 0, 0, $iWidth, $iHeight, 0.0, 640, 480, 0.0, -1.0, 1.0 )
		$bResize = False
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then GUI_EVENT_CLOSE()
	If Not InitOpenGLwindow() Then GUI_EVENT_CLOSE()
	If Not CreateResources() Then GUI_EVENT_CLOSE()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 50, $WS_OVERLAPPEDWINDOW )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	GUIRegisterMsg( $WM_SIZING, "WM_SIZING" )

	CreateStatusbar()

	Return 1

EndFunc


Func CreateStatusbar()

	; Use a child window as a container for controls in the Statusbar
	$hStatusbar = GUICreate( "", $iGuiWidth - 4 , 24, 2, $iHeight + 3 , $WS_CHILD, -1, $hGui )

	; Radio buttons
	GUIStartGroup()
	$idRadio1 = GUICtrlCreateRadio( "TGA",        5, 3, 45, 18 )
	$idRadio2 = GUICtrlCreateRadio( "TGA (RLE)", 60, 3, 75, 18 )
	$idRadio3 = GUICtrlCreateRadio( "RGBA",     145, 3, 50, 18 )
	GUICtrlSetOnEvent( $idRadio1, "RadioClick" )
	GUICtrlSetOnEvent( $idRadio2, "RadioClick" )
	GUICtrlSetOnEvent( $idRadio3, "RadioClick" )
	GUICtrlSetState( $idRadio1, $GUI_CHECKED )

	; Vertical splitterbar
	GUICtrlCreateLabel( "", 205, 2, 3, 20, $SS_ETCHEDFRAME )

	; Label in statusbar
	$idLabel = GUICtrlCreateLabel( "", 225, 5, 350, 14 )
	GUICtrlSetData( $idLabel, "Use Arrow up/down, Page up/down, Home and End to scroll text." )

	GUISetState( @SW_SHOW, $hStatusbar )
	GUISwitch( $hGui )

	; Label for $hStatusbar to make it look nicer
	GUICtrlCreateLabel( "", 0, $iHeight + 1, $iGuiWidth + 4, $iStatusBarHeight, $SS_ETCHEDFRAME )
	GUICtrlSetResizing( -1, $GUI_DOCKSTATEBAR )

EndFunc


Func RadioClick()
	Switch @GUI_CTRLID
		Case $idRadio1
			$bNewImage = 1
			GUICtrlSetData( $idLabel, "Images\Text and fonts\Font.tga" )
		Case $idRadio2
			$bNewImage = 2
			GUICtrlSetData( $idLabel, "Images\Text and fonts\Font(rle).tga" )
		Case $idRadio3
			$bNewImage = 3
			GUICtrlSetData( $idLabel, "Images\Text and fonts\Font.rgba" )
	EndSwitch
EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 0.0, 0.0, 0.0, 0.5 )
	glClearDepth( 1.0 )

	glShadeModel( $GL_SMOOTH )
	glEnable( $GL_LINE_SMOOTH )

	GUISwitch( $hGui )

	Return 1

EndFunc


Func CreateResources()

	Local $w = 256, $h = 256
	If Not LoadTextureImage( $iTexture1, $images & "\Text and fonts\Font.tga", $w, $h ) Then Return 0
	If Not LoadTextureImage( $iTexture2, $images & "\Text and fonts\Font(rle).tga", $w, $h ) Then Return 0
	If Not LoadTextureImage( $iTexture3, $images & "\Text and fonts\Font.rgba", $w, $h ) Then Return 0
	glBindTexture( $GL_TEXTURE_2D, $iTexture1 )

	If Not CreateOpenGLTextureFont() Then Return 0

	Local $sExtensions
	$sExtensions = glGetString( $GL_EXTENSIONS )
	$aExtensions = StringSplit( $sExtensions, " " )
	ReDim $aExtensions[$aExtensions[0]]
	$aExtensions[0] -= 1
	If $aExtensions[0] > 9 Then
		$iExtensions = 9
	Else
		$iExtensions = $aExtensions[0]
	EndIf

	GUISetState()

	Return 1

EndFunc


Func CreateOpenGLTextureFont()

	Local $cx, $cy

	$iListBase = glGenLists( 256 )

	For $i = 0 To 255																				; 256 characters arranged in a 16 x 16 grid
		$cx = Mod( $i, 16 )  / 16															; X position of current character
		$cy = Int( $i / 16 ) / 16															; Y position of current character
		glNewList( $iListBase + $i, $GL_COMPILE )							; A display list for each character
			glBegin( $GL_QUADS )																; Use a quad for each character
				glTexCoord2f( $cx, 1.0 - $cy - 0.0625 )						; Texture coord, bottom left ( 0.0625 = 1 / 16 )
				glVertex2i( 0, 16 )																; Vertex coord, bottom left
				glTexCoord2f( $cx + 0.0625, 1.0 - $cy - 0.0625 )	; Texture coord, bottom right
				glVertex2i( 16, 16 )															; Vertex coord, bottom right
				glTexCoord2f( $cx + 0.0625, 1.0 - $cy - 0.001 )		;	Texture coord, top right
				glVertex2i( 16, 0 )																; Vertex coord, top right
				glTexCoord2f( $cx, 1.0 - $cy - 0.001 )						; Texture coord, top left
				glVertex2i( 0, 0 )																; Vertex coord, top left
			glEnd()																							; Done building our quad character
			glTranslatef( 14, 0, 0 )														; Move to the right of the character
		glEndList()																						; Display list done
	Next

	Return 1

EndFunc


Func PrintText( $x, $y, $text, $bItalic = False )
	glEnable( $GL_TEXTURE_2D )
	glLoadIdentity()
	glTranslatef( $x, $y, 0 )
	glScalef( 1.0, 2.0, 1.0 )
	glListBase( $iListBase - 32 + ( 128 * $bItalic ) )
	glCallLists( StringLen( $text ), $GL_UNSIGNED_BYTE, $text )
	glDisable( $GL_TEXTURE_2D )
EndFunc


Func DisplayFunction()

	HandleKeystrokes()

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )

	glColor3f( 1.0, 0.5, 0.5 )
	PrintText( 66, 16, "Version", True )
	PrintText( 80, 48, "Vendor", True )
	PrintText( 50, 80, "Renderer", True )

	glColor3f( 1.0, 0.7, 0.4 )
	PrintText( 200, 16, glGetString( $GL_VERSION ), True )
	PrintText( 200, 48, glGetString( $GL_VENDOR ), True )
	PrintText( 200, 80, glGetString( $GL_RENDERER ), True )

	glColor3f( 0.5, 0.5, 1.0 )
	PrintText( 184, 432, "OpenGL Extensions", True )

	glLoadIdentity()
	glColor3f( 1.0, 1.0, 1.0 )
	glBegin( $GL_LINE_STRIP )
		glVertex2i(   0, 127 )
		glVertex2i( 640, 127 )
	glEnd()
	glBegin( $GL_LINE_STRIP )
		glVertex2i(   0, 417 )
		glVertex2i( 640, 417 )
	glEnd()

	; Scissor region
	glEnable( $GL_SCISSOR_TEST )

	glScissor( 1, Int( 0.135416 * $iScissorHeight ), $iScissorWidth - 2, Int( 0.597916 * $iScissorHeight ) )

	For $i = 1 To $iExtensions
		glColor3f( 0.5, 1.0, 0.5 )
		PrintText( 0, 96 + $i * 32, String( $i + $iScroll ) )
		glColor3f( 1.0, 1.0, 0.5 )
		PrintText( 50, 96 + $i * 32, $aExtensions[$i+$iScroll])
	Next

	glDisable( $GL_SCISSOR_TEST )

	SwapBuffers( $hDC )

EndFunc


Func HandleKeystrokes()

	Local Static $bScroll = 0, $nScroll = 0
	Local Static $nStrokes = 0

	Select
		Case _IsPressed( "21", $dllUser32 ) ; Page up
			$bScroll = 1
			if Not $nScroll Then $nScroll = 1
		Case _IsPressed( "22", $dllUser32 ) ; Page down
			$bScroll = 2
			if Not $nScroll Then $nScroll = 1
		Case _IsPressed( "23", $dllUser32 ) ; End
			$bScroll = 3
			if Not $nScroll Then $nScroll = 1
		Case _IsPressed( "24", $dllUser32 ) ; Home
			$bScroll = 4
			if Not $nScroll Then $nScroll = 1
		Case _IsPressed( "26", $dllUser32 ) ; Up arrow
			$bScroll = 5
			if Not $nScroll Then $nScroll = 1
		Case _IsPressed( "28", $dllUser32 ) ; Down arrow
			$bScroll = 6
			if Not $nScroll Then $nScroll = 1
		Case Else
			$nScroll = 0
			Return
	EndSelect

	If $nScroll = 1 Or $nScroll = 3 Then
		Switch $bScroll
			Case 1 ; Page up
				If $aExtensions[0] > 9 Then
					If $iScroll > 9 Then
						$iScroll -= 9
					ElseIf $iScroll Then
						$iScroll = 0
					EndIf
				EndIf
			Case 2 ; Page down
				If $aExtensions[0] > 9 Then
					If $iScroll < $aExtensions[0] - 18 Then
						$iScroll += 9
					ElseIf $iScroll < $aExtensions[0] - 9 Then
						$iScroll = $aExtensions[0] - 9
					EndIf
				EndIf
			Case 3 ; End
				If $aExtensions[0] > 9 Then _
					$iScroll = $aExtensions[0] - 9
			Case 4 ; Home
				If $aExtensions[0] > 9 Then _
					$iScroll = 0
			Case 5 ; Up arrow
				If $iScroll Then $iScroll -= 1
			Case 6 ; Down arrow
				If $iScroll < $aExtensions[0] - 9 Then $iScroll += 1
		EndSwitch
		$bScroll = 0
	EndIf

	Switch $nScroll
		Case 1
			; First keypress
			$nScroll = 2
			$nStrokes = 0
		Case 2
			; Wait a few loops for next keypress
			; To get time to release the key for a single keypress
			$nStrokes += 1
			If $nStrokes > 3 Then _
				$nScroll = 3
	EndSwitch

EndFunc


Func WM_SIZING( $hWnd, $iMsg, $iwParam, $ilParam )
	#forceref $iMsg, $iwParam
	Switch $hWnd
		Case $hGui
			Local $tRECT = DllStructCreate( $tagRECT, $ilParam )
			Local $y = DllStructGetData( $tRECT, "Top" )
			Local $iHeight = DllStructGetData( $tRECT, "Bottom" ) - $y - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight
			WinMove( $hStatusbar, "", 2, $iHeight + $iHeightXP )
	EndSwitch
	Return $GUI_RUNDEFMSG
EndFunc


Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	WinMove( $hStatusbar, "", 2, $iHeight + $iHeightXP )
	$bResize = True
EndFunc


Func GUI_EVENT_CLOSE()
	glDeleteLists( $iListBase, 256 )
	glDeleteTextures( 1, $iTexture1 )
	glDeleteTextures( 1, $iTexture2 )
	glDeleteTextures( 1, $iTexture3 )
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
