#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>

Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- OpenGL globals ---

Global $iTexture, $iCallList

; OpenGL window
Global $hOpenGL, $hDC, $hRC, $iWidth = 640, $iHeight = 299

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30
Global Const $iStatusBarHeight = 28

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight + $iStatusBarHeight

Global $hGui, $idLabel

; --- Project includes ---

#include "utilities\includes.au3"
#include "utilities\images.au3"

; --- Program code ---

MainFunc()

While 1
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitOpenGLwindow() Then GUI_EVENT_CLOSE()
	If Not CreateResources() Then GUI_EVENT_CLOSE()
EndFunc


Func InitOpenGLwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth-2, $iGuiHeight-2, 400, 100 )

	; Statusbar
	Local $idStatus
	$idStatus = GUICtrlCreateLabel( "", 0, $iHeight - 1, $iGuiWidth + 2, $iStatusBarHeight, $SS_ETCHEDFRAME )
	$idLabel = GUICtrlCreateLabel( "", 3, $iHeight - 1 + 6, $iGuiWidth - 6, $iStatusBarHeight - 10 )
	GUICtrlSetResizing( $idStatus, $GUI_DOCKSTATEBAR )
	GUICtrlSetResizing( $idLabel, $GUI_DOCKSTATEBAR )

	GUISetOnEvent( $GUI_EVENT_CLOSE, "GUI_EVENT_CLOSE" )

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If $CmdLine[0] = 0 Then
		If Not GetOpenGLContext( $hOpenGL, $hDC, $hRC ) Then Return 0
		GUICtrlSetData( $idLabel, "Homer Simpson" )
	Else
		Switch $CmdLine[1]
			Case 1
				If Not OpenGLSimpleContext( $hOpenGL, $hDC, $hRC ) Then Return 0
				GUICtrlSetData( $idLabel, "Context (simple), highest supported version" )
			Case 2
				If Not OpenGLProperContext( $hOpenGL, $hDC, $hRC, -1, -1 ) Then Return 0
				GUICtrlSetData( $idLabel, "Context (proper), highest supported version" )
			Case 3
				If Not OpenGLProperContext( $hOpenGL, $hDC, $hRC, 3, 3 ) Then Return 0
				GUICtrlSetData( $idLabel, "Context (proper) version 3.3" )
			Case 4
				; Proper context, version 3.2
				If Not OpenGLProperContext( $hOpenGL, $hDC, $hRC, 3, 2 ) Then Return 0
				GUICtrlSetData( $idLabel, "Context (proper) version 3.2" )
			Case 5
				; Proper context, version 3.1
				If Not OpenGLProperContext( $hOpenGL, $hDC, $hRC, 3, 1 ) Then Return 0
				GUICtrlSetData( $idLabel, "Context (proper) version 3.1" )
			Case 6
				; Proper context, version 3.0
				If Not OpenGLProperContext( $hOpenGL, $hDC, $hRC, 3, 0 ) Then Return 0
				GUICtrlSetData( $idLabel, "Context (proper) version 3.0" )
			Case 7
				; Proper context, version 2.1
				If Not OpenGLProperContext( $hOpenGL, $hDC, $hRC, 2, 1 ) Then Return 0
				GUICtrlSetData( $idLabel, "Context (proper) version 2.1" )
			Case 8
				; Generic context, version 1.1
				If Not OpenGLProperContext( $hOpenGL, $hDC, $hRC, 1, 1 ) Then Return 0
				GUICtrlSetData( $idLabel, "Context (generic) version 1.1" )
		EndSwitch
	EndIf
	GUISetState()

	glViewport( 0, 0, $iWidth, $iHeight )
	glClearColor( 0.0, 0.0, 0.0, 0.0 )

	GUISwitch( $hGui )

	Return 1

EndFunc


Func CreateResources()
	Local $w, $h
	If Not LoadTextureImage( $iTexture, $images & "\Homer Simpson\Homer Simpson.jpg", $w, $h ) Then Return 0
	glBindTexture( $GL_TEXTURE_2D, $iTexture )

	$iCallList = glGenLists( 1 )
	glNewList( $iCallList, $GL_COMPILE )
		glBegin( $GL_QUADS )
			glTexCoord2f(  0.0,  0.0 )
			glVertex2f  ( -1.0, -1.0 )
			glTexCoord2f(  1.0,  0.0 )
			glVertex2f  ( +1.0, -1.0 )
			glTexCoord2f(  1.0,  1.0 )
			glVertex2f  ( +1.0, +1.0 )
			glTexCoord2f(  0.0,  1.0 )
			glVertex2f  ( -1.0, +1.0 )
		glEnd()
	glEndList()

	GUISetState()

	Return 1
EndFunc


Func DisplayFunction()
	glClear( $GL_COLOR_BUFFER_BIT )
	glCallList( $iCallList )
	SwapBuffers( $hDC )
EndFunc


Func GUI_EVENT_CLOSE()
	glDeleteLists( $iCallList, 1 )
	glDeleteTextures( 1, $iTexture )
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
