#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>
#include <EditConstants.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- OpenGL globals ---

Global $iTexture1, $iTexture2, $iCallList

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 400, $iHeight = 300

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30
Global Const $iStatusBarHeight = 28

Global $hStatusbar, $idRadio1, $idRadio2

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight + $iStatusBarHeight

Global $hGui, $bResize = True

; --- Project includes ---

#include "utilities\includes.au3"
#include "utilities\images.au3"

; --- Images ---

Global $bImage1 = True
Global $sImage1 = $images & "\Modern OpenGL\hello1.tga" ; Sunrise
Global $sImage2 = $images & "\Modern OpenGL\hello2.tga" ; Sunset

; --- Program code ---

MainFunc()

While 1
	If $bResize Then
		; An OpenGL window (viewport) should not be resized in the middle of a drawing cycle.
		; In GUI_EVENT_RESIZE() $bResize is set to True when the GUI is resized. At this point
		; when the DisplayFunction() isn't running it's safe to resize the OpenGL window.
		glViewport( 0, 0, $iWidth, $iHeight )
		$bResize = False
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then GUI_EVENT_CLOSE()
	If Not InitOpenGLwindow() Then GUI_EVENT_CLOSE()
	If Not CreateResources() Then GUI_EVENT_CLOSE()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 100, $WS_OVERLAPPEDWINDOW )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	GUIRegisterMsg( $WM_SIZING, "WM_SIZING" )

	CreateStatusbar()

	Return 1

EndFunc


Func CreateStatusbar()

	; Use a child window as a container for controls in the Statusbar
	$hStatusbar = GUICreate( "", $iGuiWidth - 4 , 24, 2, $iHeight + 3 , $WS_CHILD, -1, $hGui )

	; Radio buttons
	GUIStartGroup()
	$idRadio1 = GUICtrlCreateRadio( "Sunrise", 5, 3, 55, 18 )
	$idRadio2 = GUICtrlCreateRadio( "Sunset", 70, 3, 55, 18 )
	GUICtrlSetOnEvent( $idRadio1, "RadioClick" )
	GUICtrlSetOnEvent( $idRadio2, "RadioClick" )
	GUICtrlSetState( $idRadio1, $GUI_CHECKED )

	GUISetState()
	GUISwitch( $hGui )

	; Label for $hStatusbar to make it look nicer
	GUICtrlCreateLabel( "", 0, $iHeight + 1, $iGuiWidth + 2, $iStatusBarHeight, $SS_ETCHEDFRAME )
	GUICtrlSetResizing( -1, $GUI_DOCKSTATEBAR )

EndFunc


Func RadioClick()
	If @GUI_CTRLID = $idRadio1 Then
		$bImage1 = True
	ElseIf @GUI_CTRLID = $idRadio2 Then
		$bImage1 = False
	EndIf
EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC, 0 ) = 0 Then Return 0
	GUISetState()

	glClearColor( 0.0, 0.0, 0.0, 0.0 )

	GUISwitch( $hGui )

	Return 1

EndFunc


Func CreateResources()

	Local $w, $h
	If Not LoadTextureImage( $iTexture1, $sImage1, $w, $h ) Then Return 0
	If Not LoadTextureImage( $iTexture2, $sImage2, $w, $h ) Then Return 0

	$iCallList = glGenLists( 1 )
	glNewList( $iCallList, $GL_COMPILE )
		glBegin( $GL_QUADS )
			glTexCoord2f(  0.0,  0.0 )
			glVertex2f  ( -1.0, -1.0 )
			glTexCoord2f(  1.0,  0.0 )
			glVertex2f  ( +1.0, -1.0 )
			glTexCoord2f(  1.0,  1.0 )
			glVertex2f  ( +1.0, +1.0 )
			glTexCoord2f(  0.0,  1.0 )
			glVertex2f  ( -1.0, +1.0 )
		glEnd()
	glEndList()

	GUISetState()

	Return 1

EndFunc


Func DisplayFunction()
	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )

	If $bImage1 Then
		glBindTexture( $GL_TEXTURE_2D, $iTexture1 )
	Else
		glBindTexture( $GL_TEXTURE_2D, $iTexture2 )
	EndIf

	glCallList( $iCallList )

	SwapBuffers( $hDC )
EndFunc


Func WM_SIZING( $hWnd, $iMsg, $iwParam, $ilParam )
	#forceref $iMsg, $iwParam
	Switch $hWnd
		Case $hGui
			Local $tRECT = DllStructCreate( $tagRECT, $ilParam )
			Local $y = DllStructGetData( $tRECT, "Top" )
			Local $iHeight = DllStructGetData( $tRECT, "Bottom" ) - $y - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight
			WinMove( $hStatusbar, "", 2, $iHeight + $iHeightXP )
	EndSwitch
	Return $GUI_RUNDEFMSG
EndFunc


Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	WinMove( $hStatusbar, "", 2, $iHeight + $iHeightXP )
	$bResize = True ; The OpenGL window (viewport) is resized in top of the main loop
EndFunc


Func GUI_EVENT_CLOSE()
	glDeleteLists( $iCallList, 1 )
	glDeleteTextures( 1, $iTexture1 )
	glDeleteTextures( 1, $iTexture2 )
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
