#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>
#include <EditConstants.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- OpenGL globals ---

Global $iTexture, $iCallList

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 740, $iHeight = 510, $w, $h

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30
Global Const $iStatusBarHeight = 28

Global $hStatusbar, $idRadio1, $idRadio2, $idRadio3, $idRadio4, $idRadio5
Global $idRadio6, $idRadio7, $idRadio8, $idRadio9, $idRadio10, $idRadio11, $idRadio12
Global $sImage = "", $sNewImage = "bmp"

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight + $iStatusBarHeight

Global $hGui, $bResize = True

; --- Project includes ---

#include "utilities\includes.au3"
#include "utilities\images.au3"

; --- Program code ---

MainFunc()

While 1
	If $sImage <> $sNewImage Then
		$sImage = $sNewImage
		Switch $sImage
			Case "bmp"
				LoadTextureImage( $iTexture, $images & "\Homer Simpson\Homer Simpson.bmp", $w, $h )
				glBindTexture( $GL_TEXTURE_2D, $iTexture )
				$bResize = True
			Case "gif"
				LoadTextureImage( $iTexture, $images & "\Homer Simpson\Homer Simpson.gif", $w, $h )
				glBindTexture( $GL_TEXTURE_2D, $iTexture )
				$bResize = True
			Case "ico"
				LoadTextureImage( $iTexture, $images & "\right.ico", $w, $h )
				glBindTexture( $GL_TEXTURE_2D, $iTexture )
				$bResize = True
				$w *= 20
				$h *= 20
			Case "jpg"
				LoadTextureImage( $iTexture, $images & "\WikiBooks\Earthmap720x360_grid.jpg", $w, $h )
				glBindTexture( $GL_TEXTURE_2D, $iTexture )
				$bResize = True
			Case "png"
				LoadTextureImage( $iTexture, $images & "\WikiBooks\Land_shallow_topo_alpha_2048.png", $w, $h )
				glBindTexture( $GL_TEXTURE_2D, $iTexture )
				$bResize = True
				$w = 512
				$h = 256
			Case "rgb"
				$w = 640
				$h = 299
				LoadTextureImage( $iTexture, $images & "\Homer Simpson\Homer Simpson.rgb", $w, $h )
				glBindTexture( $GL_TEXTURE_2D, $iTexture )
				$bResize = True
			Case "rgba"
				$w = 256
				$h = 256
				LoadTextureImage( $iTexture, $images & "\Tunnels\metalframe.rgba", $w, $h )
				glBindTexture( $GL_TEXTURE_2D, $iTexture )
				$bResize = True
			Case "tga"
				LoadTextureImage( $iTexture, $images & "\Modern OpenGL\hello1.tga", $w, $h )
				glBindTexture( $GL_TEXTURE_2D, $iTexture )
				$bResize = True
			Case "tga (rle)"
				LoadTextureImage( $iTexture, $images & "\Tunnels\metalframe(rle).tga", $w, $h )
				glBindTexture( $GL_TEXTURE_2D, $iTexture )
				$bResize = True
			Case "tif"
				LoadTextureImage( $iTexture, $images & "\Homer Simpson\Homer Simpson.tif", $w, $h )
				glBindTexture( $GL_TEXTURE_2D, $iTexture )
				$bResize = True
			Case "url"
				Local $image = "http://aut1.autoit-cdn.com/site/wp-content/themes/TheCorporation/images/custom/featured-autoit.jpg"
				LoadTextureImage( $iTexture, $image, $w, $h )
				glBindTexture( $GL_TEXTURE_2D, $iTexture )
				$bResize = True
			Case "wmf"
				LoadTextureImage( $iTexture, $images & "\firewrks.wmf", $w, $h )
				glBindTexture( $GL_TEXTURE_2D, $iTexture )
				$bResize = True
				$w = 480
				$h = 480
		EndSwitch
	EndIf
	If $bResize Then
		; An OpenGL window (viewport) should not be resized in the middle of a drawing cycle.
		; In GUI_EVENT_RESIZE() $bResize is set to True when the GUI is resized. At this point
		; when the DisplayFunction() isn't running it's safe to resize the OpenGL window.
		glViewport( ( $iWidth - $w ) / 2, ( $iHeight - $h ) / 2, $w, $h )
		$bResize = False
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then GUI_EVENT_CLOSE()
	If Not InitOpenGLwindow() Then GUI_EVENT_CLOSE()
	If Not CreateResources() Then GUI_EVENT_CLOSE()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 100, $WS_OVERLAPPEDWINDOW )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	GUIRegisterMsg( $WM_GETMINMAXINFO, "WM_GETMINMAXINFO" )
	GUIRegisterMsg( $WM_SIZING,        "WM_SIZING" )

	CreateStatusbar()

	Return 1

EndFunc


Func CreateStatusbar()

	; Use a child window as a container for controls in the Statusbar
	$hStatusbar = GUICreate( "", $iGuiWidth - 4 , 24, 2, $iHeight + 3 , $WS_CHILD, -1, $hGui )

	; Radio buttons
	GUIStartGroup()
	$idRadio1  = GUICtrlCreateRadio( "BMP",        5, 3, 45, 18 )
	$idRadio2  = GUICtrlCreateRadio( "GIF",       60, 3, 45, 18 )
	$idRadio3  = GUICtrlCreateRadio( "ICO",      110, 3, 45, 18 )
	$idRadio4  = GUICtrlCreateRadio( "JPG",      160, 3, 45, 18 )
	$idRadio5  = GUICtrlCreateRadio( "PNG",      215, 3, 45, 18 )
	$idRadio6  = GUICtrlCreateRadio( "RGB",      270, 3, 45, 18 )
	$idRadio7  = GUICtrlCreateRadio( "RGBA",     325, 3, 50, 18 )
	$idRadio8  = GUICtrlCreateRadio( "TGA",      385, 3, 45, 18 )
	$idRadio9  = GUICtrlCreateRadio( "TGA (RLE)",440, 3, 75, 18 )
	$idRadio10 = GUICtrlCreateRadio( "TIF",      525, 3, 45, 18 )
	$idRadio11 = GUICtrlCreateRadio( "URL",      575, 3, 45, 18 )
	$idRadio12 = GUICtrlCreateRadio( "WMF",      630, 3, 45, 18 )
	GUICtrlSetOnEvent( $idRadio1,  "RadioClick" )
	GUICtrlSetOnEvent( $idRadio2,  "RadioClick" )
	GUICtrlSetOnEvent( $idRadio3,  "RadioClick" )
	GUICtrlSetOnEvent( $idRadio4,  "RadioClick" )
	GUICtrlSetOnEvent( $idRadio5,  "RadioClick" )
	GUICtrlSetOnEvent( $idRadio6,  "RadioClick" )
	GUICtrlSetOnEvent( $idRadio7,  "RadioClick" )
	GUICtrlSetOnEvent( $idRadio8,  "RadioClick" )
	GUICtrlSetOnEvent( $idRadio9,  "RadioClick" )
	GUICtrlSetOnEvent( $idRadio10, "RadioClick" )
	GUICtrlSetOnEvent( $idRadio11, "RadioClick" )
	GUICtrlSetOnEvent( $idRadio12, "RadioClick" )
	GUICtrlSetState( $idRadio1, $GUI_CHECKED )

	GUISetState()
	GUISwitch( $hGui )

	; Label for $hStatusbar to make it look nicer
	GUICtrlCreateLabel( "", 0, $iHeight + 1, $iGuiWidth + 4, $iStatusBarHeight, $SS_ETCHEDFRAME )
	GUICtrlSetResizing( -1, $GUI_DOCKSTATEBAR )

EndFunc


Func RadioClick()
	Switch @GUI_CTRLID
		Case $idRadio1
			$sNewImage = "bmp"
		Case $idRadio2
			$sNewImage = "gif"
		Case $idRadio3
			$sNewImage = "ico"
		Case $idRadio4
			$sNewImage = "jpg"
		Case $idRadio5
			$sNewImage = "png"
		Case $idRadio6
			$sNewImage = "rgb"
		Case $idRadio7
			$sNewImage = "rgba"
		Case $idRadio8
			$sNewImage = "tga"
		Case $idRadio9
			$sNewImage = "tga (rle)"
		Case $idRadio10
			$sNewImage = "tif"
		Case $idRadio11
			$sNewImage = "url"
		Case $idRadio12
			$sNewImage = "wmf"
	EndSwitch
EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 0.8, 0.8, 0.8, 0.0 )

	glBlendFunc( $GL_SRC_ALPHA, $GL_ONE_MINUS_SRC_ALPHA )
	glEnable( $GL_BLEND )

	GUISwitch( $hGui )

	Return 1

EndFunc


Func CreateResources()

	$iCallList = glGenLists( 1 )
	glNewList( $iCallList, $GL_COMPILE )
		glColor3f( 1.0, 1.0, 1.0 )
		glBegin( $GL_QUADS )
			glTexCoord2f(  0.0,  0.0 )
			glVertex2f  ( -1.0, -1.0 )
			glTexCoord2f(  1.0,  0.0 )
			glVertex2f  ( +1.0, -1.0 )
			glTexCoord2f(  1.0,  1.0 )
			glVertex2f  ( +1.0, +1.0 )
			glTexCoord2f(  0.0,  1.0 )
			glVertex2f  ( -1.0, +1.0 )
		glEnd()
	glEndList()

	GUISetState()

	Return 1

EndFunc


Func DisplayFunction()
	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )
	glCallList( $iCallList )
	SwapBuffers( $hDC )
EndFunc


Func WM_GETMINMAXINFO( $hWnd, $iMsg, $iwParam, $ilParam )
	#forceref $iMsg, $iwParam
	Switch $hWnd
		Case $hGui
			Local $tMINMAXINFO = DllStructCreate( "int;int;int;int;int;int;int;int;int;int", $ilParam )
			DllStructSetData( $tMINMAXINFO, 7, 740 + 2 * $iBorderWidth )																		; Min width
			DllStructSetData( $tMINMAXINFO, 8, 510 + $iTitleBarHeight + $iBorderWidth + $iStatusBarHeight )	; Min height
	EndSwitch
	Return $GUI_RUNDEFMSG
EndFunc


Func WM_SIZING( $hWnd, $iMsg, $iwParam, $ilParam )
	#forceref $iMsg, $iwParam
	Switch $hWnd
		Case $hGui
			Local $tRECT = DllStructCreate( $tagRECT, $ilParam )
			Local $y = DllStructGetData( $tRECT, "Top" )
			Local $iHeight = DllStructGetData( $tRECT, "Bottom" ) - $y - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight
			WinMove( $hStatusbar, "", 2, $iHeight + $iHeightXP )
	EndSwitch
	Return $GUI_RUNDEFMSG
EndFunc


Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	WinMove( $hStatusbar, "", 2, $iHeight + $iHeightXP )
	$bResize = True ; The OpenGL window (viewport) is resized in top of the main loop
EndFunc


Func GUI_EVENT_CLOSE()
	glDeleteLists( $iCallList, 1 )
	glDeleteTextures( 1, $iTexture )
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
