; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

Global $BlockCountX =  5 ; Number of buildings, x direction, 5
Global $BlockCountY = 30 ; Number of buildings, y direction, 10 - 30
Global $BlockCountZ =  3 ; Height of buildings, number of floors, 1 - 5

Global $aBuildings[1], $iMaxBuildings, $iBuildings = 0

; --- OpenGL globals ---

Global $GroundTexture, $GroundCallList[$BlockCountX]
Global $BuildingTexture, $BuildingsCallList[$BlockCountX]
Global $iLists = 0

; --- Statusbar ---

Global $iStatusBarRows = 1
Global $idSpeed, $iSpeed = 3, $minSteps = 0, $iSteps = 0

; --- Project includes ---

#include "utilities\utilities1.au3"

; --- Program code ---

MainFunc()

While 1
	If $fResize Then
		ResizeOpenGLwindow( 0, 0, $iWidth, $iHeight, 45.0, 0.1, 100.0 )
		$fResize = False
	EndIf
	If $fRebuild Then _
		RebuildCity()
	DisplayFunction()
	Sleep(10)
WEnd

Func MainFunc()
	If Not InitOpenGLwindow() Then GUI_EVENT_CLOSE()
	If Not CreateResources() Then GUI_EVENT_CLOSE()
EndFunc

Func CreateResources()
	Local $w = 64, $h = 64
	If Not LoadTextureImage( $GroundTexture, $images & "\Foggy city\Foggy city 5.rgb", $w, $h ) Then Return 0
	If Not LoadTextureImage( $BuildingTexture, $images & "\Foggy city\Foggy city 1.rgb", $w, $h ) Then Return 0
	If Not GenerateCallLists() Then Return 0
	If Not CalculateHeights() Then Return 0
	Return 1
EndFunc

Func GenerateCallLists()
	Local $iLists
	$iLists = glGenLists( $BlockCountX )
	For $i = 0 To $BlockCountX - 1
		$GroundCallList[$i] = $iLists + $i
	Next
	$iLists = glGenLists( $BlockCountX )
	For $i = 0 To $BlockCountX - 1
		$BuildingsCallList[$i] = $iLists + $i
	Next
	Return 1
EndFunc

Func CalculateHeights()
	Local $iHeight
	$iLists = 0
	$iBuildings = 0
	$iMaxBuildings = $BlockCountX * $BlockCountY
	ReDim $aBuildings[$iMaxBuildings]
	For $i = 0 To $iMaxBuildings - 1
		$iHeight = Random( 0.0, 1.0 )											; Building height, number of blocks or floors
		$iHeight = Round( ( $iHeight^3 ) * $BlockCountZ )	; Not too many high buildings and not too small distance between buildings for a small number of floors
		$aBuildings[$i] = $iHeight												; If $iHeight > 0 it's a building, else it's ground with space for streets and squares
	Next
	Return 1
EndFunc

Func RebuildCity()
	$BlockCountY = $newBlockCountY
	$BlockCountZ = $newBlockCountZ
	$minSteps = 2 * ( 3 - $iSpeed )
	$iSteps = $minSteps
	CalculateHeights()
	$fRebuild = False
EndFunc

Func CreateGround( $CallList, $minBuildings, $maxBuildings, _
								 $StartPosX = 0, $StartPosY = 0, $StartPosZ = 0, _
								 $BlockLength = 4, $BlockHeight = 2.5, $BlockWidth = 3, _
								 $BlockCountX = 20, $BlockCountY = 40, $BlockCountZ = 5 )
	$BlockWidth = -$BlockWidth

	Local $x1 = $StartPosX, $x2 = $x1 + $BlockLength			; $x1, $x2 = left, right of a building
	Local $y1 = $StartPosY, $y2 = $y1 + $BlockHeight			; $y1, $y2 = bottom, top of a block/floor
	Local $z1 = $StartPosZ, $z2 = $z1 + $BlockWidth				; $z1, $z2 = front, back of a building
	Local $iHeight, $iBuilding = 0

	glNewList( $CallList, $GL_COMPILE )										; Create the $CallList used for this row of buildings, it's given by $iLists

	For $i = 1 To $BlockCountX														; For $i = 1 the leftmost row of buildings in the city are created
		$z1 = $StartPosZ																		; For $i = $BlockCountX the rightmost row of buildings in the city are created
		$z2 = $z1 + $BlockWidth

		For $j = 1 To $BlockCountY													; For $j = 1 the frontmost building in the row given by $i is created
			$y1 = $StartPosY																	; For $j = $BlockCountY the backmost building in the row given by $i is created
			$y2 = $y1 + $BlockHeight

			$iHeight = $aBuildings[$iBuilding]								; Building height, number of blocks or floors

			; In this If-statement inside the 2 For-loops a row
			; of buildings is created and added to the $CallList.
			If $iBuilding >= $minBuildings Then								; $minBuildings is calculated as $iLists * $BlockCountY, $iLists starts with 0
																												; $minBuildings is 0 or equals one or more full rows of buildings
				If $iHeight = 0 Then
					glBegin( $GL_QUADS )													; If $iHeight = 0 space is created between the buildings for streets and squares
						glNormal3f( 0, 1, 0 )
						glTexCoord2f( 0, 0 )
						glVertex3f( $x2, $y1, $z1 )
						glTexCoord2f( 0, 1 )
						glVertex3f( $x2, $y1, $z2 )
						glTexCoord2f( 1, 1 )
						glVertex3f( $x1, $y1, $z2 )
						glTexCoord2f( 1, 0 )
						glVertex3f( $x1, $y1, $z1 )
					glEnd()
				EndIf

			EndIf

			$iBuilding += 1																		; Increase the number of buildings
			If $iBuilding = $maxBuildings Then ExitLoop				; $maxBuildings is increased by one at a time in DisplayFunction()

			$z1 += $BlockWidth																; The front position is increased with the block width for the next building
			$z2 += $BlockWidth																; The back position is increased with the block width for the next building
		Next

		If $iBuilding = $maxBuildings Then ExitLoop

		$x1 += $BlockLength																	; The left position is increased with the block length for the next row of buildings
		$x2 += $BlockLength																	; The right position is increased with the block length for the next row of buildings
	Next

	glEndList()

	Return 1

EndFunc

; Create buildings of a specific type
; The OpenGl commands are saved in $CallList
Func CreateBuildings( $CallList, $minBuildings, $maxBuildings, _
								 $StartPosX = 0, $StartPosY = 0, $StartPosZ = 0, _
								 $BlockLength = 4, $BlockHeight = 2.5, $BlockWidth = 3, _
								 $BlockCountX = 20, $BlockCountY = 40, $BlockCountZ = 5 )
	$BlockWidth = -$BlockWidth

	Local $x1 = $StartPosX, $x2 = $x1 + $BlockLength			; $x1, $x2 = left, right of a building
	Local $y1 = $StartPosY, $y2 = $y1 + $BlockHeight			; $y1, $y2 = bottom, top of a block/floor
	Local $z1 = $StartPosZ, $z2 = $z1 + $BlockWidth				; $z1, $z2 = front, back of a building
	Local $iHeight, $iBuilding = 0

	glNewList( $CallList, $GL_COMPILE )

	For $i = 1 To $BlockCountX														; For $i = 1 the leftmost row of buildings in the city are created
		$z1 = $StartPosZ																		; For $i = $BlockCountX the rightmost row of buildings in the city are created
		$z2 = $z1 + $BlockWidth

		For $j = 1 To $BlockCountY													; For $j = 1 the frontmost building in the row given by $i is created
			$y1 = $StartPosY																	; For $j = $BlockCountY the backmost building in the row given by $i is created
			$y2 = $y1 + $BlockHeight

			$iHeight = $aBuildings[$iBuilding]								; Building height, number of blocks or floors

			; In this If-statement inside the 2 For-loops a row
			; of buildings is created and added to the $CallList.
			If $iBuilding >= $minBuildings Then								; $minBuildings is calculated as $iLists * $BlockCountY, $iLists starts with 0
																												; $minBuildings is 0 or equals one or more full rows of buildings
				If $iHeight > 0 Then _
					CreateBuilding( $x1, $x2, $y1, $y2, $z1, $z2, $iHeight, $BlockHeight )

			EndIf

			$iBuilding += 1																		; Increase the number of buildings
			If $iBuilding = $maxBuildings Then ExitLoop				; $maxBuildings is increased by one at a time in DisplayFunction()

			$z1 += $BlockWidth																; The front position is increased with the block width for the next building
			$z2 += $BlockWidth																; The back position is increased with the block width for the next building
		Next

		If $iBuilding = $maxBuildings Then ExitLoop

		$x1 += $BlockLength																	; The left position is increased with the block length for the next row of buildings
		$x2 += $BlockLength																	; The right position is increased with the block length for the next row of buildings
	Next

	glEndList()

	Return 1

EndFunc

Func DisplayFunction()
	Local Static $FrameRate = 1000 / 30, $Timer = TimerInit()

	If TimerDiff( $Timer ) >= $FrameRate Then

		; In this If-statement the city is builded
		; - by one row of buildings at a time
		; - and by one building/groundspace at a time
		If $iBuildings < $iMaxBuildings Then
			If $iSteps = $minSteps Then																; $minSteps controls how fast the city is builded
					$iBuildings += 1																			; $iBuildings is increased by one at a time
					$iLists = Floor( ( $iBuildings - 1 ) / $BlockCountY )	; Index in the arrays of CallLists
					Local $minBuildings = $iLists * $BlockCountY					; $minBuildings is 0 or equals one or more full rows of buildings
					; In the function definitions for these two functions the third parameter with the value $iBuildings is named $maxBuildings
					CreateGround( $GroundCallList[$iLists], $minBuildings, $iBuildings, 0, 0, 0, 4, 2.5, 3, $BlockCountX, $BlockCountY, $BlockCountZ )
					CreateBuildings( $BuildingsCallList[$iLists], $minBuildings, $iBuildings, 0, 0, 0, 4, 2.5, 3, $BlockCountX, $BlockCountY, $BlockCountZ )
					$iSteps = 0
			Else
				$iSteps += 1
			EndIf
		EndIf

		glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )

		glLoadIdentity()

		; Rotate city
		glRotatef( 25, 1, 0, 0 )

		; Move camera
		glTranslatef( -10, -20, -20 )

		; Light
		Local $curPos[4] = [ -10, -20, -20 ]
		glLightfv( $GL_LIGHT0, $GL_POSITION, $curPos )

		; Show city
		glBindTexture( $GL_TEXTURE_2D, $GroundTexture )		; Ground with streets and squares
		For $i = 0 To $iLists
			glCallList( $GroundCallList[$i] )
		Next

		glBindTexture( $GL_TEXTURE_2D, $BuildingTexture )	; Buildings
		For $i = 0 To $iLists
			glCallList( $BuildingsCallList[$i] )
		Next

		SwapBuffers( $hDC )
	
		$Timer = TimerInit()
	
	EndIf
EndFunc

Func GUI_EVENT_CLOSE()
	glDeleteLists( $GroundCallList[0], $BlockCountX )
	glDeleteLists( $BuildingsCallList[0], $BlockCountX )
	glDeleteTextures( 1, $BuildingTexture )
	glDeleteTextures( 1, $GroundTexture )
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc

; --- Statusbar ---

Func CreateStatusbar()
	; Use a child window as a container for controls in the Statusbar
	$hStatusbar = GUICreate( "", 500, 24, 2, $iHeight + 3, $WS_CHILD, -1, $hGui )

	; Number of buildings, y direction, 10 - 30
	GUICtrlCreateLabel( "Buildings, y direction:", 2, 5, 100, 14 )
	$idBlockCountY = GUICtrlCreateInput( "30", 105, 2, 36, 20 )
	GUICtrlSetStyle( $idBlockCountY, $ES_READONLY )
	GUICtrlSetBkColor( $idBlockCountY, 0xFFFFFF )
	GUICtrlCreateUpdown( $idBlockCountY )
	GUICtrlSetLimit( -1, 30, 10 )

	; Number of floors, 1 - 5
	GUICtrlCreateLabel( "Max floors:", 161, 5, 50, 14, $SS_LEFTNOWORDWRAP )
	$idBlockCountZ = GUICtrlCreateInput( "3", 216, 2, 36, 20 )
	GUICtrlSetStyle( $idBlockCountZ, $ES_READONLY )
	GUICtrlSetBkColor( $idBlockCountZ, 0xFFFFFF )
	GUICtrlCreateUpdown( $idBlockCountZ )
	GUICtrlSetLimit( -1, 5, 1 )

	; Build speed, 1 - 3
	GUICtrlCreateLabel( "Speed:", 272, 5, 35, 14 )
	$idSpeed = GUICtrlCreateInput( "3", 310, 2, 36, 20 )
	GUICtrlSetStyle( $idSpeed, $ES_READONLY )
	GUICtrlSetBkColor( $idSpeed, 0xFFFFFF )
	GUICtrlCreateUpdown( $idSpeed )
	GUICtrlSetLimit( -1, 3, 1 )

	; Rebuild
	$idRebuild = GUICtrlCreateButton( "Rebuild", 370, 2, 55, 20 )
	GUICtrlSetOnEvent( $idRebuild, "Rebuild" )

	GUISetState( @SW_SHOW, $hStatusbar )
	GUISwitch( $hGui )

	; The child window is put on top of a label to make it look nicer
	GUICtrlCreateLabel( "", 0, $iHeight + 1, $iGuiWidth + 2, $iStatusBarHeight, $SS_ETCHEDFRAME )
	GUICtrlSetResizing( -1, $GUI_DOCKSTATEBAR )
EndFunc

Func Rebuild()
	$newBlockCountY = GUICtrlRead( $idBlockCountY )
	$newBlockCountZ = GUICtrlRead( $idBlockCountZ )
	$iSpeed = GUICtrlRead( $idSpeed )
	$fRebuild = True
EndFunc
