; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

Global $BlockCountX = 20 ; Number of buildings, x direction, 20 - 100
Global $BlockCountY = 40 ; Number of buildings, y direction, 40 - 100
Global $BlockCountZ =  5 ; Height of buildings, number of floors, 1 - 8
Global $aBuildings[1]

; --- OpenGL globals ---

Global $GroundTexture, $GroundCallList
Global $BuildingTexture, $BuildingsCallList

; --- Statusbar ---

Global $iStatusBarRows = 2

; --- Project includes ---

#include "utilities\utilities1.au3"
#include "utilities\utilities2.au3"

; --- Program code ---

MainFunc()

While 1
	If $fResize Then
		ResizeOpenGLwindow( 0, 0, $iWidth, $iHeight, 45.0, 0.1, 100.0 )
		$fResize = False
	EndIf
	If $fRebuild Then _
		RebuildCity()
	DisplayFunction()
	Sleep(10)
WEnd

Func MainFunc()
	If Not InitOpenGLwindow() Then GUI_EVENT_CLOSE()
	If Not CreateResources() Then GUI_EVENT_CLOSE()
EndFunc

Func CreateResources()
	Local $w = 64, $h = 64
	If Not LoadTextureImage( $GroundTexture, $images & "\Foggy city\Foggy city 5.rgb", $w, $h ) Then Return 0
	If Not LoadTextureImage( $BuildingTexture, $images & "\Foggy city\Foggy city 1.rgb", $w, $h ) Then Return 0
	$BuildingsCallList = glGenLists( 1 )
	$GroundCallList = glGenLists( 1 )
	RebuildCity()
	Return 1
EndFunc

Func RebuildCity()
	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )
	SwapBuffers( $hDC )

	$BlockCountX = $newBlockCountX
	$BlockCountY = $newBlockCountY
	$BlockCountZ = $newBlockCountZ
	ReDim $aBuildings[$BlockCountX * $BlockCountY]

	; Create the ground between the buildings, it's used for streets and squares
	CreateGround( $GroundCallList, 0, 0, 0, 4, 2.5, 3, $BlockCountX, $BlockCountY, $BlockCountZ )

	; Create all buildings
	CreateAllBuildings( $BuildingsCallList, 0, 0, 0, 4, 2.5, 3, $BlockCountX, $BlockCountY, $BlockCountZ )

	$fRebuild = False
EndFunc

Func CreateGround( $CallList, _
									 $StartPosX = 0, $StartPosY = 0, $StartPosZ = 0, _
									 $BlockLength = 4, $BlockHeight = 2.5, $BlockWidth = 3, _
									 $BlockCountX = 20, $BlockCountY = 40, $BlockCountZ = 5 )
	$BlockWidth = -$BlockWidth

	Local $x1 = $StartPosX, $x2 = $x1 + $BlockLength			; $x1, $x2 = left, right of a building
	Local $y1 = $StartPosY, $y2 = $y1 + $BlockHeight			; $y1, $y2 = bottom, top of a block/floor
	Local $z1 = $StartPosZ, $z2 = $z1 + $BlockWidth				; $z1, $z2 = front, back of a building
	Local $iHeight, $iBuildingType, $iLight, $iBuilding = 0

	glNewList( $CallList, $GL_COMPILE )

	UpdateProgressBar( 0, $BlockCountX )

	For $i = 1 To $BlockCountX														; For $i = 1 the leftmost row of buildings in the city are created
		$z1 = $StartPosZ																		; For $i = $BlockCountX the rightmost row of buildings in the city are created
		$z2 = $z1 + $BlockWidth

		For $j = 1 To $BlockCountY													; For $j = 1 the frontmost building in the row given by $i is created
			$y1 = $StartPosY																	; For $j = $BlockCountY the backmost building in the row given by $i is created
			$y2 = $y1 + $BlockHeight

			$iHeight = Random( 0.0, 1.0 )											; Building height, number of blocks or floors
			$iHeight = Round( ( $iHeight^3 ) * $BlockCountZ )	; Not too many high buildings and not too small distance between buildings for a small number of floors

			$aBuildings[$iBuilding] = $iHeight								; Save the building to be created later
																												; If $iHeight > 0 it's a building
			If $iHeight = 0 Then															; If $iHeight = 0 space is created between the buildings for streets and squares
				glBegin( $GL_QUADS )
					glNormal3f( 0, 1, 0 )
					glTexCoord2f( 0, 0 )
					glVertex3f( $x2, $y1, $z1 )
					glTexCoord2f( 0, 1 )
					glVertex3f( $x2, $y1, $z2 )
					glTexCoord2f( 1, 1 )
					glVertex3f( $x1, $y1, $z2 )
					glTexCoord2f( 1, 0 )
					glVertex3f( $x1, $y1, $z1 )
				glEnd()
			EndIf

			$iBuilding += 1																		; Increase the number of buildings

			$z1 += $BlockWidth																; The front position is increased with the block width for the next building
			$z2 += $BlockWidth																; The back position is increased with the block width for the next building
		Next

		$x1 += $BlockLength																	; The left position is increased with the block length for the next row of buildings
		$x2 += $BlockLength																	; The right position is increased with the block length for the next row of buildings
	Next

	glEndList()

EndFunc

Func CreateAllBuildings( $CallList, _
												 $StartPosX = 0, $StartPosY = 0, $StartPosZ = 0, _
												 $BlockLength = 4, $BlockHeight = 2.5, $BlockWidth = 3, _
												 $BlockCountX = 20, $BlockCountY = 40, $BlockCountZ = 5 )
	$BlockWidth = -$BlockWidth

	Local $x1 = $StartPosX, $x2 = $x1 + $BlockLength			; $x1, $x2 = left, right of a building
	Local $y1 = $StartPosY, $y2 = $y1 + $BlockHeight			; $y1, $y2 = bottom, top of a block/floor
	Local $z1 = $StartPosZ, $z2 = $z1 + $BlockWidth				; $z1, $z2 = front, back of a building
	Local $iHeight, $iBuilding = 0

	glNewList( $CallList, $GL_COMPILE )

	For $i = 1 To $BlockCountX														; For $i = 1 the leftmost row of buildings in the city are created
		UpdateProgressBar( $i, $BlockCountX )								; For $i = $BlockCountX the rightmost row of buildings in the city are created

		$z1 = $StartPosZ
		$z2 = $z1 + $BlockWidth

		For $j = 1 To $BlockCountY													; For $j = 1 the frontmost building in the row given by $i is created
			$y1 = $StartPosY																	; For $j = $BlockCountY the backmost building in the row given by $i is created
			$y2 = $y1 + $BlockHeight

			$iHeight = $aBuildings[$iBuilding]								; Building height, number of blocks or floors

			If $iHeight Then _																; If $iHeight > 0 it's a building
				CreateBuilding( $x1, $x2, $y1, $y2, $z1, $z2, $iHeight, $BlockHeight )

			$iBuilding += 1																		; Increase the number of buildings

			$z1 += $BlockWidth																; The front position is increased with the block width for the next building
			$z2 += $BlockWidth																; The back position is increased with the block width for the next building
		Next

		$x1 += $BlockLength																	; The left position is increased with the block length for the next row of buildings
		$x2 += $BlockLength																	; The right position is increased with the block length for the next row of buildings
	Next

	glEndList()

EndFunc

Func DisplayCity()
	; Show city
	glBindTexture( $GL_TEXTURE_2D, $GroundTexture )		; Ground with streets and squares
	glCallList( $GroundCallList )

	glBindTexture( $GL_TEXTURE_2D, $BuildingTexture )	; Buildings
	glCallList( $BuildingsCallList )
EndFunc

Func GUI_EVENT_CLOSE()
	glDeleteLists( $GroundCallList, 1 )
	glDeleteLists( $BuildingsCallList, 1 )
	glDeleteTextures( 1, $BuildingTexture )
	glDeleteTextures( 1, $GroundTexture )
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
