; Globals and functions common to Foggy city examples (all examples)

#include-once

#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>
#include <EditConstants.au3>

; --- OpenGL globals ---

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 800, $iHeight = 600

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30
Global Const $iStatusBarHeight = 28

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight + $iStatusBarHeight * $iStatusBarRows

Global $hGui, $fResize = True

; --- Statusbar ---

Global $hStatusbar

Global $idBlockCountX, $idBlockCountY, $idBlockCountZ, $idRebuild, $idProgress
Global $newBlockCountX = $BlockCountX, $newBlockCountY = $BlockCountY, $newBlockCountZ = $BlockCountZ, $fRebuild = False

; --- Project includes ---

#include "includes.au3"
#include "images.au3"

Func InitOpenGLwindow()

	; --- GUI window ---

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 50, $WS_OVERLAPPEDWINDOW )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	GUIRegisterMsg( $WM_GETMINMAXINFO, "WM_GETMINMAXINFO" )
	GUIRegisterMsg( $WM_SIZING,        "WM_SIZING"        )

	CreateStatusbar()

	; --- OpenGL window ---

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 0.4, 0.4, 0.5, 0 )

	; --- OpenGL initializations ---

	Local $lAmbient[4] = [ 1, 1, 1, 1 ]       ; Ambientes light color
	Local $lDiffuse[4] = [ 0.4, 0.4, 0.5, 1 ] ; Diffuses light color

	glShadeModel( $GL_SMOOTH )

	glPolygonMode( $GL_FRONT_AND_BACK, $GL_FILL )

	glHint( $GL_PERSPECTIVE_CORRECTION_HINT, $GL_NICEST )

	; Depth buffer
	glClearDepth( 1 )
	glEnable( $GL_DEPTH_TEST )
	glDepthFunc( $GL_LESS )

	; Light
	glEnable( $GL_LIGHTING )
	glEnable( $GL_LIGHT0 )
	glLightfv( $GL_LIGHT0, $GL_AMBIENT, $lAmbient )
	glLightfv( $GL_LIGHT0, $GL_DIFFUSE, $lDiffuse )
	glLightf( $GL_LIGHT0, $GL_LINEAR_ATTENUATION, 0.003 )

	; Fog
	glEnable( $GL_FOG )
	glFogi( $GL_FOG_MODE, $GL_LINEAR )
	glFogfv( $GL_FOG_COLOR, $lDiffuse )
	glFogf( $GL_FOG_DENSITY, 0.03 )
	glHint( $GL_FOG_HINT, $GL_NICEST )
	glFogf( $GL_FOG_START, 10.0 )
	glFogf( $GL_FOG_END, 100.0 )

	GUISwitch( $hGui )
	GUISetState()

	Return 1

EndFunc

Func CreateBuilding( $x1, $x2, $y1, $y2, $z1, $z2, $iHeight, $BlockHeight )
	For $k = 1 To $iHeight								; For $k = 1 the bottom floor of the building given by $i and $j is created
		glBegin( $GL_QUADS )								; For $k = $iHeight the top floor of the building given by $i and $j is created
			; Front														; In the innermost loop a single building is created from bottom to top
			glNormal3f( 0, 0, 1 )							; If $iHeight = 0 there is created space for a street or square
			glTexCoord2f( 0, 0 )
			glVertex3f( $x2, $y2, $z1 )
			glTexCoord2f( 0, 1 )
			glVertex3f( $x1, $y2, $z1 )
			glTexCoord2f( 1, 1 )
			glVertex3f( $x1, $y1, $z1 )
			glTexCoord2f( 1, 0 )
			glVertex3f( $x2, $y1, $z1 )
			; Back
			glNormal3f( 0, 0, -1 )
			glTexCoord2f( 0, 0 )
			glVertex3f( $x2, $y2, $z2 )
			glTexCoord2f( 0, 1 )
			glVertex3f( $x1, $y2, $z2 )
			glTexCoord2f( 1, 1 )
			glVertex3f( $x1, $y1, $z2 )
			glTexCoord2f( 1, 0 )
			glVertex3f( $x2, $y1, $z2 )
			; Left
			glNormal3f( -1, 0, 0 )
			glTexCoord2f( 0, 0 )
			glVertex3f( $x1, $y2, $z1 )
			glTexCoord2f( 0, 1 )
			glVertex3f( $x1, $y2, $z2 )
			glTexCoord2f( 1, 1 )
			glVertex3f( $x1, $y1, $z2 )
			glTexCoord2f( 1, 0 )
			glVertex3f( $x1, $y1, $z1 )
			; Right
			glNormal3f( 1, 0, 0 )
			glTexCoord2f( 0, 0 )
			glVertex3f( $x2, $y2, $z1 )
			glTexCoord2f( 0, 1 )
			glVertex3f( $x2, $y2, $z2 )
			glTexCoord2f( 1, 1 )
			glVertex3f( $x2, $y1, $z2 )
			glTexCoord2f( 1, 0 )
			glVertex3f( $x2, $y1, $z1 )
		glEnd()

		$y1 += $BlockHeight									; The bottom position is increased with the block height for the next floor
		$y2 += $BlockHeight									; The top position is increased with the block height for the next floor
	Next

	glBegin( $GL_QUADS )									; If $iHeight > 0 the roof of the building is created
		glNormal3f( 0, 1, 0 )								; If $iHeight = 0 space is created for a street or square
		glVertex3f( $x2, $y1, $z1 )
		glVertex3f( $x2, $y1, $z2 )
		glVertex3f( $x1, $y1, $z2 )
		glVertex3f( $x1, $y1, $z1 )
	glEnd()
EndFunc

; --- Resizing ---

Func WM_SIZING( $hWnd, $iMsg, $iwParam, $ilParam )
	#forceref $iMsg, $iwParam
	Switch $hWnd
		Case $hGui
			Local $tRECT = DllStructCreate( $tagRECT, $ilParam )
			Local $y = DllStructGetData( $tRECT, "Top" )
			Local $iHeight = DllStructGetData( $tRECT, "Bottom" ) - $y - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight
			WinMove( $hStatusbar, "", 2, $iHeight + $iHeightXP )
	EndSwitch
	Return $GUI_RUNDEFMSG
EndFunc

Func WM_GETMINMAXINFO( $hWnd, $iMsg, $iwParam, $ilParam )
	#forceref $iMsg, $iwParam
	Switch $hWnd
		Case $hGui
			Local $tMINMAXINFO = DllStructCreate( "int;int;int;int;int;int;int;int;int;int", $ilParam )
			DllStructSetData( $tMINMAXINFO, 7, 800 + 2 * $iBorderWidth )																											; Min width
			DllStructSetData( $tMINMAXINFO, 8, 600 + $iTitleBarHeight + $iBorderWidth + $iStatusBarHeight * $iStatusBarRows )	; Min height
	EndSwitch
	Return $GUI_RUNDEFMSG
EndFunc

Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight * $iStatusBarRows
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	WinMove( $hStatusbar, "", 2, $iHeight + $iHeightXP + $iStatusBarHeight * ( $iStatusBarRows - 1 ) )
	$fResize = True
EndFunc
