; Globals and functions common to Foggy city examples (not "Build foggy city.au3")

#include-once

#include <Misc.au3>

Func DisplayFunction()
	Local Static $fMoveCamera = True
	Local Static $PosX = -40, $PosY = -20, $PosZ = 30, $RotX = 0, $RotY = 25, $Speed = 0.3
	Local Static $FrameRate = 1000 / 30, $Timer = TimerInit(), $Angle = 3.141592 / 2
	Local $fKeyPress = False

	If TimerDiff( $Timer ) >= $FrameRate Then

		If $fMoveCamera Then
			$PosX += ( Sin( $Angle ) * $Speed ) ; Camera x position
			$PosZ += ( Cos( $Angle ) * $Speed ) ; Camera z position
			$Angle += $Speed / 20
		EndIf

		; Stop/start camera with <End> key
		If _IsPressed( "23", $dllUser32 ) Then
			$fMoveCamera = Not $fMoveCamera
			$fKeyPress = True
		EndIf

		; Move camera with arrow keys
		If _IsPressed( "26", $dllUser32 ) Then $RotY -= 0.6 ; Up
		If _IsPressed( "28", $dllUser32 ) Then $RotY += 0.6 ; Down
		If _IsPressed( "25", $dllUser32 ) Then $RotX -= 0.6 ; Left
		If _IsPressed( "27", $dllUser32 ) Then $RotX += 0.6 ; Right

		; Initialize camera position with <Home> key
		If _IsPressed( "24", $dllUser32 ) Then
			$PosX = -40
			$PosY = -20
			$PosZ = 30
			$RotX = 0
			$RotY = 25
		EndIf

		glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )

		glLoadIdentity()

		; Rotate city
		glRotatef( $RotY, 1, 0, 0 )
		glRotatef( $RotX, 0, 1, 0 )

		; Move camera
		glTranslatef( $PosX, $PosY, $PosZ )

		; Light
		Local $curPos[4] = [ $PosX, $PosY, $PosZ ]
		glLightfv( $GL_LIGHT0, $GL_POSITION, $curPos )

		; Display ground and buildings
		DisplayCity()

		SwapBuffers( $hDC )

		; Wait until <End> is released
		If $fKeyPress Then
			While _IsPressed( "23", $dllUser32 )
			WEnd
		EndIf

		$Timer = TimerInit()

	EndIf
EndFunc

; --- Statusbar ---

Func CreateStatusbar()
	; Use a child window as a container for controls in the Statusbar
	$hStatusbar = GUICreate( "", 800, 24, 2, $iHeight + $iStatusBarHeight + 3, $WS_CHILD, -1, $hGui )

	; Number of buildings, x direction, 20 - 99
	GUICtrlCreateLabel( "Buildings, x direction:", 2, 5, 100, 14 )
	$idBlockCountX = GUICtrlCreateInput( "30", 105, 2, 36, 20 )
	GUICtrlSetStyle( $idBlockCountX, $ES_READONLY )
	GUICtrlSetBkColor( $idBlockCountX, 0xFFFFFF )
	GUICtrlCreateUpdown( $idBlockCountX )
	GUICtrlSetLimit( -1, 99, 20 )

	; Number of buildings, y direction, 40 - 99
	GUICtrlCreateLabel( "Buildings, y direction:", 161, 5, 100, 14 )
	$idBlockCountY = GUICtrlCreateInput( "40", 264, 2, 36, 20 )
	GUICtrlSetStyle( $idBlockCountY, $ES_READONLY )
	GUICtrlSetBkColor( $idBlockCountY, 0xFFFFFF )
	GUICtrlCreateUpdown( $idBlockCountY )
	GUICtrlSetLimit( -1, 99, 40 )

	; Number of floors, 1 - 8
	GUICtrlCreateLabel( "Max floors:", 320, 5, 50, 14, $SS_LEFTNOWORDWRAP )
	$idBlockCountZ = GUICtrlCreateInput( "5", 375, 2, 36, 20 )
	GUICtrlSetStyle( $idBlockCountZ, $ES_READONLY )
	GUICtrlSetBkColor( $idBlockCountZ, 0xFFFFFF )
	GUICtrlCreateUpdown( $idBlockCountZ )
	GUICtrlSetLimit( -1, 8, 1 )

	; Rebuild
	$idRebuild = GUICtrlCreateButton( "Rebuild", 431, 2, 55, 20 )
	GUICtrlSetOnEvent( $idRebuild, "Rebuild" )

	; Vertical splitterbar
	GUICtrlCreateLabel( "", 506, 2, 3, 20, $SS_ETCHEDFRAME )

	; Progressbar
	$idProgress = GUICtrlCreateProgress( 529, 2, 197, 20 )

	GUISetState( @SW_SHOW, $hStatusbar )
	GUISwitch( $hGui )

	; Label for $hStatusbar to make it look nicer
	GUICtrlCreateLabel( "", 0, $iHeight + 1, $iGuiWidth + 2, $iStatusBarHeight + 1, $SS_ETCHEDFRAME )
	GUICtrlSetResizing( -1, $GUI_DOCKSTATEBAR )
	GUICtrlCreateLabel( "<End>: Stop/start camera  |  <Home>: Initialize camera position  |  Arrow keys: Move camera", _
											3, $iHeight + 1 + 6, $iGuiWidth - 6, $iStatusBarHeight - 10 )
	GUICtrlSetResizing( -1, $GUI_DOCKSTATEBAR )
EndFunc

Func Rebuild()
	$newBlockCountX = GUICtrlRead( $idBlockCountX )
	$newBlockCountY = GUICtrlRead( $idBlockCountY )
	$newBlockCountZ = GUICtrlRead( $idBlockCountZ )
	$fRebuild = True
EndFunc

Func UpdateProgressBar( $n, $max )
	Local $pct = $n / $max * 100
	GUICtrlSetData( $idProgress, $pct )
EndFunc
