#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

Global $nSectors = 24
Global $nVertices = 1, $aVertices[$nVertices][5] ; x, y, z, u, v

; --- OpenGL globals ---

Global $iTexture, $iCallList

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 500, $iHeight = 500

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30
Global Const $iStatusBarHeight = 28

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight + $iStatusBarHeight

Global $hGui, $bResize = True

; --- Project includes ---

#include "utilities\images.au3"
#include "utilities\includes.au3"
#include "utilities\statusbar.au3"

; --- Program code ---

MainFunc()

While 1
	If $bNewSphere Then
		$nSectors = GUICtrlRead( $idSectors ) * 12
		CreateSphereAndCallList()
		$bNewSphere = False
	EndIf
	If $bResize Then
		ResizeOpenGLwindow( 0, 0, $iWidth, $iHeight, 60.0, 0.1, 100.0 )
		$bResize = False
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then GUI_EVENT_CLOSE()
	If Not InitOpenGLwindow() Then GUI_EVENT_CLOSE()
	If Not CreateResources() Then GUI_EVENT_CLOSE()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 50, $WS_OVERLAPPEDWINDOW )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	GUIRegisterMsg( $WM_SIZING, "WM_SIZING" )

	CreateStatusbar()

	Return 1

EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 0.0, 0.0, 0.0, 1.0 )

	glEnable( $GL_DEPTH_TEST )
	glDepthFunc( $GL_LEQUAL )

	glEnable( $GL_CULL_FACE )
	glFrontFace( $GL_CCW )
	glCullFace( $GL_BACK )

	GUISwitch( $hGui )
	GUISetState()

	Return 1

EndFunc


Func CreateResources()

	Local $w, $h
	If Not LoadTextureImage( $iTexture, $images & "\Spheres\earth.jpg", $w, $h ) Then Return 0
	glBindTexture( $GL_TEXTURE_2D, $iTexture )

	$iCallList = glGenLists( 1 )

	CreateSphereAndCallList()

	Return 1
EndFunc


Func CreateSphereAndCallList()
	Local $fDegrees = 360 / $nSectors
	$nVertices = ( 90 / $fDegrees + 1 ) * ( 360 / $fDegrees + 1 ) * 4
	ReDim $aVertices[$nVertices][5] ; x, y, z, u, v
	CreateSphere( 70, 0, 0, 0, $fDegrees )
	CreateCallList()
EndFunc


Func CreateSphere( $radius, $xCenter, $yCenter, $zCenter, $fDegrees )

	Local Const $PI = 3.14159265358979323846
	Local $n = -1

	for $i = 0 To 90 - $fDegrees Step $fDegrees

		for $j = 0 To 360 - $fDegrees Step $fDegrees

			$n += 1
			$aVertices[$n][0] = $radius * Sin( ( $j ) / 180 * $PI ) * Sin( ( $i ) / 180 * $PI ) - $xCenter
			$aVertices[$n][1] = $radius * Cos( ( $j ) / 180 * $PI ) * Sin( ( $i ) / 180 * $PI ) + $yCenter
			$aVertices[$n][2] = $radius * Cos( ( $i ) / 180 * $PI ) - $zCenter
			$aVertices[$n][3] = ( $j ) / 360
			$aVertices[$n][4] = ( 2 * $i ) / 360

			$n += 1
			$aVertices[$n][0] = $radius * Sin( ( $j ) / 180 * $PI ) * Sin( ( $i + $fDegrees ) / 180 * $PI ) - $xCenter
			$aVertices[$n][1] = $radius * Cos( ( $j ) / 180 * $PI ) * Sin( ( $i + $fDegrees ) / 180 * $PI ) + $yCenter
			$aVertices[$n][2] = $radius * Cos( ( $i + $fDegrees ) / 180 * $PI ) - $zCenter
			$aVertices[$n][3] = ( $j ) / 360
			$aVertices[$n][4] = ( 2 * ( $i + $fDegrees ) ) / 360

			$n += 1
			$aVertices[$n][0] = $radius * Sin( ( $j + $fDegrees ) / 180 * $PI ) * Sin( ( $i ) / 180 * $PI ) - $xCenter
			$aVertices[$n][1] = $radius * Cos( ( $j + $fDegrees ) / 180 * $PI ) * Sin( ( $i ) / 180 * $PI ) + $yCenter
			$aVertices[$n][2] = $radius * Cos( ( $i ) / 180 * $PI ) - $zCenter
			$aVertices[$n][3] = ( $j + $fDegrees ) / 360
			$aVertices[$n][4] = ( 2 * $i ) / 360

			$n += 1
			$aVertices[$n][0] = $radius * Sin( ( $j + $fDegrees ) / 180 * $PI ) * Sin( ( $i + $fDegrees ) / 180 * $PI ) - $xCenter
			$aVertices[$n][1] = $radius * Cos( ( $j + $fDegrees ) / 180 * $PI ) * Sin( ( $i + $fDegrees ) / 180 * $PI ) + $yCenter
			$aVertices[$n][2] = $radius * Cos( ( $i + $fDegrees ) / 180 * $PI ) - $zCenter
			$aVertices[$n][3] = ( $j + $fDegrees ) / 360
			$aVertices[$n][4] = ( 2 * ( $i + $fDegrees ) ) / 360

		Next

	Next

EndFunc


Func CreateCallList()
	glNewList( $iCallList, $GL_COMPILE )
		glBegin( $GL_TRIANGLE_STRIP )
			For $i = 0 To $nVertices - 1
				glTexCoord2f( -$aVertices[$i][3], $aVertices[$i][4] )
				glVertex3f( -$aVertices[$i][0], $aVertices[$i][1], $aVertices[$i][2] )
			Next
			For $i = 0 to $nVertices - 1
				glTexCoord2f( $aVertices[$i][3], -$aVertices[$i][4] )
				glVertex3f( $aVertices[$i][0], $aVertices[$i][1], -$aVertices[$i][2] )
			Next
		glEnd()
	glEndList()
EndFunc


Func DisplayFunction()
	Local Static $fAngle = 0.0
	Local $fScale = 5 * 0.0125

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )

	glLoadIdentity()
	glTranslatef( 0, 0, -10 )
	glRotatef( $fAngle, 0.25, 0.8, 0 )
	glScalef( $fScale, $fScale, $fScale )
	glRotatef( 90, 1, 0, 0 )

	glCallList( $iCallList )

	SwapBuffers( $hDC )

	$fAngle += 0.25
EndFunc


Func GUI_EVENT_CLOSE()
	glDeleteLists( $iCallList, 1 )
	glDeleteTextures( 1, $iTexture )
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
