#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

Global $nSectors = 24

; --- OpenGL globals ---

Global $iTexture, $pQuadObj

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 500, $iHeight = 500

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30
Global Const $iStatusBarHeight = 28

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight + $iStatusBarHeight

Global $hGui, $bResize = True

; --- Project includes ---

#include "utilities\images.au3"
#include "utilities\includes.au3"
#include "utilities\statusbar.au3"

; --- Program code ---

MainFunc()

While 1
	If $bNewSphere Then
		$nSectors = GUICtrlRead( $idSectors ) * 12
		$bNewSphere = False
	EndIf
	If $bResize Then
		ResizeOpenGLwindow( 0, 0, $iWidth, $iHeight, 60.0, 0.1, 100.0 )
		$bResize = False
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then GUI_EVENT_CLOSE()
	If Not InitOpenGLwindow() Then GUI_EVENT_CLOSE()
	If Not CreateResources() Then GUI_EVENT_CLOSE()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 50, $WS_OVERLAPPEDWINDOW )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	GUIRegisterMsg( $WM_SIZING, "WM_SIZING" )

	CreateStatusbar()

	Return 1

EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 0.0, 0.0, 0.0, 1.0 )

	glEnable( $GL_DEPTH_TEST )
	glDepthFunc( $GL_LEQUAL )

	glEnable( $GL_CULL_FACE )
	glFrontFace( $GL_CCW )
	glCullFace( $GL_BACK )

	GUISwitch( $hGui )
	GUISetState()

	Return 1

EndFunc


Func CreateResources()
	Local $w, $h
	If Not LoadTextureImage( $iTexture, $images & "\Spheres\earth.jpg", $w, $h ) Then Return 0
	glBindTexture( $GL_TEXTURE_2D, $iTexture )

	$pQuadObj = gluNewQuadric()
	gluQuadricTexture( $pQuadObj, $GL_TRUE )

	Return 1
EndFunc


Func DisplayFunction()
	Local Static $fAngle = 0.0
	Local $fScale = 5 * 0.0125

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )

	glLoadIdentity()
	glTranslatef( 0, 0, -10 )
	glRotatef( $fAngle, 0.25, 0.8, 0 )
	glScalef( $fScale, $fScale, $fScale )
	glRotatef( -90, 1, 0, 0 )

	gluQuadricDrawStyle( $pQuadObj, $GLU_FILL )
	gluSphere( $pQuadObj, 70, $nSectors, $nSectors )

	SwapBuffers( $hDC )

	$fAngle += 0.25
EndFunc


Func GUI_EVENT_CLOSE()
	gluDeleteQuadric( $pQuadObj )
	glDeleteTextures( 1, $iTexture )
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
