#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

; --- OpenGL globals ---

Global $iTexture1, $iTexture2, $iTexture3, $iCallList

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 600, $iHeight = 600

; --- GUI globals ---

Global Const $iBorderWidth		 =	4
Global Const $iTitleBarHeight	 = 30
Global Const $iStatusBarHeight = 28

Global $hStatusbar, $idRadio1, $idRadio2, $idRadio3, $idLabel
Global $bImage = 1, $bNewImage = 1

Global $iGuiWidth	= $iWidth
Global $iGuiHeight = $iHeight + $iStatusBarHeight

Global $hGui, $bResize = True

; --- Project includes ---

#include "utilities\includes.au3"
#include "utilities\images.au3"

; --- Program code ---

MainFunc()

While 1
	If $bNewImage <> $bImage Then
		Switch $bNewImage
			Case 1
				glBindTexture( $GL_TEXTURE_2D, $iTexture1 )
			Case 2
				glBindTexture( $GL_TEXTURE_2D, $iTexture2 )
			Case 3
				glBindTexture( $GL_TEXTURE_2D, $iTexture3 )
		EndSwitch
		$bImage = $bNewImage
	EndIf
	If $bResize Then
		ResizeOpenGLwindow( 0, 0, $iWidth, $iHeight, 45.0, 0.1, 500.0 )
		$bResize = False
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then GUI_EVENT_CLOSE()
	If Not InitOpenGLwindow() Then GUI_EVENT_CLOSE()
	If Not CreateResources() Then GUI_EVENT_CLOSE()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 50, $WS_OVERLAPPEDWINDOW )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,	"GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,	"GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,		"GUI_EVENT_CLOSE"	)

	GUIRegisterMsg( $WM_SIZING, "WM_SIZING" )

	CreateStatusbar()

	Return 1

EndFunc


Func CreateStatusbar()

	; Use a child window as a container for controls in the Statusbar
	$hStatusbar = GUICreate( "", $iGuiWidth - 4 , 24, 2, $iHeight + 3 , $WS_CHILD, -1, $hGui )

	; Radio buttons
	GUIStartGroup()
	$idRadio1 = GUICtrlCreateRadio( "TGA",        5, 3, 45, 18 )
	$idRadio2 = GUICtrlCreateRadio( "TGA (RLE)", 60, 3, 75, 18 )
	$idRadio3 = GUICtrlCreateRadio( "RGBA",     145, 3, 50, 18 )
	GUICtrlSetOnEvent( $idRadio1, "RadioClick" )
	GUICtrlSetOnEvent( $idRadio2, "RadioClick" )
	GUICtrlSetOnEvent( $idRadio3, "RadioClick" )
	GUICtrlSetState( $idRadio1, $GUI_CHECKED )

	; Vertical splitterbar
	GUICtrlCreateLabel( "", 207, 2, 3, 20, $SS_ETCHEDFRAME )

	; Label in statusbar
	$idLabel = GUICtrlCreateLabel( "", 222, 5, 230, 14 )
	GUICtrlSetData( $idLabel, "Tunnel effect with a transparent texture and fog." )

	GUISetState( @SW_SHOW, $hStatusbar )
	GUISwitch( $hGui )

	; Label for $hStatusbar to make it look nicer
	GUICtrlCreateLabel( "", 0, $iHeight + 1, $iGuiWidth + 4, $iStatusBarHeight, $SS_ETCHEDFRAME )
	GUICtrlSetResizing( -1, $GUI_DOCKSTATEBAR )

EndFunc


Func RadioClick()
	Switch @GUI_CTRLID
		Case $idRadio1
			$bNewImage = 1
			GUICtrlSetData( $idLabel, "Images\Tunnels\metalframe.tga" )
		Case $idRadio2
			$bNewImage = 2
			GUICtrlSetData( $idLabel, "Images\Tunnels\metalframe(rle).tga" )
		Case $idRadio3
			$bNewImage = 3
			GUICtrlSetData( $idLabel, "Images\Tunnels\metalframe.rgba" )
	EndSwitch
EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 0.8, 0.8, 0.9, 0.0 )

	glShadeModel( $GL_FLAT )

	glEnable( $GL_DEPTH_TEST )
	glDepthFunc( $GL_LESS )
	glClearDepth( 1.0 )

	glEnable( $GL_ALPHA_TEST )
	glAlphaFunc( $GL_GREATER, 0.4 )

	glEnable( $GL_BLEND )
	glBlendFunc( $GL_SRC_ALPHA, $GL_ONE_MINUS_SRC_ALPHA )

	glHint( $GL_PERSPECTIVE_CORRECTION_HINT, $GL_NICEST )

	Local $FogColor[4] = [ 0.9, 0.9, 1.0, 0.0 ]
	glFogi( $GL_FOG_MODE, $GL_EXP )
	glFogf( $GL_FOG_DENSITY, 0.10 )
	glFogfv( $GL_FOG_COLOR, $FogColor )
	glEnable( $GL_FOG )

	GUISwitch( $hGui )

	Return 1

EndFunc


Func CreateResources()

	Local $w = 256, $h = 256
	If Not LoadTextureImage( $iTexture1, $images & "\Tunnels\metalframe.tga",      $w, $h ) Then Return 0
	If Not LoadTextureImage( $iTexture2, $images & "\Tunnels\metalframe(rle).tga", $w, $h ) Then Return 0
	If Not LoadTextureImage( $iTexture3, $images & "\Tunnels\metalframe.rgba",     $w, $h ) Then Return 0
	glBindTexture( $GL_TEXTURE_2D, $iTexture1 )

	$iCallList = glGenLists( 1 )
	glNewList( $iCallList, $GL_COMPILE )
		glBegin( $GL_QUADS )
			; Front Face
			glTexCoord2f( 0.0, 0.0 )
			glVertex3f( -2.0, -1.0,	1.0 )
			glTexCoord2f( 2.0, 0.0 )
			glVertex3f(	2.0, -1.0,	1.0 )
			glTexCoord2f( 2.0, 1.0 )
			glVertex3f(	2.0,	1.0,	1.0 )
			glTexCoord2f( 0.0, 1.0 )
			glVertex3f( -2.0,	1.0,	1.0 )
			; Back Face
			glTexCoord2f( 2.0, 0.0 )
			glVertex3f( -2.0, -1.0, -1.0 )
			glTexCoord2f( 2.0, 1.0 )
			glVertex3f( -2.0,	1.0, -1.0 )
			glTexCoord2f( 0.0, 1.0 )
			glVertex3f(	2.0,	1.0, -1.0 )
			glTexCoord2f( 0.0, 0.0 )
			glVertex3f(	2.0, -1.0, -1.0 )
			; Top Face
			glTexCoord2f( 0.0, 1.0 )
			glVertex3f( -2.0,	1.0, -1.0 )
			glTexCoord2f( 0.0, 0.0 )
			glVertex3f( -2.0,	1.0,	1.0 )
			glTexCoord2f( 2.0, 0.0 )
			glVertex3f(	2.0,	1.0,	1.0 )
			glTexCoord2f( 2.0, 1.0 )
			glVertex3f(	2.0,	1.0, -1.0 )
			; Bottom Face
			glTexCoord2f( 2.0, 1.0 )
			glVertex3f( -2.0, -1.0, -1.0 )
			glTexCoord2f( 0.0, 1.0 )
			glVertex3f(	2.0, -1.0, -1.0 )
			glTexCoord2f( 0.0, 0.0 )
			glVertex3f(	2.0, -1.0,	1.0 )
			glTexCoord2f( 2.0, 0.0 )
			glVertex3f( -2.0, -1.0,	1.0 )
			; Left Face
			glTexCoord2f( 0.0, 0.0 )
			glVertex3f(	2.0, -1.0, -1.0 )
			glTexCoord2f( 1.0, 0.0 )
			glVertex3f(	2.0, -1.0,	1.0 )
			glTexCoord2f( 1.0, 1.0 )
			glVertex3f(	2.0,	1.0,	1.0 )
			glTexCoord2f( 0.0, 1.0 )
			glVertex3f(	2.0,	1.0, -1.0 )
		glEnd()
	glEndList()

	GUI_EVENT_RESIZE()
	GUISetState()

	Return 1

EndFunc


Func DisplayFunction()

	Local Static $iStartTime = TimerInit(), $fLastTime = 0
	Local $fElapsedTime = TimerDiff( $iStartTime )
	$fElapsedTime = ( $fLastTime + $fElapsedTime ) / 2
	$fLastTime = $fElapsedTime

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )
	glLoadIdentity()

	glRotatef( $fElapsedTime / 70, 0, 0, 1 )
	glRotatef( 90, 0, 1, 0 )
	glTranslatef( -$fElapsedTime / 400, 0, 0 )

	Local $iTime = Floor( ( ( $fElapsedTime - 400 ) / 400 ) / 4 )
	For $i = 1 To 5
		glPushMatrix()
		glTranslatef( $i * 4 + $iTime * 4, 0, 0 )
		glCallList( $iCallList )
		glPopMatrix()
	Next

	SwapBuffers( $hDC )

EndFunc


Func WM_SIZING( $hWnd, $iMsg, $iwParam, $ilParam )
	#forceref $iMsg, $iwParam
	Switch $hWnd
		Case $hGui
			Local $tRECT = DllStructCreate( $tagRECT, $ilParam )
			Local $y = DllStructGetData( $tRECT, "Top" )
			Local $iHeight = DllStructGetData( $tRECT, "Bottom" ) - $y - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight
			WinMove( $hStatusbar, "", 2, $iHeight + $iHeightXP )
	EndSwitch
	Return $GUI_RUNDEFMSG
EndFunc


Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	WinMove( $hStatusbar, "", 2, $iHeight + $iHeightXP )
	$bResize = True
EndFunc


Func GUI_EVENT_CLOSE()
	glDeleteLists( $iCallList, 1 )
	glDeleteTextures( 1, $iTexture1 )
	glDeleteTextures( 1, $iTexture2 )
	glDeleteTextures( 1, $iTexture3 )
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
