#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

Global $aTunnel[13][33][3]

; --- OpenGL globals ---

Global $iTexture

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 600, $iHeight = 600

; --- GUI globals ---

Global Const $iBorderWidth		 =	4
Global Const $iTitleBarHeight	 = 30
Global Const $iStatusBarHeight = 28

Global $iGuiWidth	= $iWidth
Global $iGuiHeight = $iHeight + $iStatusBarHeight

Global $hGui, $idLabel, $bResize = True, $bClose = False

; --- Project includes ---

#include "utilities\includes.au3"
#include "utilities\images.au3"

; --- Program code ---

MainFunc()

While 1
	If $bResize Then
		ResizeOpenGLwindow( 0, 0, $iWidth, $iHeight, 45.0, 1.0, 100.0 )
		$bResize = False
	EndIf
	If $bClose Then
		; Because most program running time is spent in the loop in DisplayFunction(), it must be taken
		; into account that the program is not shut down in the middle of an OpenGL drawing cycle.
		CloseProgram()
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then CloseProgram()
	If Not InitOpenGLwindow() Then CloseProgram()
	If Not CreateResources() Then CloseProgram()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 50, $WS_OVERLAPPEDWINDOW )

	; Statusbar
	Local $idStatus
	$idStatus = GUICtrlCreateLabel( "", 0, $iHeight + 1, $iGuiWidth + 2, $iStatusBarHeight, $SS_ETCHEDFRAME )
	$idLabel = GUICtrlCreateLabel( "", 3, $iHeight + 1 + 6, $iGuiWidth - 6, $iStatusBarHeight - 10 )
	GUICtrlSetData( $idLabel, "Tunnel effect with texture." )
	GUICtrlSetResizing( $idStatus, $GUI_DOCKSTATEBAR )
	GUICtrlSetResizing( $idLabel, $GUI_DOCKSTATEBAR )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,	"GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,	"GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,		"GUI_EVENT_CLOSE"	)

	Return 1

EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 0.0, 0.0, 0.0, 0.0 )

	glShadeModel( $GL_SMOOTH )

	glEnable( $GL_DEPTH_TEST )
	glDepthFunc( $GL_LESS )
	glClearDepth( 1.0 )

	glHint( $GL_PERSPECTIVE_CORRECTION_HINT,	$GL_NICEST )

	GUISwitch( $hGui )

	Return 1

EndFunc


Func CreateResources()

	Local $w, $h
	If Not LoadTextureImage( $iTexture, $images & "\Tunnels\tunnel.bmp", $w, $h ) Then Return 0
	glBindTexture( $GL_TEXTURE_2D, $iTexture )
	GUI_EVENT_RESIZE()

	; Tunnel z-coordinates
	For $i = 0 To 12
		For $j = 0 To 32
			$aTunnel[$i][$j][2] = -$j
		Next
	Next

	GUISetState()
	Return 1

EndFunc


Func DisplayFunction()

	Local Const $PI = 3.14159265358979323846

	Local Static $iStartTime = TimerInit(), $fLastTime = 0
	Local $fElapsedTime = TimerDiff( $iStartTime )
	$fElapsedTime = ( $fLastTime + $fElapsedTime ) / 2
	$fLastTime = $fElapsedTime

	;Local $fAngle = $fElapsedTime / 14
	Local $fAngle = $fElapsedTime / 28
	;Local $fAngle = $fElapsedTime / 56

	; Tunnel x,y-coordinates
	For $i = 0 To 12
		For $j = 0 To 32
			$aTunnel[$i][$j][0] = ( 3 - $j / 12 ) * Cos( 2 * $PI / 12 * $i ) + 2 * Sin( ( $fAngle + 2 * $j ) / 29 ) + Cos( ( $fAngle + 2 * $j ) / 13 ) - 2 * Sin( $fAngle / 29 ) - Cos( $fAngle / 13 )
			$aTunnel[$i][$j][1] = ( 3 - $j / 12 ) * Sin( 2 * $PI / 12 * $i ) + 2 * Cos( ( $fAngle + 2 * $j ) / 33 ) + Sin( ( $fAngle + 2 * $j ) / 17 ) - 2 * Cos( $fAngle / 33 ) - Sin( $fAngle / 17 )
		Next
	Next

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )
	glLoadIdentity()

	glTranslatef( 0.0, 0.0, -4.2 )

	; Draw tunnel
	Local $j1, $j2, $c
	For $j = 0 To 30
		$j1 = $j / 32 + $fAngle * 0.02
		$j2 = ( $j + 1 ) / 32 + $fAngle * 0.02

		; Fade the effect away near the end of the tunnel
		If $j > 24 Then
			$c = 1.0 - ( $j - 24 ) / 10
		Else
			$c = 1.0
		EndIf
		glColor3f( $c, $c, $c )

		glBegin( $GL_QUADS )
			For $i = 0 To 11
				glTexCoord2f( ( $i - 3 ) / 12, $j1 )
				glVertex3f( $aTunnel[$i  ][$j  ][0], $aTunnel[$i  ][$j  ][1], $aTunnel[$i  ][$j  ][2] )
				glTexCoord2f( ( $i - 2 ) / 12, $j1 )
				glVertex3f( $aTunnel[$i+1][$j  ][0], $aTunnel[$i+1][$j  ][1], $aTunnel[$i+1][$j  ][2] )
				glTexCoord2f( ( $i - 2 ) / 12, $j2 )
				glVertex3f( $aTunnel[$i+1][$j+1][0], $aTunnel[$i+1][$j+1][1], $aTunnel[$i+1][$j+1][2] )
				glTexCoord2f( ( $i - 3 ) / 12, $j2 )
				glVertex3f( $aTunnel[$i  ][$j+1][0], $aTunnel[$i  ][$j+1][1], $aTunnel[$i  ][$j+1][2] )
			Next
		glEnd()
	Next

	SwapBuffers( $hDC )

EndFunc


Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	$bResize = True
EndFunc


Func GUI_EVENT_CLOSE()
	$bClose = True
EndFunc


Func CloseProgram()
	glDeleteTextures( 1, $iTexture )
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
