#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>
#include <Misc.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

Global $aTunnel, $pTunnel

; --- FASM globals ---

Global $pData1, $pCode1
Global $pData2, $pCode2

; --- OpenGL globals ---

Global $iTexture

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 600, $iHeight = 600

; --- GUI globals ---

Global Const $iBorderWidth		 =	4
Global Const $iTitleBarHeight	 = 30
Global Const $iStatusBarHeight = 28

Global $iGuiWidth	= $iWidth
Global $iGuiHeight = $iHeight + $iStatusBarHeight

Global $hGui, $bResize = True

; --- Project includes ---

#include "utilities\includes.au3"
#include "utilities\images.au3"

; --- Program code ---

MainFunc()

While 1
	If $bResize Then
		ResizeOpenGLwindow( 0, 0, $iWidth, $iHeight, 45.0, 1.0, 100.0 )
		$bResize = False
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then GUI_EVENT_CLOSE()
	If Not InitOpenGLwindow() Then GUI_EVENT_CLOSE()
	If Not CreateResources() Then GUI_EVENT_CLOSE()
	If Not GetFASMcode() Then GUI_EVENT_CLOSE()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 50, $WS_OVERLAPPEDWINDOW )

	; Statusbar
	Local $idStatus, $idLabel
	$idStatus = GUICtrlCreateLabel( "", 0, $iHeight + 1, $iGuiWidth + 2, $iStatusBarHeight, $SS_ETCHEDFRAME )
	$idLabel = GUICtrlCreateLabel( "", 3, $iHeight + 1 + 6, $iGuiWidth - 6, $iStatusBarHeight - 10 )
	GUICtrlSetData( $idLabel, "Tunnel effect with texture.  |  Use Arrow up/down to increase/decrease speed." )
	GUICtrlSetResizing( $idStatus, $GUI_DOCKSTATEBAR )
	GUICtrlSetResizing( $idLabel, $GUI_DOCKSTATEBAR )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,	"GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,	"GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,		"GUI_EVENT_CLOSE"	)

	Return 1

EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 0.0, 0.0, 0.0, 0.0 )

	glShadeModel( $GL_SMOOTH )

	glEnable( $GL_DEPTH_TEST )
	glDepthFunc( $GL_LESS )
	glClearDepth( 1.0 )

	glHint( $GL_PERSPECTIVE_CORRECTION_HINT,	$GL_NICEST )

	GUISwitch( $hGui )

	Return 1

EndFunc


Func CreateResources()

	Local $w, $h
	If Not LoadTextureImage( $iTexture, $images & "\Tunnels\tunnel.bmp", $w, $h ) Then Return 0
	glBindTexture( $GL_TEXTURE_2D, $iTexture )
	GUI_EVENT_RESIZE()

	$aTunnel = DllStructCreate( "float[" & 13 * 33 * 3 & "]" )
	$pTunnel = DllStructGetPtr( $aTunnel )

	; Tunnel z-coordinates
	Local $k = 3
	For $i = 0 To 12
		For $j = 0 To 32
			DllStructSetData( $aTunnel, 1, -$j, $k )
			$k += 3
		Next
	Next

	GUISetState()
	Return 1

EndFunc


Func GetFASMcode()

	; --- Fasm1 ---

	Local Static $oFasm1, $aData1, $sCode1, $tCode1

	Local Const $PI = 3.14159265358979323846
	Local $x = 2 * $PI / 12

	; Fasm1 data

	; Tunnel x,y-coordinates
	; For $i = 0 To 12
	; 	For $j = 0 To 32
	; 		$aTunnel[$i][$j][0] = ( 3 - $j / 12 ) * Cos( 2 * $PI / 12 * $i ) + 2 * Sin( ( $fAngle + 2 * $j ) / 29 ) + Cos( ( $fAngle + 2 * $j ) / 13 ) - 2 * Sin( $fAngle / 29 ) - Cos( $fAngle / 13 )
	; 		$aTunnel[$i][$j][1] = ( 3 - $j / 12 ) * Sin( 2 * $PI / 12 * $i ) + 2 * Cos( ( $fAngle + 2 * $j ) / 33 ) + Sin( ( $fAngle + 2 * $j ) / 17 ) - 2 * Cos( $fAngle / 33 ) - Sin( $fAngle / 17 )
	; 	Next
	; Next

	$aData1 = DllStructCreate( "ushort;ushort;ushort[6];float[6];ptr" )
	; Element 1, offset  0 : ushort $i
	; Element 2, offset  2 : ushort $j
	; Element 3, offset  4 : ushort[6] for constants
	; Element 4, offset 16 : float[6] for calculations
	; Element 5, offset 40 : ptr $pTunnel

	DllStructSetData( $aData1, 1,  0 )				; Element 1, index 1 : $i
	DllStructSetData( $aData1, 2,  0 )				; Element 2, index 1 : $j
	DllStructSetData( $aData1, 3,  3, 1 )			; Element 3, index 1 : constant 3
	DllStructSetData( $aData1, 3, 12, 2 )			; Element 3, index 2 : constant 12
	DllStructSetData( $aData1, 3, 13, 3 )			; Element 3, index 3 : constant 13
	DllStructSetData( $aData1, 3, 17, 4 )			; Element 3, index 4 : constant 17
	DllStructSetData( $aData1, 3, 29, 5 )			; Element 3, index 5 : constant 29
	DllStructSetData( $aData1, 3, 33, 6 )			; Element 3, index 6 : constant 33
	DllStructSetData( $aData1, 4, $x, 1 )			; Element 4, index 1 : constant 2 * $PI / 12
	DllStructSetData( $aData1, 5, $pTunnel )	; Element 5, index 1 : $pTunnel

	$pData1 = DllStructGetPtr( $aData1 )

	; Fasm1 code

	If @AutoItX64 Then
		If Not ReadX64code( "utilities\fasm\fasm1-x64.bin", "utilities\fasm\fasm1-mk-x64-bin.au3", $sCode1 ) Then Return 0
	Else
		$oFasm1 = FasmInit()
		If Not ReadFasmCode( "utilities\fasm\fasm1.asm", $oFasm1, $sCode1 ) Then Return 0
		FasmExit( $oFasm1 )
	EndIf
	$pCode1 = _MemVirtualAlloc( 0, 1024, $MEM_COMMIT, $PAGE_EXECUTE_READWRITE )
	$tCode1 = DllStructCreate( "byte[1024]", $pCode1 )
	DllStructSetData( $tCode1, 1, $sCode1 )


	; --- Fasm2 ---

	Local Static $oFasm2, $aData2, $sCode2, $tCode2

	; Fasm2 data

	; For $j = 0 To 30
	; 	$j1 = $j / 32 + $fAngle * 0.02
	; 	$j2 = ( $j + 1 ) / 32 + $fAngle * 0.02
  ;
	; 	; Fade the effect away near the end of the tunnel
	; 	If $j > 24 Then
  ; 		$c = 1.0 - ( $j - 24 ) / 10
  ; 	Else
  ; 		$c = 1.0
  ; 	EndIf
  ; 	glColor3f( $c, $c, $c )
  ;
  ; 	glBegin( $GL_QUADS )
  ; 		For $i = 0 To 11
  ; 			glTexCoord2f( ( $i - 3 ) / 12, $j1 )
  ; 			glVertex3f( $aTunnel[$i  ][$j  ][0], $aTunnel[$i  ][$j  ][1], $aTunnel[$i  ][$j  ][2] )
  ; 			glTexCoord2f( ( $i - 2 ) / 12, $j1 )
  ; 			glVertex3f( $aTunnel[$i+1][$j  ][0], $aTunnel[$i+1][$j  ][1], $aTunnel[$i+1][$j  ][2] )
  ; 			glTexCoord2f( ( $i - 2 ) / 12, $j2 )
  ; 			glVertex3f( $aTunnel[$i+1][$j+1][0], $aTunnel[$i+1][$j+1][1], $aTunnel[$i+1][$j+1][2] )
  ; 			glTexCoord2f( ( $i - 3 ) / 12, $j2 )
  ; 			glVertex3f( $aTunnel[$i  ][$j+1][0], $aTunnel[$i  ][$j+1][1], $aTunnel[$i  ][$j+1][2] )
  ; 		Next
  ; 	glEnd()
  ; Next

	$aData2 = DllStructCreate( "ushort;ushort;ushort[6];float[6];float[7];uint;ptr;ptr[5]" )
	; Element 1, offset  0      : ushort $i
	; Element 2, offset  2      : ushort $j
	; Element 3, offset  4      : ushort[6] for constants
	; Element 4, offset 16      : float[6] for calculations: 0.02, $j1, $j2, ( $i - 2 ) / 12, ( $i - 3 ) / 12
	; Element 5, offset 40      : float[7] for 1.0 and $c = 1.0 - ( $j - 24 ) / 10, 25 <= $j <= 30
	; Element 6, offset 68      : uint $GL_QUADS
	; Element 7, offset 72      : ptr $pTunnel
	; Element 8, offset 76 (80) : ptr[5]: glColor3f, glBegin, glTexCoord2f, glVertex3f, glEnd  ; 4/8-bytes pointers in x86/x64
	; (The number in brackets is x64 offset)

	DllStructSetData( $aData2, 1,  0 )				; Element 1, index 1 : $i
	DllStructSetData( $aData2, 2,  0 )				; Element 2, index 1 : $j
	DllStructSetData( $aData2, 3,  2, 1 )			; Element 3, index 1 : constant 2
	DllStructSetData( $aData2, 3,  3, 2 )			; Element 3, index 2 : constant 3
	DllStructSetData( $aData2, 3, 12, 3 )			; Element 3, index 3 : constant 12
	DllStructSetData( $aData2, 3, 32, 4 )			; Element 3, index 4 : constant 32
	DllStructSetData( $aData2, 4, 0.02, 1 )		; Element 4, index 1 : constant 0.02
	DllStructSetData( $aData2, 5, 1.0, 1 )		; Element 5, index 1 : constant 1.0
	For $j = 25 To 30
		DllStructSetData( $aData2, 5, 1.0 - ( $j - 24 ) / 10, $j - 23 )
	Next
	DllStructSetData( $aData2,6,$GL_QUADS,1 )	; Element 6, index 1 : constant $GL_QUADS
	DllStructSetData( $aData2, 7, $pTunnel )	; Element 7, index 1 : $pTunnel
	DllStructSetData( $aData2, 8, GetProcAddress( "opengl32.dll", "glColor3f" ),    1 )
	DllStructSetData( $aData2, 8, GetProcAddress( "opengl32.dll", "glBegin" ),      2 )
	DllStructSetData( $aData2, 8, GetProcAddress( "opengl32.dll", "glTexCoord2f" ), 3 )
	DllStructSetData( $aData2, 8, GetProcAddress( "opengl32.dll", "glVertex3f" ),   4 )
	DllStructSetData( $aData2, 8, GetProcAddress( "opengl32.dll", "glEnd" ),        5 )

	$pData2 = DllStructGetPtr( $aData2 )

	; Fasm2 code

	If @AutoItX64 Then
		If Not ReadX64code( "utilities\fasm\fasm2-x64.bin", "utilities\fasm\fasm2-mk-x64-bin.au3", $sCode2 ) Then Return 0
	Else
		$oFasm2 = FasmInit()
		If Not ReadFasmCode( "utilities\fasm\fasm2.asm", $oFasm2, $sCode2 ) Then Return 0
		FasmExit( $oFasm2 )
	EndIf
	$pCode2 = _MemVirtualAlloc( 0, 1024, $MEM_COMMIT, $PAGE_EXECUTE_READWRITE )
	$tCode2 = DllStructCreate( "byte[1024]", $pCode2 )
	DllStructSetData( $tCode2, 1, $sCode2 )

	Return 1

EndFunc


Func DisplayFunction()

	Local Static $fAngle, $fSpeed = 0.5

	Select
		Case _IsPressed( "26", $dllUser32 ) ; Up arrow
			if $fSpeed < 3.0 Then $fSpeed += 0.01
		Case _IsPressed( "28", $dllUser32 ) ; Down arrow
			if $fSpeed > -3.0 Then $fSpeed -= 0.01
	EndSelect

	$fAngle += $fSpeed

	; Tunnel x,y-coordinates
	DllCallAddress( "none", $pCode1, _
		"float", $fAngle, "ptr", $pData1 )

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )
	glLoadIdentity()

	glTranslatef( 0.0, 0.0, -4.2 )

	; Draw tunnel
	DllCallAddress( "none", $pCode2, _
		"float", $fAngle, "ptr", $pData2 )

	SwapBuffers( $hDC )

EndFunc


Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	$bResize = True
EndFunc


Func GUI_EVENT_CLOSE()
	glDeleteTextures( 1, $iTexture )
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
