#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

Global $aVertices[8]  = [ -1.0, -1.0, +1.0, -1.0, +1.0, +1.0, -1.0, +1.0 ]
Global $aTexCoords[8] = [  0.0,  0.0,  1.0,  0.0,  1.0,  1.0,  0.0,  1.0 ]

; --- OpenGL globals ---

Global $iTexture

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 640, $iHeight = 299

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight

Global $hGui, $bResize = True

; --- Project includes ---

#include "utilities\includes.au3"
#include "utilities\images.au3"

; --- Program code ---

MainFunc()

While 1
	If $bResize Then
		; An OpenGL window (viewport) should not be resized in the middle of a drawing cycle.
		; In GUI_EVENT_RESIZE() $bResize is set to True when the GUI is resized. At this point
		; when the DisplayFunction() isn't running it's safe to resize the OpenGL window.
		glViewport( 0, 0, $iWidth, $iHeight )
		$bResize = False
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd

Func MainFunc()
	If Not InitGUIwindow() Then GUI_EVENT_CLOSE()
	If Not InitOpenGLwindow() Then GUI_EVENT_CLOSE()
	;If Not CheckOpenGLversion( 2, 1 ) Then GUI_EVENT_CLOSE()
	If Not CreateResources() Then GUI_EVENT_CLOSE()
EndFunc

Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 50, $WS_OVERLAPPEDWINDOW )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	Return 1

EndFunc

Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 0.0, 0.0, 0.0, 0.0 )

	GUISwitch( $hGui )
	GUISetState()

	Return 1

EndFunc

Func CreateResources()

	Local $w, $h
	If Not LoadTextureImage( $iTexture, $images & "\Homer Simpson\Homer Simpson.jpg", $w, $h ) Then Return 0
	glBindTexture( $GL_TEXTURE_2D, $iTexture )

	Return 1

EndFunc

Func DisplayFunction()

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )

	glEnableClientState( $GL_VERTEX_ARRAY )
	glEnableClientState( $GL_TEXTURE_COORD_ARRAY )

	glTexCoordPointer( 2, $GL_FLOAT, 0, $aTexCoords )
	glVertexPointer( 2, $GL_FLOAT, 0, $aVertices )
	glDrawArrays( $GL_QUADS, 0, 4 )

	glDisableClientState( $GL_TEXTURE_COORD_ARRAY )
	glDisableClientState( $GL_VERTEX_ARRAY )

	SwapBuffers( $hDC )

EndFunc

Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	$bResize = True ; The OpenGL window (viewport) is resized in top of the main loop
EndFunc

Func GUI_EVENT_CLOSE()
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
