#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>
#include <EditConstants.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

; --- OpenGL globals ---

Global $iVBOtriangle
Global $iVertexShader, $iFragmentShader
Global $iProgram, $iAttributeCoord2d

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 640, $iHeight = 480

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30
Global Const $iStatusBarHeight = 28

Global $hStatusbar, $idShader1, $idShader2
Global $iShader = 1, $iNewShader = $iShader

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight + $iStatusBarHeight

Global $hGui, $bResize = True

; --- Project includes ---

#include "utilities\includes.au3"

; --- Program code ---

MainFunc()

While 1
	If $iShader <> $iNewShader Then _
		CreateResources()
	If $bResize Then
		; An OpenGL window (viewport) should not be resized in the middle of a drawing cycle.
		; In GUI_EVENT_RESIZE() $bResize is set to True when the GUI is resized. At this point
		; when the DisplayFunction() isn't running it's safe to resize the OpenGL window.
		glViewport( 0, 0, $iWidth, $iHeight )
		$bResize = False
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then GUI_EVENT_CLOSE()
	If Not InitOpenGLwindow() Then GUI_EVENT_CLOSE()
	If Not CheckOpenGLversion( 2, 1 ) Then GUI_EVENT_CLOSE()
	If Not CreateResources() Then GUI_EVENT_CLOSE()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 50, $WS_OVERLAPPEDWINDOW )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	GUIRegisterMsg( $WM_SIZING, "WM_SIZING" )

	CreateStatusbar()

	Return 1

EndFunc


Func CreateStatusbar()

	; Use a child window as a container for controls in the Statusbar
	$hStatusbar = GUICreate( "", $iGuiWidth - 4 , 24, 2, $iHeight + 3 , $WS_CHILD, -1, $hGui )

	; Select fragment shader
	GUIStartGroup()
	GUICtrlCreateLabel( "Fragment shader:", 2, 4, 90, 17 )
	$idShader1 = GUICtrlCreateRadio( "triangle-1.f.glsl", 100, 3, 90, 18 )
	$idShader2 = GUICtrlCreateRadio( "triangle-2.f.glsl", 200, 3, 90, 18 )
	GUICtrlSetState( $idShader1, $GUI_CHECKED )
	GUICtrlSetOnEvent( $idShader1, "RadioClick" )
	GUICtrlSetOnEvent( $idShader2, "RadioClick" )

	; Vertical splitterbar
	GUICtrlCreateLabel( "", 300, 2, 3, 20, $SS_ETCHEDFRAME )

	; Label in statusbar
	GUICtrlCreateLabel( "", 318, 5, 300, 14 )
	GUICtrlSetData( -1, "Loading shaders from files. Vertex Buffer Objects (VBO)." )

	GUISetState( @SW_SHOW, $hStatusbar )
	GUISwitch( $hGui )

	; Label for $hStatusbar to make it look nicer
	GUICtrlCreateLabel( "", 0, $iHeight + 1, $iGuiWidth + 4, $iStatusBarHeight, $SS_ETCHEDFRAME )
	GUICtrlSetResizing( -1, $GUI_DOCKSTATEBAR )

EndFunc


Func RadioClick()
	Switch @GUI_CTRLID
		Case $idShader1
			$iNewShader = 1
		Case $idShader2
			$iNewShader = 2
	EndSwitch
EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 1.0, 1.0, 1.0, 1.0 )

	glBlendFunc( $GL_SRC_ALPHA, $GL_ONE_MINUS_SRC_ALPHA )

	GUISwitch( $hGui )

	Return 1

EndFunc


Func CreateResources()

	If $iShader <> $iNewShader Then
		$iShader = $iNewShader
		glDetachShader( $iProgram, $iVertexShader )
		glDeleteShader( $iVertexShader )
		glDetachShader( $iProgram, $iFragmentShader )
		glDeleteShader( $iFragmentShader )
		glDeleteProgram( $iProgram )
		glDeleteBuffers( 1, $iVBOtriangle )
	EndIf

	Local $aTriangleVertices[6] = [ _
		 0.0,  0.8, _
		-0.8, -0.8, _
		 0.8, -0.8 ]

	glGenBuffers( 1, $iVBOtriangle )
	glBindBuffer( $GL_ARRAY_BUFFER, $iVBOtriangle )
	glBufferData( $GL_ARRAY_BUFFER, $GL_FLOAT, 4*UBound($aTriangleVertices), $aTriangleVertices, $GL_STATIC_DRAW )

	If Not LoadGLSLshader( $iVertexShader, "shaders\02-triangle.v.glsl", $GL_VERTEX_SHADER ) Then Return 0
	If $iShader = 1 Then
		glDisable( $GL_BLEND )
		If Not LoadGLSLshader( $iFragmentShader, "shaders\02-triangle-1.f.glsl", $GL_FRAGMENT_SHADER ) Then Return 0
	ElseIf $iShader = 2 Then
		glEnable( $GL_BLEND )
		If Not LoadGLSLshader( $iFragmentShader, "shaders\02-triangle-2.f.glsl", $GL_FRAGMENT_SHADER ) Then Return 0
	EndIf

	Local $bStatus

	$iProgram = glCreateProgram()
	glAttachShader( $iProgram, $iVertexShader )
	glAttachShader( $iProgram, $iFragmentShader )
	glLinkProgram( $iProgram )
	glGetProgramiv( $iProgram, $GL_LINK_STATUS, $bStatus )
	If $bStatus = 0 Then
		PrintError( "Could not link program" )
		Return 0
	EndIf

	$iAttributeCoord2d = glGetAttribLocation( $iProgram, "coord2d" )
	If $iAttributeCoord2d = -1 Then
		PrintError( "Could not bind attribute ""coord2d""" )
		Return 0
	EndIf

	GUISetState()

	Return 1

EndFunc


Func DisplayFunction()

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )

	glUseProgram( $iProgram )
	glEnableVertexAttribArray( $iAttributeCoord2d )
	glBindBuffer( $GL_ARRAY_BUFFER, $iVBOtriangle )
	; Describe our vertices array to OpenGL (it can't guess its format automatically)
  glVertexAttribPointer( _
		$iAttributeCoord2d, _ ; attribute
		2, _                  ; number of elements per vertex, here (x,y)
		$GL_FLOAT, _          ; the type of each element
		$GL_FALSE, _          ; take our values as-is
		0, _                  ; no extra data between each position
		0 )                   ; offset of first element

  ; Push each element in triangle vertices to the vertex shader
  glDrawArrays( $GL_TRIANGLES, 0, 3 )

  glDisableVertexAttribArray( $iAttributeCoord2d )

	SwapBuffers( $hDC )

EndFunc


Func WM_SIZING( $hWnd, $iMsg, $iwParam, $ilParam )
	#forceref $iMsg, $iwParam
	Switch $hWnd
		Case $hGui
			Local $tRECT = DllStructCreate( $tagRECT, $ilParam )
			Local $y = DllStructGetData( $tRECT, "Top" )
			Local $iHeight = DllStructGetData( $tRECT, "Bottom" ) - $y - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight
			WinMove( $hStatusbar, "", 2, $iHeight + $iHeightXP )
	EndSwitch
	Return $GUI_RUNDEFMSG
EndFunc


Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	WinMove( $hStatusbar, "", 2, $iHeight + $iHeightXP )
	$bResize = True ; The OpenGL window (viewport) is resized in top of the main loop
EndFunc


Func GUI_EVENT_CLOSE()
	glDeleteProgram( $iProgram )
	glDeleteBuffers( 1, $iVBOtriangle )
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
