#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

; --- OpenGL globals ---

Global $iProgram
Global $iVBOtriangle
Global $iAttributeCoord2d, $iAttributeVcolor
Global $iUniformFade

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 640, $iHeight = 480

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30
Global Const $iStatusBarHeight = 28

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight + $iStatusBarHeight

Global $hGui, $bResize = True

; --- Project includes ---

#include "utilities\includes.au3"

; --- Program code ---

MainFunc()

While 1
	If $bResize Then
		; An OpenGL window (viewport) should not be resized in the middle of a drawing cycle.
		; In GUI_EVENT_RESIZE() $bResize is set to True when the GUI is resized. At this point
		; when the DisplayFunction() isn't running it's safe to resize the OpenGL window.
		glViewport( 0, 0, $iWidth, $iHeight )
		$bResize = False
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then GUI_EVENT_CLOSE()
	If Not InitOpenGLwindow() Then GUI_EVENT_CLOSE()
	If Not CheckOpenGLversion( 2, 1 ) Then GUI_EVENT_CLOSE()
	If Not CreateResources() Then GUI_EVENT_CLOSE()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 50, $WS_OVERLAPPEDWINDOW )

	; Statusbar
	Local $idStatus, $idLabel
	$idStatus = GUICtrlCreateLabel( "", 0, $iHeight + 1, $iGuiWidth + 2, $iStatusBarHeight, $SS_ETCHEDFRAME )
	$idLabel = GUICtrlCreateLabel( "", 3, $iHeight + 1 + 6, $iGuiWidth - 4, $iStatusBarHeight - 10 )
	GUICtrlSetData( $idLabel, "Passing uniform variables (globals) to shaders." )
	GUICtrlSetResizing( $idStatus, $GUI_DOCKSTATEBAR )
	GUICtrlSetResizing( $idLabel, $GUI_DOCKSTATEBAR )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	Return 1

EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 1.0, 1.0, 1.0, 1.0 )

	glEnable( $GL_BLEND )
	glBlendFunc( $GL_SRC_ALPHA, $GL_ONE_MINUS_SRC_ALPHA )

	GUISwitch( $hGui )

	Return 1

EndFunc


Func CreateResources()

  Local $aTriangleAttributes[15] = [ _
     0.0,  0.8,   1.0, 1.0, 0.0, _
    -0.8, -0.8,   0.0, 0.0, 1.0, _
     0.8, -0.8,   1.0, 0.0, 0.0 ]

	glGenBuffers( 1, $iVBOtriangle )
	glBindBuffer( $GL_ARRAY_BUFFER, $iVBOtriangle )
	glBufferData( $GL_ARRAY_BUFFER, $GL_FLOAT, 4*UBound($aTriangleAttributes), $aTriangleAttributes, $GL_STATIC_DRAW )

	Local $iVertexShader, $iFragmentShader, $bStatus

	If Not LoadGLSLshader( $iVertexShader, "shaders\03-triangle.v.glsl", $GL_VERTEX_SHADER ) Then Return 0
	If Not LoadGLSLshader( $iFragmentShader, "shaders\03-triangle.f.glsl", $GL_FRAGMENT_SHADER ) Then Return 0

	$iProgram = glCreateProgram()
	glAttachShader( $iProgram, $iVertexShader )
	glAttachShader( $iProgram, $iFragmentShader )
	glLinkProgram( $iProgram )
	glGetProgramiv( $iProgram, $GL_LINK_STATUS, $bStatus )
	If $bStatus = 0 Then
		PrintError( "Could not link program" )
		Return 0
	EndIf

	$iAttributeCoord2d = glGetAttribLocation( $iProgram, "coord2d" )
	If $iAttributeCoord2d = -1 Then
		PrintError( "Could not bind attribute ""coord2d""" )
		Return 0
	EndIf

	$iAttributeVcolor = glGetAttribLocation( $iProgram, "v_color" )
	If $iAttributeVcolor = -1 Then
		PrintError( "Could not bind attribute ""v_color""" )
		Return 0
	EndIf

	$iUniformFade = glGetUniformLocation( $iProgram, "fade" )
	If $iUniformFade = -1 Then
		PrintError( "Could not bind uniform ""fade""" )
		Return 0
	EndIf

	GUISetState()

	Return 1

EndFunc


Func DisplayFunction()

	Local Static $iStartTime = TimerInit(), $fLastTime = 0
	Local $fElapsedTime = TimerDiff( $iStartTime )
	$fElapsedTime = ( $fLastTime + $fElapsedTime ) / 2
	$fLastTime = $fElapsedTime

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )

	glUseProgram( $iProgram )

	Local $fCurFade = Sin( $fElapsedTime / 1000.0 * 2 * $PI / 5 ) / 2 + 0.5 ; 0 -> 1 every 5 seconds
	glUniform1f( $iUniformFade, $fCurFade )

	glBindBuffer( $GL_ARRAY_BUFFER, $iVBOtriangle )

	glEnableVertexAttribArray( $iAttributeCoord2d )
  glVertexAttribPointer( _
		$iAttributeCoord2d, _ ; attribute
		2, _                  ; number of elements per vertex, here (x,y)
		$GL_FLOAT, _          ; the type of each element
		$GL_FALSE, _          ; take our values as-is
		5 * 4, _              ; next coord2d appears every 5 floats
		0 )                   ; offset of first element

  glEnableVertexAttribArray( $iAttributeVcolor )
  glVertexAttribPointer( _
    $iAttributeVcolor, _  ; attribute
    3, _                  ; number of elements per vertex, here (r,g,b)
    $GL_FLOAT, _          ; the type of each element
    $GL_FALSE, _          ; take our values as-is
		5 * 4, _              ; next coord2d appears every 5 floats
    2 * 4 )               ; offset of first element

  ; Push each element in triangle vertices to the vertex shader
  glDrawArrays( $GL_TRIANGLES, 0, 3 )

  glDisableVertexAttribArray( $iAttributeCoord2d )
  glDisableVertexAttribArray( $iAttributeVcolor )

	SwapBuffers( $hDC )

EndFunc


Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	$bResize = True ; The OpenGL window (viewport) is resized in top of the main loop
EndFunc


Func GUI_EVENT_CLOSE()
	glDeleteProgram( $iProgram )
	glDeleteBuffers( 1, $iVBOtriangle )
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
