#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

; --- OpenGL globals ---

Global $iProgram
Global $iVBOcubeVertices, $iVBOcubeColors, $iVBOcubeElements
Global $iAttributeCoord3d, $iAttributeVcolor
Global $iUniformMVP

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 800, $iHeight = 600

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30
Global Const $iStatusBarHeight = 28

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight + $iStatusBarHeight

Global $hGui, $bResize = True

; --- Project includes ---

#include "utilities\includes.au3"

; --- Program code ---

MainFunc()

While 1
	If $bResize Then
		; An OpenGL window (viewport) should not be resized in the middle of a drawing cycle.
		; In GUI_EVENT_RESIZE() $bResize is set to True when the GUI is resized. At this point
		; when the DisplayFunction() isn't running it's safe to resize the OpenGL window.
		glViewport( 0, 0, $iWidth, $iHeight )
		$bResize = False
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then GUI_EVENT_CLOSE()
	If Not InitOpenGLwindow() Then GUI_EVENT_CLOSE()
	If Not CheckOpenGLversion( 2, 1 ) Then GUI_EVENT_CLOSE()
	If Not CreateResources() Then GUI_EVENT_CLOSE()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 50, $WS_OVERLAPPEDWINDOW )

	; Statusbar
	Local $idStatus, $idLabel
	$idStatus = GUICtrlCreateLabel( "", 0, $iHeight + 1, $iGuiWidth + 2, $iStatusBarHeight, $SS_ETCHEDFRAME )
	$idLabel = GUICtrlCreateLabel( "", 3, $iHeight + 1 + 6, $iGuiWidth - 4, $iStatusBarHeight - 10 )
	GUICtrlSetData( $idLabel, "Rotating cube with global transformation matrix (Model-View-Projection matrix, MVP). More GLM-functions (OpenGL Mathematics library)." )
	GUICtrlSetResizing( $idStatus, $GUI_DOCKSTATEBAR )
	GUICtrlSetResizing( $idLabel, $GUI_DOCKSTATEBAR )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	Return 1

EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 1.0, 1.0, 1.0, 1.0 )

	glEnable( $GL_DEPTH_TEST )

	glEnable( $GL_BLEND )
	glBlendFunc( $GL_SRC_ALPHA, $GL_ONE_MINUS_SRC_ALPHA )

	GUISwitch( $hGui )

	Return 1

EndFunc


Func CreateResources()

  Local $aCubeVertices[3*8] = [ _
		-1.0, -1.0,  1.0, _    ; front
		 1.0, -1.0,  1.0, _
		 1.0,  1.0,  1.0, _
		-1.0,  1.0,  1.0, _
		-1.0, -1.0, -1.0, _    ; back
		 1.0, -1.0, -1.0, _
		 1.0,  1.0, -1.0, _
		-1.0,  1.0, -1.0 ]

	glGenBuffers( 1, $iVBOcubeVertices )
	glBindBuffer( $GL_ARRAY_BUFFER, $iVBOcubeVertices )
	glBufferData( $GL_ARRAY_BUFFER, $GL_FLOAT, 4*UBound($aCubeVertices), $aCubeVertices, $GL_STATIC_DRAW )

  Local $aCubeColors[3*8] = [ _
		1.0, 0.0, 0.0, _    ; front colors
		0.0, 1.0, 0.0, _
		0.0, 0.0, 1.0, _
		1.0, 1.0, 1.0, _
		1.0, 0.0, 0.0, _    ; back colors
		0.0, 1.0, 0.0, _
		0.0, 0.0, 1.0, _
		1.0, 1.0, 1.0 ]

	glGenBuffers( 1, $iVBOcubeColors )
	glBindBuffer( $GL_ARRAY_BUFFER, $iVBOcubeColors )
	glBufferData( $GL_ARRAY_BUFFER, $GL_FLOAT, 4*UBound($aCubeColors), $aCubeColors, $GL_STATIC_DRAW )

  Local $aCubeElements[3*12] = [ _
		0, 1, 2, _    ; front
		2, 3, 0, _
		1, 5, 6, _    ; top
		6, 2, 1, _
		7, 6, 5, _    ; back
		5, 4, 7, _
		4, 0, 3, _    ; bottom
		3, 7, 4, _
		4, 5, 1, _    ; left
		1, 0, 4, _
		3, 2, 6, _    ; right
		6, 7, 3 ]

	glGenBuffers( 1, $iVBOcubeElements )
	glBindBuffer( $GL_ELEMENT_ARRAY_BUFFER, $iVBOcubeElements )
	glBufferData( $GL_ELEMENT_ARRAY_BUFFER, $GL_UNSIGNED_SHORT, 2*UBound($aCubeElements), $aCubeElements, $GL_STATIC_DRAW )

	Local $iVertexShader, $iFragmentShader, $bStatus

	If Not LoadGLSLshader( $iVertexShader, "shaders\05-cube.v.glsl", $GL_VERTEX_SHADER ) Then Return 0
	If Not LoadGLSLshader( $iFragmentShader, "shaders\05-cube.f.glsl", $GL_FRAGMENT_SHADER ) Then Return 0

	$iProgram = glCreateProgram()
	glAttachShader( $iProgram, $iVertexShader )
	glAttachShader( $iProgram, $iFragmentShader )
	glLinkProgram( $iProgram )
	glGetProgramiv( $iProgram, $GL_LINK_STATUS, $bStatus )
	If $bStatus = 0 Then
		PrintError( "Could not link program" )
		Return 0
	EndIf

	$iAttributeCoord3d = glGetAttribLocation( $iProgram, "coord3d" )
	If $iAttributeCoord3d = -1 Then
		PrintError( "Could not bind attribute ""coord3d""" )
		Return 0
	EndIf

	$iAttributeVcolor = glGetAttribLocation( $iProgram, "v_color" )
	If $iAttributeVcolor = -1 Then
		PrintError( "Could not bind attribute ""v_color""" )
		Return 0
	EndIf

	$iUniformMVP = glGetUniformLocation( $iProgram, "mvp" )
	If $iUniformMVP = -1 Then
		PrintError( "Could not bind uniform ""mvp""" )
		Return 0
	EndIf

	GUISetState()

	Return 1

EndFunc


Func DisplayFunction()

	Local Static $iStartTime = TimerInit(), $fLastTime = 0
	Local $fElapsedTime = TimerDiff( $iStartTime )
	$fElapsedTime = ( $fLastTime + $fElapsedTime ) / 2
	$fLastTime = $fElapsedTime

	Local $mIdentity = $IDENTITY_MATRIX, $iSize

	Local $fAngle = $fElapsedTime / 1000.0 * 45 ; 45 per second
	Local $vAxisY[3] = [ 0, 1, 0 ]
	Local $mRotate = glmRotate( $mIdentity, $fAngle, $vAxisY )

	Local $vModel[3] = [ 0.0, 0.0, -4.0 ]
	Local $mModel = glmTranslate( $mIdentity, $vModel )

	Local $eye[3] = [ 0.0, 2.0, 0.0 ], $center[3] = [ 0.0, 0.0, -4.0 ], $up[3] = [ 0.0, 1.0, 0.0 ]
	Local $mView = glmLookAt( $eye, $center, $up )

	Local $mProjection = glmPerspective( 45.0, 1.0*$iWidth/$iHeight, 0.1, 10.0 )

	Local $mMult1 = glmMultiply( $mProjection, $mView )
	Local $mMult2 = glmMultiply( $mMult1, $mModel )
	Local $mMVP = glmMultiply( $mMult2, $mRotate ) ; $mProjection * $mView * $mModel * $mRotate

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )

	glUseProgram( $iProgram )
	glUniformMatrix4fv( $iUniformMVP, 1, $GL_FALSE, $mMVP )

	glEnableVertexAttribArray( $iAttributeCoord3d )
	glBindBuffer( $GL_ARRAY_BUFFER, $iVBOcubeVertices )
  glVertexAttribPointer( _
		$iAttributeCoord3d, _ ; attribute
		3, _                  ; number of elements per vertex, here (x,y,z)
		$GL_FLOAT, _          ; the type of each element
		$GL_FALSE, _          ; take our values as-is
		0, _                  ; no extra data between each position
		0 )                   ; offset of first element

  glEnableVertexAttribArray( $iAttributeVcolor )
	glBindBuffer( $GL_ARRAY_BUFFER, $iVBOcubeColors )
  glVertexAttribPointer( _
    $iAttributeVcolor, _  ; attribute
    3, _                  ; number of elements per vertex, here (r,g,b)
    $GL_FLOAT, _          ; the type of each element
    $GL_FALSE, _          ; take our values as-is
		0, _                  ; no extra data between each position
		0 )                   ; offset of first element

	; Push each element in vertices buffer to the vertex shader
	glBindBuffer( $GL_ELEMENT_ARRAY_BUFFER, $iVBOcubeElements )
	glGetBufferParameteriv( $GL_ELEMENT_ARRAY_BUFFER, $GL_BUFFER_SIZE, $iSize )
	glDrawElements( $GL_TRIANGLES, $iSize / 2, $GL_UNSIGNED_SHORT, 0 )

  glDisableVertexAttribArray( $iAttributeCoord3d )
  glDisableVertexAttribArray( $iAttributeVcolor )

	SwapBuffers( $hDC )

EndFunc


Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	$bResize = True ; The OpenGL window (viewport) is resized in top of the main loop
EndFunc


Func GUI_EVENT_CLOSE()
	glDeleteProgram( $iProgram )
	glDeleteBuffers( 1, $iVBOcubeColors )
	glDeleteBuffers( 1, $iVBOcubeVertices )
	glDeleteBuffers( 1, $iVBOcubeElements )
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
