#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>
#include <Misc.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

Global $fOffsetX = 0.0, $fScaleX = 1.0

; --- OpenGL globals ---

Global $iProgram, $iVbo
Global $iAttributeCoord2d, $iUniformOffsetX, $iUniformScaleX, $iUniformSprite
Global $iTextureId, $iTexWidth, $iTexHeight

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 640, $iHeight = 480

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30
Global Const $iStatusBarHeight = 28
Global Const $iStatusBarRows   =  2

Global $hStatusbar, $idRadio0, $idRadio1, $idRadio2, $idLabel
Global $bMode = 0, $bNewMode = $bMode

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight + $iStatusBarHeight * $iStatusBarRows

Global $hGui, $idLabel, $bResize = True

; --- Project includes ---

#include "utilities\includes.au3"
#include "utilities\images.au3"

; --- Program code ---

MainFunc()

While 1
	If $bMode <> $bNewMode Then
		$bMode = $bNewMode
	EndIf
	If $bResize Then
		glViewport( 0, 0, $iWidth, $iHeight )
		$bResize = False
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then GUI_EVENT_CLOSE()
	If Not InitOpenGLwindow() Then GUI_EVENT_CLOSE()
	If Not CheckOpenGLversion( 2, 1 ) Then GUI_EVENT_CLOSE()
	If Not CheckVideoCard() Then GUI_EVENT_CLOSE()
	If Not CreateResources() Then GUI_EVENT_CLOSE()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 50, $WS_OVERLAPPEDWINDOW )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	GUIRegisterMsg( $WM_SIZING, "WM_SIZING" )

	CreateStatusbar()

	Return 1

EndFunc


Func CreateStatusbar()

	; Use a child window as a container for controls in the Statusbar
	$hStatusbar = GUICreate( "", $iGuiWidth - 4 , 24, 2, $iHeight + $iStatusBarHeight + 3 , $WS_CHILD, -1, $hGui )

	; Graph
	GUIStartGroup()
	GUICtrlCreateLabel( "Graph:", 2, 5, 40, 14 )
	$idRadio0 = GUICtrlCreateRadio( "Line",     45, 3, 50, 18 )
	$idRadio1 = GUICtrlCreateRadio( "Points",   98, 3, 50, 18 )
	$idRadio2 = GUICtrlCreateRadio( "Sprites", 158, 3, 50, 18 )
	GUICtrlSetOnEvent( $idRadio0, "RadioClick" )
	GUICtrlSetOnEvent( $idRadio1, "RadioClick" )
	GUICtrlSetOnEvent( $idRadio2, "RadioClick" )
	GUICtrlSetState( $idRadio0, $GUI_CHECKED )

	; Vertical splitterbar
	GUICtrlCreateLabel( "", 225, 2, 3, 20, $SS_ETCHEDFRAME )

	; Label in statusbar
	GUICtrlCreateLabel( "", 245, 5, 250, 14 )
	GUICtrlSetData( -1, "Use arrow keys to move and zoom. Home to reset." )

	GUISetState( @SW_SHOW, $hStatusbar )
	GUISwitch( $hGui )

	; Label for $hStatusbar to make it look nicer
	GUICtrlCreateLabel( "", 0, $iHeight + 1, $iGuiWidth + 4, $iStatusBarHeight + 1, $SS_ETCHEDFRAME )
	GUICtrlSetResizing( -1, $GUI_DOCKSTATEBAR )
	GUICtrlCreateLabel( "Plotting a simple function, using vertex buffer objects and point sprites.", _
											3, $iHeight + 1 + 6, $iGuiWidth - 6, $iStatusBarHeight - 10 )
	GUICtrlSetResizing( -1, $GUI_DOCKSTATEBAR )

EndFunc


Func RadioClick()
	Switch @GUI_CTRLID
		Case $idRadio0
			$bNewMode = 0
		Case $idRadio1
			$bNewMode = 1
		Case $idRadio2
			$bNewMode = 2
	EndSwitch
EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 0.0, 0.0, 0.0, 0.0 )

  glEnable( $GL_BLEND )
  glBlendFunc( $GL_SRC_ALPHA, $GL_ONE_MINUS_SRC_ALPHA )

  glEnable( $GL_VERTEX_PROGRAM_POINT_SIZE )

  glEnable( $GL_POINT_SPRITE )

	GUISwitch( $hGui )

	Return 1

EndFunc


Func CreateResources()

	; --- Shaders ---

	Local $VertexShaderId, $FragmentShaderId, $bStatus

	If Not LoadGLSLshader( $VertexShaderId, "shaders\graph01.v.glsl", $GL_VERTEX_SHADER ) Then Return 0
	If Not LoadGLSLshader( $FragmentShaderId, "shaders\graph01.f.glsl", $GL_FRAGMENT_SHADER ) Then Return 0

	$iProgram = glCreateProgram()
	glAttachShader( $iProgram, $VertexShaderId )
	glAttachShader( $iProgram, $FragmentShaderId )
	glLinkProgram( $iProgram )
	glGetProgramiv( $iProgram, $GL_LINK_STATUS, $bStatus )
	If $bStatus = 0 Then
		PrintError( "Could not link program" )
		Return 0
	EndIf

	$iAttributeCoord2d = glGetAttribLocation( $iProgram, "coord2d" )
	If $iAttributeCoord2d = -1 Then
		PrintError( "Could not bind attribute ""coord2d""" )
		Return 0
	EndIf

  $iUniformOffsetX = glGetUniformLocation( $iProgram, "offset_x" )
	If $iUniformOffsetX = -1 Then
		PrintError( "Could not bind uniform ""offset_x""" )
		Return 0
	EndIf

  $iUniformScaleX = glGetUniformLocation( $iProgram, "scale_x" )
	If $iUniformScaleX = -1 Then
		PrintError( "Could not bind uniform ""scale_x""" )
		Return 0
	EndIf

  $iUniformSprite = glGetUniformLocation( $iProgram, "sprite" )
	If $iUniformSprite = -1 Then
		PrintError( "Could not bind uniform ""sprite""" )
		Return 0
	EndIf

	; --- Upload texture for point sprites ---

	If Not LoadTextureImage( $iTextureId, $images & "\WikiBooks\res_texture.png", $iTexWidth, $iTexHeight ) Then Return 0

	; --- Create array with graph data ---

  Local $aGraph[2*2000]
	Local $x, $y, $j = 0
	For $i = 0 To 1999
		$x = ( $i - 1000.0 ) / 100.0
		$aGraph[$j+0] = $x
		$aGraph[$j+1] = Sin( $x * 10.0 ) / ( 1.0 + $x * $x )
		$j += 2
	Next

	; --- Create buffer object and copy graph data ---

  glGenBuffers( 1, $iVbo )
  glBindBuffer( $GL_ARRAY_BUFFER, $iVbo )
	glBufferData( $GL_ARRAY_BUFFER, $GL_FLOAT, 4*UBound($aGraph), $aGraph, $GL_STATIC_DRAW )

	GUISetState()

	Return 1

EndFunc


Func DisplayFunction()

	HandleKeystrokes()

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )

	glUseProgram( $iProgram )

	glUniform1f( $iUniformScaleX, $fScaleX )
	glUniform1f( $iUniformOffsetX, $fOffsetX )

	; Draw using vertices in the vertex buffer object
	glBindBuffer( $GL_ARRAY_BUFFER, $iVbo )

	glEnableVertexAttribArray( $iAttributeCoord2d )
	glVertexAttribPointer( _
		$iAttributeCoord2d,  _ ; attribute
		2,                   _ ; number of elements per vertex, here (x,y)
		$GL_FLOAT,           _ ; the type of each element
		$GL_FALSE,           _ ; take our values as-is
		0,                   _ ; no space between values
		0 )                    ; use the vertex buffer object

  ; Push each element in buffer_vertices to the vertex shader
	Switch $bMode
		Case 0
			glUniform1f( $iUniformSprite, 0 )
			glDrawArrays( $GL_LINE_STRIP, 0, 2000 )
		Case 1
			glUniform1f( $iUniformSprite, 1 )
			glDrawArrays( $GL_POINTS, 0, 2000 )
		Case 2
			glUniform1f( $iUniformSprite, $iTexWidth )
			glDrawArrays( $GL_POINTS, 0, 2000 )
	EndSwitch

	; Stop using the vertex buffer object
	glDisableVertexAttribArray( $iAttributeCoord2d )
	glBindBuffer( $GL_ARRAY_BUFFER, 0 )

	SwapBuffers( $hDC )

EndFunc


Func HandleKeystrokes()

	Select
		Case _IsPressed( "24", $dllUser32 ) ; Home
			$fScaleX = 1.0
			$fOffsetX = 0.0
		Case _IsPressed( "25", $dllUser32 ) ; Left
			$fOffsetX -= 0.1
		Case _IsPressed( "27", $dllUser32 ) ; Right
			$fOffsetX += 0.1
		Case _IsPressed( "26", $dllUser32 ) ; Up
			$fScaleX *= 1.1
		Case _IsPressed( "28", $dllUser32 ) ; Down
			$fScaleX /= 1.1
	EndSelect

EndFunc


Func WM_SIZING( $hWnd, $iMsg, $iwParam, $ilParam )
	#forceref $iMsg, $iwParam
	Switch $hWnd
		Case $hGui
			Local $tRECT = DllStructCreate( $tagRECT, $ilParam )
			Local $y = DllStructGetData( $tRECT, "Top" )
			Local $iHeight = DllStructGetData( $tRECT, "Bottom" ) - $y - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight
			WinMove( $hStatusbar, "", 2, $iHeight + $iHeightXP )
	EndSwitch
	Return $GUI_RUNDEFMSG
EndFunc


Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight * $iStatusBarRows
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	WinMove( $hStatusbar, "", 2, $iHeight + $iHeightXP + $iStatusBarHeight * ( $iStatusBarRows - 1 ) )
	$bResize = True
EndFunc


Func GUI_EVENT_CLOSE()
	glDeleteBuffers( 1, $iVbo )
	glDeleteProgram( $iProgram )
  glDeleteTextures( 1, $iTextureId )
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
