#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>
#include <Misc.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

Global $fOffsetX = 0.0, $fScaleX = 1.0

; --- OpenGL globals ---

Global $iProgram, $iVbo
Global $iAttributeCoord1d, $iUniformOffsetX, $iUniformScaleX
Global $iTextureId

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 640, $iHeight = 480

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30
Global Const $iStatusBarHeight = 28
Global Const $iStatusBarRows   =  2

Global $hStatusbar, $idShowPoints
Global $bShowPoints = False

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight + $iStatusBarHeight * $iStatusBarRows

Global $hGui, $idLabel, $bResize = True

; --- Project includes ---

#include "utilities\includes.au3"

; --- Program code ---

MainFunc()

While 1
	If $bResize Then
		glViewport( 0, 0, $iWidth, $iHeight )
		$bResize = False
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then GUI_EVENT_CLOSE()
	If Not InitOpenGLwindow() Then GUI_EVENT_CLOSE()
	If Not CheckOpenGLversion( 2, 1 ) Then GUI_EVENT_CLOSE()
	If Not CheckVideoCard() Then GUI_EVENT_CLOSE()
	If Not CreateResources() Then GUI_EVENT_CLOSE()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 50, $WS_OVERLAPPEDWINDOW )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	GUIRegisterMsg( $WM_SIZING, "WM_SIZING" )

	CreateStatusbar()

	Return 1

EndFunc


Func CreateStatusbar()

	; Use a child window as a container for controls in the Statusbar
	$hStatusbar = GUICreate( "", $iGuiWidth - 4 , 24, 2, $iHeight + $iStatusBarHeight + 3 , $WS_CHILD, -1, $hGui )

	; Show points
	$idShowPoints = GUICtrlCreateCheckbox( "Show points", 2, 3, 80, 18 )
	GUICtrlSetOnEvent( $idShowPoints, "ShowPoints" )

	; Vertical splitterbar
	GUICtrlCreateLabel( "", 100, 2, 3, 20, $SS_ETCHEDFRAME )

	; Label in statusbar
	GUICtrlCreateLabel( "", 120, 5, 250, 14 )
	GUICtrlSetData( -1, "Use arrow keys to move and zoom. Home to reset." )

	GUISetState( @SW_SHOW, $hStatusbar )
	GUISwitch( $hGui )

	; Label for $hStatusbar to make it look nicer
	GUICtrlCreateLabel( "", 0, $iHeight + 1, $iGuiWidth + 4, $iStatusBarHeight + 1, $SS_ETCHEDFRAME )
	GUICtrlSetResizing( -1, $GUI_DOCKSTATEBAR )
	GUICtrlCreateLabel( "Plotting a graph from data in a texture.", _
											3, $iHeight + 1 + 6, $iGuiWidth - 6, $iStatusBarHeight - 10 )
	GUICtrlSetResizing( -1, $GUI_DOCKSTATEBAR )

EndFunc


Func ShowPoints()
	$bShowPoints = Not $bShowPoints
EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 0.0, 0.0, 0.0, 0.0 )

	GUISwitch( $hGui )

	Return 1

EndFunc


Func CreateResources()

	; --- Shaders ---

	Local $VertexShaderId, $FragmentShaderId, $bStatus

	If Not LoadGLSLshader( $VertexShaderId, "shaders\graph02.v.glsl", $GL_VERTEX_SHADER ) Then Return 0
	If Not LoadGLSLshader( $FragmentShaderId, "shaders\graph02.f.glsl", $GL_FRAGMENT_SHADER ) Then Return 0

	$iProgram = glCreateProgram()
	glAttachShader( $iProgram, $VertexShaderId )
	glAttachShader( $iProgram, $FragmentShaderId )
	glLinkProgram( $iProgram )
	glGetProgramiv( $iProgram, $GL_LINK_STATUS, $bStatus )
	If $bStatus = 0 Then
		PrintError( "Could not link program" )
		Return 0
	EndIf

	$iAttributeCoord1d = glGetAttribLocation( $iProgram, "coord1d" )
	If $iAttributeCoord1d = -1 Then
		PrintError( "Could not bind attribute ""coord1d""" )
		Return 0
	EndIf

  $iUniformOffsetX = glGetUniformLocation( $iProgram, "offset_x" )
	If $iUniformOffsetX = -1 Then
		PrintError( "Could not bind uniform ""offset_x""" )
		Return 0
	EndIf

  $iUniformScaleX = glGetUniformLocation( $iProgram, "scale_x" )
	If $iUniformScaleX = -1 Then
		PrintError( "Could not bind uniform ""scale_x""" )
		Return 0
	EndIf

	; --- Create texture with y values for the graph ---

  Local $graph = DllStructCreate( "byte[2048]" ), $x, $y
	For $i = 0 To 2047
		$x = ( $i - 1024.0 ) / 100.0
		$y = Sin( $x * 10.0 ) / ( 1.0 + $x * $x )
		DllStructSetData( $graph, 1, Roundf( $y * 128 + 128 ), $i+1 )
	Next

	If Not CreateTexture( $iTextureId, 2048, 1, $GL_LUMINANCE, DllStructGetPtr( $graph ) ) Then Return 0

	; --- Create vertex buffer object with x values ---

  Local $line[101]
	For $i = 0 To 100
    $line[$i] = ( $i - 50 ) / 50.0
	Next

  glGenBuffers( 1, $iVbo )
  glBindBuffer( $GL_ARRAY_BUFFER, $iVbo )
	glBufferData( $GL_ARRAY_BUFFER, $GL_FLOAT, 4*UBound($line), $line, $GL_STATIC_DRAW )

  glEnable( $GL_VERTEX_PROGRAM_POINT_SIZE ) ; Enable point size control in vertex shader

	GUISetState()

	Return 1

EndFunc


Func DisplayFunction()

	HandleKeystrokes()

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )

	glUseProgram( $iProgram )

	glUniform1f( $iUniformScaleX, $fScaleX )
	glUniform1f( $iUniformOffsetX, $fOffsetX )

	; Draw using vertices in the vertex buffer object
	glBindBuffer( $GL_ARRAY_BUFFER, $iVbo )

	glEnableVertexAttribArray( $iAttributeCoord1d )
	glVertexAttribPointer( _
		$iAttributeCoord1d,  _ ; attribute
		1,                   _ ; number of elements per vertex, here just x
		$GL_FLOAT,           _ ; the type of each element
		$GL_FALSE,           _ ; take our values as-is
		0,                   _ ; no space between values
		0 )                    ; use the vertex buffer object

	; Draw the line
	glDrawArrays( $GL_LINE_STRIP, 0, 101 )

	; Draw points if requested
	If $bShowPoints Then _
		glDrawArrays( $GL_POINTS, 0, 101 )

	; Stop using the vertex buffer object
	glDisableVertexAttribArray( $iAttributeCoord1d )
	glBindBuffer( $GL_ARRAY_BUFFER, 0 )

	SwapBuffers( $hDC )

EndFunc


Func HandleKeystrokes()

	Select
		Case _IsPressed( "24", $dllUser32 ) ; Home
			$fScaleX = 1.0
			$fOffsetX = 0.0
		Case _IsPressed( "25", $dllUser32 ) ; Left
			$fOffsetX -= 0.1
		Case _IsPressed( "27", $dllUser32 ) ; Right
			$fOffsetX += 0.1
		Case _IsPressed( "26", $dllUser32 ) ; Up
			$fScaleX *= 1.1
		Case _IsPressed( "28", $dllUser32 ) ; Down
			$fScaleX /= 1.1
	EndSelect

EndFunc


Func WM_SIZING( $hWnd, $iMsg, $iwParam, $ilParam )
	#forceref $iMsg, $iwParam
	Switch $hWnd
		Case $hGui
			Local $tRECT = DllStructCreate( $tagRECT, $ilParam )
			Local $y = DllStructGetData( $tRECT, "Top" )
			Local $iHeight = DllStructGetData( $tRECT, "Bottom" ) - $y - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight
			WinMove( $hStatusbar, "", 2, $iHeight + $iHeightXP )
	EndSwitch
	Return $GUI_RUNDEFMSG
EndFunc


Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight * $iStatusBarRows
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	WinMove( $hStatusbar, "", 2, $iHeight + $iHeightXP + $iStatusBarHeight * ( $iStatusBarRows - 1 ) )
	$bResize = True
EndFunc


Func GUI_EVENT_CLOSE()
	glDeleteBuffers( 1, $iVbo )
	glDeleteProgram( $iProgram )
  glDeleteTextures( 1, $iTextureId )
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
