#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>
#include <Misc.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

Global $fOffsetX = 0.0, $fScaleX = 1.0, $fPixelX, $fPixelY
Global Const $iBorder = 10, $iTickSize = 10

; --- OpenGL globals ---

Global $iProgram, $aVbo[3]
Global $iAttributeCoord2d, $iUniformTransform, $iUniformColor

Global $mIdentity, $mVPtransform

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 640, $iHeight = 480

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30
Global Const $iStatusBarHeight = 28

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight + $iStatusBarHeight

Global $hGui, $bResize = True

; --- Project includes ---

#include "utilities\includes.au3"

; --- Program code ---

MainFunc()

While 1
	If $bResize Then
		; Only necessary to do this calculation when the OpenGL window is resized
	  $mVPtransform = ViewportTransform( _
	    $iBorder + $iTickSize, _
	    $iBorder + $iTickSize, _
	    $iWidth - $iBorder * 2 - $iTickSize, _
	    $iHeight - $iBorder * 2 - $iTickSize, _
	    $fPixelX, $fPixelY )
		$bResize = False
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then GUI_EVENT_CLOSE()
	If Not InitOpenGLwindow() Then GUI_EVENT_CLOSE()
	If Not CheckOpenGLversion( 2, 1 ) Then GUI_EVENT_CLOSE()
	If Not CheckVideoCard() Then GUI_EVENT_CLOSE()
	If Not CreateResources() Then GUI_EVENT_CLOSE()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 50, $WS_OVERLAPPEDWINDOW )

	; Statusbar
	Local $idStatus, $idLabel
	$idStatus = GUICtrlCreateLabel( "", 0, $iHeight + 1, $iGuiWidth + 2, $iStatusBarHeight, $SS_ETCHEDFRAME )
	$idLabel = GUICtrlCreateLabel( "", 3, $iHeight + 1 + 6, $iGuiWidth - 4, $iStatusBarHeight - 10 )
	GUICtrlSetData( $idLabel, "Plotting borders and axes, clipping.  |  Use arrow keys to move and zoom. Home to reset." )
	GUICtrlSetResizing( $idStatus, $GUI_DOCKSTATEBAR )
	GUICtrlSetResizing( $idLabel, $GUI_DOCKSTATEBAR )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	Return 1

EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 1.0, 1.0, 1.0, 1.0 )

	GUISwitch( $hGui )

	Return 1

EndFunc


Func CreateResources()

	; --- Shaders ---

	Local $VertexShaderId, $FragmentShaderId, $bStatus

	If Not LoadGLSLshader( $VertexShaderId, "shaders\graph03.v.glsl", $GL_VERTEX_SHADER ) Then Return 0
	If Not LoadGLSLshader( $FragmentShaderId, "shaders\graph03.f.glsl", $GL_FRAGMENT_SHADER ) Then Return 0

	$iProgram = glCreateProgram()
	glAttachShader( $iProgram, $VertexShaderId )
	glAttachShader( $iProgram, $FragmentShaderId )
	glLinkProgram( $iProgram )
	glGetProgramiv( $iProgram, $GL_LINK_STATUS, $bStatus )
	If $bStatus = 0 Then
		PrintError( "Could not link program" )
		Return 0
	EndIf

	$iAttributeCoord2d = glGetAttribLocation( $iProgram, "coord2d" )
	If $iAttributeCoord2d = -1 Then
		PrintError( "Could not bind attribute ""coord2d""" )
		Return 0
	EndIf

  $iUniformTransform = glGetUniformLocation( $iProgram, "transform" )
	If $iUniformTransform = -1 Then
		PrintError( "Could not bind uniform ""transform""" )
		Return 0
	EndIf

  $iUniformColor = glGetUniformLocation( $iProgram, "color" )
	If $iUniformColor = -1 Then
		PrintError( "Could not bind uniform ""color""" )
		Return 0
	EndIf

	; --- Create array with graph data ---

  Local $aGraph[2*2000]
	Local $x, $y, $j = 0
	For $i = 0 To 1999
		$x = ( $i - 1000.0 ) / 100.0
		$aGraph[$j+0] = $x
		$aGraph[$j+1] = Sin( $x * 10.0 ) / ( 1.0 + $x * $x )
		$j += 2
	Next

	; --- Create buffer objects and copy data ---

  glGenBuffers( 3, $aVbo )

	; Copy graph data to the buffer object
  glBindBuffer( $GL_ARRAY_BUFFER, $aVbo[0] )
	glBufferData( $GL_ARRAY_BUFFER, $GL_FLOAT, 4*UBound($aGraph), $aGraph, $GL_STATIC_DRAW )

  ; VBO for the $iBorder
  Local $aBorder[8] = [ -1, -1, 1, -1, 1, 1, -1, 1 ]
  glBindBuffer( $GL_ARRAY_BUFFER, $aVbo[1] )
  glBufferData( $GL_ARRAY_BUFFER, $GL_FLOAT, 4*UBound($aBorder), $aBorder, $GL_STATIC_DRAW )

	GUISetState()

	Return 1

EndFunc


; Create a projection matrix that has the same effect as glViewport
; Return scaling factors to easily convert normalized device coordinates to pixels
Func ViewportTransform( $x, $y, $width, $height, ByRef $fPixelX, ByRef $fPixelY )

	$mIdentity = $IDENTITY_MATRIX

	; Map OpenGL coordinates (-1,-1) to window coordinates (x,y)
	; Map OpenGL coordinates ( 1, 1) to (x + width, y + height)

	; Calculate how to translate the x and y coordinates:
	Local $fOffsetX = ( 2.0 * $x + ( $width  - $iWidth  ) ) / $iWidth
	Local $fOffsetY = ( 2.0 * $y + ( $height - $iHeight ) ) / $iHeight
	Local $vOffset[3] = [ $fOffsetX, $fOffsetY, 0 ]

	; Calculate how to rescale the x and y coordinates
	Local $fScaleX = $width  / $iWidth
	Local $fScaleY = $height / $iHeight
	Local $vScale[3] = [ $fScaleX, $fScaleY, 0 ]

	; Calculate size of pixels in OpenGL coordinates
	$fPixelX = 2.0 / $width
	$fPixelY = 2.0 / $height

	Local $mTranslate = glmTranslate( $mIdentity, $vOffset )
	Local $mScale = glmScale( $mTranslate, $vScale )
	Return $mScale

EndFunc


Func DisplayFunction()

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )

	HandleKeystrokes()

	glUseProgram( $iProgram )

	; --- Draw the graph ---

	; Set our viewport, this will clip geometry
  glViewport( _
		$iBorder + $iTickSize, _
		$iBorder + $iTickSize, _
		$iWidth - $iBorder * 2 - $iTickSize, _
		$iHeight - $iBorder * 2 - $iTickSize )

	; Set the scissor rectangle, this will clip fragments
  glScissor( _
		$iBorder + $iTickSize, _
		$iBorder + $iTickSize, _
		$iWidth - $iBorder * 2 - $iTickSize, _
		$iHeight - $iBorder * 2 - $iTickSize )

  glEnable( $GL_SCISSOR_TEST )

  ; Set our coordinate transformation matrix
	Local $vScale[3] = [ $fScaleX, 1, 1 ], $vOffset[3] = [ $fOffsetX, 0, 0 ]
	Local $mScale = glmScale( $mIdentity, $vScale ), $mTransform = glmTranslate( $mScale, $vOffset )
	glUniformMatrix4fv( $iUniformTransform, 1, $GL_FALSE, $mTransform )

  ; Set the color to red
	Local $red[4] = [ 1, 0, 0, 1 ]
	glUniform4fv( $iUniformColor, 1, $red )

	glEnableVertexAttribArray( $iAttributeCoord2d )

	; Draw using the vertices in our vertex buffer object
	glBindBuffer( $GL_ARRAY_BUFFER, $aVbo[0] )
	glVertexAttribPointer( _
		$iAttributeCoord2d,  _ ; attribute
		2,                   _ ; number of elements per vertex, here (x,y)
		$GL_FLOAT,           _ ; the type of each element
		$GL_FALSE,           _ ; take our values as-is
		0,                   _ ; no space between values
		0 )                    ; use the vertex buffer object
	glDrawArrays( $GL_LINE_STRIP, 0, 2000 )

	; Stop clipping
  glViewport( 0, 0, $iWidth, $iHeight )
  glDisable( $GL_SCISSOR_TEST )

	; --- Draw the borders ---

	; Tell the vertex shader about the transformation matrix that gives us normalized device coordinates
	glUniformMatrix4fv( $iUniformTransform, 1, $GL_FALSE, $mVPtransform )

	; Set the color to black
	Local $black[4] = [ 0, 0, 0, 1 ]
	glUniform4fv( $iUniformColor, 1, $black )

	; Draw a border around our graph
  glBindBuffer( $GL_ARRAY_BUFFER, $aVbo[1] )
	glVertexAttribPointer( _
		$iAttributeCoord2d,  _ ; attribute
		2,                   _ ; number of elements per vertex, here (x,y)
		$GL_FLOAT,           _ ; the type of each element
		$GL_FALSE,           _ ; take our values as-is
		0,                   _ ; no space between values
		0 )                    ; use the vertex buffer object
  glDrawArrays( $GL_LINE_LOOP, 0, 4 )

	; --- Draw the y tick marks ---

  Local $aTicks[2*42], $fTickScale, $y, $j = 0

	For $i = 0 To 20
		$y = -1 + $i * 0.1
		If Mod( $i, 10 ) Then
			$fTickScale = 0.5
		Else
			$fTickScale = 1.0
		EndIf
		$aTicks[$j + 0] = -1
		$aTicks[$j + 1] = $y
		$aTicks[$j + 2] = -1 - $iTickSize * $fTickScale * $fPixelX
		$aTicks[$j + 3] = $y
		$j += 4
	Next

  glBindBuffer( $GL_ARRAY_BUFFER, $aVbo[2] )

	glBufferData( $GL_ARRAY_BUFFER, $GL_FLOAT, 4*UBound($aTicks), $aTicks, $GL_DYNAMIC_DRAW )
	glVertexAttribPointer( _
		$iAttributeCoord2d,  _ ; attribute
		2,                   _ ; number of elements per vertex, here (x,y)
		$GL_FLOAT,           _ ; the type of each element
		$GL_FALSE,           _ ; take our values as-is
		0,                   _ ; no space between values
		0 )                    ; use the vertex buffer object
  glDrawArrays( $GL_LINES, 0, 42 )

	; --- Draw the x tick marks ---

	Local $fTickSpacing = 0.1 * 10^-Floor( Log10( $fScaleX ) )	; desired space between ticks, in graph coordinates
	Local $fLeft = -1.0 / $fScaleX - $fOffsetX									; left edge, in graph coordinates
	Local $fRight = 1.0 / $fScaleX - $fOffsetX									; right edge, in graph coordinates
	Local $iLeft = Ceiling( $fLeft  / $fTickSpacing )						; index of left tick, counted from the origin
	Local $iRight =  Floor( $fRight / $fTickSpacing )						; index of right tick, counted from the origin
	Local $rem = $iLeft * $fTickSpacing - $fLeft								; space between left edge of graph and the first tick

  Local $fFirstTick = -1.0 + $rem * $fScaleX									; first tick in device coordinates

  Local $iTicks = $iRight - $iLeft + 1												; number of ticks to show

	Local $x, $j = 0
	For $i = 0 To $iTicks - 1
		$x = $fFirstTick + $i * $fTickSpacing * $fScaleX
		If Mod( $i + $iLeft, 10 ) Then
			$fTickScale = 0.5
		Else
			$fTickScale = 1.0
		EndIf
		$aTicks[$j + 0] = $x
		$aTicks[$j + 1] = -1
		$aTicks[$j + 2] = $x
		$aTicks[$j + 3] = -1 - $iTickSize * $fTickScale * $fPixelY
		$j += 4
	Next

	glBufferData( $GL_ARRAY_BUFFER, $GL_FLOAT, 4*UBound($aTicks), $aTicks, $GL_DYNAMIC_DRAW )
	glVertexAttribPointer( _
		$iAttributeCoord2d,  _ ; attribute
		2,                   _ ; number of elements per vertex, here (x,y)
		$GL_FLOAT,           _ ; the type of each element
		$GL_FALSE,           _ ; take our values as-is
		0,                   _ ; no space between values
		0 )                    ; use the vertex buffer object
  glDrawArrays( $GL_LINES, 0, $iTicks * 2 )

	; --- We are done ---

	glDisableVertexAttribArray( $iAttributeCoord2d )

	SwapBuffers( $hDC )

EndFunc


Func HandleKeystrokes()

	Select
		Case _IsPressed( "24", $dllUser32 ) ; Home
			$fScaleX = 1.0
			$fOffsetX = 0.0
		Case _IsPressed( "25", $dllUser32 ) ; Left
			$fOffsetX -= 0.1
		Case _IsPressed( "27", $dllUser32 ) ; Right
			$fOffsetX += 0.1
		Case _IsPressed( "26", $dllUser32 ) ; Up
			$fScaleX *= 1.1
		Case _IsPressed( "28", $dllUser32 ) ; Down
			$fScaleX /= 1.1
	EndSelect

EndFunc


Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	$bResize = True
EndFunc


Func GUI_EVENT_CLOSE()
	glDeleteBuffers( 3, $aVbo )
	glDeleteProgram( $iProgram )
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
