#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>
#include <Misc.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

Global $fOffsetX = 0.0, $fOffsetY = 0.0, $fScale = 1.0
Global $fLastTime = 0.0, $fLastElapsed = 0.0
Global $iStartTime = 0, $fElapsedTime = 0

; --- OpenGL globals ---

Global $iProgram, $aVbo[2]
Global $iAttributeCoord2d, $iUniformVertexTransform, $iUniformTextureTransform
Global $iTextureId

Global $mIdentity, $mModel, $mView, $mProjection, $mVP
Global $aAxisZ[3] = [ 0.0, 0.0, 1.0 ]

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 640, $iHeight = 480

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30
Global Const $iStatusBarHeight = 28
Global Const $iStatusBarRows   =  2

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight + $iStatusBarHeight * $iStatusBarRows

Global $hGui, $bResize = True

; --- Project includes ---

#include "utilities\includes.au3"
#include "utilities\utils0405.au3"

; --- Program code ---

MainFunc()

While 1
	If $bResize Then
		glViewport( 0, 0, $iWidth, $iHeight )
		$mProjection = glmPerspective( 45.0, 1.0*$iWidth/$iHeight, 0.1, 10.0 )
		$mVP = glmMultiply( $mProjection, $mView )
		$bResize = False
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then GUI_EVENT_CLOSE()
	If Not InitOpenGLwindow() Then GUI_EVENT_CLOSE()
	If Not CheckOpenGLversion( 2, 1 ) Then GUI_EVENT_CLOSE()
	If Not CheckVideoCard() Then GUI_EVENT_CLOSE()
	If Not CreateResources() Then GUI_EVENT_CLOSE()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 50, $WS_OVERLAPPEDWINDOW )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	GUIRegisterMsg( $WM_SIZING, "WM_SIZING" )

	CreateProgressbar()

	CreateStatusbar( 4 )

	Return 1

EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 0.0, 0.0, 0.0, 0.0 )

	glEnable( $GL_DEPTH_TEST )

	GUISwitch( $hGui )
	GUISetState()

	Return 1

EndFunc


Func CreateResources()

	; --- Shaders ---

	Local $VertexShaderId, $FragmentShaderId, $bStatus

	If Not LoadGLSLshader( $VertexShaderId, "shaders\graph04.v.glsl", $GL_VERTEX_SHADER ) Then Return 0
	If Not LoadGLSLshader( $FragmentShaderId, "shaders\graph04.f.glsl", $GL_FRAGMENT_SHADER ) Then Return 0

	$iProgram = glCreateProgram()
	glAttachShader( $iProgram, $VertexShaderId )
	glAttachShader( $iProgram, $FragmentShaderId )
	glLinkProgram( $iProgram )
	glGetProgramiv( $iProgram, $GL_LINK_STATUS, $bStatus )
	If $bStatus = 0 Then
		PrintError( "Could not link program" )
		Return 0
	EndIf

	$iAttributeCoord2d = glGetAttribLocation( $iProgram, "coord2d" )
	If $iAttributeCoord2d = -1 Then
		PrintError( "Could not bind attribute ""coord2d""" )
		Return 0
	EndIf

	$iUniformVertexTransform = glGetUniformLocation( $iProgram, "vertex_transform" )
	If $iUniformVertexTransform = -1 Then
		PrintError( "Could not bind uniform ""vertex_transform""" )
		Return 0
	EndIf

	$iUniformTextureTransform = glGetUniformLocation( $iProgram, "texture_transform" )
	If $iUniformTextureTransform = -1 Then
		PrintError( "Could not bind uniform ""texture_transform""" )
		Return 0
	EndIf

	; --- Create texture with z values for the graph ---

	Local $n = 256, $x, $y, $d, $z, $graph
	$graph = DllStructCreate( "byte[" & $n * $n & "]" )
	For $i = 0 To $n - 1
		$x = ( $i - $n / 2.0 ) / ( $n / 2.0 )
		For $j = 0 To $n - 1
			$y = ( $j - $n / 2.0 ) / ( $n / 2.0 )
			$d = Hypot( $x, $y ) * 4.0
			$d *= $d
			$z = ( 1 - $d ) * Exp( $d / -2.0 )
			DllStructSetData( $graph, 1, Roundf( $z * 127 + 128 ), $i * $n + $j + 1 )
		Next
		UpdateProgressBar( $idProgress2, $i, $n - 1 )
	Next
	UpdateProgressBar( $idProgress1, 90, 100 )

	If Not CreateTexture( $iTextureId, $n, $n, $GL_LUMINANCE, DllStructGetPtr( $graph ) ) Then Return 0

	; --- Create vertex buffer objects ---

	glGenBuffers( 2, $aVbo )

	; Create an array for 101 * 101 vertices
	Local $n = 101, $k = 0
	Local $vertices[ $n * $n * 2 + 1 ]

	For $i = 0 To $n - 1
		For $j = 0 To $n - 1
			$vertices[$k]   = ( $j - 50 ) / 50.0
			$vertices[$k+1] = ( $i - 50 ) / 50.0
			$k += 2
		Next
	Next
	UpdateProgressBar( $idProgress2, 33, 100 )

	glBindBuffer( $GL_ARRAY_BUFFER, $aVbo[0] )
	glBufferData( $GL_ARRAY_BUFFER, $GL_FLOAT, 4*UBound($vertices), $vertices, $GL_STATIC_DRAW )

	; Create an array of indices into the vertex array that traces both horizontal and vertical lines
	Local $indices[ ($n-1) * $n * 4 ], $k = 0

	For $y = 0 To $n - 2
		For $x = 0 To $n - 2
			$indices[$k]   = $y * $n + $x
			$indices[$k+1] = $y * $n + $x + 1
			$k += 2
		Next
	Next
	UpdateProgressBar( $idProgress2, 66, 100 )

	For $x = 0 To $n - 1
		For $y = 0 To $n - 2
			$indices[$k]   = $y * $n + $x
			$indices[$k+1] = ( $y + 1 ) * $n + $x
			$k += 2
		Next
	Next
	UpdateProgressBar( $idProgress2, 100, 100 )
	UpdateProgressBar( $idProgress1, 100, 100 )

	glBindBuffer( $GL_ELEMENT_ARRAY_BUFFER, $aVbo[1] )
	glBufferData( $GL_ELEMENT_ARRAY_BUFFER, $GL_UNSIGNED_SHORT, 2*UBound($indices), $indices, $GL_STATIC_DRAW )

	GUISetState( @SW_HIDE, $hProgress )
	GUICtrlDelete( $idProLabel )
	GUIDelete( $hProgress )
	$hProgress = 0

	CalculateMatrices()

	Return 1

EndFunc


Func CalculateMatrices()

	$mIdentity = $IDENTITY_MATRIX

	;Local $eye[3] = [ 0.0, -2.0, 2.0 ], $center[3] = [ 0.0, 0.0, 0.0 ], $up[3] = [ 0.0, 0.0, 1.0 ]
	Local $eye[3] = [ 0.0, -2.0, 2.0 ], $center[3] = [ 0.0, 0.35, 0.0 ], $up[3] = [ 0.0, 0.0, 1.0 ]
	$mView = glmLookAt( $eye, $center, $up )

EndFunc


Func DisplayFunction()

	HandleKeystrokes()

	CalcElapsedTime()
  $mModel = glmRotate( $mIdentity, $fElapsedTime / 100.0, $aAxisZ )
	Local $mVertexTransform = glmMultiply( $mVP, $mModel )

	Local $v1[3] = [ $fScale, $fScale, 1 ], $v2[3] = [ $fOffsetX, $fOffsetY, 0 ]
	Local $mScale = glmScale( $mIdentity, $v1 )
	Local $mTextureTransform = glmTranslate( $mScale, $v2 )

	glUseProgram( $iProgram )

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )

	glUniformMatrix4fv( $iUniformVertexTransform, 1, $GL_FALSE, $mVertexTransform )
	glUniformMatrix4fv( $iUniformTextureTransform, 1, $GL_FALSE, $mTextureTransform )

	; Draw using vertices in the vertex buffer object
	glBindBuffer( $GL_ARRAY_BUFFER, $aVbo[0] )

	glEnableVertexAttribArray( $iAttributeCoord2d )
	glVertexAttribPointer( _
		$iAttributeCoord2d,  _ ; attribute
		2,                   _ ; number of elements per vertex
		$GL_FLOAT,           _ ; the type of each element
		$GL_FALSE,           _ ; take our values as-is
		0,                   _ ; no space between values
		0 )                    ; use the vertex buffer object

	glBindBuffer( $GL_ELEMENT_ARRAY_BUFFER, $aVbo[1] )
	glDrawElements( $GL_LINES, 100 * 101 * 4, $GL_UNSIGNED_SHORT, 0 )

	; Stop using the vertex buffer object
	glDisableVertexAttribArray( $iAttributeCoord2d )
	glBindBuffer( $GL_ELEMENT_ARRAY_BUFFER, 0 )
	glBindBuffer( $GL_ARRAY_BUFFER, 0 )

	SwapBuffers( $hDC )

EndFunc
