#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>
#include <EditConstants.au3>
#include <File.au3>
#include <Misc.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

Global $sObjectFile
Global $sVshader, $sFshader

Global $aInfo[6] = [ _
	"Diffuse Reflection: About per-vertex diffuse lighting and multiple light sources of different kinds.", _
	"Specular Highlights: About per-vertex lighting.", _
	"Two-Sided Surfaces: About two-sided per-vertex lighting.", _
	"Smooth Specular Highlights: About per-pixel lighting.", _
	"Two-Sided Smooth Surfaces: About two-sided per-pixel lighting.", _
	"Multiple Lights: About for-loops for handling multiple light sources." ]

Global $sInfo = "Use arrow keys to rotate/move. Page up/down to move along z-axis. Home to reset."

; --- OpenGL globals ---

Global $aObject[7], $aLightBox[7], $aGround[7]
; 0 - 2: $iVboVertices, $iVboNormals, $iIboElements
;     3: $aVertices[], $aVec[4]
;     4: $aNormals[], $aVec[3]
;     5: $aElements[], Ushort
;     6: $mObject2world[16]

Global $iProgram
Global $iAttributeVcoord = -1, $iAttributeVnormal = -1
Global $iUniformM = -1, $iUniformV = -1, $iUniformP = -1
Global $iUniformM3x3InvTransp = -1, $iUniformVinv = -1

Global $mTransforms
Global $iRotXdirection = 0, $iRotYdirection = 0, $iTransZdirection = 0

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 880, $iHeight = 660

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30
Global Const $iStatusBarHeight = 28
Global Const $iStatusBarRows   =  2

Global $hStatusbar, $idMonkey, $idTwoSided, $idObject, $idWorld, $idPos1, $idPos2, $idLeft, $idRight, $idLabel2
Global $iMonkeyObejct = 1, $iTwoSidedObejct = 2, $iObejctView = 1, $iWorldView = 2, $iPos1 = 1, $iPos2 = 2
Global $bObject = $iMonkeyObejct, $bNewObject = $bObject, $bView = $iObejctView, $bNewView = $bView, $bPos = 0, $bNewPos = $bPos

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight + $iStatusBarHeight * $iStatusBarRows

Global $hGui, $bResize = True

; --- Project includes ---

#include "utilities\includes.au3"
#include "utilities\arcball.au3"

; --- Program code ---

MainFunc()

While 1
	If $bObject <> $bNewObject Then
		$bObject = $bNewObject
		If $bObject = $iMonkeyObejct Then
			$sObjectFile = "monkey.obj"
		ElseIf $bObject = $iTwoSidedObejct Then
			$sObjectFile = "two-sided.obj"
		EndIf
		LoadObject( "models\" & $sObjectFile, $aObject )
		UploadObject( $aObject )
		InitView()
	EndIf
	If $bView <> $bNewView Then
		$bView = $bNewView
	EndIf
	If $bPos <> $bNewPos Then
		$bPos = $bNewPos
		SetPosition( $bPos )
	EndIf
	If $bResize Then
		glViewport( 0, 0, $iWidth, $iHeight )
		$bResize = False
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then GUI_EVENT_CLOSE()
	If Not InitOpenGLwindow() Then GUI_EVENT_CLOSE()
	If Not CheckOpenGLversion( 2, 1 ) Then GUI_EVENT_CLOSE()
	If Not CreateResources() Then GUI_EVENT_CLOSE()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 10, $WS_OVERLAPPEDWINDOW )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	GUIRegisterMsg( $WM_SIZING, "WM_SIZING" )

	CreateStatusbar()

	Return 1

EndFunc


Func CreateStatusbar()

	; Use a child window as a container for controls in the Statusbar
	$hStatusbar = GUICreate( "", $iGuiWidth - 4, 24, 2, $iHeight + $iStatusBarHeight + 3, $WS_CHILD, -1, $hGui )

	; Object buttons
	GUIStartGroup()
	GUICtrlCreateLabel( "Object:", 3, 5, 40, 14 )
	$idMonkey = GUICtrlCreateRadio( "Monkey",     45, 3, 55, 18 )
	$idTwoSided = GUICtrlCreateRadio( "2-sided", 110, 3, 55, 18 )
	GUICtrlSetOnEvent( $idMonkey, "ObjectButtons" )
	GUICtrlSetOnEvent( $idTwoSided, "ObjectButtons" )
	GUICtrlSetState( $idMonkey, $GUI_CHECKED )

	; Vertical splitterbar
	GUICtrlCreateLabel( "", 171, 2, 3, 20, $SS_ETCHEDFRAME )

	Local $w = 181

	; View buttons
	GUIStartGroup()
	GUICtrlCreateLabel( "View:", $w+3, 5, 30, 14 )
	$idObject = GUICtrlCreateRadio( "Object", $w+35, 3, 55, 18 )
	$idWorld = GUICtrlCreateRadio( "World",   $w+95, 3, 55, 18 )
	GUICtrlSetOnEvent( $idObject, "ViewButtons" )
	GUICtrlSetOnEvent( $idWorld, "ViewButtons" )
	GUICtrlSetState( $idObject, $GUI_CHECKED )

	; Vertical splitterbar
	GUICtrlCreateLabel( "", $w+151, 2, 3, 20, $SS_ETCHEDFRAME )

	; Pos buttons
	GUIStartGroup()
	$idPos1 = GUICtrlCreateRadio( "Pos 1", $w+164, 3, 45, 18 )
	$idPos2 = GUICtrlCreateRadio( "Pos 2", $w+219, 3, 45, 18 )
	GUICtrlSetOnEvent( $idPos1, "PosButtons" )
	GUICtrlSetOnEvent( $idPos2, "PosButtons" )

	; Vertical splitterbar
	GUICtrlCreateLabel( "", $w+275, 2, 3, 20, $SS_ETCHEDFRAME )

	; Arcball buttons
	GUIStartGroup()
	GUICtrlCreateLabel( "Arcball mouse button:", $w+290, 5, 110, 14 )
	$idLeft  = GUICtrlCreateRadio( "Left",  $w+400, 3, 40, 18 )
	$idRight = GUICtrlCreateRadio( "Right", $w+445, 3, 50, 18 )
	GUICtrlSetOnEvent( $idLeft, "ArcballButtons" )
	GUICtrlSetOnEvent( $idRight, "ArcballButtons" )
	GUICtrlSetState( $idLeft, $GUI_CHECKED )

	GUISetState( @SW_SHOW, $hStatusbar )
	GUISwitch( $hGui )

	Local $idLabel1

	; Label for $hStatusbar to make it look nicer
	$idLabel1 = GUICtrlCreateLabel( "", 0, $iHeight + 1, $iGuiWidth + 4, $iStatusBarHeight, $SS_ETCHEDFRAME )
	GUICtrlSetResizing( $idLabel1, $GUI_DOCKSTATEBAR )
	$idLabel2 = GUICtrlCreateLabel( "", 3, $iHeight + 1 + 6, $iGuiWidth - 6, $iStatusBarHeight - 10 )
	GUICtrlSetResizing( $idLabel2, $GUI_DOCKSTATEBAR )

EndFunc


Func ObjectButtons()
	If @GUI_CTRLID = $idMonkey Then
		$bNewObject = $iMonkeyObejct
	ElseIf @GUI_CTRLID = $idTwoSided Then
		$bNewObject = $iTwoSidedObejct
	EndIf
EndFunc


Func ViewButtons()
	If @GUI_CTRLID = $idObject Then
		$bNewView = $iObejctView
	ElseIf @GUI_CTRLID = $idWorld Then
		$bNewView = $iWorldView
	EndIf
EndFunc


Func PosButtons()
	If @GUI_CTRLID = $idPos1 Then
		$bNewPos = $iPos1
	ElseIf @GUI_CTRLID = $idPos2 Then
		$bNewPos = $iPos2
	EndIf
EndFunc


Func ResetPosButtons()
	$bPos = 0
	$bNewPos = $bPos
	GUICtrlSetState( $idPos1, $GUI_UNCHECKED )
	GUICtrlSetState( $idPos2, $GUI_UNCHECKED )
EndFunc


Func ArcballButtons()
	If @GUI_CTRLID = $idLeft Then
		$fSwapMouseButtons = False
	ElseIf @GUI_CTRLID = $idRight Then
		$fSwapMouseButtons = True
	EndIf
EndFunc


Func SetPosition( $bPos )
	Local $mIdentity = $IDENTITY_MATRIX
	$mTransforms = $mIdentity
	If $bPos = $iPos1 Then
		$aObject[6] = $mIdentity
		$mTransforms[14] = -4.85
	ElseIf $bPos = $iPos2 Then
		Local $mMatrix = $mIdentity
		$mMatrix[5]  =  0.75
		$mMatrix[6]  =  0.68
		$mMatrix[9]  = -0.68
		$mMatrix[10] =  0.75
		$mMatrix[13] =  4.27
		$mMatrix[14] = -3.12
		$aObject[6] = $mMatrix
		$mTransforms[6]  =  -0.13
		$mTransforms[9]  =   0.13
		$mTransforms[13] =  -2.64
		$mTransforms[14] = -16.70
	EndIf
EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 0.45, 0.45, 0.45, 1.0 )

	glEnable( $GL_BLEND )
	glBlendFunc( $GL_SRC_ALPHA, $GL_ONE_MINUS_SRC_ALPHA )

	glEnable( $GL_DEPTH_TEST )

	GUISwitch( $hGui )

	Return 1

EndFunc


Func CreateResources()

	If $CmdLine[0] = 0 Then
		$sObjectFile = "monkey.obj"
		$sVshader = "01-gouraud-shading.v.glsl"
		$sFshader = "01-gouraud-shading.f.glsl"
		GUICtrlSetData( $idLabel2, $aInfo[0] & "  |  " & $sInfo )
	Else
		WinSetTitle( $hGui, "", $CmdLine[1] )
		$sObjectFile = $CmdLine[2]
		$sVshader = $CmdLine[3]
		$sFshader = $CmdLine[4]
		If $sObjectFile = "two-sided.obj" Then
			$bObject = $iTwoSidedObejct
			$bNewObject = $iTwoSidedObejct
			GUICtrlSetState( $idMonkey, $GUI_UNCHECKED )
			GUICtrlSetState( $idTwoSided, $GUI_CHECKED )
		EndIf
		Local $i = StringRight( $CmdLine[1], 2 ) - 1
		GUICtrlSetData( $idLabel2, $aInfo[$i] & "  |  " & $sInfo )
	EndIf

	; === Create objects ===

	; --- Main object ---

	LoadObject( "models\" & $sObjectFile, $aObject )

	; --- Ground ---

	Local $iGroundSize = 20, $aVert[$iGroundSize*$iGroundSize*6*4], $aNorm[$iGroundSize*$iGroundSize*6*3], $l = 0, $m = 0
	For $i = -$iGroundSize / 2 To $iGroundSize / 2 - 1
		For $j = -$iGroundSize / 2 To $iGroundSize / 2 - 1
			$aVert[$l+00] = $i		; (i,   0.0,  j+1, 1.0)
			$aVert[$l+01] = 0.0
			$aVert[$l+02] = $j+1
			$aVert[$l+03] = 1.0

			$aVert[$l+04] = $i+1	; (i+1, 0.0,  j+1, 1.0)
			$aVert[$l+05] = 0.0
			$aVert[$l+06] = $j+1
			$aVert[$l+07] = 1.0

			$aVert[$l+08] = $i		; (i,   0.0,  j,   1.0)
			$aVert[$l+09] = 0.0
			$aVert[$l+10] = $j
			$aVert[$l+11] = 1.0

			$aVert[$l+12] = $i		; (i,   0.0,  j,   1.0)
			$aVert[$l+13] = 0.0
			$aVert[$l+14] = $j
			$aVert[$l+15] = 1.0

			$aVert[$l+16] = $i+1	; (i+1, 0.0,  j+1, 1.0)
			$aVert[$l+17] = 0.0
			$aVert[$l+18] = $j+1
			$aVert[$l+19] = 1.0

			$aVert[$l+20] = $i+1	; (i+1, 0.0,  j,   1.0)
			$aVert[$l+21] = 0.0
			$aVert[$l+22] = $j
			$aVert[$l+23] = 1.0

			$l += 24

			For $k = 0 To 5
				$aNorm[$m+0] = 0.0
				$aNorm[$m+1] = 1.0
				$aNorm[$m+2] = 0.0
				$m += 3
			Next
		Next
	Next

	$aGround[3] = $aVert
	$aGround[4] = $aNorm
	$aGround[6] = $IDENTITY_MATRIX

	; --- Lightbox ---

	Local $aVert[8*4] = [ _
	  -0.1, -0.1, -0.1, 0.0, _
	   0.1, -0.1, -0.1, 0.0, _
	   0.1,  0.1, -0.1, 0.0, _
	  -0.1,  0.1, -0.1, 0.0, _
	  -0.1, -0.1,  0.1, 0.0, _
	   0.1, -0.1,  0.1, 0.0, _
	   0.1,  0.1,  0.1, 0.0, _
	  -0.1,  0.1,  0.1, 0.0 ]

	Local $mIdentity = $IDENTITY_MATRIX, $aLightPos[3] = [ 0.0, 1.0, 2.0 ]
	Local $mObject2world = glmTranslate( $mIdentity, $aLightPos )

	$aLightBox[3] = $aVert
	$aLightBox[6] = $mObject2world

	; === Upload objects ===

	UploadObject( $aObject )
	UploadObject( $aGround )
	UploadObject( $aLightBox )

	; === Load, compile and link shaders ===

	Local $iVertexShader, $iFragmentShader, $bStatus

	If Not LoadGLSLshader( $iVertexShader, "shaders\" & $sVshader, $GL_VERTEX_SHADER ) Then Return 0
	If Not LoadGLSLshader( $iFragmentShader, "shaders\" & $sFshader, $GL_FRAGMENT_SHADER ) Then Return 0

	$iProgram = glCreateProgram()
	glAttachShader( $iProgram, $iVertexShader )
	glAttachShader( $iProgram, $iFragmentShader )
	glLinkProgram( $iProgram )
	glGetProgramiv( $iProgram, $GL_LINK_STATUS, $bStatus )
	If $bStatus = 0 Then
		PrintError( "Could not link program" )
		Return 0
	EndIf
	glValidateProgram( $iProgram )
	glGetProgramiv( $iProgram, $GL_VALIDATE_STATUS, $bStatus )
	If $bStatus = 0 Then
		PrintError( "Could not validate program" )
		Return 0
	EndIf

	$iAttributeVcoord = glGetAttribLocation( $iProgram, "v_coord" )
	If $iAttributeVcoord = -1 Then
		PrintError( "Could not bind attribute ""v_coord""" )
		Return 0
	EndIf

	$iAttributeVnormal = glGetAttribLocation( $iProgram, "v_normal" )
	If $iAttributeVnormal = -1 Then
		PrintError( "Could not bind attribute ""v_normal""" )
		Return 0
	EndIf

	$iUniformM = glGetUniformLocation( $iProgram, "m" )
	If $iUniformM = -1 Then
		PrintError( "Could not bind uniform ""m""" )
		Return 0
	EndIf

	$iUniformV = glGetUniformLocation( $iProgram, "v" )
	If $iUniformV = -1 Then
		PrintError( "Could not bind uniform ""v""" )
		Return 0
	EndIf

	$iUniformP = glGetUniformLocation( $iProgram, "p" )
	If $iUniformP = -1 Then
		PrintError( "Could not bind uniform ""p""" )
		Return 0
	EndIf

	$iUniformM3x3InvTransp = glGetUniformLocation( $iProgram, "m_3x3_inv_transp" )
	If $iUniformM3x3InvTransp = -1 Then
		PrintError( "Could not bind uniform ""m_3x3_inv_transp""" )
		Return 0
	EndIf

	$iUniformVinv = glGetUniformLocation( $iProgram, "v_inv" )
	If $iUniformVinv = -1 Then
		PrintError( "Could not bind uniform ""v_inv""" )
		Return 0
	EndIf

	InitView()

	GUISetState()

	Return 1

EndFunc


Func LoadObject( $sObjectFile, ByRef $aObject )

	Local $aLines, $line, $aWords, $aVert[400], $aElem[300], $v = 0, $e = 0
	_FileReadToArray( $sObjectFile, $aLines )
	For $i = 1 To $aLines[0]
		$line = $aLines[$i]
		If StringLeft( $line, 2 ) = "v " Then
			$aWords = StringSplit( $line, " ", 2 )
			$aVert[$v+0] = $aWords[1]
			$aVert[$v+1] = $aWords[2]
			$aVert[$v+2] = $aWords[3]
			$aVert[$v+3] = 1.0
			If Mod( $v+4, 400 ) = 0 Then Redim $aVert[$v+4+400]
			$v += 4
		ElseIf StringLeft( $line, 2 ) = "f " Then
			$aWords = StringSplit( $line, " ", 2 )
			$aElem[$e+0] = $aWords[1] - 1
			$aElem[$e+1] = $aWords[2] - 1
			$aElem[$e+2] = $aWords[3] - 1
			If Mod( $e+3, 300 ) = 0 Then Redim $aElem[$e+3+300]
			$e += 3
		EndIf
	Next
	ReDim $aVert[$v]
	ReDim $aElem[$e]

	Local $iVertSize = UBound( $aVert ) / 4, $aNorm[3*$iVertSize], $aNbSeen[$iVertSize]
	For $i = 0 To $iVertSize - 1
		$aNorm[3*$i+0] = 0.0
		$aNorm[3*$i+1] = 0.0
		$aNorm[3*$i+2] = 0.0
		$aNbSeen[$i] = 0
	Next

	Local $iElemSize = UBound( $aElem ), $ia, $ib, $ic
	Local $vec1, $vec2, $vec3, $norm
	For $i = 0 To $iElemSize - 1 Step 3
		$ia = $aElem[$i+0]
		$ib = $aElem[$i+1]
		$ic = $aElem[$i+2]
		Local $vecA[3] = [ $aVert[$ia*4+0], $aVert[$ia*4+1], $aVert[$ia*4+2] ]
		Local $vecB[3] = [ $aVert[$ib*4+0], $aVert[$ib*4+1], $aVert[$ib*4+2] ]
		Local $vecC[3] = [ $aVert[$ic*4+0], $aVert[$ic*4+1], $aVert[$ic*4+2] ]
		$vec1 = glmSubtractVec3( $vecB, $vecA )
		$vec2 = glmSubtractVec3( $vecC, $vecA )
		$vec3 = glmCrossVec3( $vec1, $vec2 )
		$norm = glmNormalizeVec3( $vec3 )

		Local $v[3] = [ $ia, $ib, $ic ], $curV, $vec5
		For $j = 0 To 2
			$curV = $v[$j]
			$aNbSeen[$curV] += 1
			If $aNbSeen[$curV] = 1 Then
				$aNorm[$curV*3+0] = $norm[0]
				$aNorm[$curV*3+1] = $norm[1]
				$aNorm[$curV*3+2] = $norm[2]
			Else
				Local $x = 1.0 / $aNbSeen[$curV], $y = 1.0 - $x
				$aNorm[$curV*3+0] = $aNorm[$curV*3+0] * $y + $norm[0] * $x
				$aNorm[$curV*3+1] = $aNorm[$curV*3+1] * $y + $norm[1] * $x
				$aNorm[$curV*3+2] = $aNorm[$curV*3+2] * $y + $norm[2] * $x
				Local $vec4[3] = [ $aNorm[$curV*3+0], $aNorm[$curV*3+1], $aNorm[$curV*3+2] ]
				$vec5 = glmNormalizeVec3( $vec4 )
				$aNorm[$curV*3+0] = $vec5[0]
				$aNorm[$curV*3+1] = $vec5[1]
				$aNorm[$curV*3+2] = $vec5[2]
			EndIf
		Next
	Next

	$aObject[3] = $aVert
	$aObject[4] = $aNorm
	$aObject[5] = $aElem

EndFunc


Func UploadObject( ByRef $aObject )

	Local $iVbo, $aObj

	If $aObject[0] Then glDeleteBuffers( 1, $aObject[0] )
	If $aObject[1] Then glDeleteBuffers( 1, $aObject[1] )
	If $aObject[2] Then glDeleteBuffers( 1, $aObject[2] )

	; Store object vertices, normals and elements in video card buffers

	If IsArray( $aObject[3] ) Then
		$aObj = $aObject[3]
		glGenBuffers( 1, $iVbo )
		glBindBuffer( $GL_ARRAY_BUFFER, $iVbo )
		glBufferData( $GL_ARRAY_BUFFER, $GL_FLOAT, 4*UBound($aObj), $aObj, $GL_STATIC_DRAW )
		$aObject[0] = $iVbo
	EndIf

	If IsArray( $aObject[4] ) Then
		$aObj = $aObject[4]
		glGenBuffers( 1, $iVbo )
		glBindBuffer( $GL_ARRAY_BUFFER, $iVbo )
		glBufferData( $GL_ARRAY_BUFFER, $GL_FLOAT, 4*UBound($aObj), $aObj, $GL_STATIC_DRAW )
		$aObject[1] = $iVbo
	EndIf

	If IsArray( $aObject[5] ) Then
		$aObj = $aObject[5]
		glGenBuffers( 1, $iVbo )
		glBindBuffer( $GL_ELEMENT_ARRAY_BUFFER, $iVbo )
		glBufferData( $GL_ELEMENT_ARRAY_BUFFER, $GL_UNSIGNED_SHORT, 2*UBound($aObj), $aObj, $GL_STATIC_DRAW )
		$aObject[2] = $iVbo
	EndIf

EndFunc


Func DrawObject( ByRef $aObject )

	Local $iSize

	; Vertices
	If $aObject[0] Then
	  glEnableVertexAttribArray( $iAttributeVcoord )
	  glBindBuffer( $GL_ARRAY_BUFFER, $aObject[0] )
	  glVertexAttribPointer( _
			$iAttributeVcoord,   _ ; attribute
			4,                   _ ; number of elements per vertex, here (x,y,z,w)
			$GL_FLOAT,           _ ; the type of each element
			$GL_FALSE,           _ ; take our values as-is
			0,                   _ ; no extra data between each position
			0 )                    ; offset of first element
	EndIf

	; Normals
	If $aObject[1] Then
	  glEnableVertexAttribArray( $iAttributeVnormal )
	  glBindBuffer( $GL_ARRAY_BUFFER, $aObject[1] )
	  glVertexAttribPointer( _
			$iAttributeVnormal,  _ ; attribute
			3,                   _ ; number of elements per vertex, here (x,y,z)
			$GL_FLOAT,           _ ; the type of each element
			$GL_FALSE,           _ ; take our values as-is
			0,                   _ ; no extra data between each position
			0 )                    ; offset of first element
	EndIf

	; Model
	; Apply object's transformation matrix
	Local $mMatrix = $aObject[6]
	glUniformMatrix4fv( $iUniformM, 1, $GL_FALSE, $mMatrix )

	; Transform normal vectors with transpose of inverse of upper left 3x3 model matrix
	Local $mMat3[9] = [ $mMatrix[0], $mMatrix[1], $mMatrix[2], _
	                    $mMatrix[4], $mMatrix[5], $mMatrix[6], _
	                    $mMatrix[8], $mMatrix[9], $mMatrix[10] ]
	Local $mInv = glmInverseMat3( $mMat3 )
	Local $mTransp = glmTransposeMat3( $mInv )
	glUniformMatrix3fv( $iUniformM3x3InvTransp, 1, $GL_FALSE, $mTransp )

	; Push each element in vertices buffer to the vertex shader
	If $aObject[2] Then
		glBindBuffer( $GL_ELEMENT_ARRAY_BUFFER, $aObject[2] )
		glGetBufferParameteriv( $GL_ELEMENT_ARRAY_BUFFER, $GL_BUFFER_SIZE, $iSize )
		glDrawElements( $GL_TRIANGLES, $iSize / 2, $GL_UNSIGNED_SHORT, 0 )
	Else
		Local $aVert = $aObject[3]
		Local $iVertSize = UBound( $aVert ) / 4
		glDrawArrays( $GL_TRIANGLES, 0, $iVertSize )
	EndIf

	If $aObject[0] Then _
		glDisableVertexAttribArray( $iAttributeVcoord )
	If $aObject[1] Then _
		glDisableVertexAttribArray( $iAttributeVnormal )

EndFunc


Func DrawBoundingBox( ByRef $aObject )

	If Not IsArray( $aObject[3] ) Then Return

	; Cube 1x1x1, centered on origin
	Local $aVertices[8*4] = [ _
		-0.5, -0.5, -0.5, 1.0, _
		 0.5, -0.5, -0.5, 1.0, _
		 0.5,  0.5, -0.5, 1.0, _
		-0.5,  0.5, -0.5, 1.0, _
		-0.5, -0.5,  0.5, 1.0, _
		 0.5, -0.5,  0.5, 1.0, _
		 0.5,  0.5,  0.5, 1.0, _
		-0.5,  0.5,  0.5, 1.0 ]

	Local $iVboVertices
	glGenBuffers( 1, $iVboVertices )
	glBindBuffer( $GL_ARRAY_BUFFER, $iVboVertices )
	glBufferData( $GL_ARRAY_BUFFER, $GL_FLOAT, 4*UBound($aVertices), $aVertices, $GL_STATIC_DRAW )
	glBindBuffer( $GL_ARRAY_BUFFER, 0 )

	Local $aElements[2*4+8] = [ _
		0, 1, 2, 3, _
		4, 5, 6, 7, _
		0, 4, 1, 5, 2, 6, 3, 7 ]

	Local $iIboElements
	glGenBuffers( 1, $iIboElements )
	glBindBuffer( $GL_ELEMENT_ARRAY_BUFFER, $iIboElements )
	glBufferData( $GL_ELEMENT_ARRAY_BUFFER, $GL_UNSIGNED_SHORT, 2*UBound($aElements), $aElements, $GL_STATIC_DRAW )
	glBindBuffer( $GL_ELEMENT_ARRAY_BUFFER, 0 )

	Local $aVert = $aObject[3], $k = 4
	Local $iVertSize = UBound( $aVert ) / 4
	Local $minX = $aVert[0], $maxX = $minX
	Local $minY = $aVert[1], $maxY = $minY
	Local $minZ = $aVert[2], $maxZ = $minZ
	For $i = 1 To $iVertSize - 1
		If $aVert[$k+0] < $minX Then $minX = $aVert[$k+0]
		If $aVert[$k+0] > $maxX Then $maxX = $aVert[$k+0]
		If $aVert[$k+1] < $minY Then $minY = $aVert[$k+1]
		If $aVert[$k+1] > $maxY Then $maxY = $aVert[$k+1]
		If $aVert[$k+2] < $minZ Then $minZ = $aVert[$k+2]
		If $aVert[$k+2] > $maxZ Then $maxZ = $aVert[$k+2]
		$k += 4
	Next
	Local $mIdentity = $IDENTITY_MATRIX
	Local $aSize[3] = [ $maxX - $minX, $maxY - $minY, $maxZ - $minZ ]
	Local $aCenter[3] = [ ( $minX + $maxX ) / 2, ( $minY + $maxY ) / 2, ( $minZ + $maxZ ) / 2 ]
	Local $mScale = glmScale( $mIdentity, $aSize ), $mTranslate = glmTranslate( $mIdentity, $aCenter )
	Local $mTransform = glmMultiply( $mScale, $mTranslate )

	; Apply object's transformation matrix
	Local $mMatrix = $aObject[6]
	$mMatrix = glmMultiply( $mMatrix, $mTransform )
	glUniformMatrix4fv( $iUniformM, 1, $GL_FALSE, $mMatrix )

	glBindBuffer( $GL_ARRAY_BUFFER, $iVboVertices )
	glEnableVertexAttribArray( $iAttributeVcoord )
	glVertexAttribPointer( _
		$iAttributeVcoord,   _ ; attribute
		4,                   _ ; number of elements per vertex, here (x,y,z,w)
		$GL_FLOAT,           _ ; the type of each element
		$GL_FALSE,           _ ; take our values as-is
		0,                   _ ; no extra data between each position
		0 )                    ; offset of first element

	glBindBuffer( $GL_ELEMENT_ARRAY_BUFFER, $iIboElements )
	glDrawElements( $GL_LINE_LOOP, 4, $GL_UNSIGNED_SHORT, 0 )
	glDrawElements( $GL_LINE_LOOP, 4, $GL_UNSIGNED_SHORT, 4*2 )
	glDrawElements( $GL_LINES, 8, $GL_UNSIGNED_SHORT, 2*4*2 )
	glBindBuffer( $GL_ELEMENT_ARRAY_BUFFER, 0 )

	glDisableVertexAttribArray( $iAttributeVcoord )
	glBindBuffer( $GL_ARRAY_BUFFER, 0 )

	glDeleteBuffers( 1, $iVboVertices )
	glDeleteBuffers( 1, $iIboElements )

EndFunc


Func InitView()
	$aObject[6] = $IDENTITY_MATRIX
	Local $eye[3] = [ 0.0, 0.0, 4.0 ], $center[3] = [ 0.0, 0.0, 0.0 ], $up[3] = [ 0.0, 1.0, 0.0 ]
	$mTransforms = glmLookAt( $eye, $center, $up )
	If $bPos Then
		$bPos = 0
		$bNewPos = $bPos
		GUICtrlSetState( $idPos1, $GUI_UNCHECKED )
		GUICtrlSetState( $idPos2, $GUI_UNCHECKED )
	EndIf
EndFunc


Func DisplayFunction()

	Local Static $iStartTime = TimerInit(), $fLastTime = 0
	Local $fElapsedTime = TimerDiff( $iStartTime ) / 1000
	$fElapsedTime = ( $fLastTime + $fElapsedTime ) / 2
	$fLastTime = $fElapsedTime

	HandleKeystrokes()

	If $iRotXdirection <> 0 Or _
	   $iRotYdirection <> 0 Or _
	   $iTransZdirection <> 0 Then
	     CalcTransforms( $fElapsedTime )
	     If $bPos Then ResetPosButtons()
	EndIf

	If $bArcballOn And _
	   ( $thisMousePos[0] <> $lastMousePos[0] Or _
	     $thisMousePos[1] <> $lastMousePos[1] ) Then
	       HandleArcball()
	       If $bPos Then _
	         ResetPosButtons()
	EndIf

	; Model
	; Set in DrawObject()

	; View
	Local $mWorld2camera = $mTransforms

	; Projection
  Local $mCamera2screen = glmPerspective( 45.0, 1.0*$iWidth/$iHeight, 0.1, 100.0 )

	glUseProgram( $iProgram )

	glUniformMatrix4fv( $iUniformV, 1, $GL_FALSE, $mWorld2camera )
	glUniformMatrix4fv( $iUniformP, 1, $GL_FALSE, $mCamera2screen )

	Local $mInv = glmInverse( $mWorld2camera )
	glUniformMatrix4fv( $iUniformVinv, 1, $GL_FALSE, $mInv )

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )

	DrawObject( $aObject )
	DrawObject( $aGround )
	DrawBoundingBox( $aLightBox )

	SwapBuffers( $hDC )

EndFunc


Func CalcTransforms( $fElapsedTime )

	Local $fDeltaRotX = -$iRotXdirection * $fElapsedTime / 1000.0 * 60			; 60 per second
	Local $fDeltaRotY =  $iRotYdirection * $fElapsedTime / 1000.0 * 60			; 60 per second
	Local $fDeltaTransZ = $iTransZdirection * $fElapsedTime / 1000.0 * 2.5	; 2.5 units per second
	Local $axisX[3] = [ 1.0, 0.0, 0.0 ], $axisY[3] = [ 0.0, 1.0, 0.0 ], $vec[3] = [ 0.0, 0.0, $fDeltaTransZ ]

	If $bView = $iObejctView Then

		Local $mMatrix = $aObject[6]
		$mMatrix = glmRotate( $mMatrix, $fDeltaRotX / 2, $axisX )
		$mMatrix = glmRotate( $mMatrix, $fDeltaRotY / 2, $axisY )
		$aObject[6] = glmTranslate( $mMatrix, $vec )

	ElseIf $bView = $iWorldView Then

		Local $mMatrix = $mTransforms, $mIdentity = $IDENTITY_MATRIX
		Local $mMat3[9] = [ $mMatrix[0], $mMatrix[1], $mMatrix[2], _
												$mMatrix[4], $mMatrix[5], $mMatrix[6], _
												$mMatrix[8], $mMatrix[9], $mMatrix[10] ]
		Local $axisYworld = glmMultiplyMat3Vec( $mMat3, $axisY )
		Local $mMatrix1 = glmRotate( $mIdentity, -$fDeltaRotY / 50, $axisYworld )
		$mMatrix = glmMultiply( $mMatrix1, $mMatrix )
		Local $mMatrix2 = glmTranslate( $mIdentity, $vec )
		$mMatrix = glmMultiply( $mMatrix2, $mMatrix )
		Local $mMatrix3 = glmRotate( $mIdentity, $fDeltaRotX / 100, $axisX )
		$mTransforms = glmMultiply( $mMatrix3, $mMatrix )

	EndIf

	$iTransZdirection = 0
	$iRotYdirection = 0
	$iRotXdirection = 0

EndFunc


Func HandleKeystrokes()

	Local Static $bKey = 0, $nKey = 0
	Local Static $nStrokes = 0

	Select
		Case _IsPressed( "21", $dllUser32 ) ; Page up
			$bKey = 1
			if Not $nKey Then $nKey = 1
		Case _IsPressed( "22", $dllUser32 ) ; Page down
			$bKey = 2
			if Not $nKey Then $nKey = 1
		Case _IsPressed( "24", $dllUser32 ) ; Home
			$bKey = 3
			if Not $nKey Then $nKey = 1
		Case _IsPressed( "25", $dllUser32 ) ; Left
			$bKey = 4
			if Not $nKey Then $nKey = 1
		Case _IsPressed( "27", $dllUser32 ) ; Right
			$bKey = 5
			if Not $nKey Then $nKey = 1
		Case _IsPressed( "26", $dllUser32 ) ; Up
			$bKey = 6
			if Not $nKey Then $nKey = 1
		Case _IsPressed( "28", $dllUser32 ) ; Down
			$bKey = 7
			if Not $nKey Then $nKey = 1
		Case Else
			$nKey = 0
			Return
	EndSelect

	If $nKey = 1 Or $nKey = 3 Then
		Switch $bKey
			Case 1 ; Page up
				$iTransZdirection = -1
			Case 2 ; Page down
				$iTransZdirection = 1
			Case 3 ; Home
				InitView()
			Case 4 ; Left
				$iRotYdirection = -1
			Case 5 ; Right
				$iRotYdirection = 1
			Case 6 ; Up
				$iRotXdirection = -1
			Case 7 ; Down
				$iRotXdirection = 1
		EndSwitch
		$bKey = 0
	EndIf

	Switch $nKey
		Case 1
			; First keypress
			$nKey = 2
			$nStrokes = 0
		Case 2
			; Wait a few loops for next keypress
			; To get time to release the key for a single keypress
			$nStrokes += 1
			If $nStrokes > 3 Then _
				$nKey = 3
	EndSwitch

EndFunc


Func WM_SIZING( $hWnd, $iMsg, $iwParam, $ilParam )
	#forceref $iMsg, $iwParam
	Switch $hWnd
		Case $hGui
			Local $tRECT = DllStructCreate( $tagRECT, $ilParam )
			Local $y = DllStructGetData( $tRECT, "Top" )
			Local $iHeight = DllStructGetData( $tRECT, "Bottom" ) - $y - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight
			WinMove( $hStatusbar, "", 2, $iHeight + $iHeightXP )
	EndSwitch
	Return $GUI_RUNDEFMSG
EndFunc


Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight * $iStatusBarRows
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	WinMove( $hStatusbar, "", 2, $iHeight + $iHeightXP + $iStatusBarHeight * ( $iStatusBarRows - 1 ) )
	$bResize = True
EndFunc


Func GUI_EVENT_CLOSE()
	glDeleteProgram( $iProgram )
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
