#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

Global $nSectors = 60

Global $aDemos[9][4] = [ _
	[ "Earthmap720x360_grid.jpg", "0-sphere.v.glsl", "0-sphere.f.glsl", "Textured sphere" ], _                                ; 0: Textured Spheres
	[ "Earthmap720x360_grid.jpg", "1-sphere.v.glsl", "1-sphere_ST.f.glsl", "Tiling and offsets" ], _
	[ "Earthmap720x360_grid.jpg", "2-sphere-gouraud.v.glsl", "2-sphere-gouraud.f.glsl", "Light source" ], _                   ; 2: Lighting Textured Surfaces
	[ "Land_shallow_topo_alpha_2048.png", "3-sphere-gouraud.v.glsl", "3-sphere-gouraud-glossy.f.glsl", "Glossy surface" ], _  ; 3: Glossy Textures
	[ "Land_shallow_topo_alpha_2048.png", "4-sphere-phong.v.glsl", "4-sphere-phong.f.glsl", "Glossy surface" ], _
	[ "Land_shallow_topo_alpha_2048.png", "5-sphere.v.glsl", "5-sphere_discard.f.glsl", "Transparency" ], _                   ; 5: Transparent Textures
	[ "Land_shallow_topo_alpha_2048.png", "6-sphere.v.glsl", "6-sphere.f.glsl", "Transparency" ], _
	[ "Land_shallow_topo_alpha_2048.png", "7-sphere.v.glsl", "7-sphere_oceans.f.glsl", "Transparency" ], _
	[ "Earth_lights_lrg.jpg", "8-sphere-sunlit.v.glsl", "8-sphere-sunlit.f.glsl", "Layers of textures" ] ]                    ; 8: Layers of Textures

; --- OpenGL globals ---

Global $iProgram, $iVBOsphere
Global $iMyTextureId, $iMyTextureSunLitId
Global $iAttributeVcoord, $iAttributeVnormal
Global $iUniformM, $iUniformV, $iUniformP
Global $iUniformM3x3InvTransp, $iUniformVinv = -1
Global $iUniformMyTexture, $iUniformMyTextureSunLit, $iUniformMyTextureST = -1

Global $mIdentity, $vAxisX[3] = [ 1, 0, 0 ], $vAxisY[3] = [ 0, 1, 0 ]
Global $mModel, $mView, $mProjection, $mFixOrientation

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 800, $iHeight = 600

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30
Global Const $iStatusBarHeight = 28

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight + $iStatusBarHeight

Global $hGui, $bResize = True

; --- Project includes ---

#include "utilities\images.au3"
#include "utilities\includes.au3"
#include "utilities\statusbar.au3"

; --- Program code ---

MainFunc()

While 1
	If $bNewSphere Then
		glDeleteBuffers( 1, $iVBOsphere )
		$nSectors = GUICtrlRead( $idSectors ) * 15
		$iVBOsphere = sphere( $nSectors, $nSectors )
		$bNewSphere = False
	EndIf
	If $bResize Then
		glViewport( 0, 0, $iWidth, $iHeight )
		$mProjection = glmPerspective( 45.0, 1.0*$iWidth/$iHeight, 0.1, 10.0 )
		$bResize = False
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then GUI_EVENT_CLOSE()
	If Not InitOpenGLwindow() Then GUI_EVENT_CLOSE()
	If Not CheckOpenGLversion( 2, 1 ) Then GUI_EVENT_CLOSE()
	If Not CreateResources() Then GUI_EVENT_CLOSE()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 50, $WS_OVERLAPPEDWINDOW )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	GUIRegisterMsg( $WM_SIZING, "WM_SIZING" )

	CreateStatusbar()

	Return 1

EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 1.0, 1.0, 1.0, 1.0 )

	glEnable( $GL_BLEND )
	glBlendFunc( $GL_SRC_ALPHA, $GL_ONE_MINUS_SRC_ALPHA )

	glEnable( $GL_CULL_FACE )

	GUISwitch( $hGui )

	Return 1

EndFunc


Func CreateResources()

	Local $iCurDemo, $sTexture, $sVshader, $sFshader

	If $CmdLine[0] = 0 Then
		$iCurDemo = 0
	Else
		$iCurDemo = $CmdLine[1]
		WinSetTitle( $hGui, "", $CmdLine[2] )
	EndIf

	$sTexture = $aDemos[$iCurDemo][0]
	$sVshader = $aDemos[$iCurDemo][1]
	$sFshader = $aDemos[$iCurDemo][2]

	GUICtrlSetData( $idLabel, $aDemos[$iCurDemo][3] )

	Local $w, $h
	If Not LoadTextureImage( $iMyTextureId, $images & "\WikiBooks\" & $sTexture, $w, $h ) Then Return 0
	If Not LoadTextureImage( $iMyTextureSunLitId, $images & "\WikiBooks\Land_shallow_topo_2048.jpg", $w, $h ) Then Return 0

	Local $iVertexShader, $iFragmentShader, $bStatus

	If Not LoadGLSLshader( $iVertexShader, "shaders\" & $sVshader, $GL_VERTEX_SHADER ) Then Return 0
	If Not LoadGLSLshader( $iFragmentShader, "shaders\" & $sFshader, $GL_FRAGMENT_SHADER ) Then Return 0

	$iProgram = glCreateProgram()
	glAttachShader( $iProgram, $iVertexShader )
	glAttachShader( $iProgram, $iFragmentShader )
	glLinkProgram( $iProgram )
	glGetProgramiv( $iProgram, $GL_LINK_STATUS, $bStatus )
	If $bStatus = 0 Then
		PrintError( "Could not link program" )
		Return 0
	EndIf

	$iAttributeVcoord = glGetAttribLocation( $iProgram, "v_coord" )
	If $iAttributeVcoord = -1 Then
		PrintError( "Could not bind attribute ""v_coord""" )
		Return 0
	EndIf

	Switch $iCurDemo
		Case 0, 1, 5 To 7
		Case Else
			$iAttributeVnormal = glGetAttribLocation( $iProgram, "v_normal" )
			If $iAttributeVnormal = -1 Then
				PrintError( "Could not bind attribute ""v_normal""" )
				Return 0
			EndIf
	EndSwitch

	$iUniformM = glGetUniformLocation( $iProgram, "m" )
	If $iUniformM = -1 Then
		PrintError( "Could not bind uniform ""m""" )
		Return 0
	EndIf

	$iUniformV = glGetUniformLocation( $iProgram, "v" )
	If $iUniformV = -1 Then
		PrintError( "Could not bind uniform ""v""" )
		Return 0
	EndIf

	$iUniformP = glGetUniformLocation( $iProgram, "p" )
	If $iUniformP = -1 Then
		PrintError( "Could not bind uniform ""p""" )
		Return 0
	EndIf

	Switch $iCurDemo
		Case 0, 1, 5 To 7
		Case Else
			$iUniformM3x3InvTransp = glGetUniformLocation( $iProgram, "m_3x3_inv_transp" )
			If $iUniformM3x3InvTransp = -1 Then
				PrintError( "Could not bind uniform ""m_3x3_inv_transp""" )
				Return 0
			EndIf
	EndSwitch

	Switch $iCurDemo
		Case 0, 1, 5 To 8
		Case Else
			$iUniformVinv = glGetUniformLocation( $iProgram, "v_inv" )
			If $iUniformVinv = -1 Then
				PrintError( "Could not bind uniform ""v_inv""" )
				Return 0
			EndIf
	EndSwitch

	$iUniformMyTexture = glGetUniformLocation( $iProgram, "mytexture" )
	If $iUniformMyTexture = -1 Then
		PrintError( "Could not bind uniform ""mytexture""" )
		Return 0
	EndIf

	Switch $iCurDemo
		Case 0 To 7
		Case Else
			$iUniformMyTextureSunLit = glGetUniformLocation( $iProgram, "mytexture_sunlit" )
			If $iUniformMyTextureSunLit = -1 Then
				PrintError( "Could not bind uniform ""mytexture_sunlit""" )
				Return 0
			EndIf
	EndSwitch

	If StringInStr( $sFshader, "_ST" ) Then
		$iUniformMyTextureST = glGetUniformLocation( $iProgram, "mytexture_ST" )
		If $iUniformMyTextureST = -1 Then
			PrintError( "Could not bind uniform ""mytexture_ST""" )
			Return 0
		EndIf
	EndIf

	$iVBOsphere = sphere( $nSectors, $nSectors )

	CalculateMatrices()

	GUISetState()

	Return 1

EndFunc


Func sphere( $slices, $stacks )

	Local $n = 2 * ( $slices + 1 ) * $stacks, $aPoints[3*$n], $i = 0

	For $theta = -$PI / 2 To $PI / 2 - 0.0001 Step $PI / $stacks
		For $phi = -$PI To $PI + 0.0001 Step 2 * $PI / $slices
			$aPoints[$i+0] =  Cos( $theta ) * Sin( $phi )
			$aPoints[$i+1] = -Sin( $theta )
			$aPoints[$i+2] =  Cos( $theta ) * Cos( $phi )
			$aPoints[$i+3] =  Cos( $theta + $PI / $stacks ) * Sin( $phi )
			$aPoints[$i+4] = -Sin( $theta + $PI / $stacks )
			$aPoints[$i+5] =  Cos( $theta + $PI / $stacks ) * Cos( $phi )
			$i += 6
		Next
	Next

	Local $iVBO
	glGenBuffers( 1, $iVBO )
	glBindBuffer( $GL_ARRAY_BUFFER, $iVBO )
	glBufferData( $GL_ARRAY_BUFFER, $GL_FLOAT, 4*UBound($aPoints), $aPoints, $GL_STATIC_DRAW )

	Return $iVBO

EndFunc


Func CalculateMatrices()

	$mIdentity = $IDENTITY_MATRIX

	$mFixOrientation = glmRotate( $mIdentity, -90, $vAxisX )

	Local $vModel[3] = [ 0.0, 0.0, -2.0 ]
	$mModel = glmTranslate( $mIdentity, $vModel )

	Local $eye[3] = [ 0.0, 2.0, 0.0 ], $center[3] = [ 0.0, 0.0, -2.0 ], $up[3] = [ 0.0, 1.0, 0.0 ]
	$mView = glmLookAt( $eye, $center, $up )

EndFunc


Func DisplayFunction()

	Local Static $iStartTime = TimerInit(), $fLastTime = 0
	Local $fElapsedTime = TimerDiff( $iStartTime )
	$fElapsedTime = ( $fLastTime + $fElapsedTime ) / 2
	$fLastTime = $fElapsedTime

	Local $fAngle = $fElapsedTime / 1000.0 * 30 ; 30 per second
	Local $mRotate = glmRotate( $mIdentity, $fAngle, $vAxisY )

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )

	glUseProgram( $iProgram )

	Local $mMult = glmMultiply( $mModel, $mRotate )
	Local $mMatrix = glmMultiply( $mMult, $mFixOrientation ) ; $mModel * $mRotate * $mFixOrientation
	glUniformMatrix4fv( $iUniformM, 1, $GL_FALSE, $mMatrix )

	Local $mMat3[9] = [ $mMatrix[0], $mMatrix[1], $mMatrix[2], _
	                    $mMatrix[4], $mMatrix[5], $mMatrix[6], _
	                    $mMatrix[8], $mMatrix[9], $mMatrix[10] ]
	Local $mInv = glmInverseMat3( $mMat3 )
	Local $mTransp = glmTransposeMat3( $mInv )
	glUniformMatrix3fv( $iUniformM3x3InvTransp, 1, $GL_FALSE, $mTransp )

	glUniformMatrix4fv( $iUniformV, 1, $GL_FALSE, $mView )

	If $iUniformVinv >= 0 Then
		$mInv = glmInverse( $mView )
		glUniformMatrix4fv( $iUniformVinv, 1, $GL_FALSE, $mInv )
	EndIf

	glUniformMatrix4fv( $iUniformP, 1, $GL_FALSE, $mProjection )

	If $iUniformMyTextureST >= 0 Then _
		glUniform4f( $iUniformMyTextureST, 2, 1, 0, -0.05 )

  glActiveTexture( $GL_TEXTURE0 )
  glBindTexture( $GL_TEXTURE_2D, $iMyTextureId )
  glUniform1i( $iUniformMyTexture, 0 ) ; $GL_TEXTURE0

  glActiveTexture( $GL_TEXTURE1 )
  glBindTexture( $GL_TEXTURE_2D, $iMyTextureSunLitId )
  glUniform1i( $iUniformMyTextureSunLit, 1 ) ; $GL_TEXTURE1

  glEnableVertexAttribArray( $iAttributeVcoord )
  glEnableVertexAttribArray( $iAttributeVnormal )
  glBindBuffer( $GL_ARRAY_BUFFER, $iVBOsphere )
  glVertexAttribPointer( $iAttributeVcoord, 3, $GL_FLOAT, $GL_FALSE, 0, 0 )
  glVertexAttribPointer( $iAttributeVnormal, 3, $GL_FLOAT, $GL_FALSE, 0, 0 )

	Local $n = 2 * ( $nSectors + 1 ) * $nSectors

  glCullFace( $GL_FRONT )
  glDrawArrays( $GL_TRIANGLE_STRIP, 0, $n )

  glCullFace( $GL_BACK )
  glDrawArrays( $GL_TRIANGLE_STRIP, 0, $n )

	SwapBuffers( $hDC )

EndFunc


Func GUI_EVENT_CLOSE()
	glDeleteProgram( $iProgram )
	glDeleteBuffers( 1, $iVBOsphere )
  glDeleteTextures( 1, $iMyTextureId )
  glDeleteTextures( 1, $iMyTextureSunLitId )
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
