#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

; --- OpenGL globals ---

Global $iVertexBufferId, $iElementBufferId
Global $iVertexShaderId, $iFragmentShaderId, $iProgramId, $VaoId
Global $iTexture0, $iTexture1, $iTexture0UniformLocation, $iTexture1UniformLocation
Global $Timer = 0, $Position, $iTimerUniformLocation, $iPositionAttribLocation

Global $VertexBufferData[16] = [ _
	-1.0, -1.0, 0.0, 1.0, _
	 1.0, -1.0, 0.0, 1.0, _
	-1.0,  1.0, 0.0, 1.0, _
	 1.0,  1.0, 0.0, 1.0 ]

Global $ElementBufferData[4] = [ 0, 1, 2, 3 ]

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 400, $iHeight = 300

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30
Global Const $iStatusBarHeight = 28

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight + $iStatusBarHeight

Global $hGui, $idLabel, $bResize = True

; --- Project includes ---

#include "utilities\includes.au3"
#include "utilities\images.au3"

; --- Program code ---

MainFunc()

While 1
	If $bResize Then
		glViewport( 0, 0, $iWidth, $iHeight )
		$bResize = False
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then GUI_EVENT_CLOSE()
	If Not InitOpenGLwindow() Then GUI_EVENT_CLOSE()
	If Not CheckOpenGLversion( 2, 1 ) Then GUI_EVENT_CLOSE()
	If Not CreateResources() Then GUI_EVENT_CLOSE()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 100, $WS_OVERLAPPEDWINDOW )

	; Statusbar
	Local $idStatus
	$idStatus = GUICtrlCreateLabel( "", 0, $iHeight + 1, $iGuiWidth + 2, $iStatusBarHeight, $SS_ETCHEDFRAME )
	$idLabel = GUICtrlCreateLabel( "", 3, $iHeight + 1 + 6, $iGuiWidth - 4, $iStatusBarHeight - 10 )
	GUICtrlSetResizing( $idStatus, $GUI_DOCKSTATEBAR )
	GUICtrlSetResizing( $idLabel, $GUI_DOCKSTATEBAR )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	Return 1

EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 0.8, 0.8, 0.8, 0.0 )

	glEnable( $GL_DEPTH_TEST )
	glDepthFunc( $GL_LESS )

	GUISwitch( $hGui )

	Return 1

EndFunc


Func CreateResources()

	Local $VertexShaderFile

	If $CmdLine[0] = 0 Then
		$VertexShaderFile = "shaders\1-hello-gl.v.glsl"
	Else
		$VertexShaderFile = $CmdLine[1]
		WinSetTitle( $hGui, "", $CmdLine[2] )
	EndIf

	If Not MakeResources( $VertexShaderFile ) Then Return 0

	GUICtrlSetData( $idLabel, "Vertex shader: " & $VertexShaderFile )

	GUISetState()

	Return 1

EndFunc


Func MakeResources( $VertexShaderFile )

	glGenVertexArrays( 1, $VaoId )
	glBindVertexArray( $VaoId )

	If Not MakeBuffer( $iVertexBufferId, $GL_ARRAY_BUFFER, $GL_FLOAT, $VertexBufferData ) Then Return 0
	If Not MakeBuffer( $iElementBufferId, $GL_ELEMENT_ARRAY_BUFFER, $GL_INT, $ElementBufferData ) Then Return 0

	Local $w, $h
	If Not LoadTextureImage( $iTexture0, $images & "\Modern OpenGL\hello1.tga", $w, $h ) Then Return 0
	If Not LoadTextureImage( $iTexture1, $images & "\Modern OpenGL\hello2.tga", $w, $h ) Then Return 0

	If Not LoadGLSLshader( $iVertexShaderId, $VertexShaderFile, $GL_VERTEX_SHADER ) Then Return 0
	If Not LoadGLSLshader( $iFragmentShaderId, "shaders\hello-gl.f.glsl", $GL_FRAGMENT_SHADER ) Then Return 0

	If Not MakeProgram( $iProgramId, $iVertexShaderId, $iFragmentShaderId ) Then Return 0
	glUseProgram( $iProgramId )

	$iTimerUniformLocation = glGetUniformLocation( $iProgramId, "timer" )
	$iTexture0UniformLocation = glGetUniformLocation( $iProgramId, "textures[0]" )
	$iTexture1UniformLocation = glGetUniformLocation( $iProgramId, "textures[1]" )
	$iPositionAttribLocation = glGetAttribLocation( $iProgramId, "position" )

	Return 1

EndFunc


Func MakeBuffer( ByRef $bufferId, $target, $type, $bufferData )

	Local $iOpenGLerror = glGetError()

	glGenBuffers( 1, $bufferId )
	glBindBuffer( $target, $bufferId )
	glBufferData( $target, $type, 4*UBound($bufferData), $bufferData, $GL_STATIC_DRAW )

	$iOpenGLerror = glGetError()
	If $iOpenGLerror <> $GL_NO_ERROR Then
		PrintOpenGLerror( "MakeBuffer", "glGenBuffers...", $iOpenGLerror )
		Return 0
	EndIf

	Return 1

EndFunc

Func MakeProgram( ByRef $programId, $vertexShader, $fragmentShader )

	Local $iOpenGLerror = glGetError()

	$programId = glCreateProgram()
	glAttachShader( $programId, $vertexShader )
	glAttachShader( $programId, $fragmentShader )
	glLinkProgram( $programId )

	$iOpenGLerror = glGetError()
	If $iOpenGLerror <> $GL_NO_ERROR Then
		PrintOpenGLerror( "MakeProgram", "glCreateProgram...", $iOpenGLerror )
		Return 0
	EndIf

	Return 1

EndFunc


Func DisplayFunction()

	Local Static $LastTime = TimerInit()
	Local $mSec = TimerDiff( $LastTime )
	$Timer = $mSec * 0.001

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )

	glUniform1f( $iTimerUniformLocation, $Timer )

	glActiveTexture( $GL_TEXTURE0 )
	glBindTexture( $GL_TEXTURE_2D, $iTexture0 )
	glUniform1i( $iTexture0UniformLocation, 0 )

	glActiveTexture( $GL_TEXTURE1 )
	glBindTexture( $GL_TEXTURE_2D, $iTexture1 )
	glUniform1i( $iTexture1UniformLocation, 1 )

	glBindBuffer( $GL_ARRAY_BUFFER, $iVertexBufferId )
	glVertexAttribPointer( _
		$iPositionAttribLocation, _ ; attribute
		4, _                       ; size
		$GL_FLOAT, _               ; type
		$GL_FALSE, _               ; normalized?
		16, _                      ; stride
		0 )                        ; array buffer offset
	glEnableVertexAttribArray( $iPositionAttribLocation )

	glBindBuffer( $GL_ELEMENT_ARRAY_BUFFER, $iElementBufferId )
	glDrawElements( _
		$GL_TRIANGLE_STRIP, _  ; mode
		4, _                   ; count
		$GL_UNSIGNED_INT, _    ; type
		0 )                    ; element array buffer offset

	glDisableVertexAttribArray( $iPositionAttribLocation )

	SwapBuffers( $hDC )

EndFunc


Func DeleteResources()

	glUseProgram(0)

	glDetachShader( $iProgramId, $iVertexShaderId )
	glDetachShader( $iProgramId, $iFragmentShaderId )

	glDeleteShader( $iFragmentShaderId )
	glDeleteShader( $iVertexShaderId )

	glDeleteProgram( $iProgramId )

	glBindTexture( $GL_TEXTURE_2D, 0 )
	glDeleteTextures( 1, $iTexture0 )
	glDeleteTextures( 1, $iTexture1 )

	glBindBuffer( $GL_ARRAY_BUFFER, 0 )
	glDeleteBuffers( 1, $iVertexBufferId )

	glBindBuffer( $GL_ELEMENT_ARRAY_BUFFER, 0 )
	glDeleteBuffers( 1, $iElementBufferId )

	glBindVertexArray(0)
	glDeleteVertexArrays( 1, $VaoId )

EndFunc


Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	$bResize = True
EndFunc


Func GUI_EVENT_CLOSE()
	DeleteResources()
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
