#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

; --- OpenGL globals ---

Global $iProgram, $iVertexShader, $iFragmentShader, $iPositionBufferObject, $iVao

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 500, $iHeight = 500

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight

Global $hGui, $bResize = True

; --- Project includes ---

#include "utilities\includes.au3"

; --- Program code ---

MainFunc()

While 1
	If $bResize Then
		; An OpenGL window (viewport) should not be resized in the middle of a drawing cycle.
		; In GUI_EVENT_RESIZE() $bResize is set to True when the GUI is resized. At this point
		; when the DisplayFunction() isn't running it's safe to resize the OpenGL window.
		glViewport( 0, 0, $iWidth, $iHeight )
		$bResize = False
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then GUI_EVENT_CLOSE()
	If Not InitOpenGLwindow() Then GUI_EVENT_CLOSE()
	If Not CheckOpenGLversion( 3, 1 ) Then GUI_EVENT_CLOSE()
	If Not CreateResources() Then GUI_EVENT_CLOSE()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 100, $WS_OVERLAPPEDWINDOW )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	Return 1

EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 0.0, 0.0, 0.0, 0.0 )

	GUISwitch( $hGui )

	Return 1

EndFunc


Func CreateResources()

	InitializeProgram()

	InitializeVertexBuffer()

	glGenVertexArrays( 1, $iVao )
	glBindVertexArray( $iVao )

	GUISetState()

	Return 1

EndFunc


Func InitializeProgram()

	Local $sVertexShader = _
		"#version 330" & @CRLF & _
		"layout(location = 0) in vec4 position;" & @CRLF & _
		"void main()" & @CRLF & _
		"{" & @CRLF & _
		"    gl_Position = position;" & @CRLF & _
		"}" & @CRLF

	Local $sFragmentShader = _
		"#version 330" & @CRLF & _
		"out vec4 outputColor;" & @CRLF & _
		"void main()" & @CRLF & _
		"{" & @CRLF & _
		"    outputColor = vec4(1.0f, 1.0f, 1.0f, 1.0f);" & @CRLF & _
		"}" & @CRLF

	$iVertexShader = glCreateShader( $GL_VERTEX_SHADER )
	glShaderSource( $iVertexShader, 1, $sVertexShader, StringLen( $sVertexShader ) )
	glCompileShader( $iVertexShader )

	$iFragmentShader = glCreateShader( $GL_FRAGMENT_SHADER )
	glShaderSource( $iFragmentShader, 1, $sFragmentShader, StringLen( $sFragmentShader ) )
	glCompileShader( $iFragmentShader )

	$iProgram = glCreateProgram()
	glAttachShader( $iProgram, $iVertexShader )
	glAttachShader( $iProgram, $iFragmentShader )
	glLinkProgram( $iProgram )
	glUseProgram(0)

EndFunc


Func DestroyProgram()

	glUseProgram(0)

	glDetachShader( $iProgram, $iVertexShader )
	glDetachShader( $iProgram, $iFragmentShader )

	glDeleteShader( $iFragmentShader )
	glDeleteShader( $iVertexShader )

	glDeleteProgram( $iProgram )

EndFunc


Func InitializeVertexBuffer()

	Local $aVertexPositions[12] = [ _
		 0.75,  0.75, 0.0, 1.0, _
		 0.75, -0.75, 0.0, 1.0, _
		-0.75, -0.75, 0.0, 1.0 ]

	glGenBuffers( 1, $iPositionBufferObject )
	glBindBuffer( $GL_ARRAY_BUFFER, $iPositionBufferObject )
	glBufferData( $GL_ARRAY_BUFFER, $GL_FLOAT, 4*UBound($aVertexPositions), $aVertexPositions, $GL_STATIC_DRAW )
	glBindBuffer( $GL_ARRAY_BUFFER, 0 )

EndFunc

Func DestroyVertexBuffer()

	glDisableVertexAttribArray(0)

	glBindBuffer( $GL_ARRAY_BUFFER, 0 )
	glDeleteBuffers( 1, $iPositionBufferObject )

EndFunc


Func DisplayFunction()

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )

	glUseProgram( $iProgram )

	glBindBuffer( $GL_ARRAY_BUFFER, $iPositionBufferObject )
	glEnableVertexAttribArray(0)
	glVertexAttribPointer( 0, 4, $GL_FLOAT, $GL_FALSE, 0, 0 )

	glDrawArrays( $GL_TRIANGLES, 0, 3 )

	glDisableVertexAttribArray(0)
	glUseProgram(0)

	SwapBuffers( $hDC )

EndFunc


Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	$bResize = True ; The OpenGL window (viewport) is resized in top of the main loop
EndFunc


Func GUI_EVENT_CLOSE()
	DestroyProgram()
	DestroyVertexBuffer()
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
