#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

; --- OpenGL globals ---

Global $iProgram, $iVertexBufferObject, $iVao

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 500, $iHeight = 500

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight

Global $hGui, $bResize = True, $bClose = False

; --- Project includes ---

#include "utilities\includes.au3"

; --- Program code ---

MainFunc()

While 1
	If $bResize Then
		glViewport( 0, 0, $iWidth, $iHeight )
		$bResize = False
	EndIf
	If $bClose Then
		CloseProgram()
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then CloseProgram()
	If Not InitOpenGLwindow() Then CloseProgram()
	If Not CheckOpenGLversion( 3, 1 ) Then CloseProgram()
	If Not CreateResources() Then CloseProgram()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 100, $WS_OVERLAPPEDWINDOW )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	Return 1

EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 0.0, 0.0, 0.0, 0.0 )

	GUISwitch( $hGui )

	Return 1

EndFunc


Func CreateResources()

	If Not InitializeProgram() Then Return 0

	If Not InitializeVertexBuffer() Then Return 0

	glGenVertexArrays( 1, $iVao )
	glBindVertexArray( $iVao )

	GUISetState()

	Return 1

EndFunc


Func InitializeProgram()

	Local $iVertexShader, $iFragmentShader
	If Not LoadGLSLshader( $iVertexShader, "shaders\VertexColors.vert", $GL_VERTEX_SHADER ) Then Return 0
	If Not LoadGLSLshader( $iFragmentShader, "shaders\VertexColors.frag", $GL_FRAGMENT_SHADER ) Then Return 0

	$iProgram = glCreateProgram()
	glAttachShader( $iProgram, $iVertexShader )
	glAttachShader( $iProgram, $iFragmentShader )
	glLinkProgram( $iProgram )
	glUseProgram(0)

	Return 1

EndFunc


Func InitializeVertexBuffer()

	Local $aVertexData[24] = [ _
		 0.0,    0.5, 0.0, 1.0, _
		 0.5, -0.366, 0.0, 1.0, _
		-0.5, -0.366, 0.0, 1.0, _
		 1.0,    0.0, 0.0, 1.0, _
		 0.0,    1.0, 0.0, 1.0, _
		 0.0,    0.0, 1.0, 1.0 ]

	glGenBuffers( 1, $iVertexBufferObject )
	glBindBuffer( $GL_ARRAY_BUFFER, $iVertexBufferObject )
	glBufferData( $GL_ARRAY_BUFFER, $GL_FLOAT, 4*UBound($aVertexData), $aVertexData, $GL_STATIC_DRAW )
	glBindBuffer( $GL_ARRAY_BUFFER, 0 )

	Return 1

EndFunc


Func DisplayFunction()

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )

	glUseProgram( $iProgram )

	glBindBuffer( $GL_ARRAY_BUFFER, $iVertexBufferObject )
	glEnableVertexAttribArray(0)
	glEnableVertexAttribArray(1)
	glVertexAttribPointer( 0, 4, $GL_FLOAT, $GL_FALSE, 0, 0 )
	glVertexAttribPointer( 1, 4, $GL_FLOAT, $GL_FALSE, 0, 12*4 )

	glDrawArrays( $GL_TRIANGLES, 0, 3 )

	glDisableVertexAttribArray(0)
	glDisableVertexAttribArray(1)
	glUseProgram(0)

	SwapBuffers( $hDC )

EndFunc


Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	$bResize = True
EndFunc


Func GUI_EVENT_CLOSE()
	$bClose = True
EndFunc


Func CloseProgram()
	glDeleteProgram( $iProgram )
	;DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
