#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

; --- OpenGL globals ---

Global $iProgram, $iVao, $iVertexBufferObject, $iIndexBufferObject
Global $iOffsetUniform, $iPerspectiveMatrixUnif
Global $fFrustumScale = 1.0
Global $mPerspectiveMatrix[16] = [ _
	0, 0, 0, 0, _
	0, 0, 0, 0, _
	0, 0, 0, 0, _
	0, 0, 0, 0 ]

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 500, $iHeight = 500

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30
Global Const $iStatusBarHeight = 28

Global $hStatusbar, $idRadio1, $idRadio2, $idRadio3
Global $iOverlap = 0, $iNewOverlap = $iOverlap

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight + $iStatusBarHeight

Global $hGui, $bResize = True, $bClose = False

; --- Project includes ---

#include "utilities\includes.au3"
#include "utilities\VertexData1.au3"

; --- Program code ---

MainFunc()

While 1
	If $iOverlap <> $iNewOverlap Then
		$iOverlap = $iNewOverlap
	EndIf
	If $bResize Then
		$mPerspectiveMatrix[0] = $fFrustumScale / ( $iWidth / $iHeight )
		$mPerspectiveMatrix[5] = $fFrustumScale

		glUseProgram( $iProgram )
		glUniformMatrix4fv( $iPerspectiveMatrixUnif, 1, $GL_FALSE, $mPerspectiveMatrix )
		glUseProgram(0)

		glViewport( 0, 0, $iWidth, $iHeight )
		$bResize = False
	EndIf
	If $bClose Then
		CloseProgram()
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then CloseProgram()
	If Not InitOpenGLwindow() Then CloseProgram()
	If Not CheckOpenGLversion( 3, 1 ) Then CloseProgram()
	If Not CreateResources() Then CloseProgram()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 100, $WS_OVERLAPPEDWINDOW )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	GUIRegisterMsg( $WM_SIZING, "WM_SIZING" )

	CreateStatusbar()

	Return 1

EndFunc


Func CreateStatusbar()

	; Use a child window as a container for controls in the Statusbar
	$hStatusbar = GUICreate( "", $iGuiWidth - 4 , 24, 2, $iHeight + 3 , $WS_CHILD, -1, $hGui )

	; Radio buttons
	GUIStartGroup()
	GUICtrlCreateLabel( "Overlap:", 5, 5, 50, 14 )
	$idRadio1 = GUICtrlCreateRadio( "None",   60, 3, 50, 18 )
	$idRadio2 = GUICtrlCreateRadio( "Minor", 120, 3, 50, 18 )
	$idRadio3 = GUICtrlCreateRadio( "Major", 180, 3, 50, 18 )
	GUICtrlSetOnEvent( $idRadio1, "RadioClick" )
	GUICtrlSetOnEvent( $idRadio2, "RadioClick" )
	GUICtrlSetOnEvent( $idRadio3, "RadioClick" )
	GUICtrlSetState( $idRadio1, $GUI_CHECKED )

	GUISetState( @SW_SHOW, $hStatusbar )
	GUISwitch( $hGui )

	; Label for $hStatusbar to make it look nicer
	GUICtrlCreateLabel( "", 0, $iHeight + 1, $iGuiWidth + 4, $iStatusBarHeight, $SS_ETCHEDFRAME )
	GUICtrlSetResizing( -1, $GUI_DOCKSTATEBAR )

EndFunc


Func RadioClick()
	Switch @GUI_CTRLID
		Case $idRadio1
			$iNewOverlap = 0
		Case $idRadio2
			$iNewOverlap = 1
		Case $idRadio3
			$iNewOverlap = 2
	EndSwitch
EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 0.0, 0.0, 0.0, 0.0 )

	glEnable( $GL_CULL_FACE )
	glCullFace( $GL_BACK )
	glFrontFace( $GL_CW )

	glEnable( $GL_DEPTH_TEST )
	glDepthMask( $GL_TRUE )
	glDepthFunc( $GL_LEQUAL )
	glDepthRange( 0.0, 1.0 )

	GUISwitch( $hGui )

	Return 1

EndFunc


Func CreateResources()

	If Not InitializeProgram() Then Return 0

	If Not InitializeVertexBuffer() Then Return 0

	If Not InitializeVertexArrayObject() Then Return 0

	GUISetState()

	Return 1

EndFunc


Func InitializeProgram()

	Local $iVertexShader, $iFragmentShader
	If Not LoadGLSLshader( $iVertexShader, "shaders\Standard.vert", $GL_VERTEX_SHADER ) Then Return 0
	If Not LoadGLSLshader( $iFragmentShader, "shaders\Standard.frag", $GL_FRAGMENT_SHADER ) Then Return 0

	$iProgram = glCreateProgram()
	glAttachShader( $iProgram, $iVertexShader )
	glAttachShader( $iProgram, $iFragmentShader )
	glLinkProgram( $iProgram )

	$iOffsetUniform = glGetUniformLocation( $iProgram, "offset" )

	$iPerspectiveMatrixUnif = glGetUniformLocation( $iProgram, "perspectiveMatrix" )

	Local $zNear = 1.0, $zFar = 3.0

	$mPerspectiveMatrix[0]  = $fFrustumScale
	$mPerspectiveMatrix[5]  = $fFrustumScale
	$mPerspectiveMatrix[10] = ( $zFar + $zNear ) / ( $zNear - $zFar )
	$mPerspectiveMatrix[14] = ( 2 * $zFar * $zNear ) / ( $zNear - $zFar )
	$mPerspectiveMatrix[11] = -1.0

	glUseProgram( $iProgram )
	glUniformMatrix4fv( $iPerspectiveMatrixUnif, 1, $GL_FALSE, $mPerspectiveMatrix )
	glUseProgram(0)

	Return 1

EndFunc


Func InitializeVertexBuffer()

	glGenBuffers( 1, $iVertexBufferObject )
	glBindBuffer( $GL_ARRAY_BUFFER, $iVertexBufferObject )
	glBufferData( $GL_ARRAY_BUFFER, $GL_FLOAT, 4*UBound($aVertexData), $aVertexData, $GL_STATIC_DRAW )
	glBindBuffer( $GL_ARRAY_BUFFER, 0 )

	glGenBuffers( 1, $iIndexBufferObject )
	glBindBuffer( $GL_ELEMENT_ARRAY_BUFFER, $iIndexBufferObject )
	glBufferData( $GL_ELEMENT_ARRAY_BUFFER, $GL_INT, 4*UBound($aIndexData), $aIndexData, $GL_STATIC_DRAW )
	glBindBuffer( $GL_ELEMENT_ARRAY_BUFFER, 0 )

	Return 1

EndFunc


Func InitializeVertexArrayObject()

	glGenVertexArrays( 1, $iVao )
	glBindVertexArray( $iVao )

	Local $ColorDataOffset = 4 * 3 * $NumberOfVertices

	glBindBuffer( $GL_ARRAY_BUFFER, $iVertexBufferObject )
	glEnableVertexAttribArray(0)
	glEnableVertexAttribArray(1)
	glVertexAttribPointer( 0, 3, $GL_FLOAT, $GL_FALSE, 0, 0 )
	glVertexAttribPointer( 1, 4, $GL_FLOAT, $GL_FALSE, 0, $ColorDataOffset )
	glBindBuffer( $GL_ELEMENT_ARRAY_BUFFER, $iIndexBufferObject )

	glBindVertexArray(0)

	Return 1

EndFunc


Func DisplayFunction()

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )

	glUseProgram( $iProgram )

	glBindVertexArray( $iVao )

	Local $zOverlap
	Switch $iOverlap
		Case 0
			$zOverlap = 0.0
		Case 1
			$zOverlap = -0.75
		Case 2
			$zOverlap = -1.0
	EndSwitch
	glUniform3f( $iOffsetUniform, 0.0, 0.0, $zOverlap )
	glDrawElements( $GL_TRIANGLES, UBound($aIndexData), $GL_UNSIGNED_INT, 0 )

	glUniform3f( $iOffsetUniform, 0.0, 0.0, -1.0 )
	glDrawElementsBaseVertex( $GL_TRIANGLES, UBound($aIndexData), $GL_UNSIGNED_INT, 0, $NumberOfVertices / 2 )

	glUseProgram(0)

	SwapBuffers( $hDC )

EndFunc


Func WM_SIZING( $hWnd, $iMsg, $iwParam, $ilParam )
	#forceref $iMsg, $iwParam
	Switch $hWnd
		Case $hGui
			Local $tRECT = DllStructCreate( $tagRECT, $ilParam )
			Local $y = DllStructGetData( $tRECT, "Top" )
			Local $iHeight = DllStructGetData( $tRECT, "Bottom" ) - $y - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight
			WinMove( $hStatusbar, "", 2, $iHeight + $iHeightXP )
	EndSwitch
	Return $GUI_RUNDEFMSG
EndFunc


Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	WinMove( $hStatusbar, "", 2, $iHeight + $iHeightXP )
	$bResize = True
EndFunc


Func GUI_EVENT_CLOSE()
	$bClose = True
EndFunc


Func CloseProgram()
	glDeleteProgram( $iProgram )
	;DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
