#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

; --- OpenGL globals ---

Global $iProgram, $iVao, $iVertexBufferObject, $iIndexBufferObject
Global $iModelToCameraMatrixUnif, $iCameraToClipMatrixUnif
Global $fFrustumScale, $mCameraToClipMatrix

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 500, $iHeight = 500

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30
Global Const $iStatusBarHeight = 28

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight + $iStatusBarHeight

Global $hGui, $bResize = True, $bClose = False

; --- Project includes ---

#include "utilities\includes.au3"
#include "..\utilities\mathematics.au3"
#include "utilities\VertexData1.au3"

; --- Program code ---

MainFunc()

While 1
	If $bResize Then
		$mCameraToClipMatrix[0] = $fFrustumScale * ( $iHeight / $iWidth )
		$mCameraToClipMatrix[5] = $fFrustumScale

		glUseProgram( $iProgram )
		glUniformMatrix4fv( $iCameraToClipMatrixUnif, 1, $GL_FALSE, $mCameraToClipMatrix )
		glUseProgram(0)

		glViewport( 0, 0, $iWidth, $iHeight )
		$bResize = False
	EndIf
	If $bClose Then
		CloseProgram()
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then CloseProgram()
	If Not InitOpenGLwindow() Then CloseProgram()
	If Not CheckOpenGLversion( 3, 1 ) Then CloseProgram()
	If Not CreateResources() Then CloseProgram()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 100, $WS_OVERLAPPEDWINDOW )

	; Statusbar
	Local $idStatus, $idLabel
	$idStatus = GUICtrlCreateLabel( "", 0, $iHeight + 1, $iGuiWidth + 2, $iStatusBarHeight, $SS_ETCHEDFRAME )
	$idLabel = GUICtrlCreateLabel( "", 3, $iHeight + 1 + 6, $iGuiWidth - 4, $iStatusBarHeight - 10 )
	GUICtrlSetData( $idLabel, "Translation project." )
	GUICtrlSetResizing( $idStatus, $GUI_DOCKSTATEBAR )
	GUICtrlSetResizing( $idLabel, $GUI_DOCKSTATEBAR )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	Return 1

EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 0.0, 0.0, 0.0, 0.0 )
	glClearDepth( 1.0 )

	glEnable( $GL_CULL_FACE )
	glCullFace( $GL_BACK )
	glFrontFace( $GL_CW )

	glEnable( $GL_DEPTH_TEST )
	glDepthMask( $GL_TRUE )
	glDepthFunc( $GL_LEQUAL )
	glDepthRange( 0.0, 1.0 )

	GUISwitch( $hGui )

	Return 1

EndFunc


Func CreateResources()

	If Not InitializeProgram() Then Return 0

	If Not InitializeVertexBuffer() Then Return 0

	If Not InitializeVertexArrayObject() Then Return 0

	GUISetState()

	Return 1

EndFunc


Func InitializeProgram()

	Local $iVertexShader, $iFragmentShader
	If Not LoadGLSLshader( $iVertexShader, "shaders\PosColorLocalTransform.vert", $GL_VERTEX_SHADER ) Then Return 0
	If Not LoadGLSLshader( $iFragmentShader, "shaders\ColorPassthrough.frag", $GL_FRAGMENT_SHADER ) Then Return 0

	$iProgram = glCreateProgram()
	glAttachShader( $iProgram, $iVertexShader )
	glAttachShader( $iProgram, $iFragmentShader )
	glLinkProgram( $iProgram )

	$iModelToCameraMatrixUnif = glGetUniformLocation( $iProgram, "modelToCameraMatrix" )
	$iCameraToClipMatrixUnif = glGetUniformLocation( $iProgram, "cameraToClipMatrix" )

	Local $zNear = 1.0, $zFar = 45.0
	$fFrustumScale = CalcFrustumScale( 45 )

	$mCameraToClipMatrix = $NULL_MATRIX
	$mCameraToClipMatrix[0]  = $fFrustumScale
	$mCameraToClipMatrix[5]  = $fFrustumScale
	$mCameraToClipMatrix[10] = ( $zFar + $zNear ) / ( $zNear - $zFar )
	$mCameraToClipMatrix[11] = -1.0
	$mCameraToClipMatrix[14] = ( 2 * $zFar * $zNear ) / ( $zNear - $zFar )

	glUseProgram( $iProgram )
	glUniformMatrix4fv( $iCameraToClipMatrixUnif, 1, $GL_FALSE, $mCameraToClipMatrix )
	glUseProgram(0)

	Return 1

EndFunc


Func InitializeVertexBuffer()

	glGenBuffers( 1, $iVertexBufferObject )
	glBindBuffer( $GL_ARRAY_BUFFER, $iVertexBufferObject )
	glBufferData( $GL_ARRAY_BUFFER, $GL_FLOAT, 4*UBound($aVertexData), $aVertexData, $GL_STATIC_DRAW )
	glBindBuffer( $GL_ARRAY_BUFFER, 0 )

	glGenBuffers( 1, $iIndexBufferObject )
	glBindBuffer( $GL_ELEMENT_ARRAY_BUFFER, $iIndexBufferObject )
	glBufferData( $GL_ELEMENT_ARRAY_BUFFER, $GL_SHORT, 2*UBound($aIndexData), $aIndexData, $GL_STATIC_DRAW )
	glBindBuffer( $GL_ELEMENT_ARRAY_BUFFER, 0 )

	Return 1

EndFunc


Func InitializeVertexArrayObject()

	glGenVertexArrays( 1, $iVao )
	glBindVertexArray( $iVao )

	Local $ColorDataOffset = 4 * 3 * $NumberOfVertices

	glBindBuffer( $GL_ARRAY_BUFFER, $iVertexBufferObject )
	glEnableVertexAttribArray(0)
	glEnableVertexAttribArray(1)
	glVertexAttribPointer( 0, 3, $GL_FLOAT, $GL_FALSE, 0, 0 )
	glVertexAttribPointer( 1, 4, $GL_FLOAT, $GL_FALSE, 0, $ColorDataOffset )
	glBindBuffer( $GL_ELEMENT_ARRAY_BUFFER, $iIndexBufferObject )

	glBindVertexArray(0)

	Return 1

EndFunc


Func DisplayFunction()

	Local Static $iStartTime = TimerInit(), $fLastTime = 0
	Local $fElapsedTime = TimerDiff( $iStartTime ) / 1000, $m
	$fElapsedTime = ( $fLastTime + $fElapsedTime ) / 2
	$fLastTime = $fElapsedTime

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )

	glUseProgram( $iProgram )

	glBindVertexArray( $iVao )

	$m = StationaryOffset( $fElapsedTime )
	glUniformMatrix4fv( $iModelToCameraMatrixUnif, 1, $GL_FALSE, $m )
	glDrawElements( $GL_TRIANGLES, UBound($aIndexData), $GL_UNSIGNED_SHORT, 0 )

	$m = OvalOffset( $fElapsedTime )
	glUniformMatrix4fv( $iModelToCameraMatrixUnif, 1, $GL_FALSE, $m )
	glDrawElements( $GL_TRIANGLES, UBound($aIndexData), $GL_UNSIGNED_SHORT, 0 )

	$m = BottomCircleOffset( $fElapsedTime )
	glUniformMatrix4fv( $iModelToCameraMatrixUnif, 1, $GL_FALSE, $m )
	glDrawElements( $GL_TRIANGLES, UBound($aIndexData), $GL_UNSIGNED_SHORT, 0 )

	glBindVertexArray(0)
	glUseProgram(0)

	SwapBuffers( $hDC )

EndFunc


Func StationaryOffset( $fElapsedTime )
	Local $m = $IDENTITY_MATRIX
	$m[12] =   0.0
	$m[13] =   0.0
	$m[14] = -20.0
	$m[15] =   1.0
	Return $m
EndFunc


Func OvalOffset( $fElapsedTime )
	Local $m = $IDENTITY_MATRIX
	Local $fLoopDuration = 3, $fScale = $PI * 2 / $fLoopDuration
	Local $fCurrTimeThroughLoop = Mod( $fElapsedTime, $fLoopDuration )
	$m[12] = Cos( $fCurrTimeThroughLoop * $fScale ) * 4
	$m[13] = Sin( $fCurrTimeThroughLoop * $fScale ) * 6
	$m[14] = -20.0
	$m[15] = 1.0
	Return $m
EndFunc


Func BottomCircleOffset( $fElapsedTime )
	Local $m = $IDENTITY_MATRIX
	Local $fLoopDuration = 12, $fScale = $PI * 2 / $fLoopDuration
	Local $fCurrTimeThroughLoop = Mod( $fElapsedTime, $fLoopDuration )
	$m[12] = Cos( $fCurrTimeThroughLoop * $fScale ) * 5
	$m[13] = -3.5
	$m[14] = Sin( $fCurrTimeThroughLoop * $fScale ) * 5 - 20
	$m[15] = 1.0
	Return $m
EndFunc


Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	$bResize = True
EndFunc


Func GUI_EVENT_CLOSE()
	$bClose = True
EndFunc


Func CloseProgram()
	glDeleteProgram( $iProgram )
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
