#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

Global $centerOffset      = 0
Global $posBase[3]        = [ 3.0, -5.0, -40.0 ]
Global $angBase           = -45.0
Global $posBaseLeft[3]    = [ 2.0, 0.0, 0.0 ]
Global $posBaseRight[3]   = [ -2.0, 0.0, 0.0 ]
Global $scaleBaseZ        = 3.0
Global $angUpperArm       = -51.75
Global $lenUpperArm       = 9.0
Global $posLowerArm[3]    = [ 0.0, 0.0, 8.0 ]
Global $angLowerArm       = 47.25
Global $lenLowerArm       = 5.0
Global $widthLowerArm     = 1.5
Global $posWrist[3]       = [ 0.0, 0.0, 5.0 ]
Global $angWristRoll      = 0.0
Global $angWristPitch     = 67.5
Global $lenWrist          = 2.0
Global $widthWrist        = 2.0
Global $posLeftFinger[3]  = [ 1.0, 0.0, 1.0 ]
Global $posRightFinger[3] = [ -1.0, 0.0, 1.0 ]
Global $angFingerOpen     = 50.0
Global $lenFinger         = 2.0
Global $widthFinger       = 0.5
Global $angLowerFinger    = 45.0

Global $StandardAngelIncrement = 11.25
Global $SmallAngelIncrement    =  9.0
Global $Speed                  =  0.1

; --- OpenGL globals ---

Global $iProgram, $iVao, $iVertexBufferObject, $iIndexBufferObject
Global $iModelToCameraMatrixUnif, $iCameraToClipMatrixUnif
Global $fFrustumScale, $mCameraToClipMatrix
Global $iPositionAttrib, $iColorAttrib

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 550, $iHeight = 550

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30
Global Const $iStatusBarHeight = 28

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight + $iStatusBarHeight

Global $hGui, $idInput, $bResize = True, $bClose = False, $bFocus = True

; --- Project includes ---

#include "utilities\includes.au3"
#include "..\utilities\mathematics.au3"
#include "utilities\VertexData2.au3"
#include "utilities\Hierarchy.au3"

; --- Program code ---

MainFunc()

While 1
	If $bResize Then
		$mCameraToClipMatrix[0] = $fFrustumScale * ( $iHeight / $iWidth )
		$mCameraToClipMatrix[5] = $fFrustumScale

		glUseProgram( $iProgram )
		glUniformMatrix4fv( $iCameraToClipMatrixUnif, 1, $GL_FALSE, $mCameraToClipMatrix )
		glUseProgram(0)

		glViewport( 0, 0, $iWidth, $iHeight )
		$bResize = False
	EndIf
	If $bClose Then
		CloseProgram()
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then CloseProgram()
	If Not InitOpenGLwindow() Then CloseProgram()
	If Not CheckOpenGLversion( 3, 1 ) Then CloseProgram()
	If Not CreateResources() Then CloseProgram()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 100, $WS_OVERLAPPEDWINDOW )

	; Statusbar
	Local $idStatus, $idLabel
	$idInput = GUICtrlCreateInput( "", 0, 0, 1, 1, 2048 ) ; 2048 = $ES_READONLY
	$idStatus = GUICtrlCreateLabel( "", 0, $iHeight + 1, $iGuiWidth + 2, $iStatusBarHeight, $SS_ETCHEDFRAME )
	$idLabel = GUICtrlCreateLabel( "", 3, $iHeight + 1 + 6, $iGuiWidth - 4, $iStatusBarHeight - 10 )
	GUICtrlSetData( $idLabel, "Keyboard:  Base(Q,A W,S)  Arms(E,D R,F)  Wrist(T,G Y,H)  Fingers(U,J)  Speed(Shift)  Reset(O)  Demo(M)" )
	GUICtrlSetResizing( $idStatus, $GUI_DOCKSTATEBAR )
	GUICtrlSetResizing( $idLabel, $GUI_DOCKSTATEBAR )
	GUICtrlSetBkColor( $idInput, 0x000000 )
	GUICtrlSetState( $idInput, $GUI_FOCUS )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	GUIRegisterMsg( $WM_ACTIVATE, "WM_ACTIVATE" )

	Return 1

EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 0.0, 0.0, 0.0, 0.0 )
	glClearDepth( 1.0 )

	glEnable( $GL_CULL_FACE )
	glCullFace( $GL_BACK )
	glFrontFace( $GL_CW )

	glEnable( $GL_DEPTH_TEST )
	glDepthMask( $GL_TRUE )
	glDepthFunc( $GL_LEQUAL )
	glDepthRange( 0.0, 1.0 )

	GUISwitch( $hGui )

	Return 1

EndFunc


Func CreateResources()

	If Not InitializeProgram() Then Return 0

	If Not InitializeVertexBuffer() Then Return 0

	If Not InitializeVertexArrayObject() Then Return 0

	GUISetState()

	Return 1

EndFunc


Func InitializeProgram()

	Local $iVertexShader, $iFragmentShader
	If Not LoadGLSLshader( $iVertexShader, "shaders\PosColorLocalTransform.vert", $GL_VERTEX_SHADER ) Then Return 0
	If Not LoadGLSLshader( $iFragmentShader, "shaders\ColorPassthrough.frag", $GL_FRAGMENT_SHADER ) Then Return 0

	$iProgram = glCreateProgram()
	glAttachShader( $iProgram, $iVertexShader )
	glAttachShader( $iProgram, $iFragmentShader )
	glLinkProgram( $iProgram )

	$iColorAttrib = glGetAttribLocation( $iProgram, "color" )
	$iPositionAttrib = glGetAttribLocation( $iProgram, "position" )

	$iModelToCameraMatrixUnif = glGetUniformLocation( $iProgram, "modelToCameraMatrix" )
	$iCameraToClipMatrixUnif = glGetUniformLocation( $iProgram, "cameraToClipMatrix" )

	Local $zNear = 1.0, $zFar = 100.0
	$fFrustumScale = CalcFrustumScale( 45 )

	$mCameraToClipMatrix = $NULL_MATRIX
	$mCameraToClipMatrix[0]  = $fFrustumScale
	$mCameraToClipMatrix[5]  = $fFrustumScale
	$mCameraToClipMatrix[10] = ( $zFar + $zNear ) / ( $zNear - $zFar )
	$mCameraToClipMatrix[11] = -1.0
	$mCameraToClipMatrix[14] = ( 2 * $zFar * $zNear ) / ( $zNear - $zFar )

	glUseProgram( $iProgram )
	glUniformMatrix4fv( $iCameraToClipMatrixUnif, 1, $GL_FALSE, $mCameraToClipMatrix )
	glUseProgram(0)

	Return 1

EndFunc


Func InitializeVertexBuffer()

	glGenBuffers( 1, $iVertexBufferObject )
	glBindBuffer( $GL_ARRAY_BUFFER, $iVertexBufferObject )
	glBufferData( $GL_ARRAY_BUFFER, $GL_FLOAT, 4*UBound($aVertexData), $aVertexData, $GL_STATIC_DRAW )
	glBindBuffer( $GL_ARRAY_BUFFER, 0 )

	glGenBuffers( 1, $iIndexBufferObject )
	glBindBuffer( $GL_ELEMENT_ARRAY_BUFFER, $iIndexBufferObject )
	glBufferData( $GL_ELEMENT_ARRAY_BUFFER, $GL_SHORT, 2*UBound($aIndexData), $aIndexData, $GL_STATIC_DRAW )
	glBindBuffer( $GL_ELEMENT_ARRAY_BUFFER, 0 )

	Return 1

EndFunc


Func InitializeVertexArrayObject()

	glGenVertexArrays( 1, $iVao )
	glBindVertexArray( $iVao )

	Local $ColorDataOffset = 4 * 3 * $NumberOfVertices

	glBindBuffer( $GL_ARRAY_BUFFER, $iVertexBufferObject )
	glEnableVertexAttribArray(0)
	glEnableVertexAttribArray(1)
	glVertexAttribPointer( 0, 3, $GL_FLOAT, $GL_FALSE, 0, 0 )
	glVertexAttribPointer( 1, 4, $GL_FLOAT, $GL_FALSE, 0, $ColorDataOffset )
	glBindBuffer( $GL_ELEMENT_ARRAY_BUFFER, $iIndexBufferObject )

	glBindVertexArray(0)

	Return 1

EndFunc


Func DisplayFunction()

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )

	glUseProgram( $iProgram )

	glBindVertexArray( $iVao )

	HandleKeystrokes()

	If $bDemo Then _
		RobotDemo()

	RobotDraw()

	glBindVertexArray(0)
	glUseProgram(0)

	SwapBuffers( $hDC )

EndFunc


Func RobotDraw()

	; --- Base ---

	; Main base
	Local $mMainBase = $IDENTITY_MATRIX
	$mMainBase = glmTranslate( $mMainBase, $posBase )
	$mMainBase = glmRotate( $mMainBase, $angBase, $axisY )

	; Draw left base
	Local $mBaseLeft = $mMainBase, $scale[3] = [ 1.0, 1.0, $scaleBaseZ ]
	$mBaseLeft = glmTranslate( $mBaseLeft, $posBaseLeft )
	$mBaseLeft = glmScale( $mBaseLeft, $scale )
	glUniformMatrix4fv( $iModelToCameraMatrixUnif, 1, $GL_FALSE, $mBaseLeft )
	glDrawElements( $GL_TRIANGLES, UBound($aIndexData), $GL_UNSIGNED_SHORT, 0 )

	; Draw right base
	Local $mBaseRight = $mMainBase
	$mBaseRight = glmTranslate( $mBaseRight, $posBaseRight )
	$mBaseRight = glmScale( $mBaseRight, $scale )
	glUniformMatrix4fv( $iModelToCameraMatrixUnif, 1, $GL_FALSE, $mBaseRight )
	glDrawElements( $GL_TRIANGLES, UBound($aIndexData), $GL_UNSIGNED_SHORT, 0 )

	; --- Arms ---

	; Main arm
	Local $mMainArm = $mMainBase
	$mMainArm = glmRotate( $mMainArm, $angUpperArm, $axisX )

	; Draw upper arm
	Local $mUpperArm = $mMainArm, $trans[3] = [ 0.0, 0.0, ( $lenUpperArm / 2.0 ) - 1.0 ], $scale[3] = [ 1.0, 1.0, $lenUpperArm / 2.0 ]
	$mUpperArm = glmTranslate( $mUpperArm, $trans )
	$mUpperArm = glmScale( $mUpperArm, $scale )
	glUniformMatrix4fv( $iModelToCameraMatrixUnif, 1, $GL_FALSE, $mUpperArm )
	glDrawElements( $GL_TRIANGLES, UBound($aIndexData), $GL_UNSIGNED_SHORT, 0 )

	$mMainArm = glmTranslate( $mMainArm, $posLowerArm )
	$mMainArm = glmRotate( $mMainArm, $angLowerArm, $axisX )

	; Draw lower arm
	Local $mLowerArm = $mMainArm, $trans[3] = [ 0.0, 0.0, $lenLowerArm / 2.0 ], $scale[3] = [ $widthLowerArm / 2.0, $widthLowerArm / 2.0, $lenLowerArm / 2.0 ]
	$mLowerArm = glmTranslate( $mLowerArm, $trans )
	$mLowerArm = glmScale( $mLowerArm, $scale )
	glUniformMatrix4fv( $iModelToCameraMatrixUnif, 1, $GL_FALSE, $mLowerArm )
	glDrawElements( $GL_TRIANGLES, UBound($aIndexData), $GL_UNSIGNED_SHORT, 0 )

	; --- Wrist ---

	; Main wrist
	Local $mMainWrist = $mMainArm
	$mMainWrist = glmTranslate( $mMainWrist, $posWrist )
	$mMainWrist = glmRotate( $mMainWrist, $angWristRoll, $axisZ )
	$mMainWrist = glmRotate( $mMainWrist, $angWristPitch, $axisX )

	; Draw wrist
	Local $mWrist = $mMainWrist, $scale[3] = [ $widthWrist / 2.0, $widthWrist / 2.0, $lenWrist / 2.0 ]
	$mWrist = glmScale( $mWrist, $scale )
	glUniformMatrix4fv( $iModelToCameraMatrixUnif, 1, $GL_FALSE, $mWrist )
	glDrawElements( $GL_TRIANGLES, UBound($aIndexData), $GL_UNSIGNED_SHORT, 0 )

	; --- Fingers ---

	Local $transFinger[3] = [ 0.0, 0.0, $lenFinger / 2.0 ], $scaleFinger[3] = [ $widthFinger / 2.0, $widthFinger / 2.0, $lenFinger / 2.0 ]

	; Main left finger
	Local $mMainLeftFinger = $mMainWrist
	$mMainLeftFinger = glmTranslate( $mMainLeftFinger, $posLeftFinger )
	$mMainLeftFinger = glmRotate( $mMainLeftFinger, $angFingerOpen, $axisY )

	; Draw left finger
	Local $mLeftFinger = $mMainLeftFinger
	$mLeftFinger = glmTranslate( $mLeftFinger, $transFinger )
	$mLeftFinger = glmScale( $mLeftFinger, $scaleFinger )
	glUniformMatrix4fv( $iModelToCameraMatrixUnif, 1, $GL_FALSE, $mLeftFinger )
	glDrawElements( $GL_TRIANGLES, UBound($aIndexData), $GL_UNSIGNED_SHORT, 0 )

	; Main left lower finger
	Local $mMainLeftLowerFinger = $mMainLeftFinger, $trans[3] = [ 0.0, 0.0, $lenFinger ]
	$mMainLeftLowerFinger = glmTranslate( $mMainLeftLowerFinger, $trans )
	$mMainLeftLowerFinger = glmRotate( $mMainLeftLowerFinger, -$angLowerFinger, $axisY )

	; Draw left lower finger
	Local $mLeftLowerFinger = $mMainLeftLowerFinger
	$mLeftLowerFinger = glmTranslate( $mLeftLowerFinger, $transFinger )
	$mLeftLowerFinger = glmScale( $mLeftLowerFinger, $scaleFinger )
	glUniformMatrix4fv( $iModelToCameraMatrixUnif, 1, $GL_FALSE, $mLeftLowerFinger )
	glDrawElements( $GL_TRIANGLES, UBound($aIndexData), $GL_UNSIGNED_SHORT, 0 )

	; Main right finger
	Local $mMainRightFinger = $mMainWrist
	$mMainRightFinger = glmTranslate( $mMainRightFinger, $posRightFinger )
	$mMainRightFinger = glmRotate( $mMainRightFinger, -$angFingerOpen, $axisY )

	; Draw right finger
	Local $mRightFinger = $mMainRightFinger
	$mRightFinger = glmTranslate( $mRightFinger, $transFinger )
	$mRightFinger = glmScale( $mRightFinger, $scaleFinger )
	glUniformMatrix4fv( $iModelToCameraMatrixUnif, 1, $GL_FALSE, $mRightFinger )
	glDrawElements( $GL_TRIANGLES, UBound($aIndexData), $GL_UNSIGNED_SHORT, 0 )

	; Main right lower finger
	Local $mMainRightLowerFinger = $mMainRightFinger
	$mMainRightLowerFinger = glmTranslate( $mMainRightLowerFinger, $trans )
	$mMainRightLowerFinger = glmRotate( $mMainRightLowerFinger, $angLowerFinger, $axisY )

	; Draw right lower finger
	Local $mRightLowerFinger = $mMainRightLowerFinger
	$mRightLowerFinger = glmTranslate( $mRightLowerFinger, $transFinger )
	$mRightLowerFinger = glmScale( $mRightLowerFinger, $scaleFinger )
	glUniformMatrix4fv( $iModelToCameraMatrixUnif, 1, $GL_FALSE, $mRightLowerFinger )
	glDrawElements( $GL_TRIANGLES, UBound($aIndexData), $GL_UNSIGNED_SHORT, 0 )

EndFunc


Func WM_ACTIVATE( $hWnd, $Msg, $wParam, $lParam )
	Switch $hWnd
		Case $hGui
			$bFocus = BitAND( $wParam, 0xFFFF )
	EndSwitch
	Return $GUI_RUNDEFMSG
EndFunc


Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	$bResize = True
EndFunc


Func GUI_EVENT_CLOSE()
	$bClose = True
EndFunc


Func CloseProgram()
	glDeleteProgram( $iProgram )
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
