#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

Global $iRobots = 3, $aRobots[$iRobots][24], $aStoreRobots[$iRobots][3], $aParkRobots

Global $posBase[3]        = [  3.0, -5.0, -40.0 ]
Global $posBaseLeft[3]    = [  2.0,  0.0,   0.0 ]
Global $posBaseRight[3]   = [ -2.0,  0.0,   0.0 ]
Global $posLowerArm[3]    = [  0.0,  0.0,   8.0 ]
Global $posWrist[3]       = [  0.0,  0.0,   5.0 ]
Global $posLeftFinger[3]  = [  1.0,  0.0,   1.0 ]
Global $posRightFinger[3] = [ -1.0,  0.0,   1.0 ]

Global $StandardAngelIncrement = 11.25
Global $SmallAngelIncrement    =  9.0
Global $Speed                  =  0.1

; --- OpenGL globals ---

Global $iProgram, $iVao, $iVertexBufferObject, $iIndexBufferObject
Global $iModelToCameraMatrixUnif, $iCameraToClipMatrixUnif
Global $fFrustumScale, $mCameraToClipMatrix
Global $iPositionAttrib, $iColorAttrib

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 550, $iHeight = 550

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30
Global Const $iStatusBarHeight = 28
Global Const $iStatusBarRows   =  2

Global $hStatusbar, $idNone, $idRobot0, $idRobot1, $idRobot2
Global $iRobot = -1, $iNewRobot = $iRobot

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight + $iStatusBarHeight * $iStatusBarRows

Global $hGui, $idInput, $bResize = True, $bClose = False, $bFocus = True

; --- Project includes ---

#include "utilities\includes.au3"
#include "..\utilities\mathematics.au3"
#include "utilities\VertexData2.au3"
#include "utilities\Three robotic arms.au3"

; --- Program code ---

MainFunc()

While 1
	If $iRobot <> $iNewRobot Then
		If $iRobot > -1 Then
			; Store robot positions
			$aStoreRobots[$iRobot][0] = $aRobots[$iRobot][0]
			$aStoreRobots[$iRobot][1] = $aRobots[$iRobot][1]
			$aStoreRobots[$iRobot][2] = $aRobots[$iRobot][2]
			; Positions to park robot
			$aRobots[$iRobot][0] = $aParkRobots[$iRobot][0]
			$aRobots[$iRobot][1] = $aParkRobots[$iRobot][1]
			$aRobots[$iRobot][2] = $aParkRobots[$iRobot][2]
		EndIf
		$bDemo = 0
		$iRobot = $iNewRobot
		If $iRobot > -1 Then
			; Restore robot positions
			$aRobots[$iRobot][0] = $aStoreRobots[$iRobot][0]
			$aRobots[$iRobot][1] = $aStoreRobots[$iRobot][1]
			$aRobots[$iRobot][2] = $aStoreRobots[$iRobot][2]
		EndIf
	EndIf
	If $bResize Then
		$mCameraToClipMatrix[0] = $fFrustumScale * ( $iHeight / $iWidth )
		$mCameraToClipMatrix[5] = $fFrustumScale

		glUseProgram( $iProgram )
		glUniformMatrix4fv( $iCameraToClipMatrixUnif, 1, $GL_FALSE, $mCameraToClipMatrix )
		glUseProgram(0)

		glViewport( 0, 0, $iWidth, $iHeight )
		$bResize = False
	EndIf
	If $bClose Then
		CloseProgram()
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then CloseProgram()
	If Not InitOpenGLwindow() Then CloseProgram()
	If Not CheckOpenGLversion( 3, 1 ) Then CloseProgram()
	If Not CreateResources() Then CloseProgram()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 100, $WS_OVERLAPPEDWINDOW )

	; Input control to receive key presses (to prevent sound alarm)
	$idInput = GUICtrlCreateInput( "", 0, 0, 1, 1, 2048 ) ; 2048 = $ES_READONLY
	GUICtrlSetBkColor( $idInput, 0x000000 )
	GUICtrlSetState( $idInput, $GUI_FOCUS )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	GUIRegisterMsg( $WM_ACTIVATE, "WM_ACTIVATE" )
	GUIRegisterMsg( $WM_SIZING,   "WM_SIZING" )

	CreateStatusbar()

	Return 1

EndFunc


Func CreateStatusbar()

	; Use a child window as a container for controls in the Statusbar
	$hStatusbar = GUICreate( "", $iGuiWidth - 4 , 24, 2, $iHeight + $iStatusBarHeight + 3 , $WS_CHILD, -1, $hGui )

	; Robots
	GUIStartGroup()
	GUICtrlCreateLabel( "Select active robot:", 2, 5, 100, 14 )
	$idNone   = GUICtrlCreateRadio( "None",    110, 3, 60, 18 )
	$idRobot0 = GUICtrlCreateRadio( "Robot 0", 180, 3, 60, 18 )
	$idRobot1 = GUICtrlCreateRadio( "Robot 1", 250, 3, 60, 18 )
	$idRobot2 = GUICtrlCreateRadio( "Robot 2", 320, 3, 60, 18 )
	GUICtrlSetOnEvent( $idNone,   "RadioClick" )
	GUICtrlSetOnEvent( $idRobot0, "RadioClick" )
	GUICtrlSetOnEvent( $idRobot1, "RadioClick" )
	GUICtrlSetOnEvent( $idRobot2, "RadioClick" )
	GUICtrlSetState( $idNone, $GUI_CHECKED )

	GUISetState( @SW_SHOW, $hStatusbar )
	GUISwitch( $hGui )

	; Label for $hStatusbar to make it look nicer
	GUICtrlCreateLabel( "", 0, $iHeight + 1, $iGuiWidth + 4, $iStatusBarHeight + 1, $SS_ETCHEDFRAME )
	GUICtrlSetResizing( -1, $GUI_DOCKSTATEBAR )
	GUICtrlCreateLabel( "Keyboard:  Base(Q,A W,S)  Arms(E,D R,F)  Wrist(T,G Y,H)  Fingers(U,J)  Speed(Shift)  Reset(O)  Demo(M)", _
											3, $iHeight + 1 + 6, $iGuiWidth - 6, $iStatusBarHeight - 10 )
	GUICtrlSetResizing( -1, $GUI_DOCKSTATEBAR )

EndFunc


Func RadioClick()
	Switch @GUI_CTRLID
		Case $idNone
			$iNewRobot = -1
		Case $idRobot0
			$iNewRobot = 0
		Case $idRobot1
			$iNewRobot = 1
		Case $idRobot2
			$iNewRobot = 2
	EndSwitch
	GUICtrlSetState( $idInput, $GUI_FOCUS )
EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 0.0, 0.0, 0.0, 0.0 )
	glClearDepth( 1.0 )

	glEnable( $GL_CULL_FACE )
	glCullFace( $GL_BACK )
	glFrontFace( $GL_CW )

	glEnable( $GL_DEPTH_TEST )
	glDepthMask( $GL_TRUE )
	glDepthFunc( $GL_LEQUAL )
	glDepthRange( 0.0, 1.0 )

	GUISwitch( $hGui )

	Return 1

EndFunc


Func CreateResources()

	If Not CreateRobots() Then Return 0

	If Not InitializeProgram() Then Return 0

	If Not InitializeVertexBuffer() Then Return 0

	If Not InitializeVertexArrayObject() Then Return 0

	GUISetState()

	Return 1

EndFunc


Func CreateRobots()

	For $i = 0 To $iRobots - 1
		$aStoreRobots[$i][0] =   0   ; $centerOffset
		$aStoreRobots[$i][1] =         $posBase
		$aStoreRobots[$i][2] = -45.0 ; $angBase
		$aRobots[$i][ 3] =          $posBaseLeft
		$aRobots[$i][ 4] =          $posBaseRight
		$aRobots[$i][ 5] =   3.0  ; $scaleBaseZ
		$aRobots[$i][ 6] = -51.75 ; $angUpperArm
		$aRobots[$i][ 7] =   9.0  ; $lenUpperArm
		$aRobots[$i][ 8] =          $posLowerArm
		$aRobots[$i][ 9] =  47.25 ; $angLowerArm
		$aRobots[$i][10] =  47.25 ; $angLowerArm
		$aRobots[$i][11] =   5.0  ; $lenLowerArm
		$aRobots[$i][12] =   1.5  ; $widthLowerArm
		$aRobots[$i][13] =          $posWrist
		$aRobots[$i][14] =   0.0  ; $angWristRoll
		$aRobots[$i][15] =  67.5  ; $angWristPitch
		$aRobots[$i][16] =   2.0  ; $lenWrist
		$aRobots[$i][17] =   2.0  ; $widthWrist
		$aRobots[$i][18] =          $posLeftFinger
		$aRobots[$i][19] =          $posRightFinger
		$aRobots[$i][20] =  50.0  ; $angFingerOpen
		$aRobots[$i][21] =   2.0  ; $lenFinger
		$aRobots[$i][22] =   0.5  ; $widthFinger
		$aRobots[$i][23] =  45.0  ; $angLowerFinger
	Next

	; Positions to park robot
	Local $posBase0[3] = [ -22.3840001435748, -5.0, -83.309266176085 ]
	Local $posBase1[3] = [ 1.03308323047067, -5.0, -90.0613747156601 ]
	Local $posBase2[3] = [ 25.7903230869251, -5.0, -84.7285275142557 ]

	; Positions to park robot
	Local $aParkRobotArms[$iRobots][3] = [ _
		[ -50.1999999999999, $posBase0,  30.375 ], _
		[ -50.0999999999999, $posBase1,   2.25  ], _
		[ -50.1999999999999, $posBase2, -27.0   ] ]

	$aParkRobots = $aParkRobotArms

	For $i = 0 To $iRobots - 1
		$aRobots[$i][0] = $aParkRobots[$i][0]
		$aRobots[$i][1] = $aParkRobots[$i][1]
		$aRobots[$i][2] = $aParkRobots[$i][2]
	Next

	Return 1

EndFunc


Func InitializeProgram()

	Local $iVertexShader, $iFragmentShader
	If Not LoadGLSLshader( $iVertexShader, "shaders\PosColorLocalTransform.vert", $GL_VERTEX_SHADER ) Then Return 0
	If Not LoadGLSLshader( $iFragmentShader, "shaders\ColorPassthrough.frag", $GL_FRAGMENT_SHADER ) Then Return 0

	$iProgram = glCreateProgram()
	glAttachShader( $iProgram, $iVertexShader )
	glAttachShader( $iProgram, $iFragmentShader )
	glLinkProgram( $iProgram )

	$iColorAttrib = glGetAttribLocation( $iProgram, "color" )
	$iPositionAttrib = glGetAttribLocation( $iProgram, "position" )

	$iModelToCameraMatrixUnif = glGetUniformLocation( $iProgram, "modelToCameraMatrix" )
	$iCameraToClipMatrixUnif = glGetUniformLocation( $iProgram, "cameraToClipMatrix" )

	Local $zNear = 1.0, $zFar = 100.0
	$fFrustumScale = CalcFrustumScale( 45 )

	$mCameraToClipMatrix = $NULL_MATRIX
	$mCameraToClipMatrix[0]  = $fFrustumScale
	$mCameraToClipMatrix[5]  = $fFrustumScale
	$mCameraToClipMatrix[10] = ( $zFar + $zNear ) / ( $zNear - $zFar )
	$mCameraToClipMatrix[11] = -1.0
	$mCameraToClipMatrix[14] = ( 2 * $zFar * $zNear ) / ( $zNear - $zFar )

	glUseProgram( $iProgram )
	glUniformMatrix4fv( $iCameraToClipMatrixUnif, 1, $GL_FALSE, $mCameraToClipMatrix )
	glUseProgram(0)

	Return 1

EndFunc


Func InitializeVertexBuffer()

	glGenBuffers( 1, $iVertexBufferObject )
	glBindBuffer( $GL_ARRAY_BUFFER, $iVertexBufferObject )
	glBufferData( $GL_ARRAY_BUFFER, $GL_FLOAT, 4*UBound($aVertexData), $aVertexData, $GL_STATIC_DRAW )
	glBindBuffer( $GL_ARRAY_BUFFER, 0 )

	glGenBuffers( 1, $iIndexBufferObject )
	glBindBuffer( $GL_ELEMENT_ARRAY_BUFFER, $iIndexBufferObject )
	glBufferData( $GL_ELEMENT_ARRAY_BUFFER, $GL_SHORT, 2*UBound($aIndexData), $aIndexData, $GL_STATIC_DRAW )
	glBindBuffer( $GL_ELEMENT_ARRAY_BUFFER, 0 )

	Return 1

EndFunc


Func InitializeVertexArrayObject()

	glGenVertexArrays( 1, $iVao )
	glBindVertexArray( $iVao )

	Local $ColorDataOffset = 4 * 3 * $NumberOfVertices

	glBindBuffer( $GL_ARRAY_BUFFER, $iVertexBufferObject )
	glEnableVertexAttribArray(0)
	glEnableVertexAttribArray(1)
	glVertexAttribPointer( 0, 3, $GL_FLOAT, $GL_FALSE, 0, 0 )
	glVertexAttribPointer( 1, 4, $GL_FLOAT, $GL_FALSE, 0, $ColorDataOffset )
	glBindBuffer( $GL_ELEMENT_ARRAY_BUFFER, $iIndexBufferObject )

	glBindVertexArray(0)

	Return 1

EndFunc


Func DisplayFunction()

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )

	glUseProgram( $iProgram )

	glBindVertexArray( $iVao )

	HandleKeystrokes()

	If $bDemo Then _
		RobotDemo()

	For $i = 0 To $iRobots - 1
		RobotDraw( $i )
	Next

	glBindVertexArray(0)
	glUseProgram(0)

	SwapBuffers( $hDC )

EndFunc


Func RobotDraw( $iRobot )

	; --- Base ---

	; Main base
	Local $mMainBase = $IDENTITY_MATRIX
	$mMainBase = glmTranslate( $mMainBase, $aRobots[$iRobot][1] )
	$mMainBase = glmRotate( $mMainBase, $aRobots[$iRobot][2], $axisY )

	; Draw left base
	Local $mBaseLeft = $mMainBase, $scale[3] = [ 1.0, 1.0, $aRobots[$iRobot][5] ]
	$mBaseLeft = glmTranslate( $mBaseLeft, $aRobots[$iRobot][3] )
	$mBaseLeft = glmScale( $mBaseLeft, $scale )
	glUniformMatrix4fv( $iModelToCameraMatrixUnif, 1, $GL_FALSE, $mBaseLeft )
	glDrawElements( $GL_TRIANGLES, UBound($aIndexData), $GL_UNSIGNED_SHORT, 0 )

	; Draw right base
	Local $mBaseRight = $mMainBase
	$mBaseRight = glmTranslate( $mBaseRight, $aRobots[$iRobot][4] )
	$mBaseRight = glmScale( $mBaseRight, $scale )
	glUniformMatrix4fv( $iModelToCameraMatrixUnif, 1, $GL_FALSE, $mBaseRight )
	glDrawElements( $GL_TRIANGLES, UBound($aIndexData), $GL_UNSIGNED_SHORT, 0 )

	; --- Arms ---

	; Main arm
	Local $mMainArm = $mMainBase
	$mMainArm = glmRotate( $mMainArm, $aRobots[$iRobot][6], $axisX )

	; Draw upper arm
	Local $mUpperArm = $mMainArm, $trans[3] = [ 0.0, 0.0, ( $aRobots[$iRobot][7] / 2.0 ) - 1.0 ], $scale[3] = [ 1.0, 1.0, $aRobots[$iRobot][7] / 2.0 ]
	$mUpperArm = glmTranslate( $mUpperArm, $trans )
	$mUpperArm = glmScale( $mUpperArm, $scale )
	glUniformMatrix4fv( $iModelToCameraMatrixUnif, 1, $GL_FALSE, $mUpperArm )
	glDrawElements( $GL_TRIANGLES, UBound($aIndexData), $GL_UNSIGNED_SHORT, 0 )

	$mMainArm = glmTranslate( $mMainArm, $aRobots[$iRobot][8] )
	$mMainArm = glmRotate( $mMainArm, $aRobots[$iRobot][9], $axisX )

	; Draw lower arm
	Local $mLowerArm = $mMainArm, $trans[3] = [ 0.0, 0.0, $aRobots[$iRobot][11] / 2.0 ], $scale[3] = [ $aRobots[$iRobot][12] / 2.0, $aRobots[$iRobot][12] / 2.0, $aRobots[$iRobot][11] / 2.0 ]
	$mLowerArm = glmTranslate( $mLowerArm, $trans )
	$mLowerArm = glmScale( $mLowerArm, $scale )
	glUniformMatrix4fv( $iModelToCameraMatrixUnif, 1, $GL_FALSE, $mLowerArm )
	glDrawElements( $GL_TRIANGLES, UBound($aIndexData), $GL_UNSIGNED_SHORT, 0 )

	; --- Wrist ---

	; Main wrist
	Local $mMainWrist = $mMainArm
	$mMainWrist = glmTranslate( $mMainWrist, $aRobots[$iRobot][13] )
	$mMainWrist = glmRotate( $mMainWrist, $aRobots[$iRobot][14], $axisZ )
	$mMainWrist = glmRotate( $mMainWrist, $aRobots[$iRobot][15], $axisX )

	; Draw wrist
	Local $mWrist = $mMainWrist, $scale[3] = [ $aRobots[$iRobot][17] / 2.0, $aRobots[$iRobot][17] / 2.0, $aRobots[$iRobot][16] / 2.0 ]
	$mWrist = glmScale( $mWrist, $scale )
	glUniformMatrix4fv( $iModelToCameraMatrixUnif, 1, $GL_FALSE, $mWrist )
	glDrawElements( $GL_TRIANGLES, UBound($aIndexData), $GL_UNSIGNED_SHORT, 0 )

	; --- Fingers ---

	Local $transFinger[3] = [ 0.0, 0.0, $aRobots[$iRobot][21] / 2.0 ], $scaleFinger[3] = [ $aRobots[$iRobot][22] / 2.0, $aRobots[$iRobot][22] / 2.0, $aRobots[$iRobot][21] / 2.0 ]

	; Main left finger
	Local $mMainLeftFinger = $mMainWrist
	$mMainLeftFinger = glmTranslate( $mMainLeftFinger, $aRobots[$iRobot][18] )
	$mMainLeftFinger = glmRotate( $mMainLeftFinger, $aRobots[$iRobot][20], $axisY )

	; Draw left finger
	Local $mLeftFinger = $mMainLeftFinger
	$mLeftFinger = glmTranslate( $mLeftFinger, $transFinger )
	$mLeftFinger = glmScale( $mLeftFinger, $scaleFinger )
	glUniformMatrix4fv( $iModelToCameraMatrixUnif, 1, $GL_FALSE, $mLeftFinger )
	glDrawElements( $GL_TRIANGLES, UBound($aIndexData), $GL_UNSIGNED_SHORT, 0 )

	; Main left lower finger
	Local $mMainLeftLowerFinger = $mMainLeftFinger, $trans[3] = [ 0.0, 0.0, $aRobots[$iRobot][21] ]
	$mMainLeftLowerFinger = glmTranslate( $mMainLeftLowerFinger, $trans )
	$mMainLeftLowerFinger = glmRotate( $mMainLeftLowerFinger, -$aRobots[$iRobot][23], $axisY )

	; Draw left lower finger
	Local $mLeftLowerFinger = $mMainLeftLowerFinger
	$mLeftLowerFinger = glmTranslate( $mLeftLowerFinger, $transFinger )
	$mLeftLowerFinger = glmScale( $mLeftLowerFinger, $scaleFinger )
	glUniformMatrix4fv( $iModelToCameraMatrixUnif, 1, $GL_FALSE, $mLeftLowerFinger )
	glDrawElements( $GL_TRIANGLES, UBound($aIndexData), $GL_UNSIGNED_SHORT, 0 )

	; Main right finger
	Local $mMainRightFinger = $mMainWrist
	$mMainRightFinger = glmTranslate( $mMainRightFinger, $aRobots[$iRobot][19] )
	$mMainRightFinger = glmRotate( $mMainRightFinger, -$aRobots[$iRobot][20], $axisY )

	; Draw right finger
	Local $mRightFinger = $mMainRightFinger
	$mRightFinger = glmTranslate( $mRightFinger, $transFinger )
	$mRightFinger = glmScale( $mRightFinger, $scaleFinger )
	glUniformMatrix4fv( $iModelToCameraMatrixUnif, 1, $GL_FALSE, $mRightFinger )
	glDrawElements( $GL_TRIANGLES, UBound($aIndexData), $GL_UNSIGNED_SHORT, 0 )

	; Main right lower finger
	Local $mMainRightLowerFinger = $mMainRightFinger
	$mMainRightLowerFinger = glmTranslate( $mMainRightLowerFinger, $trans )
	$mMainRightLowerFinger = glmRotate( $mMainRightLowerFinger, $aRobots[$iRobot][23], $axisY )

	; Draw right lower finger
	Local $mRightLowerFinger = $mMainRightLowerFinger
	$mRightLowerFinger = glmTranslate( $mRightLowerFinger, $transFinger )
	$mRightLowerFinger = glmScale( $mRightLowerFinger, $scaleFinger )
	glUniformMatrix4fv( $iModelToCameraMatrixUnif, 1, $GL_FALSE, $mRightLowerFinger )
	glDrawElements( $GL_TRIANGLES, UBound($aIndexData), $GL_UNSIGNED_SHORT, 0 )

EndFunc


Func WM_ACTIVATE( $hWnd, $Msg, $wParam, $lParam )
	Switch $hWnd
		Case $hGui
			$bFocus = BitAND( $wParam, 0xFFFF )
	EndSwitch
	Return $GUI_RUNDEFMSG
EndFunc


Func WM_SIZING( $hWnd, $iMsg, $iwParam, $ilParam )
	#forceref $iMsg, $iwParam
	Switch $hWnd
		Case $hGui
			Local $tRECT = DllStructCreate( $tagRECT, $ilParam )
			Local $y = DllStructGetData( $tRECT, "Top" )
			Local $iHeight = DllStructGetData( $tRECT, "Bottom" ) - $y - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight
			WinMove( $hStatusbar, "", 2, $iHeight + $iHeightXP )
	EndSwitch
	Return $GUI_RUNDEFMSG
EndFunc


Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight * $iStatusBarRows
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	WinMove( $hStatusbar, "", 2, $iHeight + $iHeightXP + $iStatusBarHeight * ( $iStatusBarRows - 1 ) )
	$bResize = True
EndFunc


Func GUI_EVENT_CLOSE()
	$bClose = True
EndFunc


Func CloseProgram()
	glDeleteProgram( $iProgram )
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
