#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

; Parthenon
Global $fParthenonWidth = 14.0
Global $fParthenonLength = 20.0
Global $fParthenonColumnHeight = 5.0
Global $fParthenonBaseHeight = 1.0
Global $fParthenonTopHeight = 2.0
Global $fColumnBaseHeight = 0.25

Global $vCamTarget[3] = [ 0.0, 0.4, 0.0 ]
Global $vSphereCamRelPos[3] = [ 67.5, -46.0, 150.0 ]

Global $fAngleX = 0.0, $fAngleY = 0.0, $fAngleZ = 0.0

Global $fSmallAngelIncrement = 1.0

Global $mIdenity

; --- OpenGL globals ---

; Programs
Global $aProgramData[5]
; $aProgramData[0] = $iProgram
; $aProgramData[1] = $iModelToWorldMatrixUnif
; $aProgramData[2] = $iWorldToCameraMatrixUnif
; $aProgramData[3] = $iCameraToClipMatrixUnif
; $aProgramData[4] = $iBaseColorUnif
Global $aUniformColor     = $aProgramData
Global $aObjectColor      = $aProgramData
Global $aUniformColorTint = $aProgramData

; Meshes
Global $aMesh[3]
; $aMesh[0] = $iVao
; $aMesh[1] = $iVbo
; $aMesh[2] = $iIbo
Global $aMeshConeTint     = $aMesh
Global $aMeshCylinderTint = $aMesh
Global $aMeshCubeTint     = $aMesh
Global $aMeshCubeColor    = $aMesh
Global $aMeshPlane        = $aMesh

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 500, $iHeight = 500

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30
Global Const $iStatusBarHeight = 28
Global Const $iStatusBarRows   =  2

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight + $iStatusBarHeight * $iStatusBarRows

Global $hGui, $idInput, $bResize = True, $bClose = False, $bFocus = True

; --- Project includes ---

#include "utilities\includes.au3"
#include "utilities\statusbar.au3"
#include "utilities\keystrokes.au3"
#include "..\utilities\mathematics.au3"
#include "models\models.au3"

; --- Program code ---

MainFunc()

While 1
	If $bMode <> $bNewMode Then
		$bMode = $bNewMode
	EndIf
	If $bResize Then
		Local $mPersMatrix = glmPerspective( 45, $iWidth / $iHeight, 1.0, 1000.0 )

		glUseProgram( $aUniformColor[0] )
		glUniformMatrix4fv( $aUniformColor[3], 1, $GL_FALSE, $mPersMatrix )
		glUseProgram( $aObjectColor[0] )
		glUniformMatrix4fv( $aObjectColor[3], 1, $GL_FALSE, $mPersMatrix )
		glUseProgram( $aUniformColorTint[0] )
		glUniformMatrix4fv( $aUniformColorTint[3], 1, $GL_FALSE, $mPersMatrix )
		glUseProgram(0)

		glViewport( 0, 0, $iWidth, $iHeight )
		$bResize = False
	EndIf
	If $bClose Then
		CloseProgram()
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then CloseProgram()
	If Not InitOpenGLwindow() Then CloseProgram()
	If Not CheckOpenGLversion( 3, 1 ) Then CloseProgram()
	If Not CreateResources() Then CloseProgram()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 100, $WS_OVERLAPPEDWINDOW )

	; Input control to receive key presses (to prevent sound alarm)
	$idInput = GUICtrlCreateInput( "", 0, 0, 1, 1, 2048 ) ; 2048 = $ES_READONLY
	GUICtrlSetBkColor( $idInput, 0x000000 )
	GUICtrlSetState( $idInput, $GUI_FOCUS )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	GUIRegisterMsg( $WM_ACTIVATE, "WM_ACTIVATE" )
	GUIRegisterMsg( $WM_SIZING,   "WM_SIZING" )

	CreateStatusbar( "Demonstrates the use of a mobile camera in a world-space scene." )

	Return 1

EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 0.0, 0.0, 0.0, 0.0 )
	glClearDepth( 1.0 )

	glEnable( $GL_CULL_FACE )
	glCullFace( $GL_BACK )
	glFrontFace( $GL_CW )

	glEnable( $GL_DEPTH_TEST )
	glDepthMask( $GL_TRUE )
	glDepthFunc( $GL_LEQUAL )
	glDepthRange( 0.0, 1.0 )
	glEnable( $GL_DEPTH_CLAMP )

	GUISwitch( $hGui )

	Return 1

EndFunc


Func CreateResources()

	If Not InitializeProgram() Then Return 0

	If Not InitializeVertexBuffer() Then Return 0

	If Not InitializeVertexArrayObject() Then Return 0

	$mIdenity = $IDENTITY_MATRIX

	GUISetState()

	Return 1

EndFunc


Func InitializeProgram()

	$aUniformColor = LoadProgram( "PosOnlyWorldTransform.vert", "ColorUniform.frag" )
	If Not IsArray( $aUniformColor ) Then Return 0

	$aObjectColor = LoadProgram( "PosColorWorldTransform.vert", "ColorPassthrough.frag" )
	If Not IsArray( $aObjectColor ) Then Return 0

	$aUniformColorTint = LoadProgram( "PosColorWorldTransform.vert", "ColorMultUniform.frag" )
	If Not IsArray( $aUniformColorTint ) Then Return 0

	Return 1

EndFunc


Func LoadProgram( $sVertexShader, $sFragmentShader )

	Local $aProgram = $aProgramData
	; $aProgramData[0] = $iProgram
	; $aProgramData[1] = $iModelToWorldMatrixUnif
	; $aProgramData[2] = $iWorldToCameraMatrixUnif
	; $aProgramData[3] = $iCameraToClipMatrixUnif
	; $aProgramData[4] = $iBaseColorUnif

	Local $iVertexShader, $iFragmentShader
	If Not LoadGLSLshader( $iVertexShader, "shaders\" & $sVertexShader, $GL_VERTEX_SHADER ) Then Return 0
	If Not LoadGLSLshader( $iFragmentShader, "shaders\" & $sFragmentShader, $GL_FRAGMENT_SHADER ) Then Return 0

	$aProgram[0] = glCreateProgram()
	glAttachShader( $aProgram[0], $iVertexShader )
	glAttachShader( $aProgram[0], $iFragmentShader )
	glLinkProgram( $aProgram[0] )

	$aProgram[1] = glGetUniformLocation( $aProgram[0], "modelToWorldMatrix" )
	$aProgram[2] = glGetUniformLocation( $aProgram[0], "worldToCameraMatrix" )
	$aProgram[3] = glGetUniformLocation( $aProgram[0], "cameraToClipMatrix" )
	$aProgram[4] = glGetUniformLocation( $aProgram[0], "baseColor" )

	Return $aProgram

EndFunc


Func InitializeVertexBuffer()

	InitVertexBuffer( $aMeshConeTint,     $aConeTintVerts,     $aConeTintIndices )
	InitVertexBuffer( $aMeshCylinderTint, $aCylinderTintVerts, $aCylinderTintIndices )
	InitVertexBuffer( $aMeshCubeTint,     $aCubeTintVerts,     $aCubeTintIndices )
	InitVertexBuffer( $aMeshCubeColor,    $aCubeColorVerts,    $aCubeColorIndices )
	InitVertexBuffer( $aMeshPlane,        $aPlaneData,         $aPlaneIndices )

	Return 1

EndFunc


Func InitVertexBuffer( ByRef $aMesh, ByRef $aData, ByRef $aIndices )

	glGenBuffers( 1, $aMesh[1] )
	glBindBuffer( $GL_ARRAY_BUFFER, $aMesh[1] )
	glBufferData( $GL_ARRAY_BUFFER, $GL_FLOAT, 4*UBound($aData), $aData, $GL_STATIC_DRAW )
	glBindBuffer( $GL_ARRAY_BUFFER, 0 )

	glGenBuffers( 1, $aMesh[2] )
	glBindBuffer( $GL_ELEMENT_ARRAY_BUFFER, $aMesh[2] )
	glBufferData( $GL_ELEMENT_ARRAY_BUFFER, $GL_SHORT, 2*UBound($aIndices), $aIndices, $GL_STATIC_DRAW )
	glBindBuffer( $GL_ELEMENT_ARRAY_BUFFER, 0 )

EndFunc


Func InitializeVertexArrayObject()

	InitVertexArrayObject( $aMeshConeTint,     $aConeTintVerts,     $iConeTintVerts )
	InitVertexArrayObject( $aMeshCylinderTint, $aCylinderTintVerts, $iCylinderTintVerts )
	InitVertexArrayObject( $aMeshCubeTint,     $aCubeTintVerts,     $iCubeTintVerts )
	InitVertexArrayObject( $aMeshCubeColor,    $aCubeColorVerts,    $iCubeColorVerts )
	InitVertexArrayObject( $aMeshPlane,        $aPlaneData,         $iPlanePoints )

	Return 1

EndFunc


Func InitVertexArrayObject( ByRef $aMesh, ByRef $aData, $iVertices )

	Local $iSize = UBound( $aData )

	glGenVertexArrays( 1, $aMesh[0] )
	glBindVertexArray( $aMesh[0] )

	glBindBuffer( $GL_ARRAY_BUFFER, $aMesh[1] )
	glEnableVertexAttribArray(0)
	glVertexAttribPointer( 0, 3, $GL_FLOAT, $GL_FALSE, 0, 0 )
	Select
		Case $iSize = $iVertices * 3
		Case $iSize = $iVertices * 3 + $iVertices * 4
			glEnableVertexAttribArray(1)
			glVertexAttribPointer( 1, 4, $GL_FLOAT, $GL_FALSE, 0, $iVertices*3*4 )
		Case Else
			; Error
	EndSelect
	glBindBuffer( $GL_ELEMENT_ARRAY_BUFFER, $aMesh[2] )

	glBindVertexArray(0)

EndFunc


Func DisplayFunction()

	HandleKeystrokes()

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )

	; Camera matrix
	Local $vCamPos = ResolveCamPosition()
	Local $mCamMatrix = glmLookAt( $vCamPos, $vCamTarget, $axisY )
	; Pass info to shaders
	glUseProgram( $aUniformColor[0] )
	glUniformMatrix4fv( $aUniformColor[2], 1, $GL_FALSE, $mCamMatrix )
	glUseProgram( $aObjectColor[0] )
	glUniformMatrix4fv( $aObjectColor[2], 1, $GL_FALSE, $mCamMatrix )
	glUseProgram( $aUniformColorTint[0] )
	glUniformMatrix4fv( $aUniformColorTint[2], 1, $GL_FALSE, $mCamMatrix )
	glUseProgram(0)

	; Model matrix
	Local $mModelMatrix = $mIdenity

	; --- Draw ground plane ---

	Local $vScale[3] = [ 100.0, 1.0, 100.0 ]
	$mModelMatrix = glmScale( $mModelMatrix, $vScale )
	glUseProgram( $aUniformColor[0] )
	glBindVertexArray( $aMeshPlane[0] )
	glUniform4f( $aUniformColor[4], 0.302, 0.416, 0.0589, 1.0 )
	glUniformMatrix4fv( $aUniformColor[1], 1, $GL_FALSE, $mModelMatrix )
	glDrawElements( $GL_TRIANGLES, $iPlaneIndices, $GL_UNSIGNED_SHORT, 0 )
	glUseProgram(0)

	; --- Draw trees ---

	Local $mTreeMatrix, $mTrunkMatrix, $mTopMatrix
	Local $vTrans1[3] = [ 0.0, 0.0, 0.0 ]
	Local $vScale2[3] = [ 1.0, 0.0, 1.0 ]
	Local $vTrans2[3] = [ 0.0, 0.5, 0.0 ]
	Local $vTrans3[3] = [ 0.0, 0.0, 0.0 ]
	Local $vScale3[3] = [ 3.0, 0.0, 3.0 ]
	For $i = 0 To $iTrees - 1
		$vTrans1[0] = $aTrees[$i][0]
		$vTrans1[2] = $aTrees[$i][1]
		$mTreeMatrix = glmTranslate( $mIdenity, $vTrans1 )

		; Draw trunk
		$vScale2[1] = $aTrees[$i][2]
		$mTrunkMatrix = glmScale( $mTreeMatrix, $vScale2 )
		$mTrunkMatrix = glmTranslate( $mTrunkMatrix, $vTrans2 )
		glUseProgram( $aUniformColorTint[0] )
		glBindVertexArray( $aMeshCylinderTint[0] )
		glUniform4f( $aUniformColorTint[4], 0.694, 0.4, 0.106, 1.0 )
		glUniformMatrix4fv( $aUniformColorTint[1], 1, $GL_FALSE, $mTrunkMatrix )
		For $j = 0 To 2
			glDrawElements( $aCylinderTintIndInfo[$j][0], $aCylinderTintIndInfo[$j][1], $GL_UNSIGNED_SHORT, $aCylinderTintIndInfo[$j][2] )
		Next
		glUseProgram(0)

		; Draw treetop
		$vTrans3[1] = $aTrees[$i][2]
		$vScale3[1] = $aTrees[$i][3]
		$mTopMatrix = glmTranslate( $mTreeMatrix, $vTrans3 )
		$mTopMatrix = glmScale( $mTopMatrix, $vScale3 )
		glUseProgram( $aUniformColorTint[0] )
		glBindVertexArray( $aMeshConeTint[0] )
		glUniform4f( $aUniformColorTint[4], 0.0, 1.0, 0.0, 1.0 )
		glUniformMatrix4fv( $aUniformColorTint[1], 1, $GL_FALSE, $mTopMatrix )
		For $j = 0 To 1
			glDrawElements( $aConeTintIndInfo[$j][0], $aConeTintIndInfo[$j][1], $GL_UNSIGNED_SHORT, $aConeTintIndInfo[$j][2] )
		Next
		glUseProgram(0)
	Next

	; --- Draw Parthenon ---

	Local $vTrans4[3] = [ 20.0, 0.0, -10.0 ]
	Local $mPartMatrix = glmTranslate( $mIdenity, $vTrans4 )
	DrawParthenon( $mPartMatrix )

	; --- Draw lookat point ---

	If $bMode Then
		glDisable( $GL_DEPTH_TEST )

		Local $vCameraAimVec = glmSubtractVec3( $vCamTarget, $vCamPos )
		Local $vTrans[3] = [ 0.0, 0.0, -glmLengthVec3Num( $vCameraAimVec ) ]
		Local $vScale[3] = [ 1.0, 1.0, 1.0 ]
		Local $mMatrix = glmTranslate( $mIdenity, $vTrans )
		$mMatrix = glmScale( $mMatrix, $vScale )

		glUseProgram( $aObjectColor[0] )
		glBindVertexArray( $aMeshCubeColor[0] )
		glUniformMatrix4fv( $aObjectColor[1], 1, $GL_FALSE, $mMatrix )
		glUniformMatrix4fv( $aObjectColor[2], 1, $GL_FALSE, $mIdenity )
		glDrawElements( $aCubeColorIndInfo[0][0], $aCubeColorIndInfo[0][1], $GL_UNSIGNED_SHORT, $aCubeColorIndInfo[0][2] )
		glUseProgram(0)

		glEnable( $GL_DEPTH_TEST )
	EndIf

	SwapBuffers( $hDC )

EndFunc


Func DrawParthenon( ByRef $mPartMatrix )

	Local $mMatrix

	; Draw base
	Local $vScale[3] = [ $fParthenonWidth, $fParthenonBaseHeight, $fParthenonLength ]
	Local $vTrans[3] = [ 0.0, 0.5, 0.0 ]
	$mMatrix = glmScale( $mPartMatrix, $vScale )
	$mMatrix = glmTranslate( $mMatrix, $vTrans )
	glUseProgram( $aUniformColorTint[0] )
	glBindVertexArray( $aMeshCubeTint[0] )
	glUniform4f( $aUniformColorTint[4], 0.9, 0.9, 0.9, 0.9 )
	glUniformMatrix4fv( $aUniformColorTint[1], 1, $GL_FALSE, $mMatrix )
	glDrawElements( $aCubeTintIndInfo[0][0], $aCubeTintIndInfo[0][1], $GL_UNSIGNED_SHORT, $aCubeTintIndInfo[0][2] )
	glUseProgram(0)

	; Draw top
	Local $vTrans1[3] = [ 0.0, $fParthenonColumnHeight + $fParthenonBaseHeight, 0.0 ]
	Local $vScale1[3] = [ $fParthenonWidth, $fParthenonTopHeight, $fParthenonLength ]
	Local $vTrans2[3] = [ 0.0, 0.5, 0.0 ]
	$mMatrix = glmTranslate( $mPartMatrix, $vTrans1 )
	$mMatrix = glmScale( $mMatrix, $vScale1 )
	$mMatrix = glmTranslate( $mMatrix, $vTrans2 )
	glUseProgram( $aUniformColorTint[0] )
	glBindVertexArray( $aMeshCubeTint[0] )
	glUniform4f( $aUniformColorTint[4], 0.9, 0.9, 0.9, 0.9 )
	glUniformMatrix4fv( $aUniformColorTint[1], 1, $GL_FALSE, $mMatrix )
	glDrawElements( $aCubeTintIndInfo[0][0], $aCubeTintIndInfo[0][1], $GL_UNSIGNED_SHORT, $aCubeTintIndInfo[0][2] )
	glUseProgram(0)

	; Draw columns front/back
	$vTrans[1] = $fParthenonBaseHeight
	Local $fFrontZVal = ( $fParthenonLength / 2.0 ) - 1.0
	For $iColumnNum = 0 To Int( $fParthenonWidth / 2.0 ) - 1
		$vTrans[0] = ( 2.0 * $iColumnNum ) - ( $fParthenonWidth / 2.0 ) + 1.0
		$vTrans[2] = $fFrontZVal
		$mMatrix = glmTranslate( $mPartMatrix, $vTrans )
		DrawColumn( $mMatrix, $fParthenonColumnHeight )

		$vTrans[2] = -$fFrontZVal
		$mMatrix = glmTranslate( $mPartMatrix, $vTrans )
		DrawColumn( $mMatrix, $fParthenonColumnHeight )
	Next

	; Draw columns left/right
	; Don't draw the first or last columns, since they've been drawn already
	$vTrans[1] = $fParthenonBaseHeight
	Local $fRightXVal = ( $fParthenonWidth  / 2.0 ) - 1.0
	For $iColumnNum = 1 To Int( ( $fParthenonLength - 2.0 ) / 2.0 ) - 1
		$vTrans[0] = $fRightXVal
		$vTrans[2] = ( 2.0 * $iColumnNum ) - ( $fParthenonLength / 2.0 ) + 1.0
		$mMatrix = glmTranslate( $mPartMatrix, $vTrans )
		DrawColumn( $mMatrix, $fParthenonColumnHeight )

		$vTrans[0] = -$fRightXVal
		$mMatrix = glmTranslate( $mPartMatrix, $vTrans )
		DrawColumn( $mMatrix, $fParthenonColumnHeight )
	Next

	; Draw interior
	Local $vTrans1[3] = [ 0.0, 1.0, 0.0 ]
	Local $vScale1[3] = [ $fParthenonWidth - 6.0, $fParthenonColumnHeight, $fParthenonLength - 6.0 ]
	Local $vTrans2[3] = [ 0.0, 0.5, 0.0 ]
	$mMatrix = glmTranslate( $mPartMatrix, $vTrans1 )
	$mMatrix = glmScale( $mMatrix, $vScale1 )
	$mMatrix = glmTranslate( $mMatrix, $vTrans2 )
	glUseProgram( $aObjectColor[0] )
	glBindVertexArray( $aMeshCubeColor[0] )
	glUniformMatrix4fv( $aObjectColor[1], 1, $GL_FALSE, $mMatrix )
	glDrawElements( $aCubeColorIndInfo[0][0], $aCubeColorIndInfo[0][1], $GL_UNSIGNED_SHORT, $aCubeColorIndInfo[0][2] )
	glUseProgram(0)

	; Draw headpiece
	Local $vTrans[3] = [ 0.0, $fParthenonColumnHeight + $fParthenonBaseHeight + $fParthenonTopHeight / 2.0, $fParthenonLength / 2.0 ]
	$mMatrix = glmTranslate( $mPartMatrix, $vTrans )
	$mMatrix = glmRotate( $mMatrix, -135, $axisX )
	$mMatrix = glmRotate( $mMatrix, 45, $axisY )
	glUseProgram( $aObjectColor[0] )
	glBindVertexArray( $aMeshCubeColor[0] )
	glUniformMatrix4fv( $aObjectColor[1], 1, $GL_FALSE, $mMatrix )
	glDrawElements( $aCubeColorIndInfo[0][0], $aCubeColorIndInfo[0][1], $GL_UNSIGNED_SHORT, $aCubeColorIndInfo[0][2] )
	glUseProgram(0)

EndFunc


; Columns are 1x1 in the X/Z-directions,
; and $fHeight units in the Y-direction.
Func DrawColumn( ByRef $mMatrix, $fHeight )

	; Draw bottom of the column
	Local $vScale[3] = [ 1.0, $fColumnBaseHeight, 1.0 ]
	Local $vTrans[3] = [ 0.0, 0.5, 0.0 ]
	Local $m = glmScale( $mMatrix, $vScale )
	$m = glmTranslate( $m, $vTrans )
	glUseProgram( $aUniformColorTint[0] )
	glBindVertexArray( $aMeshCubeTint[0] )
	glUniform4f( $aUniformColorTint[4], 1.0, 1.0, 1.0, 1.0 )
	glUniformMatrix4fv( $aUniformColorTint[1], 1, $GL_FALSE, $m )
	glDrawElements( $aCubeTintIndInfo[0][0], $aCubeTintIndInfo[0][1], $GL_UNSIGNED_SHORT, $aCubeTintIndInfo[0][2] )
	glUseProgram(0)

	; Draw top of the column
	Local $vTrans1[3] = [ 0.0, $fHeight - $fColumnBaseHeight, 0.0 ]
	Local $vScale1[3] = [ 1.0, $fColumnBaseHeight, 1.0 ]
	Local $vTrans2[3] = [ 0.0, 0.5, 0.0 ]
	$m = glmTranslate( $mMatrix, $vTrans1 )
	$m = glmScale( $m, $vScale1 )
	$m = glmTranslate( $m, $vTrans2 )
	glUseProgram( $aUniformColorTint[0] )
	glBindVertexArray( $aMeshCubeTint[0] )
	glUniform4f( $aUniformColorTint[4], 1.0, 1.0, 1.0, 1.0 )
	glUniformMatrix4fv( $aUniformColorTint[1], 1, $GL_FALSE, $m )
	glDrawElements( $aCubeTintIndInfo[0][0], $aCubeTintIndInfo[0][1], $GL_UNSIGNED_SHORT, $aCubeTintIndInfo[0][2] )
	glUseProgram(0)

	; Draw the main column
	Local $vTrans1[3] = [ 0.0, $fColumnBaseHeight, 0.0 ]
	Local $vScale1[3] = [ 0.8, $fHeight - $fColumnBaseHeight * 2.0, 0.8 ]
	Local $vTrans2[3] = [ 0.0, 0.5, 0.0 ]
	$m = glmTranslate( $mMatrix, $vTrans1 )
	$m = glmScale( $m, $vScale1 )
	$m = glmTranslate( $m, $vTrans2 )
	glUseProgram( $aUniformColorTint[0] )
	glBindVertexArray( $aMeshCylinderTint[0] )
	glUniform4f( $aUniformColorTint[4], 0.9, 0.9, 0.9, 0.9 )
	glUniformMatrix4fv( $aUniformColorTint[1], 1, $GL_FALSE, $m )
	For $j = 0 To 2
		glDrawElements( $aCylinderTintIndInfo[$j][0], $aCylinderTintIndInfo[$j][1], $GL_UNSIGNED_SHORT, $aCylinderTintIndInfo[$j][2] )
	Next
	glUseProgram(0)

EndFunc


Func WM_ACTIVATE( $hWnd, $Msg, $wParam, $lParam )
	Switch $hWnd
		Case $hGui
			$bFocus = BitAND( $wParam, 0xFFFF )
	EndSwitch
	Return $GUI_RUNDEFMSG
EndFunc


Func WM_SIZING( $hWnd, $iMsg, $iwParam, $ilParam )
	#forceref $iMsg, $iwParam
	Switch $hWnd
		Case $hGui
			Local $tRECT = DllStructCreate( $tagRECT, $ilParam )
			Local $y = DllStructGetData( $tRECT, "Top" )
			Local $iHeight = DllStructGetData( $tRECT, "Bottom" ) - $y - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight
			WinMove( $hStatusbar, "", 2, $iHeight + $iHeightXP )
	EndSwitch
	Return $GUI_RUNDEFMSG
EndFunc


Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight * $iStatusBarRows
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	WinMove( $hStatusbar, "", 2, $iHeight + $iHeightXP + $iStatusBarHeight * ( $iStatusBarRows - 1 ) )
	$bResize = True
EndFunc


Func GUI_EVENT_CLOSE()
	$bClose = True
EndFunc


Func CloseProgram()
	glDeleteProgram( $aObjectColor[0] )
	glDeleteProgram( $aUniformColor[0] )
	glDeleteProgram( $aUniformColorTint[0] )
	;DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
