        ; FASM assembler code

        use64

glmScale:
				; $result[16] = glmScale( $mat4[16], $vec3[3] )
				; Builds a scale 4 * 4 matrix created from 3 scalars
        ;
        ; Parameters: floats
				; in  rax: $mat4[0]
				; in  rcx: $vec3[0]
				; out rdx: $result[0]
        ;
        ; Registers: bl

				; $result[0]  = $mat4[0]  * $vec3[0]
				; $result[1]  = $mat4[1]  * $vec3[0]
				; $result[2]  = $mat4[2]  * $vec3[0]
				; $result[3]  = $mat4[3]  * $vec3[0]

				; $result[4]  = $mat4[4]  * $vec3[1]
				; $result[5]  = $mat4[5]  * $vec3[1]
				; $result[6]  = $mat4[6]  * $vec3[1]
				; $result[7]  = $mat4[7]  * $vec3[1]

				; $result[8]  = $mat4[8]  * $vec3[2]
				; $result[9]  = $mat4[9]  * $vec3[2]
				; $result[10] = $mat4[10] * $vec3[2]
				; $result[11] = $mat4[11] * $vec3[2]

				mov        bl,                 3                   ; Counter

@@:     finit                                              ; Init FPU

        fld        dword [rax + 00]                        ; $mat4[0]
        fmul       dword [rcx + 00]                        ; * $vec3[0]
				fst        dword [rdx + 00]                        ; st0 -> $result[0]

        fld        dword [rax + 04]                        ; $mat4[1]
        fmul       dword [rcx + 00]                        ; * $vec3[0]
				fst        dword [rdx + 04]                        ; st0 -> $result[1]

        fld        dword [rax + 08]                        ; $mat4[2]
        fmul       dword [rcx + 00]                        ; * $vec3[0]
				fst        dword [rdx + 08]                        ; st0 -> $result[2]

        fld        dword [rax + 12]                        ; $mat4[3]
        fmul       dword [rcx + 00]                        ; * $vec3[0]
				fst        dword [rdx + 12]                        ; st0 -> $result[3]

				add        rax,                16                  ; Next row
				add        rcx,                04                  ; Next scalar
				add        rdx,                16                  ; Next row

				dec        bl                                      ; Counter -= 1
        jne        @b                                      ; Next

        ; $result[12] = $mat4[12]
        ; $result[13] = $mat4[13]
        ; $result[14] = $mat4[14]
        ; $result[15] = $mat4[15]

				mov        ecx,                dword [rax + 00]
				mov        dword [rdx + 00],   ecx                 ; $result[12] = $mat4[12]
				mov        ecx,                dword [rax + 04]
				mov        dword [rdx + 04],   ecx                 ; $result[13] = $mat4[13]
				mov        ecx,                dword [rax + 08]
				mov        dword [rdx + 08],   ecx                 ; $result[14] = $mat4[14]
				mov        ecx,                dword [rax + 12]
				mov        dword [rdx + 12],   ecx                 ; $result[15] = $mat4[15]

				ret                                                ; Return


glmTranslate:
				; $result[16] = glmTranslate( $mat4[16], $vec3[3] )
				; Builds a translation 4 * 4 matrix created from a vector of 3 components
        ;
        ; Parameters: floats
				; in  rax: $mat4[0]
				; in  rcx: $vec3[0]
				; out rdx: $result[0]
        ;
        ; Registers: bl, r15

				; $result[0-11] = $mat4[0-11]

				mov        r15,                rcx
				mov        bl,                 3                   ; Counter

@@:     mov        ecx,                dword [rax + 00]
				mov        dword [rdx + 00],   ecx                 ; $result[0] = $mat4[0]
				mov        ecx,                dword [rax + 04]
				mov        dword [rdx + 04],   ecx                 ; $result[1] = $mat4[1]
				mov        ecx,                dword [rax + 08]
				mov        dword [rdx + 08],   ecx                 ; $result[2] = $mat4[2]
				mov        ecx,                dword [rax + 12]
				mov        dword [rdx + 12],   ecx                 ; $result[3] = $mat4[3]

				add        rax,                16                  ; Next row
				add        rdx,                16                  ; Next row

				dec        bl                                      ; Counter -= 1
        jne        @b                                      ; Next
				mov        rcx,                r15

				; $result[12] = $mat4[0] * $vec3[0] + $mat4[4] * $vec3[1] + $mat4[8]  * $vec3[2] + $mat4[12]
				; $result[13] = $mat4[1] * $vec3[0] + $mat4[5] * $vec3[1] + $mat4[9]  * $vec3[2] + $mat4[13]
				; $result[14] = $mat4[2] * $vec3[0] + $mat4[6] * $vec3[1] + $mat4[10] * $vec3[2] + $mat4[14]
				; $result[15] = $mat4[3] * $vec3[0] + $mat4[7] * $vec3[1] + $mat4[11] * $vec3[2] + $mat4[15]

				mov        bl,                 4                   ; Counter
				sub        rax,                48                  ; $mat4[0]

@@:     finit                                              ; Init FPU

        fld        dword [rax + 00]                        ; $mat4[0]
        fmul       dword [rcx + 00]                        ; * $vec3[0]

        fld        dword [rax + 16]                        ; $mat4[4], st0 -> st1
        fmul       dword [rcx + 04]                        ; * $vec3[1]
        fadd       st0,                st1                 ; + $mat4[0] * $vec3[0]

        fld        dword [rax + 32]                        ; $mat4[8], st0 -> st1
        fmul       dword [rcx + 08]                        ; * $vec3[2]
        fadd       st0,                st1                 ; + $mat4[4] * $vec3[1]

        fadd       dword [rax + 48]                        ; + $mat4[12]
        fst        dword [rdx + 00]                        ; st0 -> $result[12]

				add        rax,                4                   ; Next column
				add        rdx,                4                   ; Next result

				dec        bl                                      ; Counter -= 1
        jne        @b                                      ; Next
				ret


glmTranslate1y:
				; $result[16] = glmTranslate( $mat4[16], $vec3[3] )
				; Builds a translation 4 * 4 matrix created from a vector of 3 components
        ;
        ; Parameters: floats
				; in  rax: $mat4[0]
				; in  rcx: $vec3[0]
				; out rdx: $result[0]
        ;
        ; Registers: bl, r15
        ;
				; Prerequisites:
				; x- and z-components of the translation vector must have a value of zero

				; $result[0-11] = $mat4[0-11]

				mov        r15,                rcx
				mov        bl,                 3                   ; Counter

@@:     mov        ecx,                dword [rax + 00]
				mov        dword [rdx + 00],   ecx                 ; $result[0] = $mat4[0]
				mov        ecx,                dword [rax + 04]
				mov        dword [rdx + 04],   ecx                 ; $result[1] = $mat4[1]
				mov        ecx,                dword [rax + 08]
				mov        dword [rdx + 08],   ecx                 ; $result[2] = $mat4[2]
				mov        ecx,                dword [rax + 12]
				mov        dword [rdx + 12],   ecx                 ; $result[3] = $mat4[3]

				add        rax,                16                  ; Next row
				add        rdx,                16                  ; Next row

				dec        bl                                      ; Counter -= 1
        jne        @b                                      ; Next
				mov        rcx,                r15

				;                             0                                            0
				; $result[12] = $mat4[0] * $vec3[0] + $mat4[4] * $vec3[1] + $mat4[8]  * $vec3[2] + $mat4[12]
				; $result[13] = $mat4[1] * $vec3[0] + $mat4[5] * $vec3[1] + $mat4[9]  * $vec3[2] + $mat4[13]
				; $result[14] = $mat4[2] * $vec3[0] + $mat4[6] * $vec3[1] + $mat4[10] * $vec3[2] + $mat4[14]
				; $result[15] = $mat4[3] * $vec3[0] + $mat4[7] * $vec3[1] + $mat4[11] * $vec3[2] + $mat4[15]

				mov        bl,                 4                   ; Counter
				sub        rax,                48                  ; $mat4[0]

@@:     finit                                              ; Init FPU

        fld        dword [rax + 16]                        ; $mat4[4] -> st0
        fmul       dword [rcx + 04]                        ; * $vec3[1]

        fadd       dword [rax + 48]                        ; + $mat4[12]
        fst        dword [rdx + 00]                        ; st0 -> $result[12]

				add        rax,                4                   ; Next column
				add        rdx,                4                   ; Next result

				dec        bl                                      ; Counter -= 1
        jne        @b                                      ; Next
				ret


glmTranslate2:
				; Builds a translation 4 * 4 matrix created from a vector of 3 components
        ;
        ; Parameters: floats
				; in  rax: source matrix
				; in  rcx: translation vector
				; out rdx: destination matrix[12]
        ;
        ; Registers: bl
        ;
				; Prerequisites:
				; Destination matrix[0-11] must contain a copy of source matrix[0-11]

 				;  Result[3]  =  m[0]    *  v[0]    +  m[1]    *  v[1]    +  m[2]     *  v[2]    +  m[3];
				; $result[12] = $mat4[0] * $vec3[0] + $mat4[4] * $vec3[1] + $mat4[8]  * $vec3[2] + $mat4[12]
				; $result[13] = $mat4[1] * $vec3[0] + $mat4[5] * $vec3[1] + $mat4[9]  * $vec3[2] + $mat4[13]
				; $result[14] = $mat4[2] * $vec3[0] + $mat4[6] * $vec3[1] + $mat4[10] * $vec3[2] + $mat4[14]
				; $result[15] = $mat4[3] * $vec3[0] + $mat4[7] * $vec3[1] + $mat4[11] * $vec3[2] + $mat4[15]

				mov        bl,                 4                   ; Counter

@@:     finit                                              ; Init FPU

        fld        dword [rax + 00]                        ; $mat4[0]
        fmul       dword [rcx + 00]                        ; * $vec3[0]

        fld        dword [rax + 16]                        ; $mat4[4], st0 -> st1
        fmul       dword [rcx + 04]                        ; * $vec3[1]
        fadd       st0,                st1                 ; + $mat4[0] * $vec3[0]

        fld        dword [rax + 32]                        ; $mat4[8], st0 -> st1
        fmul       dword [rcx + 08]                        ; * $vec3[2]
        fadd       st0,                st1                 ; + $mat4[4] * $vec3[1]

        fadd       dword [rax + 48]                        ; + $mat4[12]
        fst        dword [rdx + 00]                        ; st0 -> $result[12]

				add        rax,                4                   ; Next column
				add        rdx,                4                   ; Next result

				dec        bl                                      ; Counter -= 1
        jne        @b                                      ; Next
				ret
