#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>
#include <Misc.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

; 4-component quaternion vector
Global $vOrientation[4] = [ 0.7892, 0.3700, 0.02514, 0.4895 ]

Global $SmallAngelIncrement = 3.0

; --- OpenGL globals ---

Global $iProgram, $iVAOship, $iVBOship
Global $iModelToCameraMatrixUnif, $iCameraToClipMatrixUnif, $iBaseColorUnif
Global $fFrustumScale, $mCameraToClipMatrix
Global $iPositionAttrib, $iColorAttrib

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 500, $iHeight = 500

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30
Global Const $iStatusBarHeight = 28
Global Const $iStatusBarRows   =  2

Global $hStatusbar, $idRadio0, $idRadio1
Global $bMode = 0, $bNewMode = $bMode

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight + $iStatusBarHeight * $iStatusBarRows

Global $hGui, $idInput, $bResize = True, $bClose = False, $bFocus = True

; --- Project includes ---

#include "utilities\includes.au3"
#include "..\utilities\mathematics.au3"
#include "models\SpaceShip.au3"

; --- Program code ---

MainFunc()

While 1
	If $bMode <> $bNewMode Then
		$bMode = $bNewMode
	EndIf
	If $bResize Then
		$mCameraToClipMatrix[0] = $fFrustumScale * ( $iHeight / $iWidth )
		$mCameraToClipMatrix[5] = $fFrustumScale

		glUseProgram( $iProgram )
		glUniformMatrix4fv( $iCameraToClipMatrixUnif, 1, $GL_FALSE, $mCameraToClipMatrix )
		glUseProgram(0)

		glViewport( 0, 0, $iWidth, $iHeight )
		$bResize = False
	EndIf
	If $bClose Then
		CloseProgram()
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then CloseProgram()
	If Not InitOpenGLwindow() Then CloseProgram()
	If Not CheckOpenGLversion( 3, 1 ) Then CloseProgram()
	If Not CreateResources() Then CloseProgram()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 100, $WS_OVERLAPPEDWINDOW )

	; Input control to receive key presses (to prevent sound alarm)
	$idInput = GUICtrlCreateInput( "", 0, 0, 1, 1, 2048 ) ; 2048 = $ES_READONLY
	GUICtrlSetBkColor( $idInput, 0x000000 )
	GUICtrlSetState( $idInput, $GUI_FOCUS )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	GUIRegisterMsg( $WM_ACTIVATE, "WM_ACTIVATE" )
	GUIRegisterMsg( $WM_SIZING,   "WM_SIZING" )

	CreateStatusbar()

	Return 1

EndFunc


Func CreateStatusbar()

	; Use a child window as a container for controls in the Statusbar
	$hStatusbar = GUICreate( "", $iGuiWidth - 4 , 24, 2, $iHeight + $iStatusBarHeight + 3 , $WS_CHILD, -1, $hGui )

	; Multiplication
	GUIStartGroup()
	GUICtrlCreateLabel( "Multiplication:", 2, 5, 70, 14 )
	$idRadio0 = GUICtrlCreateRadio( "Right",   75, 3, 50, 18 )
	$idRadio1 = GUICtrlCreateRadio( "Left", 130, 3, 45, 18 )
	GUICtrlSetOnEvent( $idRadio0, "RadioClick" )
	GUICtrlSetOnEvent( $idRadio1, "RadioClick" )
	GUICtrlSetState( $idRadio0, $GUI_CHECKED )

	; Vertical splitterbar
	GUICtrlCreateLabel( "", 180, 2, 3, 20, $SS_ETCHEDFRAME )

	; Label in statusbar
	GUICtrlCreateLabel( "", 200, 5, 250, 14 )
	GUICtrlSetData( -1, "Rotate: Q,E, W,S, A,D. Reset: O." )

	GUISetState( @SW_SHOW, $hStatusbar )
	GUISwitch( $hGui )

	; Label for $hStatusbar to make it look nicer
	GUICtrlCreateLabel( "", 0, $iHeight + 1, $iGuiWidth + 4, $iStatusBarHeight + 1, $SS_ETCHEDFRAME )
	GUICtrlSetResizing( -1, $GUI_DOCKSTATEBAR )
	GUICtrlCreateLabel( "Rotations with yaw, pitch, and roll (YPR) transformations and a 4-component quaternion vector.", _
											3, $iHeight + 1 + 6, $iGuiWidth - 6, $iStatusBarHeight - 10 )
	GUICtrlSetResizing( -1, $GUI_DOCKSTATEBAR )

EndFunc


Func RadioClick()
	Switch @GUI_CTRLID
		Case $idRadio0
			$bNewMode = 0
		Case $idRadio1
			$bNewMode = 1
	EndSwitch
	GUICtrlSetState( $idInput, $GUI_FOCUS )
EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 0.0, 0.0, 0.0, 0.0 )
	glClearDepth( 1.0 )

	glEnable( $GL_CULL_FACE )
	glCullFace( $GL_BACK )
	glFrontFace( $GL_CW )

	glEnable( $GL_DEPTH_TEST )
	glDepthMask( $GL_TRUE )
	glDepthFunc( $GL_LEQUAL )
	glDepthRange( 0.0, 1.0 )

	GUISwitch( $hGui )

	Return 1

EndFunc


Func CreateResources()

	If Not InitializeProgram() Then Return 0

	If Not InitializeVertexBuffer() Then Return 0

	If Not InitializeVertexArrayObject() Then Return 0

	GUISetState()

	Return 1

EndFunc


Func InitializeProgram()

	Local $iVertexShader, $iFragmentShader
	If Not LoadGLSLshader( $iVertexShader, "shaders\PosColorLocalTransform.vert", $GL_VERTEX_SHADER ) Then Return 0
	If Not LoadGLSLshader( $iFragmentShader, "shaders\ColorMultUniform.frag", $GL_FRAGMENT_SHADER ) Then Return 0

	$iProgram = glCreateProgram()
	glAttachShader( $iProgram, $iVertexShader )
	glAttachShader( $iProgram, $iFragmentShader )
	glLinkProgram( $iProgram )

	; Necessary in OpenGL 2.1, not in OpenGL 3.1
	; See InitializeVertexArrayObject() below and PosColorLocalTransform.vert
	;$iColorAttrib = glGetAttribLocation( $iProgram, "color" )
	;$iPositionAttrib = glGetAttribLocation( $iProgram, "position" )

	$iModelToCameraMatrixUnif = glGetUniformLocation( $iProgram, "modelToCameraMatrix" )
	$iCameraToClipMatrixUnif = glGetUniformLocation( $iProgram, "cameraToClipMatrix" )
	$iBaseColorUnif = glGetUniformLocation( $iProgram, "baseColor" )

	Local $zNear = 1.0, $zFar = 600.0
	$fFrustumScale = CalcFrustumScale( 20 )

	$mCameraToClipMatrix = $NULL_MATRIX
	$mCameraToClipMatrix[0]  = $fFrustumScale
	$mCameraToClipMatrix[5]  = $fFrustumScale
	$mCameraToClipMatrix[10] = ( $zFar + $zNear ) / ( $zNear - $zFar )
	$mCameraToClipMatrix[11] = -1.0
	$mCameraToClipMatrix[14] = ( 2 * $zFar * $zNear ) / ( $zNear - $zFar )

	glUseProgram( $iProgram )
	glUniformMatrix4fv( $iCameraToClipMatrixUnif, 1, $GL_FALSE, $mCameraToClipMatrix )
	glUseProgram(0)

	Return 1

EndFunc


Func InitializeVertexBuffer()

	glGenBuffers( 1, $iVBOship )
	glBindBuffer( $GL_ARRAY_BUFFER, $iVBOship )
	glBufferData( $GL_ARRAY_BUFFER, $GL_FLOAT, 4*UBound($aVertexData), $aVertexData, $GL_STATIC_DRAW )
	glBindBuffer( $GL_ARRAY_BUFFER, 0 )

	Return 1

EndFunc


Func InitializeVertexArrayObject()

	glGenVertexArrays( 1, $iVAOship )
	glBindVertexArray( $iVAOship )

	glBindBuffer( $GL_ARRAY_BUFFER, $iVBOship )
	glEnableVertexAttribArray(0) ; Positions (x,y,z)
	glEnableVertexAttribArray(1) ; Colors (r,g,b)
	glVertexAttribPointer( 0, 3, $GL_FLOAT, $GL_FALSE, 0, 0 )                      ; Positions (x,y,z)
	glVertexAttribPointer( 1, 3, $GL_FLOAT, $GL_FALSE, 0, $iNumberOfVertices*3*4 ) ; Colors (r,g,b)

	glBindVertexArray(0)

	Return 1

EndFunc


Func DisplayFunction()

	HandleKeystrokes()

	Local $mCurrMatrix = $IDENTITY_MATRIX, $mQuaternion
	Local $trans[3] = [ 0.0, 0.0, -200.0 ], $scale[3] = [ 3.0, 3.0, 3.0 ]
	$mCurrMatrix = glmTranslate( $mCurrMatrix, $trans )
	$mQuaternion = glmCast( $vOrientation )
	$mCurrMatrix = glmMultiply( $mCurrMatrix, $mQuaternion )
	$mCurrMatrix = glmScale( $mCurrMatrix, $scale )
	$mCurrMatrix = glmRotate( $mCurrMatrix, -90, $axisX )

	glUseProgram( $iProgram )
	glBindVertexArray( $iVAOship )

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )

	glUniform4f( $iBaseColorUnif, 1.0, 1.0, 1.0, 1.0 )
	glUniformMatrix4fv( $iModelToCameraMatrixUnif, 1, $GL_FALSE, $mCurrMatrix )

	glDrawArrays( $GL_TRIANGLES, 0, $iNumberOfVertices )

	glBindVertexArray(0)
	glUseProgram(0)

	SwapBuffers( $hDC )

EndFunc


Func OffsetOrientation( ByRef $axis, $fAngle )
	$fAngle = Radians( $fAngle )
	Local $vOffset = glmQuaternionVec4( $axis, $fAngle )
	Switch $bMode
		Case 0
			$vOrientation = glmQuaternionsMultVec4( $vOrientation, $vOffset )
		Case 1
			$vOrientation = glmQuaternionsMultVec4( $vOffset, $vOrientation )
	EndSwitch
	$vOrientation = glmNormalizeVec4( $vOrientation )
EndFunc


Func HandleKeystrokes()

	Local Static $bKey = 0, $nKey = 0
	Local Static $nStrokes = 0

	If Not $bFocus Then Return

	Select
		Case _IsPressed( "57", $dllUser32 ) ; W
			$bKey = 1
			if Not $nKey Then $nKey = 1
		Case _IsPressed( "53", $dllUser32 ) ; S
			$bKey = 2
			if Not $nKey Then $nKey = 1
		Case _IsPressed( "41", $dllUser32 ) ; A
			$bKey = 3
			if Not $nKey Then $nKey = 1
		Case _IsPressed( "44", $dllUser32 ) ; D
			$bKey = 4
			if Not $nKey Then $nKey = 1
		Case _IsPressed( "51", $dllUser32 ) ; Q
			$bKey = 5
			if Not $nKey Then $nKey = 1
		Case _IsPressed( "45", $dllUser32 ) ; E
			$bKey = 6
			if Not $nKey Then $nKey = 1
		Case _IsPressed( "4F", $dllUser32 ) ; O
			$bKey = 7
			if Not $nKey Then $nKey = 1
		Case Else
			$nKey = 0
			Return
	EndSelect

	If $nKey = 1 Or $nKey = 3 Then
		Switch $bKey
			Case 1  ; W
				OffsetOrientation( $axisX, $SmallAngelIncrement )
			Case 2  ; S
				OffsetOrientation( $axisX, -$SmallAngelIncrement )
			Case 3  ; A
				OffsetOrientation( $axisZ, $SmallAngelIncrement )
			Case 4  ; D
				OffsetOrientation( $axisZ, -$SmallAngelIncrement )
			Case 5  ; Q
				OffsetOrientation( $axisY, $SmallAngelIncrement )
			Case 6  ; E
				OffsetOrientation( $axisY, -$SmallAngelIncrement )
			Case 7  ; O
				$vOrientation[0] = 0.7892
				$vOrientation[1] = 0.3700
				$vOrientation[2] = 0.02514
				$vOrientation[3] = 0.4895
		EndSwitch
		$bKey = 0
	EndIf

	Switch $nKey
		Case 1
			; First keypress
			$nKey = 2
			$nStrokes = 0
		Case 2
			; Wait a few loops for next keypress
			; To get time to release the key for a single keypress
			$nStrokes += 1
			If $nStrokes > 3 Then _
				$nKey = 3
	EndSwitch

EndFunc


Func WM_ACTIVATE( $hWnd, $Msg, $wParam, $lParam )
	Switch $hWnd
		Case $hGui
			$bFocus = BitAND( $wParam, 0xFFFF )
	EndSwitch
	Return $GUI_RUNDEFMSG
EndFunc


Func WM_SIZING( $hWnd, $iMsg, $iwParam, $ilParam )
	#forceref $iMsg, $iwParam
	Switch $hWnd
		Case $hGui
			Local $tRECT = DllStructCreate( $tagRECT, $ilParam )
			Local $y = DllStructGetData( $tRECT, "Top" )
			Local $iHeight = DllStructGetData( $tRECT, "Bottom" ) - $y - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight
			WinMove( $hStatusbar, "", 2, $iHeight + $iHeightXP )
	EndSwitch
	Return $GUI_RUNDEFMSG
EndFunc


Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight * $iStatusBarRows
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	WinMove( $hStatusbar, "", 2, $iHeight + $iHeightXP + $iStatusBarHeight * ( $iStatusBarRows - 1 ) )
	$bResize = True
EndFunc


Func GUI_EVENT_CLOSE()
	$bClose = True
EndFunc


Func CloseProgram()
	glDeleteProgram( $iProgram )
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
