#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>
#include <Misc.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

Global $vOrient, $aOrientations
Global $idxOrient = 0, $bIsAnimating = 0

; --- OpenGL globals ---

Global $iProgram, $iVAOship, $iVBOship
Global $iModelToCameraMatrixUnif, $iCameraToClipMatrixUnif, $iBaseColorUnif
Global $fFrustumScale, $mCameraToClipMatrix
Global $iPositionAttrib, $iColorAttrib

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 500, $iHeight = 500

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30
Global Const $iStatusBarHeight = 28
Global Const $iStatusBarRows   =  2

Global $hStatusbar, $idRadio0, $idRadio1
Global $bMode = 0, $bNewMode = $bMode

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight + $iStatusBarHeight * $iStatusBarRows

Global $hGui, $idInput, $bResize = True, $bClose = False, $bFocus = True

; --- Project includes ---

#include "utilities\includes.au3"
#include "..\utilities\mathematics.au3"
#include "models\SpaceShip.au3"

; --- Program code ---

MainFunc()

While 1
	If $bMode <> $bNewMode Then
		$bMode = $bNewMode
	EndIf
	If $bResize Then
		$mCameraToClipMatrix[0] = $fFrustumScale * ( $iHeight / $iWidth )
		$mCameraToClipMatrix[5] = $fFrustumScale

		glUseProgram( $iProgram )
		glUniformMatrix4fv( $iCameraToClipMatrixUnif, 1, $GL_FALSE, $mCameraToClipMatrix )
		glUseProgram(0)

		glViewport( 0, 0, $iWidth, $iHeight )
		$bResize = False
	EndIf
	If $bClose Then
		CloseProgram()
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then CloseProgram()
	If Not InitOpenGLwindow() Then CloseProgram()
	If Not CheckOpenGLversion( 3, 1 ) Then CloseProgram()
	If Not CreateResources() Then CloseProgram()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 100, $WS_OVERLAPPEDWINDOW )

	; Input control to receive key presses (to prevent sound alarm)
	$idInput = GUICtrlCreateInput( "", 0, 0, 1, 1, 2048 ) ; 2048 = $ES_READONLY
	GUICtrlSetBkColor( $idInput, 0x000000 )
	GUICtrlSetState( $idInput, $GUI_FOCUS )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	GUIRegisterMsg( $WM_ACTIVATE, "WM_ACTIVATE" )
	GUIRegisterMsg( $WM_SIZING,   "WM_SIZING" )

	CreateStatusbar()

	Return 1

EndFunc


Func CreateStatusbar()

	; Use a child window as a container for controls in the Statusbar
	$hStatusbar = GUICreate( "", $iGuiWidth - 4 , 24, 2, $iHeight + $iStatusBarHeight + 3 , $WS_CHILD, -1, $hGui )

	; Interpolation
	GUIStartGroup()
	GUICtrlCreateLabel( "Interpolation:", 2, 5, 60, 14 )
	$idRadio0 = GUICtrlCreateRadio( "Lerp",   70, 3, 45, 18 )
	$idRadio1 = GUICtrlCreateRadio( "Slerp", 115, 3, 45, 18 )
	GUICtrlSetOnEvent( $idRadio0, "RadioClick" )
	GUICtrlSetOnEvent( $idRadio1, "RadioClick" )
	GUICtrlSetState( $idRadio0, $GUI_CHECKED )

	; Vertical splitterbar
	GUICtrlCreateLabel( "", 170, 2, 3, 20, $SS_ETCHEDFRAME )

	; Label in statusbar
	GUICtrlCreateLabel( "", 185, 5, 250, 14 )
	GUICtrlSetData( -1, "Positions: Q,W,E,R,T,Y,U.  Reset: Q." )

	GUISetState( @SW_SHOW, $hStatusbar )
	GUISwitch( $hGui )

	; Label for $hStatusbar to make it look nicer
	GUICtrlCreateLabel( "", 0, $iHeight + 1, $iGuiWidth + 4, $iStatusBarHeight + 1, $SS_ETCHEDFRAME )
	GUICtrlSetResizing( -1, $GUI_DOCKSTATEBAR )
	GUICtrlCreateLabel( "Interpolation between two orientations with linear (Lerp) or spherical linear (Slerp) interpolation.", _
											3, $iHeight + 1 + 6, $iGuiWidth - 6, $iStatusBarHeight - 10 )
	GUICtrlSetResizing( -1, $GUI_DOCKSTATEBAR )

EndFunc


Func RadioClick()
	Switch @GUI_CTRLID
		Case $idRadio0
			$bNewMode = 0
		Case $idRadio1
			$bNewMode = 1
	EndSwitch
	GUICtrlSetState( $idInput, $GUI_FOCUS )
EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 0.0, 0.0, 0.0, 0.0 )
	glClearDepth( 1.0 )

	glEnable( $GL_CULL_FACE )
	glCullFace( $GL_BACK )
	glFrontFace( $GL_CW )

	glEnable( $GL_DEPTH_TEST )
	glDepthMask( $GL_TRUE )
	glDepthFunc( $GL_LEQUAL )
	glDepthRange( 0.0, 1.0 )

	GUISwitch( $hGui )

	Return 1

EndFunc


Func CreateResources()

	If Not InitializeProgram() Then Return 0

	If Not InitializeVertexBuffer() Then Return 0

	If Not InitializeVertexArrayObject() Then Return 0

	Local $vAxis0[4] = [ 0.7071, 0.0, 0.0, 0.7071 ]
	Local $vAxis1[4] = [ 0.5, -0.5, 0.5, 0.5 ]
	Local $vAxis2[4] = [ -0.7892, -0.3700, -0.02514, -0.4895 ]
	Local $vAxis3[4] = [ 0.7892, 0.3700, 0.02514, 0.4895 ]
	Local $vAxis4[4] = [ -0.1591, -0.7991, -0.4344, 0.3840 ]
	Local $vAxis5[4] = [ 0.5208, 0.6483, 0.0410, 0.5537 ]
	Local $vAxis6[4] = [ 0.0, 1.0, 0.0, 0.0 ]

	Local $aOrients[7] = [ $vAxis0, $vAxis1, $vAxis2, $vAxis3, $vAxis4, $vAxis5, $vAxis6 ]

	$aOrientations = $aOrients
	$vOrient = $aOrientations[$idxOrient]

	GUISetState()

	Return 1

EndFunc


Func InitializeProgram()

	Local $iVertexShader, $iFragmentShader
	If Not LoadGLSLshader( $iVertexShader, "shaders\PosColorLocalTransform.vert", $GL_VERTEX_SHADER ) Then Return 0
	If Not LoadGLSLshader( $iFragmentShader, "shaders\ColorMultUniform.frag", $GL_FRAGMENT_SHADER ) Then Return 0

	$iProgram = glCreateProgram()
	glAttachShader( $iProgram, $iVertexShader )
	glAttachShader( $iProgram, $iFragmentShader )
	glLinkProgram( $iProgram )

	$iColorAttrib = glGetAttribLocation( $iProgram, "color" )
	$iPositionAttrib = glGetAttribLocation( $iProgram, "position" )

	$iModelToCameraMatrixUnif = glGetUniformLocation( $iProgram, "modelToCameraMatrix" )
	$iCameraToClipMatrixUnif = glGetUniformLocation( $iProgram, "cameraToClipMatrix" )
	$iBaseColorUnif = glGetUniformLocation( $iProgram, "baseColor" )

	Local $zNear = 1.0, $zFar = 600.0
	$fFrustumScale = CalcFrustumScale( 20 )

	$mCameraToClipMatrix = $NULL_MATRIX
	$mCameraToClipMatrix[0]  = $fFrustumScale
	$mCameraToClipMatrix[5]  = $fFrustumScale
	$mCameraToClipMatrix[10] = ( $zFar + $zNear ) / ( $zNear - $zFar )
	$mCameraToClipMatrix[11] = -1.0
	$mCameraToClipMatrix[14] = ( 2 * $zFar * $zNear ) / ( $zNear - $zFar )

	glUseProgram( $iProgram )
	glUniformMatrix4fv( $iCameraToClipMatrixUnif, 1, $GL_FALSE, $mCameraToClipMatrix )
	glUseProgram(0)

	Return 1

EndFunc


Func InitializeVertexBuffer()

	glGenBuffers( 1, $iVBOship )
	glBindBuffer( $GL_ARRAY_BUFFER, $iVBOship )
	glBufferData( $GL_ARRAY_BUFFER, $GL_FLOAT, 4*UBound($aVertexData), $aVertexData, $GL_STATIC_DRAW )
	glBindBuffer( $GL_ARRAY_BUFFER, 0 )

	Return 1

EndFunc


Func InitializeVertexArrayObject()

	glGenVertexArrays( 1, $iVAOship )
	glBindVertexArray( $iVAOship )

	glBindBuffer( $GL_ARRAY_BUFFER, $iVBOship )
	glEnableVertexAttribArray(0)
	glEnableVertexAttribArray(1)
	glVertexAttribPointer( 0, 3, $GL_FLOAT, $GL_FALSE, 0, 0 )
	glVertexAttribPointer( 1, 3, $GL_FLOAT, $GL_FALSE, 0, $iNumberOfVertices*3*4 )

	glBindVertexArray(0)

	Return 1

EndFunc


Func DisplayFunction()

	HandleKeystrokes()

	If $bIsAnimating Then Animate()

	Local $mCurrMatrix = $IDENTITY_MATRIX, $mQuaternion
	Local $trans[3] = [ 0.0, 0.0, -200.0 ], $scale[3] = [ 3.0, 3.0, 3.0 ]
	$mCurrMatrix = glmTranslate( $mCurrMatrix, $trans )
	$mQuaternion = glmCast( $vOrient )
	$mCurrMatrix = glmMultiply( $mCurrMatrix, $mQuaternion )
	$mCurrMatrix = glmScale( $mCurrMatrix, $scale )
	$mCurrMatrix = glmRotate( $mCurrMatrix, -90, $axisX )

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )

	glUseProgram( $iProgram )
	glBindVertexArray( $iVAOship )

	glUniform4f( $iBaseColorUnif, 1.0, 1.0, 1.0, 1.0 )
	glUniformMatrix4fv( $iModelToCameraMatrixUnif, 1, $GL_FALSE, $mCurrMatrix )

	glDrawArrays( $GL_TRIANGLES, 0, $iNumberOfVertices )

	glBindVertexArray(0)
	glUseProgram(0)

	SwapBuffers( $hDC )

EndFunc


Func Animate()
	Local Static $bTheMode
	Local Static $vStart, $vEnd
	Local Static $iStartTime, $fLastTime
	Local $fElapsedTime
	If $bIsAnimating = 1 Then
		$bIsAnimating = 2
		$bTheMode = $bMode
		$vStart = $vOrient
		$vEnd = $aOrientations[$idxOrient]
		$iStartTime = TimerInit()
		$fLastTime = 0
	EndIf
	$fElapsedTime = TimerDiff( $iStartTime ) / 1000
	$fElapsedTime = ( $fLastTime + $fElapsedTime ) / 2
	$fLastTime = $fElapsedTime
	If $fElapsedTime < 1.0 Then
		Switch $bTheMode
			Case 0
				$vOrient = Lerp( $vStart, $vEnd, $fElapsedTime )
			Case 1
				$vOrient = Slerp( $vStart, $vEnd, $fElapsedTime )
		EndSwitch
	Else
		$vOrient = $aOrientations[$idxOrient]
		$bIsAnimating = 0
	EndIf
EndFunc


Func Lerp( ByRef $v0, ByRef $v1, $fAlpha )
	Local $vInterp = glmMixVec4( $v0, $v1, $fAlpha )
	$vInterp = glmNormalizeVec4( $vInterp )
	Return $vInterp
EndFunc


Func Slerp( ByRef $v0, ByRef $v1, $fAlpha )
	Local $fDot = glmDotVec4Num( $v0, $v1 )
	If $fDot > 0.9995 Then Return Lerp( $v0, $v1, $fAlpha )
	$fDot = Clamp( $fDot, -1.0, 1.0 )
	Local $fTheta = ACos( $fDot ) * $fAlpha
	Local $v2[4] = [ _
		$v1[0] - $v0[0] * $fDot, _
		$v1[1] - $v0[1] * $fDot, _
		$v1[2] - $v0[2] * $fDot, _
		$v1[3] - $v0[3] * $fDot ]
	$v2 = glmNormalizeVec4( $v2 )
	Local $v3 = glmMultVec4( $v0, Cos( $fTheta ) )
	Local $v4 = glmMultVec4( $v2, Sin( $fTheta ) )
	Return glmAddVec4( $v3, $v4 )
EndFunc


Func HandleKeystrokes()

	Local Static $bKey = 0, $nKey = 0
	Local Static $nStrokes = 0

	If Not $bFocus Then Return

	Select
		Case _IsPressed( "51", $dllUser32 ) ; Q
			$bKey = 1
			if Not $nKey Then $nKey = 1
		Case _IsPressed( "57", $dllUser32 ) ; W
			$bKey = 2
			if Not $nKey Then $nKey = 1
		Case _IsPressed( "45", $dllUser32 ) ; E
			$bKey = 3
			if Not $nKey Then $nKey = 1
		Case _IsPressed( "52", $dllUser32 ) ; R
			$bKey = 4
			if Not $nKey Then $nKey = 1
		Case _IsPressed( "54", $dllUser32 ) ; T
			$bKey = 5
			if Not $nKey Then $nKey = 1
		Case _IsPressed( "59", $dllUser32 ) ; Y
			$bKey = 6
			if Not $nKey Then $nKey = 1
		Case _IsPressed( "55", $dllUser32 ) ; U
			$bKey = 7
			if Not $nKey Then $nKey = 1
		Case Else
			$nKey = 0
			Return
	EndSelect

	If $nKey = 1 Or $nKey = 3 Then
		Switch $bKey
			Case 1  ; Q
				If Not $bIsAnimating And $idxOrient <> 0 Then
					$bIsAnimating = 1
					$idxOrient = 0
				EndIf
			Case 2  ; W
				If Not $bIsAnimating And $idxOrient <> 1 Then
					$bIsAnimating = 1
					$idxOrient = 1
				EndIf
			Case 3  ; E
				If Not $bIsAnimating And $idxOrient <> 2 And $idxOrient <> 3 Then
					$bIsAnimating = 1
					$idxOrient = 2
				EndIf
			Case 4  ; R
				If Not $bIsAnimating And $idxOrient <> 2 And $idxOrient <> 3 Then
					$bIsAnimating = 1
					$idxOrient = 3
				EndIf
			Case 5  ; T
				If Not $bIsAnimating And $idxOrient <> 4 Then
					$bIsAnimating = 1
					$idxOrient = 4
				EndIf
			Case 6  ; Y
				If Not $bIsAnimating And $idxOrient <> 5 Then
					$bIsAnimating = 1
					$idxOrient = 5
				EndIf
			Case 7  ; U
				If Not $bIsAnimating And $idxOrient <> 6 Then
					$bIsAnimating = 1
					$idxOrient = 6
				EndIf
		EndSwitch
		$bKey = 0
	EndIf

	Switch $nKey
		Case 1
			; First keypress
			$nKey = 2
			$nStrokes = 0
		Case 2
			; Wait a few loops for next keypress
			; To get time to release the key for a single keypress
			$nStrokes += 1
			If $nStrokes > 3 Then _
				$nKey = 3
	EndSwitch

EndFunc


Func WM_ACTIVATE( $hWnd, $Msg, $wParam, $lParam )
	Switch $hWnd
		Case $hGui
			$bFocus = BitAND( $wParam, 0xFFFF )
	EndSwitch
	Return $GUI_RUNDEFMSG
EndFunc


Func WM_SIZING( $hWnd, $iMsg, $iwParam, $ilParam )
	#forceref $iMsg, $iwParam
	Switch $hWnd
		Case $hGui
			Local $tRECT = DllStructCreate( $tagRECT, $ilParam )
			Local $y = DllStructGetData( $tRECT, "Top" )
			Local $iHeight = DllStructGetData( $tRECT, "Bottom" ) - $y - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight
			WinMove( $hStatusbar, "", 2, $iHeight + $iHeightXP )
	EndSwitch
	Return $GUI_RUNDEFMSG
EndFunc


Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth - $iStatusBarHeight * $iStatusBarRows
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	WinMove( $hStatusbar, "", 2, $iHeight + $iHeightXP + $iStatusBarHeight * ( $iStatusBarRows - 1 ) )
	$bResize = True
EndFunc


Func GUI_EVENT_CLOSE()
	$bClose = True
EndFunc


Func CloseProgram()
	glDeleteProgram( $iProgram )
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
