#include <WindowsConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>

; AutoIt options
Opt( "GUIOnEventMode", 1 )
Opt( "MustDeclareVars", 1 )

; --- Program globals ---

; --- OpenGL globals ---

Global $iVertexShaderId, $iFragmentShaderId, $iProgramId
Global $iVaoId, $iBufferId, $iIndexBufferId

Global $sVertexShader = _
	"#version 330" & @CRLF & _
	"layout(location=0) in vec4 in_Position;" & @CRLF & _
	"layout(location=1) in vec4 in_Color;" & @CRLF & _
	"out vec4 ex_Color;" & @CRLF & _
	"void main(void)" & @CRLF & _
	"{" & @CRLF & _
	"  gl_Position = in_Position;" & @CRLF & _
	"  ex_Color = in_Color;" & @CRLF & _
	"}"

Global $sFragmentShader =  _
	"#version 330" & @CRLF & _
	"in vec4 ex_Color;" & @CRLF & _
	"out vec4 out_Color;" & @CRLF & _
	"void main(void)" & @CRLF & _
	"{" & @CRLF & _
	"  out_Color = ex_Color;" & @CRLF & _
	"}"

; OpenGL window
Global $hOpenGL = 0, $hDC, $hRC, $iWidth = 600, $iHeight = 600

; --- GUI globals ---

Global Const $iBorderWidth     =  4
Global Const $iTitleBarHeight  = 30

Global $iGuiWidth  = $iWidth
Global $iGuiHeight = $iHeight

Global $hGui, $bResize = True

; --- Project includes ---

#include "..\utilities\includes.au3"

; --- Program code ---

MainFunc()

While 1
	If $bResize Then
		; An OpenGL window (viewport) should not be resized in the middle of a drawing cycle.
		; In GUI_EVENT_RESIZE() $bResize is set to True when the GUI is resized. At this point
		; when the DisplayFunction() isn't running it's safe to resize the OpenGL window.
		glViewport( 0, 0, $iWidth, $iHeight )
		$bResize = False
	EndIf
	DisplayFunction()
	Sleep(10)
WEnd


Func MainFunc()
	If Not InitGUIwindow() Then GUI_EVENT_CLOSE()
	If Not InitOpenGLwindow() Then GUI_EVENT_CLOSE()
	If Not CheckOpenGLversion( 3, 1 ) Then GUI_EVENT_CLOSE()
	If Not CreateResources() Then GUI_EVENT_CLOSE()
EndFunc


Func InitGUIwindow()

	$hGui = GUICreate( ScriptName(), $iGuiWidth, $iGuiHeight, 400, 50, $WS_OVERLAPPEDWINDOW )

	GUISetOnEvent( $GUI_EVENT_MAXIMIZE, "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESIZED,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_RESTORE,  "GUI_EVENT_RESIZE" )
	GUISetOnEvent( $GUI_EVENT_CLOSE,    "GUI_EVENT_CLOSE"  )

	Return 1

EndFunc


Func InitOpenGLwindow()

	$hOpenGL = GUICreate( "", $iWidth, $iHeight, 0, 0, $WS_CHILD, -1, $hGui )
	GUISetState()

	If GetOpenGLContext( $hOpenGL, $hDC, $hRC ) = 0 Then Return 0
	GUISetState()

	glClearColor( 0.0, 0.0, 0.0, 0.0 )

	GUISwitch( $hGui )

	Return 1

EndFunc


Func CreateResources()

	If Not CreateShaders() Then Return 0
	If Not CreateVBO() Then Return 0

	GUISetState()

	Return 1

EndFunc


Func CreateShaders()

	Local $iErrorCheckValue = glGetError()

	$iVertexShaderId = glCreateShader( $GL_VERTEX_SHADER )
	glShaderSource( $iVertexShaderId, 1, $sVertexShader, StringLen( $sVertexShader ) )
	glCompileShader( $iVertexShaderId )

	$iFragmentShaderId = glCreateShader( $GL_FRAGMENT_SHADER )
	glShaderSource( $iFragmentShaderId, 1, $sFragmentShader, StringLen( $sFragmentShader ) )
	glCompileShader( $iFragmentShaderId )

	$iProgramId = glCreateProgram()
	glAttachShader( $iProgramId, $iVertexShaderId )
	glAttachShader( $iProgramId, $iFragmentShaderId )
	glLinkProgram( $iProgramId )
	glUseProgram( $iProgramId )

	$iErrorCheckValue = glGetError()
	If $iErrorCheckValue <> $GL_NO_ERROR Then
		PrintError( _
			"ERROR: Could not create the shaders: " & _
			gluErrorString( $iErrorCheckValue ) )
		Return 0
	EndIf

	Return 1

EndFunc


Func DestroyShaders()

	Local $iErrorCheckValue = glGetError()

	glUseProgram( 0 )

	glDetachShader( $iProgramId, $iVertexShaderId )
	glDetachShader( $iProgramId, $iFragmentShaderId )

	glDeleteShader( $iFragmentShaderId )
	glDeleteShader( $iVertexShaderId )

	glDeleteProgram( $iProgramId )

	$iErrorCheckValue = glGetError()
	If $iErrorCheckValue <> $GL_NO_ERROR Then
		PrintError( _
			"ERROR: Could not destroy the shaders: " & _
			gluErrorString( $iErrorCheckValue ) )
	EndIf

EndFunc


Func CreateVBO()

	Local $aVertices[8*(1+4*4)] = [ _
		 0.0,  0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0, _ ; 0
		-0.2,  0.8, 0.0, 1.0, 0.0, 1.0, 0.0, 1.0, _ ; 1		; Top
		 0.2,  0.8, 0.0, 1.0, 0.0, 0.0, 1.0, 1.0, _ ; 2
		 0.0,  0.8, 0.0, 1.0, 0.0, 1.0, 1.0, 1.0, _ ; 3
		 0.0,  1.0, 0.0, 1.0, 1.0, 0.0, 0.0, 1.0, _ ; 4
		-0.2, -0.8, 0.0, 1.0, 0.0, 0.0, 1.0, 1.0, _ ; 5		; Bottom
		 0.2, -0.8, 0.0, 1.0, 0.0, 1.0, 0.0, 1.0, _ ; 6
		 0.0, -0.8, 0.0, 1.0, 0.0, 1.0, 1.0, 1.0, _ ; 7
		 0.0, -1.0, 0.0, 1.0, 1.0, 0.0, 0.0, 1.0, _ ; 8
		-0.8, -0.2, 0.0, 1.0, 0.0, 1.0, 0.0, 1.0, _ ; 9		; Left
		-0.8,  0.2, 0.0, 1.0, 0.0, 0.0, 1.0, 1.0, _ ; 10
		-0.8,  0.0, 0.0, 1.0, 0.0, 1.0, 1.0, 1.0, _ ; 11
		-1.0,  0.0, 0.0, 1.0, 1.0, 0.0, 0.0, 1.0, _ ; 12
		 0.8, -0.2, 0.0, 1.0, 0.0, 0.0, 1.0, 1.0, _ ; 13	; Right
		 0.8,  0.2, 0.0, 1.0, 0.0, 1.0, 0.0, 1.0, _ ; 14
		 0.8,  0.0, 0.0, 1.0, 0.0, 1.0, 1.0, 1.0, _ ; 15
		 1.0,  0.0, 0.0, 1.0, 1.0, 0.0, 0.0, 1.0 ]  ; 16

	Local $aIndices[3*4*4] = [ _
		 0,  1,  3, _		; Top
		 0,  3,  2, _
		 3,  1,  4, _
		 3,  4,  2, _
		 0,  5,  7, _		; Bottom
		 0,  7,  6, _
		 7,  5,  8, _
		 7,  8,  6, _
		 0,  9, 11, _		; Left
		 0, 11, 10, _
		11,  9, 12, _
		11, 12, 10, _
		 0, 13, 15, _		; Right
		 0, 15, 14, _
		15, 13, 16, _
		15, 16, 14 ]

	Local $iErrorCheckValue = glGetError()
	Local $iBufferSize = 4 * UBound( $aVertices )
	Local $iVertexSize = 8 * 4
	Local $iRgbOffset = $iVertexSize / 2

	glGenVertexArrays( 1, $iVaoId )
	glBindVertexArray( $iVaoId )

	glGenBuffers( 1, $iBufferId )
	glBindBuffer( $GL_ARRAY_BUFFER, $iBufferId )
	glBufferData( $GL_ARRAY_BUFFER, $GL_FLOAT, $iBufferSize, $aVertices, $GL_STATIC_DRAW )

	glVertexAttribPointer( 0, 4, $GL_FLOAT, $GL_FALSE, $iVertexSize, 0 )
	glVertexAttribPointer(1, 4, $GL_FLOAT, $GL_FALSE, $iVertexSize, $iRgbOffset )

	glEnableVertexAttribArray(0)
	glEnableVertexAttribArray(1)

	glGenBuffers( 1, $iIndexBufferId )
	glBindBuffer( $GL_ELEMENT_ARRAY_BUFFER, $iIndexBufferId )
	glBufferData( $GL_ELEMENT_ARRAY_BUFFER, $GL_BYTE, UBound($aIndices), $aIndices, $GL_STATIC_DRAW )

	$iErrorCheckValue = glGetError()
	If $iErrorCheckValue <> $GL_NO_ERROR Then
		PrintError( _
			"ERROR: Could not create a VBO: " & _
			gluErrorString( $iErrorCheckValue ) )
		Return 0
	EndIf

	Return 1

EndFunc


Func DestroyVBO()

	Local $iErrorCheckValue = glGetError()

	glDisableVertexAttribArray(1)
	glDisableVertexAttribArray(0)

	glBindBuffer( $GL_ARRAY_BUFFER, 0 )
	glDeleteBuffers( 1, $iBufferId )

	glBindBuffer( $GL_ELEMENT_ARRAY_BUFFER, 0 )
	glDeleteBuffers( 1, $iIndexBufferId )

	glBindVertexArray(0)
	glDeleteVertexArrays( 1, $iVaoId )

	$iErrorCheckValue = glGetError()
	If $iErrorCheckValue <> $GL_NO_ERROR Then
		PrintError( _
			"ERROR: Could not destroy the VBO: " & _
			gluErrorString( $iErrorCheckValue ) )
	EndIf

EndFunc


Func DisplayFunction()

	glClear( $GL_COLOR_BUFFER_BIT + $GL_DEPTH_BUFFER_BIT )

	glDrawElements( $GL_TRIANGLES, 48, $GL_UNSIGNED_BYTE, 0 )

	SwapBuffers( $hDC )

EndFunc


Func GUI_EVENT_RESIZE()
	Local $aPos = WinGetPos( $hGui )
	$iWidth = $aPos[2] - 2 * $iBorderWidth
	$iHeight = $aPos[3] - $iTitleBarHeight - $iBorderWidth
	WinMove( $hOpenGL, "", 0, 0, $iWidth, $iHeight )
	$bResize = True ; The OpenGL window (viewport) is resized in top of the main loop
EndFunc


Func GUI_EVENT_CLOSE()
	DestroyVBO()
	DestroyShaders()
	DisableOpenGL( $hOpenGL, $hDC, $hRC )
	GUIDelete( $hGui )
	ExLauncherWin()
	CloseDlls()
	Exit
EndFunc
