#include-once

; LoadObjQuads( ByRef $aModel, $bOpenGL21 = True, $bNormals = 0, $bReload = False )
;   Use LoadObjQuads to force a model to be loaded as a quad mesh and rendered with $GL_QUADS draw mode
;   To use LoadObjQuads the model must be a quad mesh model
;   See LoadObjModel for an explanation of the parameters

Func LoadObjQuads( ByRef $aModel, $bOpenGL21 = True, $bNormals = 0, $bReload = False )
	; $bOpenGL21 is used to handle the OpenGL version
	;   Set $bOpenGL21 = True to use the model for OpenGL version 2.1 or higher
	;   Set $bOpenGL21 = False to use the model for version 2.0 or lower
	; $bNormals is used to handle normal vectors. Values: 0, 1, -1
	; $bNormals = 0/1 is for models which does not contain any normal vectors
	;   Set $bNormals = 0 to calculate normal vectors (no average)
	;   Set $bNormals = 1 to calculate average normal vectors
	; $bNormals = -1 is for models which does contain normal vectors
	;   Set $bNormals = -1 to exclude normal vectors when they are not needed
	; $bReload is used when a model is reloaded in the viewers because of a change in settings
	;   $bReload is set to True to prevent an update of the progress info section in the inifile
	;   We don't want to update the progress info section just because of a reload in the viewers
	;   $bReload should always be set to False; it should only be set to True in main loop in viewers
	;   (If settings are saved the inifile is updated when LoadObjQuads is called from CreateResources)

	$bModelLoadCancel = False

	; To calculate process time
	Local $iStartTime = TimerInit()

	; Full path to the OBJ-file
	Local $sObjModel = $aModel[11], $sObjPath

	; Informations per object
	Local $iObjects = 0, $aVertStart[1], $aVertCnts[1], $aMaterials[1] = [ 0 ] ; $aMaterials[0] is used to count the materials

	; Vertices, texture coordinates, normal vectors, faces
	Local $aVertices[6000], $aVerticesTmp[1][3]
	Local $aTextureUVs[4000], $aTextureUVsTmp[1][2]
	Local $aNormals[6000], $aNormalsTmp[1][3]
	Local $aFaces[2000], $aFacesTmp[1]

	; Face triangle attributes
	Local $iModelFaceAttribs = 0
	Local $iModelFaceAttribsTmp = 0
	; += 1: Vertices
	; += 2: TexCoords
	; += 4: Normals

	; Material info
	Local $aMaterialInfo, $iMaterialInfo, $bMaterials
	Local $iMaterialFiles = 0, $sMaterialFile = "", $sUseMaterial = "", $sLastMaterial = ""
	$aModel[6] = 0 ; Initialized to no material info

	; Texture images, normal vectors
	Local $bTexImage = False, $bUseNormals = Not ( $bNormals = -1 )

	; Status info
	$aModel[7] = 0 ; Initialized to no texture maps, no texture coordinates, no normal vectors

	; Calculate size of model
	Local $x, $y, $z, $m = 1000000.0
	Local $xMin = $m, $xMax = -$m
	Local $yMin = $m, $yMax = -$m
	Local $zMin = $m, $zMax = -$m

	Local $aLines, $line, $aWords

	; Read OBJ-file to $aLines array
	HourglassCursor( True )
	_FileReadToArray( $sObjModel & ".obj", $aLines )
	ConsoleWrite( @CRLF & "Model file: " & $sObjModel & ".obj" & @CRLF )
	ConsoleWrite( "Model file lines: " & $aLines[0] & @CRLF )
	HourglassCursor( False )

	; Progress info
	Local $bProgressInfo, $iMaxProgressInfo = 10000
	Local $aProgressInfo[$iMaxProgressInfo][2], $aNewProgressInfo[$iMaxProgressInfo][2], $p = 1
	Local $hProgress, $idProLabel1, $idProgress1, $idProLabel2, $idProgress2
	$bProgressInfo = ( $aLines[0] > 3000 )

	If $bProgressInfo Then

		; Try to read progress info in ini-file
		If FileExists( $sObjModel & ".ini" ) Then
			Local $aIniLines, $aIniWords, $i, $j, $k = 1, $s
			_FileReadToArray( $sObjModel & ".ini", $aIniLines )
			For $i = 1 To $aIniLines[0]
				If $aIniLines[$i] = "[ProgressInfoQuads]" Then ExitLoop
			Next
			If $i < $aIniLines[0] Then
				ReDim $aProgressInfo[$aIniLines[0]][2]
				For $j = $i + 1 To $aIniLines[0]
					;ConsoleWrite( $j & ": " & $aIniLines[$j] & @CRLF )
					$s = StringLeft( $aIniLines[$j], 1 )
					If $s = "[" Or $s = "" Then ExitLoop
					$aIniWords = StringSplit( $aIniLines[$j], "=", 2 )
					;_ArrayDisplay( $aIniWords )
					$aProgressInfo[$k][0] = $aIniWords[0]
					$aProgressInfo[$k][1] = $aIniWords[1]
					$k += 1
				Next
				$aProgressInfo[0][0] = $k - 1
				ReDim $aProgressInfo[$k+10][2]
				For $i = $k To $k + 9
					$aProgressInfo[$i][1] = $aLines[0]
				Next
				ReDim $aNewProgressInfo[$k+10][2]
				;_ArrayDisplay( $aProgressInfo )
				;Exit
			Else
				; Default progress info
				$aProgressInfo[0][0] = $iMaxProgressInfo
				For $i = 1 To $iMaxProgressInfo - 1
					$aProgressInfo[$i][1] = $aLines[0]
				Next
			EndIf

		Else

			; Default progress info
			$aProgressInfo[0][0] = $iMaxProgressInfo
			For $i = 1 To $iMaxProgressInfo - 1
				$aProgressInfo[$i][1] = $aLines[0]
			Next

		EndIf

		; Create window for progress info
		Local $name = StringRight( $sObjModel, StringLen( $sObjModel ) - StringInStr( $sObjModel, "\", 0, -1 ) )
		CreateModelProgressWin( $name, $hProgress, $idProLabel1, $idProgress1, $idProLabel2, $idProgress2 )

	EndIf


	; ====== Get model data, main loop ======

	;Local $aNotMatching[100][2]
	Local $v = 0, $v0 = 0, $vm, $vt = 0, $vt0 = 0, $vtm, $vn = 0, $vn0 = 0, $vnm
	Local $f = 0, $f0 = 0, $j = 0, $k = 1, $n = 0, $bPosRef, $bPosRef0 = -1, $pm1, $iPnts

	For $i = 1 To $aLines[0]

		; The first character in a line is a token that defines the meaning of the line
		; The tokens can be one of the following: "", "#", "g", "o", "m", "u", "v" ("v ", "vt", "vn"), "f", "l"

		Switch StringLeft( $aLines[$i], 1 )

			; --- Not handled ---

			Case "", "#", "g", "o", "s" ; Empty lines, comments, groups, objects, smoothing groups


			; --- Material file ---

			Case "m" ; Material file: mtllib <file>.mtl
				$bMaterials = True
				$line = $aLines[$i]
				If StringLeft( $line, 7 ) = "mtllib " Then
					$sMaterialFile = StringRight( $line, StringLen( $line ) - 7 )

					; Material file format
					Local $sFormat = StringRight( $sMaterialFile, StringLen( $sMaterialFile ) - StringInStr( $sMaterialFile, ".", 0, -1 ) )
					Switch $sFormat
						Case "mtl"
						Case Else
							ConsoleWrite( "Function LoadObjQuads()" & @CRLF & _
								"Warning: """ & $sFormat & """ is not a legal material file format." & @CRLF & _
								"Legal formats: MTL (Wavefront MTL file)" & @CRLF )
							$bMaterials = False
					EndSwitch

					; Material file exists?
					If $bMaterials Then
						$sObjPath = StringLeft( $sObjModel, StringInStr( $sObjModel, "\", 0, -1 ) )
						If Not FileExists( $sObjPath & $sMaterialFile ) Then
							ConsoleWrite( "Function LoadObjQuads()" & @CRLF & _
								"Warning: File """ & $sObjPath & $sMaterialFile & """ does not exists." & @CRLF )
							$bMaterials = False
						EndIf
					EndIf

					If $bMaterials Then
						If $iMaterialFiles = 1 Then
							PrintError( "Function LoadObjQuads()" & @CRLF & _
								"ERROR: File """ & $sObjModel & ".obj""" & @CRLF & _
								"Line #" & $i & ": " & StringStripWS( $aLines[$i], 2 ) & @CRLF & _
								"Two or more material files in a model is not allowed." )
							If $bProgressInfo Then GUIDelete( $hProgress )
							Return SetError(1,0,0)
						EndIf
						ConsoleWrite( "Material file: " & $sMaterialFile & @CRLF )
						$iMaterialInfo = ReadMaterials( $sObjPath, $sMaterialFile, $aMaterialInfo )

						If $iMaterialInfo Then
							;_ArrayDisplay( $aMaterialInfo )
							$aModel[6] = $aMaterialInfo
							$iMaterialFiles += 1
							Local $str
							For $m = 0 To $iMaterialInfo - 1
								$str = StringFormat( "%7s", $m )
								$str &= "|" & $aMaterialInfo[$m][0]
								$str &= "|" & $aMaterialInfo[$m][1]
								$str &= "|" & $aMaterialInfo[$m][2]
								$str &= "|" & $aMaterialInfo[$m][3]
								$str &= "|" & $aMaterialInfo[$m][4]
								$str &= "|" & $aMaterialInfo[$m][5]
								ConsoleWrite( $str & @CRLF )
							Next
							; Check if a valid image/texture file exists
							For $m = 0 To $iMaterialInfo - 1
								Switch $aMaterialInfo[$m][2]
									Case "map_Ka", "map_Kd", "map_Ks"
										If $aMaterialInfo[$m][4] = "" And $aMaterialInfo[$m][5] = "" Then
											$bTexImage = True
											$aModel[7] += 1
											ExitLoop
										EndIf
								EndSwitch
							Next
						EndIf
					EndIf
				EndIf


			; --- Use material ---

			Case "u" ; Use material: usemtl <name>
				$line = $aLines[$i]
				If StringLeft( $line, 7 ) = "usemtl " Then
					$line = StringStripWS( $line, 6 )
					$aWords = StringSplit( $line, " " )
					If $aWords[0] = 2 Then
						Switch $aWords[2]
							Case "", "None"
							Case Else
								$sUseMaterial = $aWords[2]
								ConsoleWrite( "Material (usemtl): " & $sUseMaterial & @CRLF )
						EndSwitch
					EndIf
				EndIf


			; --- Vertices, texture coordinates and normal vectors---

			Case "v" ; Tokens: "v ", "vt", "vn"
				Switch StringLeft( $aLines[$i], 2 )

					; --- Vertices ---

					; For the first v-line there will be done some error checking
					; Of performance reasons this will only be done for the first line

					; If there are more groups of v-lines in the OBJ-file, error checking will be done for the first line in each group

					Case "v " ; List of vertices: v x y z
						$line = StringStripWS( $aLines[$i], 6 )
						$aWords = StringSplit( $line, " " )

						; --- Error checking for the first v-line ---

						; Each line is a vertex point, it must contain three coordinates: x, y, z
						If $aWords[0] <> 4 Then
							PrintError( "Function LoadObjQuads()" & @CRLF & _
								"ERROR: File """ & $sObjModel & ".obj""" & @CRLF & _
								"Line #" & $i & ": " & StringStripWS( $aLines[$i], 2 ) & @CRLF & _
								"The number of vertex coordinates must be exactly three." )
							If $bProgressInfo Then GUIDelete( $hProgress )
							Return SetError(2,0,0)
						EndIf

						If $bProgressInfo Then
							GUICtrlSetData( $idProLabel1, "Vertex coordinates: " & $p - 1 & "/" & $aProgressInfo[0][0] )
							UpdateModelProgressBar( $idProgress2, 0, $aProgressInfo[$p][1] )
						EndIf

						; --- Loop through each v-line in the OBJ-file ---

						While StringLeft( $line, 2 ) = "v "
							Select
								Case $v >= 60000
									If Mod( $v, 60000 ) = 0 Then
										Redim $aVerticesTmp[$v+60000][3]
										If $bProgressInfo Then
											GUICtrlSetData( $idProLabel2, "Vertex coordinates: " & $v & "/" & $aProgressInfo[$p][1] )
											UpdateModelProgressBar( $idProgress2, $v, $aProgressInfo[$p][1] )
											If $bModelLoadCancel Then
												ConsoleWrite( "Function LoadObjQuads() cancelled!" & @CRLF & @CRLF )
												If $bProgressInfo Then GUIDelete( $hProgress )
												Return 0
											EndIf
										EndIf
									EndIf
								Case $v >= 10000
									If Mod( $v, 10000 ) = 0 Then
										Redim $aVerticesTmp[$v+10000][3]
										If $bProgressInfo Then
											GUICtrlSetData( $idProLabel2, "Vertex coordinates: " & $v & "/" & $aProgressInfo[$p][1] )
											UpdateModelProgressBar( $idProgress2, $v, $aProgressInfo[$p][1] )
											If $bModelLoadCancel Then
												ConsoleWrite( "Function LoadObjQuads() cancelled!" & @CRLF & @CRLF )
												If $bProgressInfo Then GUIDelete( $hProgress )
												Return 0
											EndIf
										EndIf
									EndIf
								Case Else
									If Mod( $v, 1000 ) = 0 Then
										Redim $aVerticesTmp[$v+1000][3]
										If $bProgressInfo Then
											GUICtrlSetData( $idProLabel2, "Vertex coordinates: " & $v & "/" & $aProgressInfo[$p][1] )
											UpdateModelProgressBar( $idProgress2, $v, $aProgressInfo[$p][1] )
										EndIf
									EndIf
							EndSelect
							$line = StringStripWS( $line, 6 )
							$aWords = StringSplit( $line, " ", 2 )
							; Calculate the size of the
							; model (the bounding box).
							$x = Number( $aWords[1] )
							$y = Number( $aWords[2] )
							$z = Number( $aWords[3] )
							$aVerticesTmp[$v][0] = $x
							$aVerticesTmp[$v][1] = $y
							$aVerticesTmp[$v][2] = $z
							If $x < $xMin Then $xMin = $x
							If $x > $xMax Then $xMax = $x
							If $y < $yMin Then $yMin = $y
							If $y > $yMax Then $yMax = $y
							If $z < $zMin Then $zMin = $z
							If $z > $zMax Then $zMax = $z
							$v += 1
							$i += 1
							$line = $aLines[$i]
						WEnd
						$i -= 1

						ConsoleWrite( "Vertex coordinates: " & $v - $v0 & @CRLF )
						ConsoleWrite( "Total coordinates:  " & $v & @CRLF )
						If $bProgressInfo Then
							UpdateModelProgressBar( $idProgress2, $v, $aProgressInfo[$p][1] )
							UpdateModelProgressBar( $idProgress1, $p, $aProgressInfo[0][0] )
							$aNewProgressInfo[$p][0] = "Vertex coordinates"
							$aNewProgressInfo[$p][1] = $v
							$p += 1
						EndIf
						$v0 = $v

						If $bModelLoadCancel Then
							ConsoleWrite( "Function LoadObjQuads() cancelled!" & @CRLF & @CRLF )
							If $bProgressInfo Then GUIDelete( $hProgress )
							Return 0
						EndIf


					; --- Texture coordinates ---

					; For the first vt-line there will be done some error checking
					; Of performance reasons this will only be done for the first line

					; If there are more groups of vt-lines in the OBJ-file, error checking will be done for the first line in each group

					Case "vt" ; Texture coordinates: vt u v [w]
						If $bProgressInfo Then
							GUICtrlSetData( $idProLabel1, "Texture coordinates: " & $p - 1 & "/" & $aProgressInfo[0][0] )
							UpdateModelProgressBar( $idProgress2, 0, $aProgressInfo[$p][1] )
						EndIf

						If $bTexImage Then

							; If a texture image exists, then read texture coordinates

							$line = StringStripWS( $aLines[$i], 6 )
							$aWords = StringSplit( $line, " " )

							; --- Error checking for the first vt-line ---

							; Each line is a texture point, it must contain two or three coordinates: u, v [w]
							; w is optional and defaults to zero, in this version of the function w is not used
							If $aWords[0] < 3 Or $aWords[0] > 4 Then ; vt u v [w]
								PrintError( "Function LoadObjQuads()" & @CRLF & _
									"ERROR: File """ & $sObjModel & ".obj""" & @CRLF & _
									"Line #" & $i & ": " & StringStripWS ( $aLines[$i], 2 ) & @CRLF & _
									"The number of texture coordinates must be two or three." )
								If $bProgressInfo Then GUIDelete( $hProgress )
								Return SetError(3,0,0)
							EndIf

							; --- Loop through each vt-line in the OBJ-file ---

							While StringLeft( $line, 3 ) = "vt "
								Select
									Case $vt >= 60000
										If Mod( $vt, 60000 ) = 0 Then
											Redim $aTextureUVsTmp[$vt+60000][2]
											If $bProgressInfo Then
												GUICtrlSetData( $idProLabel2, "Texture coordinates: " & $vt & "/" & $aProgressInfo[$p][1] )
												UpdateModelProgressBar( $idProgress2, $vt, $aProgressInfo[$p][1] )
												If $bModelLoadCancel Then
													ConsoleWrite( "Function LoadObjQuads() cancelled!" & @CRLF & @CRLF )
													If $bProgressInfo Then GUIDelete( $hProgress )
													Return 0
												EndIf
											EndIf
										EndIf
									Case $vt >= 10000
										If Mod( $vt, 10000 ) = 0 Then
											Redim $aTextureUVsTmp[$vt+10000][2]
											If $bProgressInfo Then
												GUICtrlSetData( $idProLabel2, "Texture coordinates: " & $vt & "/" & $aProgressInfo[$p][1] )
												UpdateModelProgressBar( $idProgress2, $vt, $aProgressInfo[$p][1] )
												If $bModelLoadCancel Then
													ConsoleWrite( "Function LoadObjQuads() cancelled!" & @CRLF & @CRLF )
													If $bProgressInfo Then GUIDelete( $hProgress )
													Return 0
												EndIf
											EndIf
										EndIf
									Case Else
										If Mod( $vt, 1000 ) = 0 Then
											Redim $aTextureUVsTmp[$vt+1000][2]
											If $bProgressInfo Then
												GUICtrlSetData( $idProLabel2, "Texture coordinates: " & $vt & "/" & $aProgressInfo[$p][1] )
												UpdateModelProgressBar( $idProgress2, $vt, $aProgressInfo[$p][1] )
											EndIf
										EndIf
								EndSelect
								$line = StringStripWS( $line, 6 )
								$aWords = StringSplit( $line, " ", 2 )
								$aTextureUVsTmp[$vt][0] = $aWords[1]
								$aTextureUVsTmp[$vt][1] = $aWords[2]
								$vt += 1
								$i += 1
								$line = $aLines[$i]
							WEnd
							$i -= 1

							ConsoleWrite( "Texture coordinates: " & $vt - $vt0 & @CRLF )
							ConsoleWrite( "Total coordinates:   " & $vt & @CRLF )

						Else

							; If no texture images exists, just loop through the texture coordinates

							$line = $aLines[$i]
							While StringLeft( $line, 3 ) = "vt "
								$vt += 1
								$i += 1
								$line = $aLines[$i]
							WEnd
							$i -= 1

						EndIf

						If $bProgressInfo Then
							UpdateModelProgressBar( $idProgress2, $vt, $aProgressInfo[$p][1] )
							UpdateModelProgressBar( $idProgress1, $p, $aProgressInfo[0][0] )
							$aNewProgressInfo[$p][0] = "Texture coordinates"
							$aNewProgressInfo[$p][1] = $vt
							$p += 1
						EndIf
						$vt0 = $vt

						If $bModelLoadCancel Then
							ConsoleWrite( "Function LoadObjQuads() cancelled!" & @CRLF & @CRLF )
							If $bProgressInfo Then GUIDelete( $hProgress )
							Return 0
						EndIf


					; --- Normal vectors ---

					; For the first vn-line there will be done some error checking
					; Of performance reasons this will only be done for the first line

					; If there are more groups of vn-lines in the OBJ-file, error checking will be done for the first line in each group

					Case "vn" ; Normal vectors: vn x y z
						If $bProgressInfo Then
							GUICtrlSetData( $idProLabel1, "Normal vectors: " & $p - 1 & "/" & $aProgressInfo[0][0] )
							UpdateModelProgressBar( $idProgress2, 0, $aProgressInfo[$p][1] )
						EndIf

						If $bUseNormals Then

							; Read normal vectors if they are not excluded in settings

							$line = StringStripWS( $aLines[$i], 6 )
							$aWords = StringSplit( $line, " " )

							; --- Error checking for the first vn-line ---

							; Each line is a normal vector, it must contain three coordinates: x, y, z
							If $aWords[0] <> 4 Then
								PrintError( "Function LoadObjQuads()" & @CRLF & _
									"ERROR: File """ & $sObjModel & ".obj""" & @CRLF & _
									"Line #" & $i & ": " & StringStripWS( $aLines[$i], 2 ) & @CRLF & _
									"The number of normal vector coordinates must be exactly three." )
								If $bProgressInfo Then GUIDelete( $hProgress )
								Return SetError(4,0,0)
							EndIf

							; --- Loop through each vn-line in the OBJ-file ---

							While StringLeft( $line, 3 ) = "vn "
								Select
									Case $vn >= 60000
										If Mod( $vn, 60000 ) = 0 Then
											Redim $aNormalsTmp[$vn+60000][3]
											If $bProgressInfo Then
												GUICtrlSetData( $idProLabel2, "Normal vectors: " & $vn & "/" & $aProgressInfo[$p][1] )
												UpdateModelProgressBar( $idProgress2, $vn, $aProgressInfo[$p][1] )
												If $bModelLoadCancel Then
													ConsoleWrite( "Function LoadObjQuads() cancelled!" & @CRLF & @CRLF )
													If $bProgressInfo Then GUIDelete( $hProgress )
													Return 0
												EndIf
											EndIf
										EndIf
									Case $vn >= 10000
										If Mod( $vn, 10000 ) = 0 Then
											Redim $aNormalsTmp[$vn+10000][3]
											If $bProgressInfo Then
												GUICtrlSetData( $idProLabel2, "Normal vectors: " & $vn & "/" & $aProgressInfo[$p][1] )
												UpdateModelProgressBar( $idProgress2, $vn, $aProgressInfo[$p][1] )
												If $bModelLoadCancel Then
													ConsoleWrite( "Function LoadObjQuads() cancelled!" & @CRLF & @CRLF )
													If $bProgressInfo Then GUIDelete( $hProgress )
													Return 0
												EndIf
											EndIf
										EndIf
									Case Else
										If Mod( $vn, 1000 ) = 0 Then
											Redim $aNormalsTmp[$vn+1000][3]
											If $bProgressInfo Then
												GUICtrlSetData( $idProLabel2, "Normal vectors: " & $vn & "/" & $aProgressInfo[$p][1] )
												UpdateModelProgressBar( $idProgress2, $vn, $aProgressInfo[$p][1] )
											EndIf
										EndIf
								EndSelect
								$line = StringStripWS( $line, 6 )
								$aWords = StringSplit( $line, " ", 2 )
								$aNormalsTmp[$vn][0] = $aWords[1]
								$aNormalsTmp[$vn][1] = $aWords[2]
								$aNormalsTmp[$vn][2] = $aWords[3]
								$vn += 1
								$i += 1
								$line = $aLines[$i]
							WEnd
							$i -= 1

							ConsoleWrite( "Normal vectors: " & $vn - $vn0 & @CRLF )
							ConsoleWrite( "Total vectors:  " & $vn & @CRLF )

						Else

							; If normal vectors are excluded just loop through this part of the model file

							$line = $aLines[$i]
							While StringLeft( $line, 3 ) = "vn "
								$vn += 1
								$i += 1
								$line = $aLines[$i]
							WEnd
							$i -= 1

						EndIf

						If $bProgressInfo Then
							UpdateModelProgressBar( $idProgress2, $vn, $aProgressInfo[$p][1] )
							UpdateModelProgressBar( $idProgress1, $p, $aProgressInfo[0][0] )
							$aNewProgressInfo[$p][0] = "Normal vectors"
							$aNewProgressInfo[$p][1] = $vn
							$p += 1
						EndIf
						$vn0 = $vn

						If $bModelLoadCancel Then
							ConsoleWrite( "Function LoadObjQuads() cancelled!" & @CRLF & @CRLF )
							If $bProgressInfo Then GUIDelete( $hProgress )
							Return 0
						EndIf


					; --- Not matching tokens ---

					Case Else ; "v?"
						$n += 1
						;$aNotMatching[$n][0] = $i
						;$aNotMatching[$n][1] = $aLines[$i]

				EndSwitch


			; --- Face quads ---

			; Face quad attributes
			; Four different combinations are handled
			; 1: Vertex:                  f v1 v2 v3 v4 ...
			; 2: Vertex/tex-coord:        f v1/vt1 v2/vt2 v3/vt3 v4/vt4 ...
			; 3: Vertex/tex-coord/normal: f v1/vt1/vn1 v2/vt2/vn2 v3/vt3/vn3 v4/vt4/vn4 ...
			; 4: Vertex/normal:           f v1//vn1 v2//vn2 v3//vn3 v4//vn4 ...

			Case "f"
				$line = StringStripWS( $aLines[$i], 6 )
				$aWords = StringSplit( $line, " " )

				; --- Error checking for the first f-line ---

				; Number of points
				$iPnts = $aWords[0] - 1
				; f v1 v2 v3 v4,  quad => $iPnts = 4
				; else                 => error
				; This also applies to the other types of face attributes

				If $iPnts <> 4 Then
					PrintError( "Function LoadObjQuads()" & @CRLF & _
						"ERROR: File """ & $sObjModel & ".obj""" & @CRLF & _
						"Line #" & $i & ": " & StringStripWS( $aLines[$i], 2 ) & @CRLF & _
						"A face quad line must contain exactly four points." )
					If $bProgressInfo Then GUIDelete( $hProgress )
					Return SetError(5,0,0)
				EndIf

				; Face attributes
				$aWords = StringSplit( $aWords[2], "/" )
				; f v1         => $aWords[0] = 1
				; f v1/vt1     => $aWords[0] = 2
				; f v1/vt1/vn1 => $aWords[0] = 3
				; f v1//vn1    => $aWords[0] = 3 and  $aWords[2] = ""

				; Positive or negative references
				; Positive references: v, vt and vn in f-lines are positive values
				; Negative references: v, vt and vn in f-lines are negative values
				If $aWords[1] > 0 Then
					$bPosRef = 1
					$pm1 = 1
				Else
					$bPosRef = 0
					$pm1 = -1
				EndIf

				; The references must be either positive or negative in the entire OBJ-file
				If $bPosRef0 = -1 Then
					$bPosRef0 = $bPosRef
				ElseIf $bPosRef <> $bPosRef0 Then
					PrintError( "Function LoadObjQuads()" & @CRLF & _
						"ERROR: File """ & $sObjModel & ".obj""" & @CRLF & _
						"Line #" & $i & ": " & StringStripWS( $aLines[$i], 2 ) & @CRLF & _
						"Both positive and negative references in face quads is not allowed." )
					If $bProgressInfo Then GUIDelete( $hProgress )
					Return SetError(6,0,0)
				EndIf

				; Face quad attributes
				; Four different combinations are handled
				; 1: Vertex:                  f v1 v2 v3 v4 ...
				; 2: Vertex/tex-coord:        f v1/vt1 v2/vt2 v3/vt3 v4/vt4 ...
				; 3: Vertex/tex-coord/normal: f v1/vt1/vn1 v2/vt2/vn2 v3/vt3/vn3 v4/vt4/vn4 ...
				; 4: Vertex/normal:           f v1//vn1 v2//vn2 v3//vn3 v4//vn4 ...

				Select
					Case $aWords[0] = 1 ; Vertex: f v1 v2 v3 ...
						$iModelFaceAttribsTmp = 1
						$aWords = StringSplit( $line, "/ " )
						; Find the maximum index v in the first f-line
						; The number of rows in $aVerticesTmp must at least match this maximum
						; $aVerticesTmp
						$vm = $pm1 * Number( $aWords[2] )
						If $vm < $pm1 * Number( $aWords[3] ) Then $vm = $pm1 * Number( $aWords[3] )
						If $vm < $pm1 * Number( $aWords[4] ) Then $vm = $pm1 * Number( $aWords[4] )
						If $vm < $pm1 * Number( $aWords[5] ) Then $vm = $pm1 * Number( $aWords[5] )
						If $v < $vm Then
							PrintError( "Function LoadObjQuads()" & @CRLF & _
								"ERROR: File """ & $sObjModel & ".obj""" & @CRLF & _
								"Line #" & $i & ": " & StringStripWS ( $aLines[$i], 2 ) & @CRLF & _
								"The number of vertex coordinates does not match the maximum v-index: " & $vm )
							If $bProgressInfo Then GUIDelete( $hProgress )
							Return SetError(7,0,0)
						EndIf

					Case $aWords[0] = 2 ; Vertex/texture: f v1/vt1 v2/vt2 v3/vt3 ...
						$iModelFaceAttribsTmp = 3
						$aWords = StringSplit( $line, "/ " )
						;_ArrayDisplay( $aWords )
						;Return 0
						; Find the maximum indices v, vt in the first f-line
						; The number of rows in $aVerticesTmp and $aTextureUVsTmp must at least match these maximums
						; $aVerticesTmp
						$vm = $pm1 * Number( $aWords[2] )
						If $vm < $pm1 * Number( $aWords[4] ) Then $vm = $pm1 * Number( $aWords[4] )
						If $vm < $pm1 * Number( $aWords[6] ) Then $vm = $pm1 * Number( $aWords[6] )
						If $vm < $pm1 * Number( $aWords[8] ) Then $vm = $pm1 * Number( $aWords[8] )
						If $v < $vm Then
							PrintError( "Function LoadObjQuads()" & @CRLF & _
								"ERROR: File """ & $sObjModel & ".obj""" & @CRLF & _
								"Line #" & $i & ": " & StringStripWS ( $aLines[$i], 2 ) & @CRLF & _
								"The number of vertex coordinates does not match the maximum v-index: " & $vm )
							If $bProgressInfo Then GUIDelete( $hProgress )
							Return SetError(8,0,0)
						EndIf
						; $aTextureUVsTmp
						If $bTexImage Then
							$vtm = $pm1 * Number( $aWords[3] )
							If $vtm < $pm1 * Number( $aWords[5] ) Then $vtm = $pm1 * Number( $aWords[5] )
							If $vtm < $pm1 * Number( $aWords[7] ) Then $vtm = $pm1 * Number( $aWords[7] )
							If $vtm < $pm1 * Number( $aWords[9] ) Then $vtm = $pm1 * Number( $aWords[9] )
							If $vt < $vtm Then
								PrintError( "Function LoadObjQuads()" & @CRLF & _
									"ERROR: File """ & $sObjModel & ".obj""" & @CRLF & _
									"Line #" & $i & ": " & StringStripWS ( $aLines[$i], 2 ) & @CRLF & _
									"The number of texture coordinates does not match the maximum vt-index: " & $vtm )
								If $bProgressInfo Then GUIDelete( $hProgress )
								Return SetError(9,0,0)
							EndIf
						EndIf

					Case $aWords[0] = 3 And  $aWords[2] = "" ; Vertex/normal: f v1//vn1 v2//vn2 v3//vn3 ...
						$iModelFaceAttribsTmp = 5
						$aWords = StringSplit( $line, "/ " )
						; Find the maximum indices v, vn in the first f-line
						; The number of rows in $aVerticesTmp and $aNormalsTmp must at least match these maximums
						; $aVerticesTmp
						$vm = $pm1 * Number( $aWords[2] )
						If $vm < $pm1 * Number( $aWords[5] ) Then $vm = $pm1 * Number( $aWords[5] )
						If $vm < $pm1 * Number( $aWords[8] ) Then $vm = $pm1 * Number( $aWords[8] )
						If $vm < $pm1 * Number( $aWords[11] ) Then $vm = $pm1 * Number( $aWords[11] )
						If $v < $vm Then
							PrintError( "Function LoadObjQuads()" & @CRLF & _
								"ERROR: File """ & $sObjModel & ".obj""" & @CRLF & _
								"Line #" & $i & ": " & StringStripWS ( $aLines[$i], 2 ) & @CRLF & _
								"The number of vertex coordinates does not match the maximum v-index: " & $vm )
							If $bProgressInfo Then GUIDelete( $hProgress )
							Return SetError(10,0,0)
						EndIf
						; $aNormalsTmp
						If $bUseNormals Then
							$vnm = $pm1 * Number( $aWords[4] )
							If $vnm < $pm1 * Number( $aWords[7] )  Then $vnm = $pm1 * Number( $aWords[7] )
							If $vnm < $pm1 * Number( $aWords[10] ) Then $vnm = $pm1 * Number( $aWords[10] )
							If $vnm < $pm1 * Number( $aWords[13] ) Then $vnm = $pm1 * Number( $aWords[13] )
							If $vn < $vnm Then
								PrintError( "Function LoadObjQuads()" & @CRLF & _
									"ERROR: File """ & $sObjModel & ".obj""" & @CRLF & _
									"Line #" & $i & ": " & StringStripWS ( $aLines[$i], 2 ) & @CRLF & _
									"The number of normal vectors does not match the maximum vn-index: " & $vnm )
								If $bProgressInfo Then GUIDelete( $hProgress )
								Return SetError(11,0,0)
							EndIf
						EndIf

					Case $aWords[0] = 3 ; Vertex/texture/normal: f v1/vt1/vn1 v2/vt2/vn2 v3/vt3/vn3 ...
						$iModelFaceAttribsTmp = 7
						$aWords = StringSplit( $line, "/ " )
						; Find the maximum indices v, vt, vn in the first f-line
						; The number of rows in $aVerticesTmp, $aTextureUVsTmp and $aNormalsTmp must at least match these maximums
						; $aVerticesTmp
						$vm = $pm1 * Number( $aWords[2] )
						If $vm < $pm1 * Number( $aWords[5] ) Then $vm = $pm1 * Number( $aWords[5] )
						If $vm < $pm1 * Number( $aWords[8] ) Then $vm = $pm1 * Number( $aWords[8] )
						If $vm < $pm1 * Number( $aWords[11] ) Then $vm = $pm1 * Number( $aWords[11] )
						If $v < $vm Then
							PrintError( "Function LoadObjQuads()" & @CRLF & _
								"ERROR: File """ & $sObjModel & ".obj""" & @CRLF & _
								"Line #" & $i & ": " & StringStripWS ( $aLines[$i], 2 ) & @CRLF & _
								"The number of vertex coordinates does not match the maximum v-index: " & $vm )
							If $bProgressInfo Then GUIDelete( $hProgress )
							Return SetError(12,0,0)
						EndIf
						; $aTextureUVsTmp
						If $bTexImage Then
							$vtm = $pm1 * Number( $aWords[3] )
							If $vtm < $pm1 * Number( $aWords[6] ) Then $vtm = $pm1 * Number( $aWords[6] )
							If $vtm < $pm1 * Number( $aWords[9] ) Then $vtm = $pm1 * Number( $aWords[9] )
							If $vtm < $pm1 * Number( $aWords[12] ) Then $vtm = $pm1 * Number( $aWords[12] )
							If $vt < $vtm Then
								PrintError( "Function LoadObjQuads()" & @CRLF & _
									"ERROR: File """ & $sObjModel & ".obj""" & @CRLF & _
									"Line #" & $i & ": " & StringStripWS ( $aLines[$i], 2 ) & @CRLF & _
									"The number of texture coordinates does not match the maximum vt-index: " & $vtm )
								If $bProgressInfo Then GUIDelete( $hProgress )
								Return SetError(13,0,0)
							EndIf
						EndIf
						; $aNormalsTmp
						If $bUseNormals Then
							$vnm = $pm1 * Number( $aWords[4] )
							If $vnm < $pm1 * Number( $aWords[7] )  Then $vnm = $pm1 * Number( $aWords[7] )
							If $vnm < $pm1 * Number( $aWords[10] ) Then $vnm = $pm1 * Number( $aWords[10] )
							If $vnm < $pm1 * Number( $aWords[13] ) Then $vnm = $pm1 * Number( $aWords[13] )
							If $vn < $vnm Then
								PrintError( "Function LoadObjQuads()" & @CRLF & _
									"ERROR: File """ & $sObjModel & ".obj""" & @CRLF & _
									"Line #" & $i & ": " & StringStripWS ( $aLines[$i], 2 ) & @CRLF & _
									"The number of normal vectors does not match the maximum vn-index: " & $vnm )
								If $bProgressInfo Then GUIDelete( $hProgress )
								Return SetError(14,0,0)
							EndIf
						EndIf

					Case Else
						PrintError( "Function LoadObjQuads()" & @CRLF & _
							"ERROR: File """ & $sObjModel & ".obj""" & @CRLF & _
							"Line #" & $i & ": " & StringStripWS( $aLines[$i], 2 ) & @CRLF & _
							"This is not a valid face quad line." )
						If $bProgressInfo Then GUIDelete( $hProgress )
						Return SetError(15,0,0)
				EndSelect

				; Only one type of face quad attributes is accepted in the entire OBJ-file
				If $iModelFaceAttribs = 0 Then
					$iModelFaceAttribs = $iModelFaceAttribsTmp
				ElseIf $iModelFaceAttribs <> $iModelFaceAttribsTmp Then
					PrintError( "Function LoadObjQuads()" & @CRLF & _
						"ERROR: File """ & $sObjModel & ".obj""" & @CRLF & _
						"Line #" & $i & ": " & StringStripWS( $aLines[$i], 2 ) & @CRLF & _
						"The file contains two or more different types of face quad attributes." & @CRLF & _
						"This is not handled by this version of the function." )
					If $bProgressInfo Then GUIDelete( $hProgress )
					Return SetError(16,0,0)
				EndIf

				If $bProgressInfo Then
					GUICtrlSetData( $idProLabel1, "Face quads: " & $p - 1 & "/" & $aProgressInfo[0][0] )
					UpdateModelProgressBar( $idProgress2, 0, $aProgressInfo[$p][1] )
				EndIf

				; --- Loop through each f-line in the OBJ-file ---

				Switch $iModelFaceAttribs
					Case 1    ; Vertex:                f v1 v2 v3 v4 ...
						Switch $bNormals
							Case 0
								; Get vertices to draw the model with glDrawArrays

								While StringLeft( $line, 2 ) = "f "
									Select
										Case $k > 8 And $f > ($k-8) * 50000 - 300 ; The number 300 takes into account that the number of points varies
											$k += 1                                 ; in a f-line. 300 accounts for 100 points (with x,y,z coordinates).
											Redim $aVertices[($k-8)*150000]
											If $bProgressInfo Then
												GUICtrlSetData( $idProLabel2, "Face quads: " & $f/4 & "/" & $aProgressInfo[$p][1] )
												UpdateModelProgressBar( $idProgress2, $f/4, $aProgressInfo[$p][1] )
												If $bModelLoadCancel Then
													ConsoleWrite( "Function LoadObjQuads() cancelled!" & @CRLF & @CRLF )
													If $bProgressInfo Then GUIDelete( $hProgress )
													Return 0
												EndIf
											EndIf
										Case $k < 9 And $k > 4 And $f > ($k-4) * 10000 - 300
											$k += 1
											Redim $aVertices[($k-4)*30000]
											If $bProgressInfo Then
												GUICtrlSetData( $idProLabel2, "Face quads: " & $f/4 & "/" & $aProgressInfo[$p][1] )
												UpdateModelProgressBar( $idProgress2, $f/4, $aProgressInfo[$p][1] )
												If $bModelLoadCancel Then
													ConsoleWrite( "Function LoadObjQuads() cancelled!" & @CRLF & @CRLF )
													If $bProgressInfo Then GUIDelete( $hProgress )
													Return 0
												EndIf
											EndIf
										Case $k < 5 And $f > $k * 2000 - 300
											$k += 1
											Redim $aVertices[$k*6000]
											If $bProgressInfo Then
												GUICtrlSetData( $idProLabel2, "Face quads: " & $f/4 & "/" & $aProgressInfo[$p][1] )
												UpdateModelProgressBar( $idProgress2, $f/4, $aProgressInfo[$p][1] )
												If $bModelLoadCancel Then
													ConsoleWrite( "Function LoadObjQuads() cancelled!" & @CRLF & @CRLF )
													If $bProgressInfo Then GUIDelete( $hProgress )
													Return 0
												EndIf
											EndIf
									EndSelect
									$line = StringStripWS( $line, 6 )
									$aWords = StringSplit( $line, "/ " )
									$iPnts = $aWords[0] - 1
									If $iPnts <> 4 Then
										PrintError( "Function LoadObjQuads()" & @CRLF & _
											"ERROR: File """ & $sObjModel & ".obj""" & @CRLF & _
											"Line #" & $i & ": " & StringStripWS ( $aLines[$i], 2 ) & @CRLF & _
											"The face line seems not to be in this format: ""f v1 v2 v3 v4""" )
										If $bProgressInfo Then GUIDelete( $hProgress )
										Return SetError(17,0,0)
									EndIf
									;ConsoleWrite( $i & ": " & $aLines[$i] & @CRLF )
									;_ArrayDisplay( $aWords )
									If $bPosRef Then ; Positive references
										; Add quad
										$j = 3 * $f
										$aVertices[$j+0] = $aVerticesTmp[$aWords[2]-1][0]
										$aVertices[$j+1] = $aVerticesTmp[$aWords[2]-1][1]
										$aVertices[$j+2] = $aVerticesTmp[$aWords[2]-1][2]
										$j += 3
										$aVertices[$j+0] = $aVerticesTmp[$aWords[3]-1][0]
										$aVertices[$j+1] = $aVerticesTmp[$aWords[3]-1][1]
										$aVertices[$j+2] = $aVerticesTmp[$aWords[3]-1][2]
										$j += 3
										$aVertices[$j+0] = $aVerticesTmp[$aWords[4]-1][0]
										$aVertices[$j+1] = $aVerticesTmp[$aWords[4]-1][1]
										$aVertices[$j+2] = $aVerticesTmp[$aWords[4]-1][2]
										$j += 3
										$aVertices[$j+0] = $aVerticesTmp[$aWords[5]-1][0]
										$aVertices[$j+1] = $aVerticesTmp[$aWords[5]-1][1]
										$aVertices[$j+2] = $aVerticesTmp[$aWords[5]-1][2]
										$f += 4
									Else ; Negative references
										; Add quad
										$j = 3 * $f
										$aVertices[$j+0] = $aVerticesTmp[$v+$aWords[2]][0]
										$aVertices[$j+1] = $aVerticesTmp[$v+$aWords[2]][1]
										$aVertices[$j+2] = $aVerticesTmp[$v+$aWords[2]][2]
										$j += 3
										$aVertices[$j+0] = $aVerticesTmp[$v+$aWords[3]][0]
										$aVertices[$j+1] = $aVerticesTmp[$v+$aWords[3]][1]
										$aVertices[$j+2] = $aVerticesTmp[$v+$aWords[3]][2]
										$j += 3
										$aVertices[$j+0] = $aVerticesTmp[$v+$aWords[4]][0]
										$aVertices[$j+1] = $aVerticesTmp[$v+$aWords[4]][1]
										$aVertices[$j+2] = $aVerticesTmp[$v+$aWords[4]][2]
										$j += 3
										$aVertices[$j+0] = $aVerticesTmp[$v+$aWords[5]][0]
										$aVertices[$j+1] = $aVerticesTmp[$v+$aWords[5]][1]
										$aVertices[$j+2] = $aVerticesTmp[$v+$aWords[5]][2]
										$f += 4
									EndIf
									If $i = $aLines[0] Then
										ExitLoop
									Else
										$i += 1
										$line = $aLines[$i]
										If StringLeft( $line, 2 ) = "s " Then
											$i += 1
											$line = $aLines[$i]
										EndIf
									EndIf
								WEnd

							Case 1
								; Get face quads to draw the model with glDrawElements

								While StringLeft( $line, 2 ) = "f "
									Select
										Case $k > 8 And $f > ($k-8) * 50000 - 300 ; The number 300 takes into account that the number of points varies
											$k += 1                                 ; in a f-line. 300 accounts for 100 points (with x,y,z coordinates).
											Redim $aFaces[($k-8)*50000]
											If $bProgressInfo Then
												GUICtrlSetData( $idProLabel2, "Face quads: " & $f/4 & "/" & $aProgressInfo[$p][1] )
												UpdateModelProgressBar( $idProgress2, $f/4, $aProgressInfo[$p][1] )
												If $bModelLoadCancel Then
													ConsoleWrite( "Function LoadObjQuads() cancelled!" & @CRLF & @CRLF )
													If $bProgressInfo Then GUIDelete( $hProgress )
													Return 0
												EndIf
											EndIf
										Case $k < 9 And $k > 4 And $f > ($k-4) * 10000 - 300
											$k += 1
											Redim $aFaces[($k-4)*10000]
											If $bProgressInfo Then
												GUICtrlSetData( $idProLabel2, "Face quads: " & $f/4 & "/" & $aProgressInfo[$p][1] )
												UpdateModelProgressBar( $idProgress2, $f/4, $aProgressInfo[$p][1] )
												If $bModelLoadCancel Then
													ConsoleWrite( "Function LoadObjQuads() cancelled!" & @CRLF & @CRLF )
													If $bProgressInfo Then GUIDelete( $hProgress )
													Return 0
												EndIf
											EndIf
										Case $k < 5 And $f > $k * 2000 - 300
											$k += 1
											Redim $aFaces[$k*2000]
											If $bProgressInfo Then
												GUICtrlSetData( $idProLabel2, "Face quads: " & $f/4 & "/" & $aProgressInfo[$p][1] )
												UpdateModelProgressBar( $idProgress2, $f/4, $aProgressInfo[$p][1] )
												If $bModelLoadCancel Then
													ConsoleWrite( "Function LoadObjQuads() cancelled!" & @CRLF & @CRLF )
													If $bProgressInfo Then GUIDelete( $hProgress )
													Return 0
												EndIf
											EndIf
									EndSelect
									$line = StringStripWS( $line, 6 )
									$aWords = StringSplit( $line, "/ " )
									$iPnts = $aWords[0] - 1
									If $iPnts <> 4 Then
										PrintError( "Function LoadObjQuads()" & @CRLF & _
											"ERROR: File """ & $sObjModel & ".obj""" & @CRLF & _
											"Line #" & $i & ": " & StringStripWS ( $aLines[$i], 2 ) & @CRLF & _
											"The face line seems not to be in this format: ""f v1 v2 v3 v4""" )
										If $bProgressInfo Then GUIDelete( $hProgress )
										Return SetError(18,0,0)
									EndIf
									If $bPosRef Then ; Positive references
										; Add quad
										$aFaces[$f+0] = $aWords[2]-1
										$aFaces[$f+1] = $aWords[3]-1
										$aFaces[$f+2] = $aWords[4]-1
										$aFaces[$f+3] = $aWords[5]-1
										$f += 4
									Else ; Negative references
										; Add quad
										$aFaces[$f+0] = $v+$aWords[2]
										$aFaces[$f+1] = $v+$aWords[3]
										$aFaces[$f+2] = $v+$aWords[4]
										$aFaces[$f+3] = $v+$aWords[5]
										$f += 4
									EndIf
									If $i = $aLines[0] Then
										ExitLoop
									Else
										$i += 1
										$line = $aLines[$i]
										If StringLeft( $line, 2 ) = "s " Then
											$i += 1
											$line = $aLines[$i]
										EndIf
									EndIf
								WEnd
						EndSwitch

					Case 3    ; Vertex/texture:        f v1/vt1 v2/vt2 v3/vt3 v4/vt4 ...
						While StringLeft( $line, 2 ) = "f "
							Select
								Case $k > 8 And $f > ($k-8) * 50000 - 300 ; The number 300 takes into account that the number of points varies
									$k += 1                                 ; in a f-line. 300 accounts for 100 points (with x,y,z coordinates).
									Redim $aVertices[($k-8)*150000]
									If $bTexImage Then _
										Redim $aTextureUVs[($k-8)*100000]
									If $bProgressInfo Then
										GUICtrlSetData( $idProLabel2, "Face quads: " & $f/4 & "/" & $aProgressInfo[$p][1] )
										UpdateModelProgressBar( $idProgress2, $f/4, $aProgressInfo[$p][1] )
										If $bModelLoadCancel Then
											ConsoleWrite( "Function LoadObjQuads() cancelled!" & @CRLF & @CRLF )
											If $bProgressInfo Then GUIDelete( $hProgress )
											Return 0
										EndIf
									EndIf
								Case $k < 9 And $k > 4 And $f > ($k-4) * 10000 - 300
									$k += 1
									Redim $aVertices[($k-4)*30000]
									If $bTexImage Then _
										Redim $aTextureUVs[($k-4)*20000]
									If $bProgressInfo Then
										GUICtrlSetData( $idProLabel2, "Face quads: " & $f/4 & "/" & $aProgressInfo[$p][1] )
										UpdateModelProgressBar( $idProgress2, $f/4, $aProgressInfo[$p][1] )
										If $bModelLoadCancel Then
											ConsoleWrite( "Function LoadObjQuads() cancelled!" & @CRLF & @CRLF )
											If $bProgressInfo Then GUIDelete( $hProgress )
											Return 0
										EndIf
									EndIf
								Case $k < 5 And $f > $k * 2000 - 300
									$k += 1
									Redim $aVertices[$k*6000]
									If $bTexImage Then _
										Redim $aTextureUVs[$k*4000]
									If $bProgressInfo Then
										GUICtrlSetData( $idProLabel2, "Face quads: " & $f/4 & "/" & $aProgressInfo[$p][1] )
										UpdateModelProgressBar( $idProgress2, $f/4, $aProgressInfo[$p][1] )
										If $bModelLoadCancel Then
											ConsoleWrite( "Function LoadObjQuads() cancelled!" & @CRLF & @CRLF )
											If $bProgressInfo Then GUIDelete( $hProgress )
											Return 0
										EndIf
									EndIf
							EndSelect
							$line = StringStripWS( $line, 6 )
							$aWords = StringSplit( $line, "/ " )
							;_ArrayDisplay( $aWords )
							$iPnts = ( $aWords[0] - 1 ) / 2
							If $iPnts <> 4 Then
								PrintError( "Function LoadObjQuads()" & @CRLF & _
									"ERROR: File """ & $sObjModel & ".obj""" & @CRLF & _
									"Line #" & $i & ": " & StringStripWS ( $aLines[$i], 2 ) & @CRLF & _
									"The face line seems not to be in this format: ""f v1/vt1 v2/vt2 v3/vt3 v4/vt4""" )
								If $bProgressInfo Then GUIDelete( $hProgress )
								Return SetError(19,0,0)
							EndIf
							If $bPosRef Then ; Positive references
								; Add quad
								$j = 3 * $f
								$aVertices[$j+0] = $aVerticesTmp[$aWords[2]-1][0]
								$aVertices[$j+1] = $aVerticesTmp[$aWords[2]-1][1]
								$aVertices[$j+2] = $aVerticesTmp[$aWords[2]-1][2]
								$j += 3
								$aVertices[$j+0] = $aVerticesTmp[$aWords[4]-1][0]
								$aVertices[$j+1] = $aVerticesTmp[$aWords[4]-1][1]
								$aVertices[$j+2] = $aVerticesTmp[$aWords[4]-1][2]
								$j += 3
								$aVertices[$j+0] = $aVerticesTmp[$aWords[6]-1][0]
								$aVertices[$j+1] = $aVerticesTmp[$aWords[6]-1][1]
								$aVertices[$j+2] = $aVerticesTmp[$aWords[6]-1][2]
								$j += 3
								$aVertices[$j+0] = $aVerticesTmp[$aWords[8]-1][0]
								$aVertices[$j+1] = $aVerticesTmp[$aWords[8]-1][1]
								$aVertices[$j+2] = $aVerticesTmp[$aWords[8]-1][2]
								If $bTexImage Then
									$j = 2 * $f
									$aTextureUVs[$j+0] = $aTextureUVsTmp[$aWords[3]-1][0]
									$aTextureUVs[$j+1] = $aTextureUVsTmp[$aWords[3]-1][1]
									$j += 2
									$aTextureUVs[$j+0] = $aTextureUVsTmp[$aWords[5]-1][0]
									$aTextureUVs[$j+1] = $aTextureUVsTmp[$aWords[5]-1][1]
									$j += 2
									$aTextureUVs[$j+0] = $aTextureUVsTmp[$aWords[7]-1][0]
									$aTextureUVs[$j+1] = $aTextureUVsTmp[$aWords[7]-1][1]
									$j += 2
									$aTextureUVs[$j+0] = $aTextureUVsTmp[$aWords[9]-1][0]
									$aTextureUVs[$j+1] = $aTextureUVsTmp[$aWords[9]-1][1]
								EndIf
								$f += 4
							Else ; Negative references
								; Add quad
								$j = 3 * $f
								$aVertices[$j+0] = $aVerticesTmp[$v+$aWords[2]][0]
								$aVertices[$j+1] = $aVerticesTmp[$v+$aWords[2]][1]
								$aVertices[$j+2] = $aVerticesTmp[$v+$aWords[2]][2]
								$j += 3
								$aVertices[$j+0] = $aVerticesTmp[$v+$aWords[4]][0]
								$aVertices[$j+1] = $aVerticesTmp[$v+$aWords[4]][1]
								$aVertices[$j+2] = $aVerticesTmp[$v+$aWords[4]][2]
								$j += 3
								$aVertices[$j+0] = $aVerticesTmp[$v+$aWords[6]][0]
								$aVertices[$j+1] = $aVerticesTmp[$v+$aWords[6]][1]
								$aVertices[$j+2] = $aVerticesTmp[$v+$aWords[6]][2]
								$j += 3
								$aVertices[$j+0] = $aVerticesTmp[$v+$aWords[8]][0]
								$aVertices[$j+1] = $aVerticesTmp[$v+$aWords[8]][1]
								$aVertices[$j+2] = $aVerticesTmp[$v+$aWords[8]][2]
								If $bTexImage Then
									$j = 2 * $f
									$aTextureUVs[$j+0] = $aTextureUVsTmp[$vt+$aWords[3]][0]
									$aTextureUVs[$j+1] = $aTextureUVsTmp[$vt+$aWords[3]][1]
									$j += 2
									$aTextureUVs[$j+0] = $aTextureUVsTmp[$vt+$aWords[5]][0]
									$aTextureUVs[$j+1] = $aTextureUVsTmp[$vt+$aWords[5]][1]
									$j += 2
									$aTextureUVs[$j+0] = $aTextureUVsTmp[$vt+$aWords[7]][0]
									$aTextureUVs[$j+1] = $aTextureUVsTmp[$vt+$aWords[7]][1]
									$j += 2
									$aTextureUVs[$j+0] = $aTextureUVsTmp[$vt+$aWords[9]][0]
									$aTextureUVs[$j+1] = $aTextureUVsTmp[$vt+$aWords[9]][1]
								EndIf
								$f += 4
							EndIf
							If $i = $aLines[0] Then
								ExitLoop
							Else
								$i += 1
								$line = $aLines[$i]
								If StringLeft( $line, 2 ) = "s " Then
									$i += 1
									$line = $aLines[$i]
								EndIf
							EndIf
						WEnd

					Case 5    ; Vertex/normal:         f v1//vn1 v2//vn2 v3//vn3 v4//vn4 ...
						While StringLeft( $line, 2 ) = "f "
							Select
								Case $k > 8 And $f > ($k-8) * 50000 - 300 ; The number 300 takes into account that the number of points varies
									$k += 1                                 ; in a f-line. 300 accounts for 100 points (with x,y,z coordinates).
									Redim $aVertices[($k-8)*150000]
									If $bUseNormals Then _
										Redim $aNormals[($k-8)*150000]
									If $bProgressInfo Then
										GUICtrlSetData( $idProLabel2, "Face quads: " & $f/4 & "/" & $aProgressInfo[$p][1] )
										UpdateModelProgressBar( $idProgress2, $f/4, $aProgressInfo[$p][1] )
										If $bModelLoadCancel Then
											ConsoleWrite( "Function LoadObjQuads() cancelled!" & @CRLF & @CRLF )
											If $bProgressInfo Then GUIDelete( $hProgress )
											Return 0
										EndIf
									EndIf
								Case $k < 9 And $k > 4 And $f > ($k-4) * 10000 - 300
									$k += 1
									Redim $aVertices[($k-4)*30000]
									If $bUseNormals Then _
										Redim $aNormals[($k-4)*30000]
									If $bProgressInfo Then
										GUICtrlSetData( $idProLabel2, "Face quads: " & $f/4 & "/" & $aProgressInfo[$p][1] )
										UpdateModelProgressBar( $idProgress2, $f/4, $aProgressInfo[$p][1] )
										If $bModelLoadCancel Then
											ConsoleWrite( "Function LoadObjQuads() cancelled!" & @CRLF & @CRLF )
											If $bProgressInfo Then GUIDelete( $hProgress )
											Return 0
										EndIf
									EndIf
								Case $k < 5 And $f > $k * 2000 - 300
									$k += 1
									Redim $aVertices[$k*6000]
									If $bUseNormals Then _
										Redim $aNormals[$k*6000]
									If $bProgressInfo Then
										GUICtrlSetData( $idProLabel2, "Face quads: " & $f/4 & "/" & $aProgressInfo[$p][1] )
										UpdateModelProgressBar( $idProgress2, $f/4, $aProgressInfo[$p][1] )
										If $bModelLoadCancel Then
											ConsoleWrite( "Function LoadObjQuads() cancelled!" & @CRLF & @CRLF )
											If $bProgressInfo Then GUIDelete( $hProgress )
											Return 0
										EndIf
									EndIf
							EndSelect
							$line = StringStripWS( $line, 6 )
							$aWords = StringSplit( $line, "/ " )
							$iPnts = ( $aWords[0] - 1 ) / 3
							If $iPnts <> 4 Then
								PrintError( "Function LoadObjQuads()" & @CRLF & _
									"ERROR: File """ & $sObjModel & ".obj""" & @CRLF & _
									"Line #" & $i & ": " & StringStripWS ( $aLines[$i], 2 ) & @CRLF & _
									"The face line seems not to be in this format: ""f v1//vn1 v2//vn2 v3//vn3 v4//vn4""" )
								If $bProgressInfo Then GUIDelete( $hProgress )
								Return SetError(20,0,0)
							EndIf
							If $bPosRef Then ; Positive references
								; Add quad
								$j = 3 * $f
								$aVertices[$j+0] = $aVerticesTmp[$aWords[2]-1][0]
								$aVertices[$j+1] = $aVerticesTmp[$aWords[2]-1][1]
								$aVertices[$j+2] = $aVerticesTmp[$aWords[2]-1][2]
								$j += 3
								$aVertices[$j+0] = $aVerticesTmp[$aWords[5]-1][0]
								$aVertices[$j+1] = $aVerticesTmp[$aWords[5]-1][1]
								$aVertices[$j+2] = $aVerticesTmp[$aWords[5]-1][2]
								$j += 3
								$aVertices[$j+0] = $aVerticesTmp[$aWords[8]-1][0]
								$aVertices[$j+1] = $aVerticesTmp[$aWords[8]-1][1]
								$aVertices[$j+2] = $aVerticesTmp[$aWords[8]-1][2]
								$j += 3
								$aVertices[$j+0] = $aVerticesTmp[$aWords[11]-1][0]
								$aVertices[$j+1] = $aVerticesTmp[$aWords[11]-1][1]
								$aVertices[$j+2] = $aVerticesTmp[$aWords[11]-1][2]
								If $bUseNormals Then
									$j = 3 * $f
									$aNormals[$j+0] = $aNormalsTmp[$aWords[4]-1][0]
									$aNormals[$j+1] = $aNormalsTmp[$aWords[4]-1][1]
									$aNormals[$j+2] = $aNormalsTmp[$aWords[4]-1][2]
									$j += 3
									$aNormals[$j+0] = $aNormalsTmp[$aWords[7]-1][0]
									$aNormals[$j+1] = $aNormalsTmp[$aWords[7]-1][1]
									$aNormals[$j+2] = $aNormalsTmp[$aWords[7]-1][2]
									$j += 3
									$aNormals[$j+0] = $aNormalsTmp[$aWords[10]-1][0]
									$aNormals[$j+1] = $aNormalsTmp[$aWords[10]-1][1]
									$aNormals[$j+2] = $aNormalsTmp[$aWords[10]-1][2]
									$j += 3
									$aNormals[$j+0] = $aNormalsTmp[$aWords[13]-1][0]
									$aNormals[$j+1] = $aNormalsTmp[$aWords[13]-1][1]
									$aNormals[$j+2] = $aNormalsTmp[$aWords[13]-1][2]
								EndIf
								$f += 4
							Else ; Negative references
								; Add quad
								$j = 3 * $f
								$aVertices[$j+0] = $aVerticesTmp[$v+$aWords[2]][0]
								$aVertices[$j+1] = $aVerticesTmp[$v+$aWords[2]][1]
								$aVertices[$j+2] = $aVerticesTmp[$v+$aWords[2]][2]
								$j += 3
								$aVertices[$j+0] = $aVerticesTmp[$v+$aWords[5]][0]
								$aVertices[$j+1] = $aVerticesTmp[$v+$aWords[5]][1]
								$aVertices[$j+2] = $aVerticesTmp[$v+$aWords[5]][2]
								$j += 3
								$aVertices[$j+0] = $aVerticesTmp[$v+$aWords[8]][0]
								$aVertices[$j+1] = $aVerticesTmp[$v+$aWords[8]][1]
								$aVertices[$j+2] = $aVerticesTmp[$v+$aWords[8]][2]
								$j += 3
								$aVertices[$j+0] = $aVerticesTmp[$v+$aWords[11]][0]
								$aVertices[$j+1] = $aVerticesTmp[$v+$aWords[11]][1]
								$aVertices[$j+2] = $aVerticesTmp[$v+$aWords[11]][2]
								If $bUseNormals Then
									$j = 3 * $f
									$aNormals[$j+0] = $aNormalsTmp[$vn+$aWords[4]][0]
									$aNormals[$j+1] = $aNormalsTmp[$vn+$aWords[4]][1]
									$aNormals[$j+2] = $aNormalsTmp[$vn+$aWords[4]][2]
									$j += 3
									$aNormals[$j+0] = $aNormalsTmp[$vn+$aWords[7]][0]
									$aNormals[$j+1] = $aNormalsTmp[$vn+$aWords[7]][1]
									$aNormals[$j+2] = $aNormalsTmp[$vn+$aWords[7]][2]
									$j += 3
									$aNormals[$j+0] = $aNormalsTmp[$vn+$aWords[10]][0]
									$aNormals[$j+1] = $aNormalsTmp[$vn+$aWords[10]][1]
									$aNormals[$j+2] = $aNormalsTmp[$vn+$aWords[10]][2]
									$j += 3
									$aNormals[$j+0] = $aNormalsTmp[$vn+$aWords[13]][0]
									$aNormals[$j+1] = $aNormalsTmp[$vn+$aWords[13]][1]
									$aNormals[$j+2] = $aNormalsTmp[$vn+$aWords[13]][2]
								EndIf
								$f += 4
							EndIf
							If $i = $aLines[0] Then
								ExitLoop
							Else
								$i += 1
								$line = $aLines[$i]
								If StringLeft( $line, 2 ) = "s " Then
									$i += 1
									$line = $aLines[$i]
								EndIf
							EndIf
						WEnd

					Case 7    ; Vertex/texture/normal: f v1/vt1/vn1 v2/vt2/vn2 v3/vt3/vn3 ...
						While StringLeft( $line, 2 ) = "f "
							Select
								Case $k > 8 And $f > ($k-8) * 50000 - 300 ; The number 300 takes into account that the number of points varies
									$k += 1                                 ; in a f-line. 300 accounts for 100 points (with x,y,z coordinates).
									Redim $aVertices[($k-8)*150000]
									If $bTexImage Then _
										Redim $aTextureUVs[($k-8)*100000]
									If $bUseNormals Then _
										Redim $aNormals[($k-8)*150000]
									If $bProgressInfo Then
										GUICtrlSetData( $idProLabel2, "Face quads: " & $f/4 & "/" & $aProgressInfo[$p][1] )
										UpdateModelProgressBar( $idProgress2, $f/4, $aProgressInfo[$p][1] )
										If $bModelLoadCancel Then
											ConsoleWrite( "Function LoadObjQuads() cancelled!" & @CRLF & @CRLF )
											If $bProgressInfo Then GUIDelete( $hProgress )
											Return 0
										EndIf
									EndIf
								Case $k < 9 And $k > 4 And $f > ($k-4) * 10000 - 300
									$k += 1
									Redim $aVertices[($k-4)*30000]
									If $bTexImage Then _
										Redim $aTextureUVs[($k-4)*20000]
									If $bUseNormals Then _
										Redim $aNormals[($k-4)*30000]
									If $bProgressInfo Then
										GUICtrlSetData( $idProLabel2, "Face quads: " & $f/4 & "/" & $aProgressInfo[$p][1] )
										UpdateModelProgressBar( $idProgress2, $f/4, $aProgressInfo[$p][1] )
										If $bModelLoadCancel Then
											ConsoleWrite( "Function LoadObjQuads() cancelled!" & @CRLF & @CRLF )
											If $bProgressInfo Then GUIDelete( $hProgress )
											Return 0
										EndIf
									EndIf
								Case $k < 5 And $f > $k * 2000 - 300
									$k += 1
									Redim $aVertices[$k*6000]
									If $bTexImage Then _
										Redim $aTextureUVs[$k*4000]
									If $bUseNormals Then _
										Redim $aNormals[$k*6000]
									If $bProgressInfo Then
										GUICtrlSetData( $idProLabel2, "Face quads: " & $f/4 & "/" & $aProgressInfo[$p][1] )
										UpdateModelProgressBar( $idProgress2, $f/4, $aProgressInfo[$p][1] )
										If $bModelLoadCancel Then
											ConsoleWrite( "Function LoadObjQuads() cancelled!" & @CRLF & @CRLF )
											If $bProgressInfo Then GUIDelete( $hProgress )
											Return 0
										EndIf
									EndIf
							EndSelect
							$line = StringStripWS( $line, 6 )
							$aWords = StringSplit( $line, "/ " )
							$iPnts = ( $aWords[0] - 1 ) / 3
							If $iPnts <> 4 Then
								PrintError( "Function LoadObjQuads()" & @CRLF & _
									"ERROR: File """ & $sObjModel & ".obj""" & @CRLF & _
									"Line #" & $i & ": " & StringStripWS ( $aLines[$i], 2 ) & @CRLF & _
									"The face line seems not to be in this format: ""f v1/vt1/vn1 v2/vt2/vn2 v3/vt3/vn3 v4/vt4/vn4""" )
								If $bProgressInfo Then GUIDelete( $hProgress )
								Return SetError(21,0,0)
							EndIf
							;_ArrayDisplay( $aWords )
							;ConsoleWrite( $i & ": " & $aLines[$i] & @CRLF )
							If $bPosRef Then ; Positive references
								; Add quad
								$j = 3 * $f
								$aVertices[$j+0] = $aVerticesTmp[$aWords[2]-1][0]
								$aVertices[$j+1] = $aVerticesTmp[$aWords[2]-1][1]
								$aVertices[$j+2] = $aVerticesTmp[$aWords[2]-1][2]
								$j += 3
								$aVertices[$j+0] = $aVerticesTmp[$aWords[5]-1][0]
								$aVertices[$j+1] = $aVerticesTmp[$aWords[5]-1][1]
								$aVertices[$j+2] = $aVerticesTmp[$aWords[5]-1][2]
								$j += 3
								$aVertices[$j+0] = $aVerticesTmp[$aWords[8]-1][0]
								$aVertices[$j+1] = $aVerticesTmp[$aWords[8]-1][1]
								$aVertices[$j+2] = $aVerticesTmp[$aWords[8]-1][2]
								$j += 3
								$aVertices[$j+0] = $aVerticesTmp[$aWords[11]-1][0]
								$aVertices[$j+1] = $aVerticesTmp[$aWords[11]-1][1]
								$aVertices[$j+2] = $aVerticesTmp[$aWords[11]-1][2]
								If $bTexImage Then
									$j = 2 * $f
									$aTextureUVs[$j+0] = $aTextureUVsTmp[$aWords[3]-1][0]
									$aTextureUVs[$j+1] = $aTextureUVsTmp[$aWords[3]-1][1]
									$j += 2
									$aTextureUVs[$j+0] = $aTextureUVsTmp[$aWords[6]-1][0]
									$aTextureUVs[$j+1] = $aTextureUVsTmp[$aWords[6]-1][1]
									$j += 2
									$aTextureUVs[$j+0] = $aTextureUVsTmp[$aWords[9]-1][0]
									$aTextureUVs[$j+1] = $aTextureUVsTmp[$aWords[9]-1][1]
									$j += 2
									$aTextureUVs[$j+0] = $aTextureUVsTmp[$aWords[12]-1][0]
									$aTextureUVs[$j+1] = $aTextureUVsTmp[$aWords[12]-1][1]
								EndIf
								If $bUseNormals Then
									$j = 3 * $f
									$aNormals[$j+0] = $aNormalsTmp[$aWords[4]-1][0]
									$aNormals[$j+1] = $aNormalsTmp[$aWords[4]-1][1]
									$aNormals[$j+2] = $aNormalsTmp[$aWords[4]-1][2]
									$j += 3
									$aNormals[$j+0] = $aNormalsTmp[$aWords[7]-1][0]
									$aNormals[$j+1] = $aNormalsTmp[$aWords[7]-1][1]
									$aNormals[$j+2] = $aNormalsTmp[$aWords[7]-1][2]
									$j += 3
									$aNormals[$j+0] = $aNormalsTmp[$aWords[10]-1][0]
									$aNormals[$j+1] = $aNormalsTmp[$aWords[10]-1][1]
									$aNormals[$j+2] = $aNormalsTmp[$aWords[10]-1][2]
									$j += 3
									$aNormals[$j+0] = $aNormalsTmp[$aWords[13]-1][0]
									$aNormals[$j+1] = $aNormalsTmp[$aWords[13]-1][1]
									$aNormals[$j+2] = $aNormalsTmp[$aWords[13]-1][2]
								EndIf
								$f += 4
							Else ; Negative references
								; Add quad
								$j = 3 * $f
								$aVertices[$j+0] = $aVerticesTmp[$v+$aWords[2]][0]
								$aVertices[$j+1] = $aVerticesTmp[$v+$aWords[2]][1]
								$aVertices[$j+2] = $aVerticesTmp[$v+$aWords[2]][2]
								$j += 3
								$aVertices[$j+0] = $aVerticesTmp[$v+$aWords[5]][0]
								$aVertices[$j+1] = $aVerticesTmp[$v+$aWords[5]][1]
								$aVertices[$j+2] = $aVerticesTmp[$v+$aWords[5]][2]
								$j += 3
								$aVertices[$j+0] = $aVerticesTmp[$v+$aWords[8]][0]
								$aVertices[$j+1] = $aVerticesTmp[$v+$aWords[8]][1]
								$aVertices[$j+2] = $aVerticesTmp[$v+$aWords[8]][2]
								$j += 3
								$aVertices[$j+0] = $aVerticesTmp[$v+$aWords[11]][0]
								$aVertices[$j+1] = $aVerticesTmp[$v+$aWords[11]][1]
								$aVertices[$j+2] = $aVerticesTmp[$v+$aWords[11]][2]
								If $bTexImage Then
									$j = 2 * $f
									$aTextureUVs[$j+0] = $aTextureUVsTmp[$vt+$aWords[3]][0]
									$aTextureUVs[$j+1] = $aTextureUVsTmp[$vt+$aWords[3]][1]
									$j += 2
									$aTextureUVs[$j+0] = $aTextureUVsTmp[$vt+$aWords[6]][0]
									$aTextureUVs[$j+1] = $aTextureUVsTmp[$vt+$aWords[6]][1]
									$j += 2
									$aTextureUVs[$j+0] = $aTextureUVsTmp[$vt+$aWords[9]][0]
									$aTextureUVs[$j+1] = $aTextureUVsTmp[$vt+$aWords[9]][1]
									$j += 2
									$aTextureUVs[$j+0] = $aTextureUVsTmp[$vt+$aWords[12]][0]
									$aTextureUVs[$j+1] = $aTextureUVsTmp[$vt+$aWords[12]][1]
								EndIf
								If $bUseNormals Then
									$j = 3 * $f
									$aNormals[$j+0] = $aNormalsTmp[$vn+$aWords[4]][0]
									$aNormals[$j+1] = $aNormalsTmp[$vn+$aWords[4]][1]
									$aNormals[$j+2] = $aNormalsTmp[$vn+$aWords[4]][2]
									$j += 3
									$aNormals[$j+0] = $aNormalsTmp[$vn+$aWords[7]][0]
									$aNormals[$j+1] = $aNormalsTmp[$vn+$aWords[7]][1]
									$aNormals[$j+2] = $aNormalsTmp[$vn+$aWords[7]][2]
									$j += 3
									$aNormals[$j+0] = $aNormalsTmp[$vn+$aWords[10]][0]
									$aNormals[$j+1] = $aNormalsTmp[$vn+$aWords[10]][1]
									$aNormals[$j+2] = $aNormalsTmp[$vn+$aWords[10]][2]
									$j += 3
									$aNormals[$j+0] = $aNormalsTmp[$vn+$aWords[13]][0]
									$aNormals[$j+1] = $aNormalsTmp[$vn+$aWords[13]][1]
									$aNormals[$j+2] = $aNormalsTmp[$vn+$aWords[13]][2]
								EndIf
								$f += 4
							EndIf
							If $i = $aLines[0] Then
								ExitLoop
							Else
								$i += 1
								$line = $aLines[$i]
								If StringLeft( $line, 2 ) = "s " Then
									$i += 1
									$line = $aLines[$i]
								EndIf
							EndIf
						WEnd
				EndSwitch

				If $bModelLoadCancel Then
					ConsoleWrite( "Function LoadObjQuads() cancelled!" & @CRLF & @CRLF )
					If $bProgressInfo Then GUIDelete( $hProgress )
					Return 0
				EndIf

				If $i < $aLines[0] Then $i -= 1

				ConsoleWrite( "Face quads:  " & $f/4 - $f0/4 & @CRLF )
				ConsoleWrite( "Total quads: " & $f/4 & @CRLF )
				If $bProgressInfo Then
					UpdateModelProgressBar( $idProgress2, $f/4, $aProgressInfo[$p][1] )
					UpdateModelProgressBar( $idProgress1, $p, $aProgressInfo[0][0] )
					$aNewProgressInfo[$p][0] = "Face quads"
					$aNewProgressInfo[$p][1] = $f/4
					$p += 1
				EndIf

				; The end of a faces section indicates that an object in the model is finalized.
				; There can be more faces sections in an OBJ-file. In that case the model is composed of more objects.
				; (In this context an object means a group of quads drawn with a specific material and a single drawing command.)

				; If this object uses the same material as the last object, just add this object to the previous object.
				; This means add the new quads to the previous quads and draw them all with the same material and one drawing command.

				; Store information necessary to draw the current object

				If $iObjects > 0 And $sUseMaterial = $sLastMaterial Then

					; Extend previous object

					; Number of vertices for this object
					$aVertCnts[$iObjects-1] += $f - $f0

					ConsoleWrite( "Adding quads to object number " & $iObjects & @CRLF )
					If $iObjects > 9 Then
						ConsoleWrite( "--------------------------------" & @CRLF )
					Else
						ConsoleWrite( "-------------------------------" & @CRLF )
					EndIf

				Else

					; Add new object

					If Mod( $iObjects, 10 ) = 0 Then
						ReDim $aVertCnts[$iObjects+10]
						ReDim $aMaterials[$iObjects+11]
					EndIf

					; Number of vertices for this object
					$aVertCnts[$iObjects] = $f - $f0

					; Material name
					If $sUseMaterial <> "" Then $aMaterials[0] += 1
					$aMaterials[$iObjects+1] = $sUseMaterial
					$sLastMaterial = $sUseMaterial

					$iObjects += 1

					ConsoleWrite( "Object number " & $iObjects & " is finished" & @CRLF )
					If $iObjects > 9 Then
						ConsoleWrite( "----------------------------" & @CRLF )
					Else
						ConsoleWrite( "---------------------------" & @CRLF )
					EndIf

				EndIf

				$f0 = $f


			; --- Line and vertex reference numbers ---

			; Not handled
			; Just loop through this part of the OBJ-file as fast as possible

			Case "l"
				$line = $aLines[$i]
				While StringLeft( $line, 2 ) = "l "
					$i += 1
					$line = $aLines[$i]
				WEnd
				$i -= 1


			; --- Not matching tokens ---

			Case Else
				$n += 1
				;$aNotMatching[$n][0] = $i
				;$aNotMatching[$n][1] = $aLines[$i]

		EndSwitch

		If $bModelLoadCancel Then
			ConsoleWrite( "Function LoadObjQuads() cancelled!" & @CRLF & @CRLF )
			If $bProgressInfo Then GUIDelete( $hProgress )
			Return 0
		EndIf

	Next

	If $v = 0 Or $f = 0 Then
		PrintError( "Function LoadObjQuads()" & @CRLF & _
			"ERROR: File """ & $sObjModel & ".obj""" & @CRLF & _
			"The file contains no vertices or face quads." )
		If $bProgressInfo Then GUIDelete( $hProgress )
		Return SetError(22,0,0)
	EndIf

	; $aModel[7] = status
	If $bTexImage Then
		If BitAND( $iModelFaceAttribs, 2 ) Then _ ; True if any texture coordinates
			$aModel[7] += 2
	EndIf
	If $bUseNormals Then
		If BitAND( $iModelFaceAttribs, 4 ) Then _ ; True if any normal vectors
			$aModel[7] += 4
	EndIf
	$aModel[8] = $iModelFaceAttribs

	; Mesh type
	Local $sMeshType = "Quads"
	$aModel[9] = $sMeshType

	; OpenGL drawing command
	; Face attribute type
	Local $sCmd, $sAttribs
	Switch $iModelFaceAttribs
		Case 1
			Switch $bNormals
				Case 0
					$sCmd = "glDrawArrays"
				Case 1
					$sCmd = "glDrawElements"
			EndSwitch
			$sAttribs = "f v1 v2 v3 v4"
		Case 3
			$sCmd = "glDrawArrays"
			$sAttribs = "f v1/vt1 v2/vt2 v3/vt3 v4/vt4"
		Case 5
			$sCmd = "glDrawArrays"
			$sAttribs = "f v1//vn1 v2//vn2 v3//vn3 v4//vn4"
		Case 7
			$sCmd = "glDrawArrays"
			$sAttribs = "f v1/vt1/vn1 v2/vt2/vn2 v3/vt3/vn3 v4/vt4/vn4"
	EndSwitch

	; Model info
	ConsoleWrite( "Not matching lines: " & $n & @CRLF )
	ConsoleWrite( "Model mesh type: " & $sMeshType & @CRLF )
	ConsoleWrite( "Face attributes: " & $sAttribs & @CRLF )
	ConsoleWrite( "Total vertex coordinates: " & $v & @CRLF )
	If $bTexImage Then
		ConsoleWrite( "Total texture coordinates: " & $vt & @CRLF )
	Else
		ConsoleWrite( "Total texture coordinates: 0" & @CRLF )
	EndIf
	If $bUseNormals Then
		ConsoleWrite( "Total normal vectors: " & $vn & @CRLF )
	Else
		ConsoleWrite( "Total normal vectors: 0" & @CRLF )
	EndIf
	ConsoleWrite( "Total face quads: " & $f/4 & @CRLF )
	ConsoleWrite( "Objects in model: " & $iObjects & @CRLF )
	;ReDim $aNotMatching[$n+1][2]
	;_ArrayDisplay( $aNotMatching, "$aNotMatching" )
	;MsgBox( 0, "$iObjects", $aModel[13] )

	; ReDim arrays
	Local $iVertices, $iQuads = $f / 4
	If $iModelFaceAttribs = 1 Then
		Switch $bNormals
			Case 0
				$iVertices = $f
				ReDim $aVertices[3*$iVertices] ; x,y,z for vertex coordinates
				ReDim $aNormals[3*$iVertices]  ; x,y,z for normal vectors
			Case 1
				$iVertices = $v
				ReDim $aFaces[4*$iQuads]       ; 4 set of coordinates per quad
				ReDim $aVertices[3*$iVertices] ; x,y,z for vertex coordinates
				ReDim $aNormals[3*$iVertices]  ; x,y,z for normal vectors
		EndSwitch
	Else
		$iVertices = $f
		If BitAND( $iModelFaceAttribs, 1 )   Then ReDim $aVertices[3*$iVertices]   ; x,y,z for vertex coordinates
		If $bTexImage Then
			If BitAND( $iModelFaceAttribs, 2 ) Then ReDim $aTextureUVs[2*$iVertices] ; u,v for texture coordinates
		EndIf
		If $bUseNormals Then
			If BitAND( $iModelFaceAttribs, 4 ) Then ReDim $aNormals[3*$iVertices]    ; x,y,z for normal vectors
		EndIf
	EndIf

	If $bModelLoadCancel Then
		ConsoleWrite( "Function LoadObjQuads() cancelled!" & @CRLF & @CRLF )
		If $bProgressInfo Then GUIDelete( $hProgress )
		Return 0
	EndIf

	; ====== Model data, main loop finished ======


	; --- Sort objects by material ---

	; The purpose of this sort is to reduce the number of objects and optimize the model drawing function.

	; The model is composed of a number of objects. An object is a group of vertices, texture coordinates and
	; normal vectors which uses the same material and can be drawn with one single OpenGL drawing command. The
	; DrawModel function in the viewers draws a model by drawing one object at a time in a loop. This is done
	; by setting up the material definitions and then drawing the object. The more objects the more loops. If
	; the model is composed of more objects with the same material definitions, the number of objects can be
	; decreased by putting these objects together to one (big) object. And this decreases the number of loops.

	ReDim $aVertCnts[$iObjects+1]
	ReDim $aMaterials[$iObjects+2]
	;_ArrayDisplay( $aVertCnts, "$aVertCnts" )
	;_ArrayDisplay( $aVertCnts, "$aVertCnts before" )

	;#cs
	If $aMaterials[0] > 1 Then

		Local $aMtlsOrder[$iObjects+1]
		Local $aMtlsSort[$iObjects+1][2]
		Local $aMtlsIdx[$iObjects+1][$iObjects+2]

		; The two first materials
		Local $s, $n = 0
		For $i = 0 To $iObjects - 1
			$s = $aMaterials[$i+1]
			If $s = "" Then ContinueLoop
			If $n = 0 Then
				$aMtlsOrder[0] = $s
				$aMtlsSort[0][0] = $s
				$aMtlsSort[0][1] = $i
				$aMtlsIdx[0][0] = 1
				$aMtlsIdx[0][1] = $i
				$n += 1
			ElseIf $n = 1 Then
				If $s < $aMtlsSort[0][0] Then
					$aMtlsOrder[1] = $s
					$aMtlsSort[1][0] = $aMtlsSort[0][0]
					$aMtlsSort[1][1] = $aMtlsSort[0][1]
					$aMtlsIdx[1][0] = $aMtlsIdx[0][0]
					$aMtlsIdx[1][1] = $aMtlsIdx[0][1]
					$aMtlsSort[0][0] = $s
					$aMtlsSort[0][1] = $i
					$aMtlsIdx[0][0] = 1
					$aMtlsIdx[0][1] = $i
					$i += 1
					$n += 1
					ExitLoop
				ElseIf $s > $aMtlsSort[0][0] Then
					$aMtlsOrder[1] = $s
					$aMtlsSort[1][0] = $s
					$aMtlsSort[1][1] = $i
					$aMtlsIdx[1][0] = 1
					$aMtlsIdx[1][1] = $i
					$i += 1
					$n += 1
					ExitLoop
				Else
					$aMtlsIdx[0][0] += 1
					$aMtlsIdx[0][$aMtlsIdx[0][0]] = $i
				EndIf
			EndIf
		Next

		; The rest of the materials
		Local $lo, $hi, $mi, $gt, $l, $m
		For $j = $i To $iObjects - 1
			$s = $aMaterials[$j+1]
			If $s = "" Then ContinueLoop

			If $s < $aMtlsSort[0][0] Then
				; Smaller than first element
				$aMtlsOrder[$n] = $s
				For $k = $n - 1 To 0 Step -1
					$aMtlsSort[$k+1][0] = $aMtlsSort[$k][0]
					$aMtlsSort[$k+1][1] = $aMtlsSort[$k][1]
					For $m = 0 To $aMtlsIdx[$k][0]
						$aMtlsIdx[$k+1][$m] = $aMtlsIdx[$k][$m]
					Next
				Next
				$aMtlsSort[0][0] = $s
				$aMtlsSort[0][1] = $j
				$aMtlsIdx[0][0] = 1
				$aMtlsIdx[0][1] = $j
				$n += 1
				ContinueLoop
			ElseIf $s > $aMtlsSort[$n-1][0] Then
				; Larger than last element
				$aMtlsOrder[$n] = $s
				$aMtlsSort[$n][0] = $s
				$aMtlsSort[$n][1] = $j
				$aMtlsIdx[$n][0] = 1
				$aMtlsIdx[$n][1] = $j
				$n += 1
				ContinueLoop
			EndIf

			; Binary search
			$lo = 0
			$hi = $n - 1
			While $lo <= $hi
				$mi = Int( ( $lo + $hi ) / 2 )
				$l = StringCompare( $s, $aMtlsSort[$mi][0], 2 )
				If $l = -1 Then
					$gt = 0
					$hi = $mi - 1
				ElseIf $l = 1 Then
					$gt = 1
					$lo = $mi + 1
				Else
					; Material found
					; Add it to the list
					$aMtlsIdx[$mi][0] += 1
					$aMtlsIdx[$mi][$aMtlsIdx[$mi][0]] = $j
					ExitLoop
				EndIf
			WEnd
			If $l = 0 Then ContinueLoop

			; New material
			; Make space for the new material in the arrays
			For $l = $n To $mi + 1 Step -1
				$aMtlsSort[$l][0] = $aMtlsSort[$l-1][0]
				$aMtlsSort[$l][1] = $aMtlsSort[$l-1][1]
				For $m = 0 To $aMtlsIdx[$l-1][0]
					$aMtlsIdx[$l][$m] = $aMtlsIdx[$l-1][$m]
				Next
			Next

			; Insert new material
			$aMtlsOrder[$n] = $s
			$aMtlsSort[$mi+$gt][0] = $s
			$aMtlsSort[$mi+$gt][1] = $j
			$aMtlsIdx[$mi+$gt][0] = 1
			$aMtlsIdx[$mi+$gt][1] = $j
			$n += 1
		Next
		;_ArrayDisplay( $aVertCnts, "$aVertCnts" )
		;_ArrayDisplay( $aMtlsOrder, "$aMtlsOrder" )
		;_ArrayDisplay( $aMtlsSort, "$aMtlsSort" )
		;_ArrayDisplay( $aMtlsIdx, "$aMtlsIdx" )

		If $n > 1 Then

			; Check if at least one material is used more than once
			$j = False
			For $i = 0 To $n - 1
				If $aMtlsIdx[$i][0] > 1 Then
					$j = True
					ExitLoop
				EndIf
			Next

			If $j Then

				; Rebuild arrays
				; Tmp-arrays are used for rebuild

				If $bProgressInfo Then
					GUICtrlSetData( $idProLabel1, "Sorting objects: " & $p - 1 & "/" & $aProgressInfo[0][0] )
					UpdateModelProgressBar( $idProgress2, 0, $aProgressInfo[$p][1] )
				EndIf

				Local $aVertCntsTmp[$n]                 ; Use $aVertCntsTmp to rebuild $aVertCnts
				Local $aMaterialsTmp[$n+1]              ; Use $aMaterialsTmp to rebuild $aMaterials
				$aMaterialsTmp[0] = $n                  ; Number of materials

				; Vars to count the number of faces, vertices, texture coordinates and normals
				Local $f = 0, $f0 = 0, $v = 0, $v0 = 0, $vt = 0, $vt0 = 0, $vn = 0, $vn0 = 0

				If $iModelFaceAttribs = 1 And $bNormals = 1 Then
					ReDim $aFacesTmp[4*$iQuads]           ; Use $aFacesTmp to rebuild $aFaces
				Else
					ReDim $aVerticesTmp[3*$iVertices]     ; Use $aVerticesTmp to rebuild $aVertices
					If $bTexImage And BitAND( $iModelFaceAttribs, 2 ) Then _
						ReDim $aTextureUVsTmp[2*$iVertices] ; Use $aTextureUVsTmp to rebuild $aTextureUVs
					If $bUseNormals And BitAND( $iModelFaceAttribs, 4 ) Then _
						ReDim $aNormalsTmp[3*$iVertices]    ; Use $aNormalsTmp to rebuild $aNormals
				EndIf

				; Calculate start of vertices per object
				$aVertStart[0] = 0
				ReDim $aVertStart[$iObjects]
				For $i = 0 To $iObjects - 2
					$aVertStart[$i+1] = $aVertStart[$i] + $aVertCnts[$i]
				Next
				;_ArrayDisplay( $aVertCnts, "$aVertCnts" )
				;_ArrayDisplay( $aVertStart, "$aVertStart" )

				For $i = 0 To $n - 1
					$s = $aMtlsOrder[$i]
					; $aMtlsOrder is the order of the materials as they are used in
					; the model for the first time. We still want to use this order.

					; Binary search
					; Find the index of the material in $aMtlsSort and $aMtlsIdx
					$lo = 0
					$hi = $n - 1
					While $lo <= $hi
						$mi = Int( ( $lo + $hi ) / 2 )
						$l = StringCompare( $s, $aMtlsSort[$mi][0], 2 )
						If $l = -1 Then
							$hi = $mi - 1
						ElseIf $l = 1 Then
							$lo = $mi + 1
						Else
							ExitLoop
						EndIf
					WEnd

					If $aMtlsSort[$mi][1] = $aMtlsIdx[$mi][1] Then

						; All objects in the row $aMtlsIdx[$mi] uses the same
						; material. Put these objects together into one object.

						; $aVertices, $aTextureUVs, $aNormals
						For $j = 1 To $aMtlsIdx[$mi][0]
							$k = $aMtlsIdx[$mi][$j]  ; $k is index in $aVertCnts
							$l = $aVertStart[$k]     ; $l is first face/vertex
							$m = $aVertCnts[$k]      ; $m is number of faces/vertices in the object
							;ConsoleWrite( "$mi, $j, $k, $l, $m = " & $mi & ", " & $j & ", " & $k & ", " & $l & ", " & $m & @CRLF )
							If $iModelFaceAttribs = 1 And $bNormals = 1 Then
								For $a = 0 To $m - 1
									$aFacesTmp[$f] = $aFaces[$l+$a]
									$f += 1
								Next
							Else
								; $aVertices
								For $a = 0 To $m - 1
									$aVerticesTmp[3*$v+0] = $aVertices[3*($l+$a)+0]
									$aVerticesTmp[3*$v+1] = $aVertices[3*($l+$a)+1]
									$aVerticesTmp[3*$v+2] = $aVertices[3*($l+$a)+2]
									$v += 1
								Next
								If $bTexImage And BitAND( $iModelFaceAttribs, 2 ) Then
									; $aTextureUVs
									For $a = 0 To $m - 1
										$aTextureUVsTmp[2*$vt+0] = $aTextureUVs[2*($l+$a)+0]
										$aTextureUVsTmp[2*$vt+1] = $aTextureUVs[2*($l+$a)+1]
										$vt += 1
									Next
								EndIf
								If $bUseNormals And BitAND( $iModelFaceAttribs, 4 ) Then
									; $aNormals
									For $a = 0 To $m - 1
										$aNormalsTmp[3*$vn+0] = $aNormals[3*($l+$a)+0]
										$aNormalsTmp[3*$vn+1] = $aNormals[3*($l+$a)+1]
										$aNormalsTmp[3*$vn+2] = $aNormals[3*($l+$a)+2]
										$vn += 1
									Next
								EndIf
							EndIf
						Next

						; $aVertCnts
						If $iModelFaceAttribs = 1 And $bNormals = 1 Then
							$aVertCntsTmp[$i] = $f - $f0
							$f0 = $f
						Else
							$aVertCntsTmp[$i] = $v - $v0
							$v0 = $v
						EndIf

						; $aMaterials
						$aMaterialsTmp[$i+1] = $s

					Else
						PrintError( "Function LoadObjQuads()" & @CRLF & _
							"ERROR: File """ & $sObjModel & ".obj""" & @CRLF & _
							"Problem with sort/index arrays when sorting objects by material." )
						If $bProgressInfo Then GUIDelete( $hProgress )
						Return SetError(23,0,0)
					EndIf

					If $bProgressInfo Then
						GUICtrlSetData( $idProLabel2, "Sorting objects: " & $i+1 & "/" & $aProgressInfo[$p][1] )
						UpdateModelProgressBar( $idProgress2, $i+1, $aProgressInfo[$p][1] )
						If $bModelLoadCancel Then
							ConsoleWrite( "Function LoadObjQuads() cancelled!" & @CRLF & @CRLF )
							If $bProgressInfo Then GUIDelete( $hProgress )
							Return 0
						EndIf
					EndIf
				Next

				; Set arrays = Tmp-arrays, this completes the rebuild
				If $iModelFaceAttribs = 1 And $bNormals = 1 Then
					$aFaces = $aFacesTmp
				Else
					$aVertices = $aVerticesTmp
					If $bTexImage And BitAND( $iModelFaceAttribs, 2 ) Then _
						$aTextureUVs = $aTextureUVsTmp
					If $bUseNormals And BitAND( $iModelFaceAttribs, 4 ) Then _
						$aNormals = $aNormalsTmp
				EndIf
				$iObjects = $n
				$aVertCnts = $aVertCntsTmp
				$aMaterials = $aMaterialsTmp
				;_ArrayDisplay( $aVertCnts, "$aVertCnts after" )

				ConsoleWrite( "Sorting objects: " & $n & @CRLF )
				If $bProgressInfo Then
					UpdateModelProgressBar( $idProgress2, $n, $aProgressInfo[$p][1] )
					UpdateModelProgressBar( $idProgress1, $p, $aProgressInfo[0][0] )
					$aNewProgressInfo[$p][0] = "Sorting objects"
					$aNewProgressInfo[$p][1] = $n
					$p += 1
				EndIf

				If $bModelLoadCancel Then
					ConsoleWrite( "Function LoadObjQuads() cancelled!" & @CRLF & @CRLF )
					If $bProgressInfo Then GUIDelete( $hProgress )
					Return 0
				EndIf

			EndIf

		EndIf

	EndIf
	;#ce

	; Calculate start of vertices per object
	$aVertStart[0] = 0
	ReDim $aVertStart[$iObjects]
	For $i = 0 To $iObjects - 2
		$aVertStart[$i+1] = $aVertStart[$i] + $aVertCnts[$i]
	Next

	Local $aVertInfo[2] = [ $aVertStart, $aVertCnts ]

	$aModel[4]  = $aVertInfo
	$aModel[5]  = $aMaterials
	$aModel[13] = $iObjects

	;_ArrayDisplay( $aModel[4], "$aVertCnts" )
	;_ArrayDisplay( $aModel[5], "$aMaterials" )


	; --- Normal vectors ---

	If $iModelFaceAttribs = 1 Then

		Switch $bNormals
			Case 0
				; Calculate normal vectors (not average) based on face quads

				If $bProgressInfo Then
					GUICtrlSetData( $idProLabel1, "Calculate normal vectors: " & $p - 1 & "/" & $aProgressInfo[0][0] )
					UpdateModelProgressBar( $idProgress2, 0, $aProgressInfo[$p][1] )
				EndIf

				Local $p0[3], $p1[3], $p2[3], $e1, $e2, $normal
				For $i = 0 To $iQuads - 1
					; Three triangle points in the quad
					$p0[0] = $aVertices[12*$i+0] ; First point
					$p0[1] = $aVertices[12*$i+1]
					$p0[2] = $aVertices[12*$i+2]
					$p1[0] = $aVertices[12*$i+3] ; Second point
					$p1[1] = $aVertices[12*$i+4]
					$p1[2] = $aVertices[12*$i+5]
					$p2[0] = $aVertices[12*$i+6] ; Third point
					$p2[1] = $aVertices[12*$i+7]
					$p2[2] = $aVertices[12*$i+8]

					; Two edges
					$e1 = glmSubtractVec3( $p1, $p0 )
					$e2 = glmSubtractVec3( $p2, $p0 )

					; Normal vector
					$normal = glmCrossVec3( $e1, $e2 )

					; Normalize the normal vector
					$normal = glmNormalizeVec3( $normal )

					; Use the normalized normal vector for the four quad points
					$aNormals[12*$i+0]  = $normal[0]
					$aNormals[12*$i+1]  = $normal[1]
					$aNormals[12*$i+2]  = $normal[2]
					$aNormals[12*$i+3]  = $normal[0]
					$aNormals[12*$i+4]  = $normal[1]
					$aNormals[12*$i+5]  = $normal[2]
					$aNormals[12*$i+6]  = $normal[0]
					$aNormals[12*$i+7]  = $normal[1]
					$aNormals[12*$i+8]  = $normal[2]
					$aNormals[12*$i+9]  = $normal[0]
					$aNormals[12*$i+10] = $normal[1]
					$aNormals[12*$i+11] = $normal[2]

					If Mod( $i, 3600 ) = 0 Then
						If $bProgressInfo Then
							GUICtrlSetData( $idProLabel2, "Calculate normal vectors: " & $i & "/" & $aProgressInfo[$p][1] )
							UpdateModelProgressBar( $idProgress2, $i, $aProgressInfo[$p][1] )
							If $bModelLoadCancel Then
								ConsoleWrite( "Function LoadObjQuads() cancelled!" & @CRLF & @CRLF )
								If $bProgressInfo Then GUIDelete( $hProgress )
								Return 0
							EndIf
						EndIf
					EndIf
				Next

				ConsoleWrite( "Calculate normal vectors: " & $iQuads & @CRLF )

			Case 1
				; Calculate average normal vectors based on face quads
				; In this case and only in this case the model will be drawn with glDrawElements
				; (In all other cases the model will be drawn with glDrawArrays)

				If $bProgressInfo Then
					GUICtrlSetData( $idProLabel1, "Average normal vectors: " & $p - 1 & "/" & $aProgressInfo[0][0] )
					UpdateModelProgressBar( $idProgress2, 0, $aProgressInfo[$p][1] )
				EndIf

				Local $aNormalsSeen[$iVertices]

				Local $i0, $i1, $i2, $i3, $quad[4], $v0[3], $v1[3], $v2[3], $e1, $e2, $normal, $cur, $a, $b, $avg[3]
				For $i = 0 To $iQuads - 1
					; One triangle
					$i0 = $aFaces[4*$i+0] ; First vertex
					$i1 = $aFaces[4*$i+1] ; Second vertex
					$i2 = $aFaces[4*$i+2] ; Third vertex
					$i3 = $aFaces[4*$i+3] ; Fourth vertex
					$quad[0] = $i0
					$quad[1] = $i1
					$quad[2] = $i2
					$quad[3] = $i3

					; Three triangle vertices
					$v0[0] = $aVerticesTmp[$i0][0] ; First vertex
					$v0[1] = $aVerticesTmp[$i0][1]
					$v0[2] = $aVerticesTmp[$i0][2]
					$v1[0] = $aVerticesTmp[$i1][0] ; Second vertex
					$v1[1] = $aVerticesTmp[$i1][1]
					$v1[2] = $aVerticesTmp[$i1][2]
					$v2[0] = $aVerticesTmp[$i2][0] ; Third vertex
					$v2[1] = $aVerticesTmp[$i2][1]
					$v2[2] = $aVerticesTmp[$i2][2]

					; Two edges
					$e1 = glmSubtractVec3( $v1, $v0 )
					$e2 = glmSubtractVec3( $v2, $v0 )

					; Normal vector
					$normal = glmCrossVec3( $e1, $e2 )

					; Normalize the normal vector
					$normal = glmNormalizeVec3( $normal )

					For $j = 0 To 3
						$cur = $quad[$j]
						$aNormalsSeen[$cur] += 1
						If $aNormalsSeen[$cur] = 1 Then
							$aNormals[$cur*3+0] = $normal[0]
							$aNormals[$cur*3+1] = $normal[1]
							$aNormals[$cur*3+2] = $normal[2]
						Else
							; Average
							$a = 1.0 / $aNormalsSeen[$cur]
							$b = 1.0 - $a
							$avg[0] = $aNormals[$cur*3+0] * $b + $normal[0] * $a
							$avg[1] = $aNormals[$cur*3+1] * $b + $normal[1] * $a
							$avg[2] = $aNormals[$cur*3+2] * $b + $normal[2] * $a
							$avg = glmNormalizeVec3( $avg )
							$aNormals[$cur*3+0] = $avg[0]
							$aNormals[$cur*3+1] = $avg[1]
							$aNormals[$cur*3+2] = $avg[2]
						EndIf
					Next

					If Mod( $i, 3600 ) = 0 Then
						If $bProgressInfo Then
							GUICtrlSetData( $idProLabel2, "Average normal vectors: " & $i & "/" & $aProgressInfo[$p][1] )
							UpdateModelProgressBar( $idProgress2, $i, $aProgressInfo[$p][1] )
							If $bModelLoadCancel Then
								ConsoleWrite( "Function LoadObjQuads() cancelled!" & @CRLF & @CRLF )
								If $bProgressInfo Then GUIDelete( $hProgress )
								Return 0
							EndIf
						EndIf
					EndIf
				Next

				; Move vertices from $aVerticesTmp to $aVertices
				For $i = 0 To $iVertices - 1
					$aVertices[3*$i+0] = $aVerticesTmp[$i][0]
					$aVertices[3*$i+1] = $aVerticesTmp[$i][1]
					$aVertices[3*$i+2] = $aVerticesTmp[$i][2]
				Next

				ConsoleWrite( "Average normal vectors: " & $iQuads & @CRLF )
		EndSwitch

		If $bProgressInfo Then
			UpdateModelProgressBar( $idProgress2, $iQuads, $aProgressInfo[$p][1] )
			UpdateModelProgressBar( $idProgress1, $p, $aProgressInfo[0][0] )
			$aNewProgressInfo[$p][0] = "Calculate normal vectors"
			$aNewProgressInfo[$p][1] = $iQuads
			$p += 1
		EndIf

		If $bModelLoadCancel Then
			ConsoleWrite( "Function LoadObjQuads() cancelled!" & @CRLF & @CRLF )
			If $bProgressInfo Then GUIDelete( $hProgress )
			Return 0
		EndIf

	EndIf


	; --- Model normalizing ---

	; Place center of model in origin of the coordinate system and scale coordinates to [-1,+1]

	Local $dx, $dy, $dz
	$dx = ( $xMin + $xMax ) / 2
	$dy = ( $yMin + $yMax ) / 2
	$dz = ( $zMin + $zMax ) / 2

	Local $w, $h, $l, $radius, $scale
	$w = $xMax - $xMin
	$h = $yMax - $yMin
	$l = $zMax - $zMin
	$radius = Max( Max( $w, $h ), $l )
	$scale = 1 / $radius

	If $bProgressInfo Then
		GUICtrlSetData( $idProLabel1, "Quad normalizing: " & $p - 1 & "/" & $aProgressInfo[0][0] )
		UpdateModelProgressBar( $idProgress2, 0, $aProgressInfo[$p][1] )
	EndIf

	For $i = 0 To $iVertices - 1
		If $dx Then $aVertices[3*$i+0] -= $dx
		If $dy Then $aVertices[3*$i+1] -= $dy
		If $dz Then $aVertices[3*$i+2] -= $dz
		$aVertices[3*$i+0] *= $scale
		$aVertices[3*$i+1] *= $scale
		$aVertices[3*$i+2] *= $scale
		If Mod( $i, 10800 ) = 0 Then
			If $bProgressInfo Then
				GUICtrlSetData( $idProLabel2, "Quad normalizing: " & $i/4 & "/" & $aProgressInfo[$p][1] )
				UpdateModelProgressBar( $idProgress2, $i/4, $aProgressInfo[$p][1] )
				If $bModelLoadCancel Then
					ConsoleWrite( "Function LoadObjQuads() cancelled!" & @CRLF & @CRLF )
					If $bProgressInfo Then GUIDelete( $hProgress )
					Return 0
				EndIf
			EndIf
		EndIf
	Next
	ConsoleWrite( "Quad normalizing: " & $iVertices/4 & @CRLF )

	If $bProgressInfo Then
		UpdateModelProgressBar( $idProgress2, $iVertices/4, $aProgressInfo[$p][1] )
		UpdateModelProgressBar( $idProgress1, $p, $aProgressInfo[0][0] )
		$aNewProgressInfo[$p][0] = "Quad normalizing"
		$aNewProgressInfo[$p][1] = $iVertices/4
		$p += 1
	EndIf

	If $bModelLoadCancel Then
		ConsoleWrite( "Function LoadObjQuads() cancelled!" & @CRLF & @CRLF )
		If $bProgressInfo Then GUIDelete( $hProgress )
		Return 0
	EndIf

	If $bProgressInfo Then _
		GUICtrlSetState( $idModelLoadCancel, $GUI_DISABLE )


	; --- Generate buffers ---

	; Note that code above this section is independent of the OpenGL version.

	; Uploading vertices, texture coordinates, normal vectors and face quads to
	; conventional/video RAM to use for OpenGL is dependent of the OpenGL version.

	ConsoleWrite( "Draw command: " & $sCmd & @CRLF )

	Local $iVbo, $tBuf, $aBuf[2], $nBuf = 0, $pBuf = 2
	If $iModelFaceAttribs = 1 Or $iModelFaceAttribs = 7 Then $pBuf += 1

	If $bProgressInfo Then
		GUICtrlSetData( $idProLabel1, "Generate buffers: " & $p - 1 & "/" & $aProgressInfo[0][0] )
		GUICtrlSetData( $idProLabel2, "Generate buffers: " & $nBuf & "/" & $pBuf )
		UpdateModelProgressBar( $idProgress2, $nBuf, $pBuf )
	EndIf

	; Vertices
	If $bOpenGL21 Then
		ConsoleWrite( "Uploading vertex coordinates to video RAM for OpenGL 2.1 (2.1 or higher)" & @CRLF )
		glGenBuffers( 1, $iVbo )
		glBindBuffer( $GL_ARRAY_BUFFER, $iVbo )
		glBufferData( $GL_ARRAY_BUFFER, $GL_FLOAT, 4*UBound($aVertices), $aVertices, $GL_STATIC_DRAW )
		$aModel[0] = $iVbo
	Else
		ConsoleWrite( "Storing vertex coordinates in conventional RAM for OpenGL 1.1 (2.0 or lower)" & @CRLF )
		$tBuf = DllStructCreate( "float[" & 3*$iVertices & "]" )
		For $i = 0 To $iVertices - 1
			DllStructSetData( $tBuf, 1, $aVertices[3*$i+0], 3*$i+1 )
			DllStructSetData( $tBuf, 1, $aVertices[3*$i+1], 3*$i+2 )
			DllStructSetData( $tBuf, 1, $aVertices[3*$i+2], 3*$i+3 )
		Next
		$aBuf[0] = $tBuf
		$aBuf[1] = DllStructGetPtr( $tBuf )
		glEnableClientState( $GL_VERTEX_ARRAY )
		glVertexPointer( 3, $GL_FLOAT, 0, $aBuf[1] )
		$aModel[0] = $aBuf
	EndIf
	If $bProgressInfo Then
		$nBuf += 1
		GUICtrlSetData( $idProLabel2, "Generate buffers: " & $nBuf & "/" & $pBuf )
		UpdateModelProgressBar( $idProgress2, $nBuf, $pBuf )
	EndIf

	; Texture coordinates
	If BitAND( $iModelFaceAttribs, 2 ) Then
		If $bTexImage Then
			If $bOpenGL21 Then
				ConsoleWrite( "Uploading texture coordinates to video RAM for OpenGL 2.1 (2.1 or higher)" & @CRLF )
				glGenBuffers( 1, $iVbo )
				glBindBuffer( $GL_ARRAY_BUFFER, $iVbo )
				glBufferData( $GL_ARRAY_BUFFER, $GL_FLOAT, 4*UBound($aTextureUVs), $aTextureUVs, $GL_STATIC_DRAW )
				$aModel[1] = $iVbo
			Else
				ConsoleWrite( "Storing texture coordinates in conventional RAM for OpenGL 1.1 (2.0 or lower)" & @CRLF )
				$tBuf = DllStructCreate( "float[" & 2*$iVertices & "]" )
				For $i = 0 To $iVertices - 1
					DllStructSetData( $tBuf, 1, $aTextureUVs[2*$i+0], 2*$i+1 )
					DllStructSetData( $tBuf, 1, $aTextureUVs[2*$i+1], 2*$i+2 )
				Next
				$aBuf[0] = $tBuf
				$aBuf[1] = DllStructGetPtr( $tBuf )
				glEnableClientState( $GL_TEXTURE_COORD_ARRAY )
				glTexCoordPointer( 2, $GL_FLOAT, 0, $aBuf[1] )
				$aModel[1] = $aBuf
			EndIf
		Else
			$aModel[1] = 0
			If Not $bOpenGL21 Then ; OpenGL 2.0 or lower
				glDisableClientState( $GL_TEXTURE_COORD_ARRAY )
				glBindTexture( $GL_TEXTURE_2D, 0 )
			EndIf
		EndIf
		If $bProgressInfo Then
			$nBuf += 1
			GUICtrlSetData( $idProLabel2, "Generate buffers: " & $nBuf & "/" & $pBuf )
			UpdateModelProgressBar( $idProgress2, $nBuf, $pBuf )
		EndIf
	Else
		$aModel[1] = 0
		If Not $bOpenGL21 Then ; OpenGL 2.0 or lower
			glDisableClientState( $GL_TEXTURE_COORD_ARRAY )
			glBindTexture( $GL_TEXTURE_2D, 0 )
		EndIf
	EndIf

	; Normal vectors
	If $iModelFaceAttribs = 1 Or _
		 BitAND( $iModelFaceAttribs, 4 ) Then
			 If $bUseNormals Then
				 If $bOpenGL21 Then
					 ConsoleWrite( "Uploading normal vectors to video RAM for OpenGL 2.1 (2.1 or higher)" & @CRLF )
					 glGenBuffers( 1, $iVbo )
					 glBindBuffer( $GL_ARRAY_BUFFER, $iVbo )
					 glBufferData( $GL_ARRAY_BUFFER, $GL_FLOAT, 4*UBound($aNormals), $aNormals, $GL_STATIC_DRAW )
					 $aModel[2] = $iVbo
				 Else
					 ConsoleWrite( "Storing normal vectors in conventional RAM for OpenGL 1.1 (2.0 or lower)" & @CRLF )
					 $tBuf = DllStructCreate( "float[" & 3*$iVertices & "]" )
					 For $i = 0 To $iVertices - 1
						 DllStructSetData( $tBuf, 1, $aNormals[3*$i+0], 3*$i+1 )
						 DllStructSetData( $tBuf, 1, $aNormals[3*$i+1], 3*$i+2 )
						 DllStructSetData( $tBuf, 1, $aNormals[3*$i+2], 3*$i+3 )
					 Next
					 $aBuf[0] = $tBuf
					 $aBuf[1] = DllStructGetPtr( $tBuf )
					 glEnableClientState( $GL_NORMAL_ARRAY )
					 glNormalPointer( $GL_FLOAT, 0, $aBuf[1] )
					 $aModel[2] = $aBuf
				 EndIf
			 Else
				 $aModel[2] = 0
				 If Not $bOpenGL21 Then ; OpenGL 2.0 or lower
					 glDisableClientState( $GL_NORMAL_ARRAY )
				 EndIf
			 EndIf
			 If $bProgressInfo Then
				 $nBuf += 1
				 GUICtrlSetData( $idProLabel2, "Generate buffers: " & $nBuf & "/" & $pBuf )
				 UpdateModelProgressBar( $idProgress2, $nBuf, $pBuf )
			 EndIf
	Else
		 $aModel[2] = 0
			If Not $bOpenGL21 Then ; OpenGL 2.0 or lower
				glDisableClientState( $GL_NORMAL_ARRAY )
			EndIf
	EndIf

	; Face quads
	If $iModelFaceAttribs = 1 Then
		Switch $bNormals
			Case 0
				$aModel[3] = 0
			Case 1
				If $bOpenGL21 Then
					ConsoleWrite( "Uploading face quads to video RAM for OpenGL 2.1 (2.1 or higher)" & @CRLF )
					glGenBuffers( 1, $iVbo )
					glBindBuffer( $GL_ELEMENT_ARRAY_BUFFER, $iVbo )
					glBufferData( $GL_ELEMENT_ARRAY_BUFFER, $GL_UNSIGNED_SHORT, 2*UBound($aFaces), $aFaces, $GL_STATIC_DRAW )
					$aModel[3] = $iVbo
				Else
					ConsoleWrite( "Storing face quads in conventional RAM for OpenGL 1.1 (2.0 or lower)" & @CRLF )
					$tBuf = DllStructCreate( "ushort[" & $iQuads*3 & "]" )
					For $i = 0 To $iQuads * 4 - 1
						DllStructSetData( $tBuf, 1, $aFaces[$i], $i+1 )
					Next
					$aBuf[0] = $tBuf
					$aBuf[1] = DllStructGetPtr( $tBuf )
					$aModel[3] = $aBuf
				EndIf
				If $bProgressInfo Then
					$nBuf += 1
					GUICtrlSetData( $idProLabel2, "Generate buffers: " & $nBuf & "/" & $pBuf )
					UpdateModelProgressBar( $idProgress2, $nBuf, $pBuf )
				EndIf
		EndSwitch
	Else
		$aModel[3] = 0
	EndIf


	; --- Update progress info ---

	If $bProgressInfo Then
		If Not $bReload Then
			$aNewProgressInfo[0][0] = $p - 1
			ReDim $aNewProgressInfo[$p][2]
			; Save progress info in inifile?
			If $aNewProgressInfo[0][0] <> $aProgressInfo[0][0] Then
				IniDelete( $sObjModel & ".ini", "ProgressInfoQuads" )
				IniWriteSection ( $sObjModel & ".ini", "ProgressInfoQuads", $aNewProgressInfo )
				ConsoleWrite( "Progress info: Updated" & @CRLF )
			Else
				For $i = 1 To $p - 1
					If $aNewProgressInfo[$i][0] <> $aProgressInfo[$i][0] Or _
						 $aNewProgressInfo[$i][1] <> $aProgressInfo[$i][1] Then
								IniDelete( $sObjModel & ".ini", "ProgressInfoQuads" )
								IniWriteSection ( $sObjModel & ".ini", "ProgressInfoQuads", $aNewProgressInfo )
								ConsoleWrite( "Progress info: Updated" & @CRLF )
								ExitLoop
					EndIf
				Next
			EndIf
		EndIf
		; Delete progress window
		GUIDelete( $hProgress )
	EndIf

	; Elapsed time
	Local $fElapsedTime = TimerDiff( $iStartTime ) / 1000
	ConsoleWrite( "Process time: " & $fElapsedTime & @CRLF & @CRLF )

	Return 1

EndFunc


Func CheckModelFile( $sObjModel, ByRef $aModel )

	Local $szDrive, $szDir, $szFName, $szExt
	Local $aPath = _PathSplit( $sObjModel, $szDrive, $szDir, $szFName, $szExt )
	If StringLen( $aPath[1] ) = 0 Then $sObjModel = $models & "\" & $sObjModel

	; Model format
	Local $sFormat = StringRight( $sObjModel, StringLen( $sObjModel ) - StringInStr( $sObjModel, ".", 0, -1 ) )
	Switch $sFormat
		Case "obj"
		Case Else
			PrintError( "Function CheckModelFile()" & @CRLF & _
				"ERROR: """ & $sFormat & """ is not a legal model format." & @CRLF & _
				"Legal formats: OBJ (Wavefront OBJ file)", $hGui )
			Return 0
	EndSwitch

	; Model exists?
	If Not FileExists( $sObjModel ) Then
		PrintError( "Function CheckModelFile()" & @CRLF & _
			"ERROR: File """ & $sObjModel & """ does not exists.", $hGui )
		Return 0
	EndIf

	; $aModel[11] = full path to the OBJ-file WITHOUT the .obj suffix
	$aModel[11] = StringLeft( $sObjModel, StringInStr( $sObjModel, ".", 0, -1 ) - 1 )

	Return 1

EndFunc


Func GetModelSettings( ByRef $aModel, $bOpenGL21 = True )

	; Model setting defaults
	Local $bModelLight    = 1 ; Light enabled ; OpenGL 1.1 (2.0 or lower)
	Local $bModelShader   = 0 ; Shader id     ; OpenGL 2.1 (2.1 or higher)
	Local $bModelTextures = 1 ; Use textures
	Local $bModelNormals  = 0 ; Normal vectors
	Local $bModelQuads    = 0 ; Quad mesh

	Local $sSection
	If $bOpenGL21 Then
		$sSection = "Settings 2.1"
	Else
		$sSection = "Settings 1.1"
	EndIf

	; INI-file? If yes then read settings
	Local $sObjModel = $aModel[11]
	If FileExists( $sObjModel & ".ini" ) Then
		Local $aParams = IniReadSection( $sObjModel & ".ini", $sSection )
		If IsArray( $aParams ) Then
			For $i = 1 To $aParams[0][0]
				If Not $bOpenGL21 Then
					If $aParams[$i][0] = "bLight"    Then $bModelLight    = Number( $aParams[$i][1] )
				EndIf
				If $aParams[$i][0]   = "bTextures" Then $bModelTextures = Number( $aParams[$i][1] )
				If $bOpenGL21 Then
					If $aParams[$i][0] = "bShader"   Then $bModelShader   = Number( $aParams[$i][1] )
				EndIf
				If $aParams[$i][0]   = "bNormals"  Then $bModelNormals  = Number( $aParams[$i][1] )
				If $aParams[$i][0]   = "bQuads"    Then $bModelQuads    = Number( $aParams[$i][1] )
			Next
		EndIf
	Else
	  ; Use the default model settings above
	EndIf

	Local $aSettings[5] = [ $bModelLight, $bModelShader, $bModelTextures, $bModelNormals, $bModelQuads ]
	$aModel[10] = $aSettings

EndFunc


Func SetModelSettings( $sObjModel, $bOpenGL21 = True )

	Local $aData11[4][2] = [ _
		[ "bLight",    $bModelLight    ], _ ; Light enabled
		[ "bTextures", $bModelTextures ], _ ; Use textures
		[ "bNormals",  $bModelNormals  ], _ ; Normal vectors
		[ "bQuads",    $bModelQuads    ] ]  ; Quad mesh

	Local $aData21[4][2] = [ _
		[ "bShader",   $bModelShader   ], _ ; Shader id
		[ "bTextures", $bModelTextures ], _ ; Use textures
		[ "bNormals",  $bModelNormals  ], _ ; Normal vectors
		[ "bQuads",    $bModelQuads    ] ]  ; Quad mesh

	Local $aData, $sSection
	If $bOpenGL21 Then
		$aData = $aData21
		$sSection = "Settings 2.1"
	Else
		$aData = $aData11
		$sSection = "Settings 1.1"
	EndIf

	IniWriteSection( $sObjModel & ".ini", $sSection, $aData, 0 )

EndFunc


Func ReadMaterials( $sObjPath, $sMaterialFile, ByRef $aMaterialInfo )

	Local $aMatInfo[100][6], $iMatInfo = 0
	Local $aMatLines, $matLine, $sMatSection = "", $aWords, $aTmp, $iIdx
	
	; Read the material file
	If StringRight( $sObjPath, 1 ) <> "\" Then $sObjPath &= "\"
	_FileReadToArray( $sObjPath & $sMaterialFile, $aMatLines )

	For $l = 1 To $aMatLines[0]
		$matLine = StringStripWS( $aMatLines[$l], 7 )

		Select
			; MTL-file section name
			Case StringLeft( $matLine, 7 ) = "newmtl "
				$aWords = StringSplit( $matLine, " " )
				If $aWords[0] = 2 Then $sMatSection = $aWords[2]

			; Get ambient light: Ka
			Case StringLeft( $matLine, 3 ) = "Ka "
				GetMatColors( $sMaterialFile, $aMatInfo, $iMatInfo, $sMatSection, "Ka", $matLine )

			; Get diffuse light: Kd
			Case StringLeft( $matLine, 3 ) = "Kd "
				GetMatColors( $sMaterialFile, $aMatInfo, $iMatInfo, $sMatSection, "Kd", $matLine )

			; Get specular light: Ks
			Case StringLeft( $matLine, 3 ) = "Ks "
				GetMatColors( $sMaterialFile, $aMatInfo, $iMatInfo, $sMatSection, "Ks", $matLine )

			; Get shininess: Ns
			Case StringLeft( $matLine, 3 ) = "Ns "
				$aWords = StringSplit( $matLine, " " )
				If $aWords[0] = 2 Then
					$aMatInfo[$iMatInfo][0] = $sMaterialFile
					$aMatInfo[$iMatInfo][1] = $sMatSection
					$aMatInfo[$iMatInfo][2] = "Ns"
					$aMatInfo[$iMatInfo][3] = $aWords[2] / 1000 ; Wavefront MTL-file shininess is in [0,1000]
					$iMatInfo += 1                              ; Scale back to a normal OpenGL [0,1] range
				EndIf

			; Get transparency: d
			Case StringLeft( $matLine, 2 ) = "d "           ; Transparency can be defined with the tokens d and Tr
				$aWords = StringSplit( $matLine, " " )
				If $aWords[0] = 2 Then
					$aMatInfo[$iMatInfo][0] = $sMaterialFile
					$aMatInfo[$iMatInfo][1] = $sMatSection
					$aMatInfo[$iMatInfo][2] = "alpha"
					$aMatInfo[$iMatInfo][3] = $aWords[2]        ; d: alpha = d
					$iMatInfo += 1
				EndIf

			; Get transparency: Tr
			Case StringLeft( $matLine, 3 ) = "Tr "          ; Transparency can be defined with the tokens d and Tr
				$aWords = StringSplit( $matLine, " " )
				If $aWords[0] = 2 Then
					$aMatInfo[$iMatInfo][0] = $sMaterialFile
					$aMatInfo[$iMatInfo][1] = $sMatSection
					$aMatInfo[$iMatInfo][2] = "alpha"
					$aMatInfo[$iMatInfo][3] = 1.0 - $aWords[2]  ; Tr: alpha = 1.0 - Tr
					$iMatInfo += 1
				EndIf

			; Get ambient texture map: map_Ka
			Case StringLeft( $matLine, 7 ) = "map_Ka "
				GetMatTexture( $sObjPath, $sMaterialFile, $aMatInfo, $iMatInfo, $sMatSection, "map_Ka", $matLine )

			; Get diffuse texture map: map_Kd
			Case StringLeft( $matLine, 7 ) = "map_Kd "
				GetMatTexture( $sObjPath, $sMaterialFile, $aMatInfo, $iMatInfo, $sMatSection, "map_Kd", $matLine )

			; Get specular texture map: map_Ks
			Case StringLeft( $matLine, 7 ) = "map_Ks "
				GetMatTexture( $sObjPath, $sMaterialFile, $aMatInfo, $iMatInfo, $sMatSection, "map_Ks", $matLine )

		EndSelect

		If Mod( $iMatInfo, 100 ) = 0 Then _
			Redim $aMatInfo[$iMatInfo+100][6]
	Next

	Redim $aMatInfo[$iMatInfo][6]
	$aMaterialInfo = $aMatInfo

	Return $iMatInfo

EndFunc


Func GetMatColors( $sMaterialFile, ByRef $aMatInfo, ByRef $iMatInfo, $sMatSection, $sColor, $matLine )
	Local $aWords = StringSplit( $matLine, " " )
	If $aWords[0] = 4 Then
		$aMatInfo[$iMatInfo][0] = $sMaterialFile
		$aMatInfo[$iMatInfo][1] = $sMatSection
		$aMatInfo[$iMatInfo][2] = $sColor
		$aMatInfo[$iMatInfo][3] = $aWords[2] ; r
		$aMatInfo[$iMatInfo][4] = $aWords[3] ; g
		$aMatInfo[$iMatInfo][5] = $aWords[4] ; b
		$iMatInfo += 1
	EndIf
EndFunc


Func GetMatTexture( $sObjPath, $sMaterialFile, ByRef $aMatInfo, ByRef $iMatInfo, $sMatSection, $sTexture, $matLine )
	$matLine = StringRight( $matLine, StringLen( $matLine ) - 7 )
	$matLine = StringRegExpReplace( $matLine, "\\\\", "\\" )

	If StringLen( $matLine ) > 4 Then
		$aMatInfo[$iMatInfo][0] = $sMaterialFile
		$aMatInfo[$iMatInfo][1] = $sMatSection
		$aMatInfo[$iMatInfo][2] = $sTexture       ; map_Ka, map_Kd, map_Ks
		$aMatInfo[$iMatInfo][3] = $matLine        ; Path relative to $sObjPath

		; Image/texture file format
		Local $sFormat = StringRight( $matLine, StringLen( $matLine ) - StringInStr( $matLine, ".", 0, -1 ) )
		Switch $sFormat
			Case "bmp", "gif", "ico", "jpg", "wmf"  ; LoadTextureImage() --> LoadImage()
			Case "png", "tif"                       ; LoadTextureImage() --> LoadGDIimage()
			Case "rgb", "rgba"                      ; LoadTextureImage() --> LoadRGBimage()
			Case "tga" ; includes RLE-encoded TGAs  ; LoadTextureImage() --> LoadTGAimage()
			Case Else
				$aMatInfo[$iMatInfo][4] = "Bad format"
		EndSwitch

		; Image/texture file exists?
		If Not FileExists( $sObjPath & $matLine ) Then _
			$aMatInfo[$iMatInfo][5] = "No such file"

		$iMatInfo += 1
	EndIf
EndFunc


Func CreateModelProgressWin( $model, ByRef $hWnd, ByRef $label1, ByRef $prgrs1, ByRef $label2, ByRef $prgrs2 )

	Local $aPos = WinGetPos( $hGui )
	Local $w = 327, $h = 165, $x = $aPos[0] + ( $iWidth - $w ) / 2, $y = $aPos[1] + ( $iHeight - $h ) / 2

	$hWnd = GUICreate( " Generating model data: " & $model, $w, $h, $x, $y, 0, BitOR( $WS_EX_TOOLWINDOW, $WS_EX_TOPMOST ) )

	$label1 = GUICtrlCreateLabel( "", 10,  8, 300, 17 )
	$prgrs1 = GUICtrlCreateProgress(  10, 25, 300, 20 )
	$label2 = GUICtrlCreateLabel( "", 10, 58, 300, 17 )
	$prgrs2 = GUICtrlCreateProgress(  10, 75, 300, 20 )

	$idModelLoadCancel = GUICtrlCreateButton( "Cancel",   100, 108, 127, 22 )
	GUICtrlSetOnEvent( $idModelLoadCancel, "CancelModelLoad" )

	GUISetState( @SW_SHOW, $hWnd )

EndFunc


Func UpdateModelProgressBar( $id, $n, $max )
	Local $pct = $n / $max * 100
	GUICtrlSetData( $id, $pct )
EndFunc


Func CancelModelLoad()
	$bModelLoadCancel = True
EndFunc


; $aMaterialData[9]   ; Material data
; 0: tFloats          ; Float structure to hold the RGBA values for ambient, diffuse and specular reflectivity, 3 * 4 floats
; 1: pKa              ; Pointer to the ambient reflectivity in tFloats,  defaults: 0.2, 0.2, 0.2, 1.0
; 2: pKd              ; Pointer to the diffuse reflectivity in tFloats,  defaults: 0.8, 0.8, 0.8, 1.0
; 3: pKs              ; Pointer to the specular reflectivity in tFloats, defaults: 0.0, 0.0, 0.0, 1.0
; 4: Ns               ; Shininess (to set the brightness of the specular light, values in [0,128]), def: 0.0
; 5: alpha            ; Transparency of the material, the default is 1.0 (not transparent at all)
; 6: map_Ka_id        ; Texture map id for ambient reflectivity;  0 if no ambient texture map
; 7: map_Kd_id        ; Texture map id for diffuse reflectivity;  0 if no diffuse texture map
; 8: map_Ks_id        ; Texture map id for specular reflectivity; 0 if no specular texture map

; For each object in the model material reflectivities are stored in tFloats. pKa, pKd and pKs
; pointers are set. Shininess is stored in Ns and transparency in alpha. Missing values are set
; to defaults. Material data are always set. Also if the model does not contain a material file
; (MTL-file). This means that you don't have to check for materials in the display function.
; Disable light ( glDisable( $GL_LIGHTING ) ) to completely disable materials.
Func LoadMaterials( ByRef $aModel, ByRef $aMaterialData )

	Local $aMaterials    = $aModel[5]
	Local $aMaterialInfo = $aModel[6]
	Local $iObjects      = $aModel[13]
	Local $sUseMaterial

	Local $iMaterialInfo = UBound( $aMaterialInfo )

	Local $tFloats, $pKa, $pKd, $pKs, $Ns, $alpha, $index, $s

	; Material reflectivities
	For $obj = 0 To $iObjects - 1

		; --- Default materials ---

		$tFloats = DllStructCreate( "float[12]" ) ; 48 bytes

		; Ambient reflectivity
		DllStructSetData( $tFloats, 1, 0.2,  1 )  ; Offset 0
		DllStructSetData( $tFloats, 1, 0.2,  2 )
		DllStructSetData( $tFloats, 1, 0.2,  3 )
		DllStructSetData( $tFloats, 1, 1.0,  4 )
		$pKa = DllStructGetPtr( $tFloats )

		; Diffuse reflectivity
		DllStructSetData( $tFloats, 1, 0.8,  5 )  ; Offset 16
		DllStructSetData( $tFloats, 1, 0.8,  6 )
		DllStructSetData( $tFloats, 1, 0.8,  7 )
		DllStructSetData( $tFloats, 1, 1.0,  8 )
		$pKd = $pKa + 16

		; Specular reflectivity
		DllStructSetData( $tFloats, 1, 0.0,  9 )  ; Offset 32
		DllStructSetData( $tFloats, 1, 0.0, 10 )
		DllStructSetData( $tFloats, 1, 0.0, 11 )
		DllStructSetData( $tFloats, 1, 1.0, 12 )
		$pKs = $pKa + 32

		$Ns    = 0.0
		$alpha = 1.0

		; -------------------------

		; Set the default materials for each object
		$aMaterialData[$obj][0] = $tFloats
		$aMaterialData[$obj][1] = $pKa
		$aMaterialData[$obj][2] = $pKd
		$aMaterialData[$obj][3] = $pKs
		$aMaterialData[$obj][4] = $Ns
		$aMaterialData[$obj][5] = $alpha

		If $aMaterials[0] = 0 Or _
			 Not IsArray( $aMaterialInfo ) Then _
				 ContinueLoop

		; The "usemtl <name>"-name from the OBJ-file
		$sUseMaterial = $aMaterials[$obj+1]
		If $sUseMaterial = "" Then ContinueLoop

		For $m = 0 To $iMaterialInfo - 1
			; Find a "newmtl <name>"-section from the MTL-file that matches the "usemtl <name>"-name
			If $aMaterialInfo[$m][1] = $sUseMaterial Then
				While $aMaterialInfo[$m][1] = $sUseMaterial
					Switch $aMaterialInfo[$m][2]
						Case "Ka", "Kd", "Ks"
							Switch $aMaterialInfo[$m][2]
								Case "Ka"
									$index = 0
								Case "Kd"
									$index = 4
								Case "Ks"
									$index = 8
							EndSwitch
							$s = StringFormat( "%2s", $obj+1 )
							$s &= "|" & $aMaterialInfo[$m][0]
							$s &= "|" & $sUseMaterial
							$s &= "|" & $aMaterialInfo[$m][2]
							$s &= "|" & $aMaterialInfo[$m][3] & "," _
							          & $aMaterialInfo[$m][4] & "," _
												& $aMaterialInfo[$m][5] & "," _
												& "1.0"
							DllStructSetData( $tFloats, 1, $aMaterialInfo[$m][3], $index+1 )
							DllStructSetData( $tFloats, 1, $aMaterialInfo[$m][4], $index+2 )
							DllStructSetData( $tFloats, 1, $aMaterialInfo[$m][5], $index+3 )
							ConsoleWrite( "Material: " & $s & "|0x" & Hex( $pKa + 4*$index ) & @CRLF )
						Case "Ns"
							$s = StringFormat( "%2s", $obj+1 )
							$s &= "|" & $aMaterialInfo[$m][0]
							$s &= "|" & $sUseMaterial
							$s &= "|" & $aMaterialInfo[$m][2]
							$s &= "|" & $aMaterialInfo[$m][3] * 128.0
							ConsoleWrite( "Material: " & $s & @CRLF )
							$Ns = $aMaterialInfo[$m][3] * 128.0 ; OpenGL shininess is in [0,128]
						Case "alpha"
							$s = StringFormat( "%2s", $obj+1 )
							$s &= "|" & $aMaterialInfo[$m][0]
							$s &= "|" & $sUseMaterial
							$s &= "|" & $aMaterialInfo[$m][2]
							$s &= "|" & $aMaterialInfo[$m][3]
							ConsoleWrite( "Material: " & $s & @CRLF )
							$alpha = $aMaterialInfo[$m][3]
					EndSwitch
					$m += 1
					If $m = $iMaterialInfo Then _
						ExitLoop
				WEnd
				ExitLoop
			EndIf
		Next

		; Set shininess and transparency
		; (The other values are already set)
		$aMaterialData[$obj][4] = $Ns
		$aMaterialData[$obj][5] = $alpha
	Next

	ConsoleWrite( @CRLF )

EndFunc


; $aMaterialData[9]   ; Material data
; 0: tFloats          ; Float structure to hold the RGBA values for ambient, diffuse and specular reflectivity, 3 * 4 floats
; 1: pKa              ; Pointer to the ambient reflectivity in tFloats,  defaults: 0.2, 0.2, 0.2, 1.0
; 2: pKd              ; Pointer to the diffuse reflectivity in tFloats,  defaults: 0.8, 0.8, 0.8, 1.0
; 3: pKs              ; Pointer to the specular reflectivity in tFloats, defaults: 0.0, 0.0, 0.0, 1.0
; 4: Ns               ; Shininess (to set the brightness of the specular light, values in [0,128]), def: 0.0
; 5: alpha            ; Transparency of the material, the default is 1.0 (not transparent at all)
; 6: map_Ka_id        ; Texture map id for ambient reflectivity;  0 if no ambient texture map
; 7: map_Kd_id        ; Texture map id for diffuse reflectivity;  0 if no diffuse texture map
; 8: map_Ks_id        ; Texture map id for specular reflectivity; 0 if no specular texture map

; For each object in the model texture maps are loaded and the ids stored in map_Ka_id, map_Kd_id
; and map_Ks_id. Texture ids are always set. Also if the model does not contain a material file
; (MTL-file). This means that you don't have to check for texture ids in the display function.
Func LoadTextures( ByRef $aModel, ByRef $aMaterialData )

	Local $aMaterials    = $aModel[5]
	Local $aMaterialInfo = $aModel[6]
	Local $bTexture      = BitAND( $aModel[7], 3 )
	Local $iObjects      = $aModel[13]
	Local $sUseMaterial

	Local $iMaterialInfo = UBound( $aMaterialInfo )

	Local $sObjPath = StringLeft( $aModel[11], StringInStr( $aModel[11], "\", 0, -1 ) )

	Local $aTexLoaded[50][2], $iTexLoaded = 0
	Local $bTexFound, $iTexture, $w, $h, $s

	; Texture maps
	For $obj = 0 To $iObjects - 1

		; Set the default texture ids for each object
		$aMaterialData[$obj][6] = 0 ; map_Ka_id
		$aMaterialData[$obj][7] = 0 ; map_Kd_id
		$aMaterialData[$obj][8] = 0 ; map_Ks_id

		If $bTexture <> 3 Or _
			 $aMaterials[0] = 0 Or _
			 Not IsArray( $aMaterialInfo ) Then _
				 ContinueLoop

		; The "usemtl <name>"-name from the OBJ-file
		$sUseMaterial = $aMaterials[$obj+1]
		If $sUseMaterial = "" Then ContinueLoop

		For $m = 0 To $iMaterialInfo - 1
			; Find a "newmtl <name>"-section from the MTL-file that matches the "usemtl <name>"-name
			If $aMaterialInfo[$m][1] = $sUseMaterial Then
				While $aMaterialInfo[$m][1] = $sUseMaterial
					Switch $aMaterialInfo[$m][2]
						Case "map_Ka", "map_Kd", "map_Ks"
							If $aMaterialInfo[$m][4] = "" And $aMaterialInfo[$m][5] = "" Then

								$s = StringFormat( "%2s", $obj+1 )
								$s &= "|" & $aMaterialInfo[$m][0]
								$s &= "|" & $sUseMaterial
								$s &= "|" & $aMaterialInfo[$m][2]
								$s &= "|" & $aMaterialInfo[$m][3]

								; Check if this image already is loaded
								; In that case we can reuse the texture id
								$bTexFound = False
								For $l = 0 To $iTexLoaded - 1
									If $aTexLoaded[$l][0] = $aMaterialInfo[$m][3] Then
										$iTexture = $aTexLoaded[$l][1]
										ConsoleWrite( "Reuse texture: " & $s & "|" & $iTexture & @CRLF )
										$bTexFound = True
										ExitLoop
									EndIf
								Next

								If Not $bTexFound Then
									; Load texture image
									If LoadTextureImage( $iTexture, $sObjPath & $aMaterialInfo[$m][3], $w, $h ) Then
										; Save image name and texture id
										; If an image is used several times we can reuse the texture id
										$aTexLoaded[$iTexLoaded][0] = $aMaterialInfo[$m][3]
										$aTexLoaded[$iTexLoaded][1] = $iTexture
										$iTexLoaded += 1
										$bTexFound = 1
										ConsoleWrite( "Load texture:  " & $s & "|" & $iTexture & @CRLF )
									EndIf
								EndIf

								If $bTexFound Then
									Switch $aMaterialInfo[$m][2]
										Case "map_Ka"
											$aMaterialData[$obj][6] = $iTexture
										Case "map_Kd"
											$aMaterialData[$obj][7] = $iTexture
										Case "map_Ks"
											$aMaterialData[$obj][8] = $iTexture
									EndSwitch
								EndIf
							EndIf
					EndSwitch
					$m += 1
					If $m = $iMaterialInfo Then _
						ExitLoop
				WEnd
				ExitLoop
			EndIf
		Next
	Next

	ConsoleWrite( @CRLF )

EndFunc


#cs#####################
#   Cursor functions   #
#ce#####################

Func HourglassCursor( $fOn )
	Static $fLoadCursor, $hHourglass, $fHourglass = False
	If Not $fLoadCursor Then
		$fLoadCursor = True
		$hHourglass = LoadCursor( $OCR_WAIT )
	EndIf
	If ( $fOn And Not $fHourglass ) Or ( Not $fOn And $fHourglass ) Then
		SetSystemCursor( $hHourglass, $OCR_NORMAL )
		$fHourglass = Not $fHourglass
	EndIf
EndFunc

; The cursor functions below is copied from http://www.autoitscript.com/forum/index.php?showtopic=84400 created by cherdeg

Func LoadCursor( $iCursor )
	Return SetError( @error, @extended, CursorAPI( DllCall( $dllUser32, "int", "LoadCursorA", "hwnd", 0, "int", $iCursor ) ) )
EndFunc

Func SetSystemCursor( $hCursor, $iCursor )
	Return SetError( @error, @extended, CursorAPI( DllCall( $dllUser32, "int", "SetSystemCursor", "int", $hCursor, "int", $iCursor ) ) )
EndFunc

Func CursorAPI( $v_ret )
	Local $err = @error
	Local $ext = @extended
	If Not $err Then
		If IsArray( $v_ret ) Then
			Return $v_ret[0]
		Else
			Return $v_ret
		EndIf
	EndIf
	Return SetError( $err, $ext, 0 )
EndFunc
