#include-once

;#include <Array.au3>

Func LoadObjModelBins( ByRef $aModel, $bOpenGL21 = True, $bNormals = 0 )

	Return LoadObjModelBinaries( $aModel, $bOpenGL21, $bNormals, False )

EndFunc


Func LoadObjQuadsBins( ByRef $aModel, $bOpenGL21 = True, $bNormals = 0 )

	Return LoadObjModelBinaries( $aModel, $bOpenGL21, $bNormals, True )

EndFunc


Func LoadObjModelBinaries( ByRef $aModel, $bOpenGL21, $bNormals, $bQuads )

	Local $sFuncName, $sIniSection, $sFaces, $sExt
	If Not $bQuads Then
		$sFuncName = "LoadObjModelBins()"
		$sIniSection = "LoadBinFiles"
		$sFaces = "triangles"
		$sExt = ".bin"
	Else
		$sFuncName = "LoadObjQuadsBins()"
		$sIniSection = "LoadBinFiles4"
		$sFaces = "quads"
		$sExt = "4.bin"
	EndIf

	; Full path to the OBJ-file
	Local $sObjModel = $aModel[11]

	; --- Read inifile ---

	Local $iStatus, $iFaceType, $iObjects, $sMeshType, $sMtlFile, $aMaterials, $aVertStart, $aVertCnts, $aVertInfo[2]

	If FileExists( $sObjModel & ".ini" ) Then
		Local $aParams = IniReadSection( $sObjModel & ".ini", $sIniSection )
		If IsArray( $aParams ) And UBound( $aParams ) = 9 Then
			For $i = 1 To $aParams[0][0]
				If $aParams[$i][0] = "MtlFile"   Then $sMtlFile   = $aParams[$i][1]
				If $aParams[$i][0] = "MeshType"  Then $sMeshType  = $aParams[$i][1]
				If $aParams[$i][0] = "Status"    Then $iStatus    = Number( $aParams[$i][1] )
				If $aParams[$i][0] = "FaceType"  Then $iFaceType  = Number( $aParams[$i][1] )
				If $aParams[$i][0] = "Objects"   Then $iObjects   = Number( $aParams[$i][1] )
				If $aParams[$i][0] = "VertStart" Then $aVertStart = StringSplit( $aParams[$i][1], "|", 2 )
				If $aParams[$i][0] = "VertCnts"  Then $aVertCnts  = StringSplit( $aParams[$i][1], "|", 2 )
				If $aParams[$i][0] = "Materials" Then $aMaterials = StringSplit( $aParams[$i][1], "|", 2 )
			Next
		Else
			PrintError( "Function " & $sFuncName & @CRLF & _
				"ERROR: File """ & $sObjModel & ".ini""" & @CRLF & _
				"The inifile does not contain a """ & $sIniSection & """ section or the section is empty." )
			Return SetError(1,0,0)
		EndIf
	Else
		PrintError( "Function " & $sFuncName & @CRLF & _
			"ERROR: File """ & $sObjModel & ".ini""" & @CRLF & _
			"The inifile does not exists." )
		Return SetError(2,0,0)
	EndIf
	;_ArrayDisplay( $aVertStart, "$aVertStart" )
	;_ArrayDisplay( $aVertCnts, "$aVertCnts" )
	;_ArrayDisplay( $aMaterials, "$aMaterials" )

	If $iObjects And IsArray( $aVertStart ) And IsArray( $aVertCnts ) _
		 And $iObjects = UBound( $aVertStart ) And $iObjects = UBound( $aVertCnts ) Then
			 $aVertInfo[0] = $aVertStart
			 $aVertInfo[1] = $aVertCnts
	Else
		PrintError( "Function " & $sFuncName & @CRLF & _
			"ERROR: File """ & $sObjModel & ".ini""" & @CRLF & _
			"Data errors in inifile in section """ & $sIniSection & """." )
		Return SetError(3,0,0)
	EndIf

	; --- Which binary files must be loaded? Check file existance ---

	Local $bLoad = 0, $sVFile, $sVnFile

	; Vertex coordinates
	; Always load vertex coordinates
	$sVFile = "V" & $sExt

	; Texture coordinates
	If BitAND( $iStatus, 2 ) Then
		If Not FileExists( $sObjModel & "Vt" & $sExt ) Then
			PrintError( "Function " & $sFuncName & @CRLF & _
				"ERROR: File """ & $sObjModel & "Vt" & $sExt & "" & @CRLF & _
				"The binary file with texture coordinates does not exists." )
			Return SetError(4,0,0)
		EndIf
		$bLoad += 1
	EndIf

	; Normal vectors
	If BitAND( $iFaceType, 4 ) Then
		If $bNormals = 0 Then
			If Not FileExists( $sObjModel & "Vn" & $sExt ) Then
				PrintError( "Function " & $sFuncName & @CRLF & _
					"ERROR: File """ & $sObjModel & "Vn" & $sExt & "" & @CRLF & _
					"The binary file with normal vectors does not exists." )
				Return SetError(5,0,0)
			EndIf
			$bLoad += 2
			$sVnFile = "Vn" & $sExt
		ElseIf $bNormals = -1 Then
			; Normal vectors are excluded because they are not needed
			; Normal vectors should not be loaded
			$iStatus -= BitAND( $iStatus, 4 )
		EndIf
	ElseIf BitAND( $iFaceType, 2 ) = 0 Then
		; It's a model without texture coordinates and without normal vectors
		; In this case normal vectors are calulated by LoadObjModel
		If $bNormals = 0 Then
			If Not FileExists( $sObjModel & "VnCal" & $sExt ) Then
				PrintError( "Function " & $sFuncName & @CRLF & _
					"ERROR: File """ & $sObjModel & "VnCal" & $sExt & "" & @CRLF & _
					"The binary file with normal vectors does not exists." )
				Return SetError(6,0,0)
			EndIf
			$bLoad += 2
			$sVFile = "VCal" & $sExt
			$sVnFile = "VnCal" & $sExt
		ElseIf $bNormals = 1 Then
			If Not FileExists( $sObjModel & "VnAvg" & $sExt ) Then
				PrintError( "Function " & $sFuncName & @CRLF & _
					"ERROR: File """ & $sObjModel & "VnAvg" & $sExt & "" & @CRLF & _
					"The binary file with normal vectors does not exists." )
				Return SetError(7,0,0)
			EndIf
			$bLoad += 2
			$sVFile = "VAvg" & $sExt
			$sVnFile = "VnAvg" & $sExt
		EndIf
	EndIf

	; Face polygons
	If $iFaceType = 1 And $bNormals Then
		If Not FileExists( $sObjModel & "F" & $sExt ) Then
			PrintError( "Function " & $sFuncName & @CRLF & _
				"ERROR: File """ & $sObjModel & "F" & $sExt & "" & @CRLF & _
				"The binary file with face indices does not exists." )
			Return SetError(8,0,0)
		EndIf
		$bLoad += 4
	EndIf

	ConsoleWrite( @CRLF & "Model file: " & $sObjModel & ".obj" & @CRLF )

	; --- Read materials ---

	Local $sObjPath = StringLeft( $sObjModel, StringInStr( $sObjModel, "\", 0, -1 ) ), $aMtlInfo
	Local $sObjFile = StringRight( $sObjModel, StringLen( $sObjModel ) - StringInStr( $sObjModel, "\", 0, -1 ) )
	If FileExists( $sObjPath & $sMtlFile ) Then
		ReadMaterials( $sObjPath, $sMtlFile, $aMtlInfo )
	Else
		$aMtlInfo = 0
	EndIf

	; --- Update $aModel ---

	; $aModel[0-3] will be updated below

	$aModel[4]  = $aVertInfo
	$aModel[5]  = $aMaterials
	$aModel[6]  = $aMtlInfo
	$aModel[7]  = $iStatus
	$aModel[8]  = $iFaceType
	$aModel[9]  = $sMeshType
	$aModel[13] = $iObjects

	; --- Load binary files ---

	Local $hFile, $sBinData, $iBytes, $tStruct, $pFloats, $tFloats, $pShorts, $tShorts, $iVbo, $aBuf[2]

	; Vertices
	ConsoleWrite( "Loading vertex coordinates from file: " & $sObjFile & $sVFile & @CRLF )
	$hFile = FileOpen( $sObjModel & $sVFile, 16 ) ; 16 = binary mode
	$sBinData = FileRead( $hFile )
	$iBytes = @extended
	FileClose( $hFile )
	$tFloats = DllStructCreate( "float[" & $iBytes/4 & "]" )
	$pFloats = DllStructGetPtr( $tFloats )
	$tStruct = DllStructCreate( "byte[" & $iBytes & "]", $pFloats )
	DllStructSetData( $tStruct, 1, $sBinData )
	If $bOpenGL21 Then
		ConsoleWrite( "Uploading vertex coordinates to video RAM for OpenGL 2.1 (2.1 or higher)" & @CRLF )
		glGenBuffers( 1, $iVbo )
		glBindBuffer( $GL_ARRAY_BUFFER, $iVbo )
		glBufferData( $GL_ARRAY_BUFFER, $GL_FLOAT, $iBytes, $pFloats, $GL_STATIC_DRAW )
		$aModel[0] = $iVbo
	Else
		ConsoleWrite( "Storing vertex coordinates in conventional RAM for OpenGL 1.1 (2.0 or lower)" & @CRLF )
		$aBuf[0] = $tFloats
		$aBuf[1] = $pFloats
		glEnableClientState( $GL_VERTEX_ARRAY )
		glVertexPointer( 3, $GL_FLOAT, 0, $aBuf[1] )
		$aModel[0] = $aBuf
	EndIf

	; Texture coordinates
	If BitAND( $bLoad, 1 ) Then
		ConsoleWrite( "Loading texture coordinates from file: " & $sObjFile & "Vt" & $sExt & @CRLF )
		$hFile = FileOpen( $sObjModel & "Vt" & $sExt, 16 ) ; 16 = binary mode
		$sBinData = FileRead( $hFile )
		$iBytes = @extended
		FileClose( $hFile )
		$tFloats = DllStructCreate( "float[" & $iBytes/4 & "]" )
		$pFloats = DllStructGetPtr( $tFloats )
		$tStruct = DllStructCreate( "byte[" & $iBytes & "]", $pFloats )
		DllStructSetData( $tStruct, 1, $sBinData )
		If $bOpenGL21 Then
			ConsoleWrite( "Uploading texture coordinates to video RAM for OpenGL 2.1 (2.1 or higher)" & @CRLF )
			glGenBuffers( 1, $iVbo )
			glBindBuffer( $GL_ARRAY_BUFFER, $iVbo )
			glBufferData( $GL_ARRAY_BUFFER, $GL_FLOAT, $iBytes, $pFloats, $GL_STATIC_DRAW )
			$aModel[1] = $iVbo
		Else
			ConsoleWrite( "Storing texture coordinates in conventional RAM for OpenGL 1.1 (2.0 or lower)" & @CRLF )
			$aBuf[0] = $tFloats
			$aBuf[1] = $pFloats
			glEnableClientState( $GL_TEXTURE_COORD_ARRAY )
			glTexCoordPointer( 2, $GL_FLOAT, 0, $aBuf[1] )
			$aModel[1] = $aBuf
		EndIf
	Else
		$aModel[1] = 0
		If Not $bOpenGL21 Then ; OpenGL 2.0 or lower
			glDisableClientState( $GL_TEXTURE_COORD_ARRAY )
			glBindTexture( $GL_TEXTURE_2D, 0 )
		EndIf
	EndIf

	; Normal vectors
	If BitAND( $bLoad, 2 ) Then
		ConsoleWrite( "Loading normal vectors from file: " & $sObjFile & $sVnFile & @CRLF )
		$hFile = FileOpen( $sObjModel & $sVnFile, 16 ) ; 16 = binary mode
		$sBinData = FileRead( $hFile )
		$iBytes = @extended
		FileClose( $hFile )
		$tFloats = DllStructCreate( "float[" & $iBytes/4 & "]" )
		$pFloats = DllStructGetPtr( $tFloats )
		$tStruct = DllStructCreate( "byte[" & $iBytes & "]", $pFloats )
		DllStructSetData( $tStruct, 1, $sBinData )
		If $bOpenGL21 Then
			ConsoleWrite( "Uploading normal vectors to video RAM for OpenGL 2.1 (2.1 or higher)" & @CRLF )
			glGenBuffers( 1, $iVbo )
			glBindBuffer( $GL_ARRAY_BUFFER, $iVbo )
			glBufferData( $GL_ARRAY_BUFFER, $GL_FLOAT, $iBytes, $pFloats, $GL_STATIC_DRAW )
			$aModel[2] = $iVbo
		Else
			ConsoleWrite( "Storing normal vectors in conventional RAM for OpenGL 1.1 (2.0 or lower)" & @CRLF )
			$aBuf[0] = $tFloats
			$aBuf[1] = $pFloats
			glEnableClientState( $GL_NORMAL_ARRAY )
			glNormalPointer( $GL_FLOAT, 0, $aBuf[1] )
			$aModel[2] = $aBuf
		EndIf
	Else
		$aModel[2] = 0
		If Not $bOpenGL21 Then ; OpenGL 2.0 or lower
			glDisableClientState( $GL_NORMAL_ARRAY )
		EndIf
	EndIf

	; Face polygons
	If BitAND( $bLoad, 4 ) Then
		ConsoleWrite( "Loading face " & $sFaces & " from file: " & $sObjFile & "F" & $sExt & @CRLF )
		$hFile = FileOpen( $sObjModel & "F" & $sExt, 16 ) ; 16 = binary mode
		$sBinData = FileRead( $hFile )
		$iBytes = @extended
		FileClose( $hFile )
		$tShorts = DllStructCreate( "short[" & $iBytes/2 & "]" )
		$pShorts = DllStructGetPtr( $tShorts )
		$tStruct = DllStructCreate( "byte[" & $iBytes & "]", $pShorts )
		DllStructSetData( $tStruct, 1, $sBinData )
		If $bOpenGL21 Then
			ConsoleWrite( "Uploading face " & $sFaces & " to video RAM for OpenGL 2.1 (2.1 or higher)" & @CRLF )
			glGenBuffers( 1, $iVbo )
			glBindBuffer( $GL_ELEMENT_ARRAY_BUFFER, $iVbo )
			glBufferData( $GL_ELEMENT_ARRAY_BUFFER, $GL_UNSIGNED_SHORT, $iBytes, $pShorts, $GL_STATIC_DRAW )
			$aModel[3] = $iVbo
		Else
			ConsoleWrite( "Storing face " & $sFaces & " in conventional RAM for OpenGL 1.1 (2.0 or lower)" & @CRLF )
			$aBuf[0] = $tShorts
			$aBuf[1] = $pShorts
			$aModel[3] = $aBuf
		EndIf
	Else
		$aModel[3] = 0
	EndIf

	ConsoleWrite( @CRLF )

	Return 1

EndFunc
