; OpenGL shader loading function

#include-once

Func LoadGLSLshader( ByRef $shaderId, $shaderFile, $shaderType )
	; File exists?
	If Not FileExists( $shaderFile ) Then
		PrintError( "Function LoadGLSLshader()" & @CRLF & _
			"ERROR: File """ & $shaderFile & """ does not exists." )
		Return SetError(1,0,0)
	EndIf

	; Read file
	Local $glslFile = FileOpen( $shaderFile, 0 )
	Local $glslSource = FileRead( $glslFile )
	FileClose( $glslFile )

	Local $iOpenGLerror = glGetError(), $bShader

	; Create shader
	$shaderId = glCreateShader( $shaderType )
	glShaderSource( $shaderId, 1, $glslSource, StringLen( $glslSource ) )

	$iOpenGLerror = glGetError()
	If $iOpenGLerror <> $GL_NO_ERROR Then
		PrintOpenGLerror( "LoadGLSLshader()", "glCreateShader...", $iOpenGLerror )
		Return SetError(2,0,0)
	EndIf

	; Compile shader
	glCompileShader( $shaderId )

	glGetShaderiv( $shaderId, $GL_COMPILE_STATUS, $bShader )
	If $bShader = 0 Then
		PrintError( "Function LoadGLSLshader()" & @CRLF & _
			"ERROR: Failed to compile shader """ & $shaderFile & """." )
		glDeleteShader( $shaderId )
		Return SetError(3,0,0)
	EndIf

	Return 1
EndFunc

Func CreateProgram( ByRef $iProgram, $iVertexShader, $iFragmentShader )
	Local $bStatus

	; Create program
	$iProgram = glCreateProgram()

	; Attach shaders
	glAttachShader( $iProgram, $iVertexShader )
	glAttachShader( $iProgram, $iFragmentShader )

	; Link program
	glLinkProgram( $iProgram )
	glGetProgramiv( $iProgram, $GL_LINK_STATUS, $bStatus )
	If $bStatus = 0 Then
		PrintError( "Function CreateProgram()" & _
			@CRLF & "ERROR: Could not link program." )
		Return 0
	EndIf

	; Validate program
	glValidateProgram( $iProgram )
	glGetProgramiv( $iProgram, $GL_VALIDATE_STATUS, $bStatus )
	If $bStatus = 0 Then
		PrintError( "Function CreateProgram()" & _
			@CRLF & "ERROR: Could not validate program." )
		Return 0
	EndIf

	Return 1
EndFunc

Func GetAttribLocation( $iProgram, ByRef $iAttribute, $sAttribute )
	$iAttribute = glGetAttribLocation( $iProgram, $sAttribute )
	If $iAttribute = -1 Then
		PrintError( "Function GetAttribLocation()" & @CRLF & _
			"ERROR: Could not bind attribute """ & $sAttribute & """." )
		Return 0
	EndIf
	Return 1
EndFunc

Func GetUniformLocation( $iProgram, ByRef $iUniform, $sUniform )
	$iUniform = glGetUniformLocation( $iProgram, $sUniform )
	If $iUniform = -1 Then
		PrintError( "Function GetUniformLocation()" & @CRLF & _
			"ERROR: Could not bind uniform """ & $sUniform & """." )
		Return 0
	EndIf
	Return 1
EndFunc
