#include-once

#include <FontConstants.au3>
#include <WinAPI.au3>

#include "wglConstants.au3"
#include "wglFunctions.au3"

Func ResizeOpenGLwindow( $x, $y, $w, $h, $fovy, $zNear, $zFar )

	; Resize viewport
	glViewport( $x, $y, $w, $h )

	; Set up a perspective view
	glMatrixMode( $GL_PROJECTION )
	glLoadIdentity()
	gluPerspective( $fovy, $w/$h, $zNear, $zFar )
	glMatrixMode( $GL_MODELVIEW )
	glLoadIdentity()

EndFunc

Func ResizeOpenGLwindowOrtho( $x, $y, $w, $h, $left, $right, $bottom, $top, $zNear, $zFar )

	; Resize viewport
	glViewport( $x, $y, $w, $h )

	; Set up a orthographic view
	glMatrixMode( $GL_PROJECTION )
	glLoadIdentity()
	glOrtho( $left, $right, $bottom, $top, $zNear, $zFar )
	glMatrixMode( $GL_MODELVIEW )
	glLoadIdentity()

EndFunc

Func CreateOpenGLBitmapFont( _
	ByRef $iListBase, _                     ; display list start
	$iFirst = 0, _                          ; first character
	$iCount = 256, _                        ; number of characters
  $szFace = "", _                         ; typeface (font) name e.g. Arial
  $size = 8.5, _                          ; font size, default 8.5
  $fnWeight = $FW_NORMAL, _               ; font weight, defined values in FontConstants.au3
  $bItalic = False, _                     ; italic attribute option, True/False
  $bUnderline = False, _                  ; underline attribute option, True/False
  $bStrikeout = False, _                  ; strikeout attribute option, True/False
  $nCharset = $ANSI_CHARSET, _            ; character set identifier, defined values in FontConstants.au3
	$nOutputPrec = $OUT_TT_PRECIS, _        ; output precision, defined values in FontConstants.au3
  $nQuality = $ANTIALIASED_QUALITY, _     ; output quality, defined values in FontConstants.au3
  $nPitchAndFamily = 0 )                  ; pitch and family, defined values in FontConstants.au3, $DEFAULT_PITCH+$FF_DONTCARE

	; Device context for the OpenGL window
	Local $hDC = wglGetCurrentDC()

	; Calculate proper font height
	Local $LOGPIXELSY = _WinAPI_GetDeviceCaps( $hDC, 90 ) ; 90 -> LOGPIXELSY
	Local $nHeight = $size * $LOGPIXELSY / 72

	; Create the font
	Local $hFont = _WinAPI_CreateFont( _
		-$nHeight, _
		0, _ ; average character width
		0, _ ; angle of escapement
		0, _ ; base-line orientation angle
		$fnWeight, _
		$bItalic, _
		$bUnderline, _
		$bStrikeout, _
		$nCharset, _
		$nOutputPrec, _
		$CLIP_DEFAULT_PRECIS, _
		$nQuality, _
		$nPitchAndFamily, _
		$szFace )
	If @error Or Not $hFont Then Return 0

	; Set the font for the OpenGL device context
	Local $hFontOld = _WinAPI_SelectObject( $hDC, $hFont )
	If @error Or Not $hFontOld Then Return 0

	; Allocate display lists for the characters
	; One display list for each character
	$iListBase = glGenLists( $iCount )
	If @error Then Return 0

	; Create bitmap display lists to draw the characters
	wglUseFontBitmaps( $hDC, $iFirst, $iCount, $iListBase )
	If @error Then Return 0

	; Cleanup
	_WinAPI_SelectObject( $hDC, $hFontOld )
	_WinAPI_DeleteObject( $hFont )

	Return 1

EndFunc

; Used in CreateOpenGLOutlineFont()
Global Const $tagGLYPHMETRICSFLOAT = _  ; = "float[6]"
  "FLOAT  gmfBlackBoxX;" & _
  "FLOAT  gmfBlackBoxY;" & _ 
  "FLOAT  x;"            & _
  "FLOAT  y;"            & _
  "FLOAT  gmfCellIncX;"  & _
  "FLOAT  gmfCellIncY"

Func CreateOpenGLOutlineFont( _
	$pGMF, _                                ; address of buffer to receive glyph metric data, see $tagGLYPHMETRICSFLOAT above
	ByRef $iListBase, _                     ; display list start
	$iFirst = 0, _                          ; first character
	$iCount = 256, _                        ; number of characters
	$fDeviation = 0.0, _                    ; specifies the maximum chordal deviation from the true outlines
	$fExtrusion = 0.2, _                    ; extrusion value (font thickness) in the negative z direction
	$iFormat = $WGL_FONT_POLYGONS, _        ; line segments or polygons in display lists
  $szFace = "", _                         ; typeface (font) name e.g. Arial
  $size = 8.5, _                          ; font size, default 8.5
  $fnWeight = $FW_NORMAL, _               ; font weight, defined values in FontConstants.au3
  $bItalic = False, _                     ; italic attribute option, True/False
  $bUnderline = False, _                  ; underline attribute option, True/False
  $bStrikeout = False, _                  ; strikeout attribute option, True/False
  $nCharset = $ANSI_CHARSET, _            ; character set identifier, defined values in FontConstants.au3
	$nOutputPrec = $OUT_TT_PRECIS, _        ; output precision, defined values in FontConstants.au3
  $nQuality = $ANTIALIASED_QUALITY, _     ; output quality, defined values in FontConstants.au3
  $nPitchAndFamily = 0 )                  ; pitch and family, defined values in FontConstants.au3, $DEFAULT_PITCH+$FF_DONTCARE

	; Device context for the OpenGL window
	Local $hDC = wglGetCurrentDC()

	; Calculate proper font height
	Local $LOGPIXELSY = _WinAPI_GetDeviceCaps( $hDC, 90 ) ; 90 -> LOGPIXELSY
	Local $nHeight = $size * $LOGPIXELSY / 72

	; Create the font
	Local $hFont = _WinAPI_CreateFont( _
		-$nHeight, _
		0, _ ; average character width
		0, _ ; angle of escapement
		0, _ ; base-line orientation angle
		$fnWeight, _
		$bItalic, _
		$bUnderline, _
		$bStrikeout, _
		$nCharset, _
		$nOutputPrec, _
		$CLIP_DEFAULT_PRECIS, _
		$nQuality, _
		$nPitchAndFamily, _
		$szFace )
	If @error Or Not $hFont Then Return 0

	; Set the font for the OpenGL device context
	Local $hFontOld = _WinAPI_SelectObject( $hDC, $hFont )
	If @error Or Not $hFontOld Then Return 0

	; Allocate display lists for the characters
	; One display list for each character
	$iListBase = glGenLists( $iCount )
	If @error Then Return 0

	; Create bitmap display lists to draw the characters
	wglUseFontOutlines( $hDC, $iFirst, $iCount, $iListBase, $fDeviation, $fExtrusion, $iFormat, $pGMF )
	If @error Then Return 0

	; Cleanup
	_WinAPI_SelectObject( $hDC, $hFontOld )
	_WinAPI_DeleteObject( $hFont )

	Return 1

EndFunc

Func CheckOpenGLextension( $sExtension )
	Local Static $aExtensions, $iExtensions = 0

	If Not $iExtensions Then
		Local $sExtensions = glGetString( $GL_EXTENSIONS )
		Local $aExtensions0 = StringSplit( $sExtensions, " " )
		ReDim $aExtensions0[$aExtensions0[0]]
		$iExtensions = $aExtensions0[0] - 1
		$aExtensions = $aExtensions0
	EndIf

	; Check extension
	For $i = 1 To $iExtensions
		If $aExtensions[$i] = $sExtension Then Return 1
	Next
	Return 0
EndFunc

Func CheckOpenGLversion( $iMajor, $iMinor )
	Local $aVer = GetOpenGLversion()
	Select
		Case $aVer[0] > $iMajor
			Return 1
		Case $aVer[0] = $iMajor And $aVer[1] >= $iMinor
			Return 1
		Case Else
			PrintError( "Current OpenGL version (context) " & $aVer[0] & "." & $aVer[1] & @CRLF & _
				"does not match the expected version " & $iMajor & "." & $iMinor & "." )
			Return 0
	EndSelect
EndFunc

Func GetOpenGLversion()
	Local $sVersion, $iMajor, $iMinor, $aVer[2]
	$sVersion = glGetString( $GL_VERSION )
	$sVersion = StringLeft( $sVersion, 3 )
	$iMajor = Number( StringLeft( $sVersion, 1 ) )   ; Major version
	$iMinor = Number( StringRight( $sVersion, 1 ) )  ; Minor version
	$aVer[0] = $iMajor
	$aVer[1] = $iMinor
	Return $aVer
EndFunc

Func CheckVideoCard( $fPointSprite = False )
	Local $error = 0

  Local $vertexTextureUnits
  glGetIntegerv( $GL_MAX_VERTEX_TEXTURE_IMAGE_UNITS, $vertexTextureUnits )
	If Not $vertexTextureUnits Then
		PrintError( "ERROR: Your graphics card does not support texture lookups in the vertex shader." )
		$error = 1
	EndIf

  Local $maxUnits
  glGetIntegerv( $GL_MAX_VERTEX_TEXTURE_IMAGE_UNITS, $maxUnits )
	If $maxUnits < 1 Then
		PrintError( "ERROR: Your GPU does not have any vertex texture image units." )
		$error = 1
	EndIf

	If $fPointSprite Then
	  Local $range[2]
	  glGetFloatv( $GL_ALIASED_POINT_SIZE_RANGE, $range )
		If $range[1] < 5 Then
			ConsoleWrite( "WARNING: Point sprite range (" & $range[0] & "-" & $range[1] & ") too small." & @CRLF )
		EndIf
	EndIf

	If $error Then
		Return 0
	Else
		Return 1
	EndIf
EndFunc

Func CheckWGLextension( $sExtension, $hDeviceContext = 0 )
	Local Static $aExtensions, $iExtensions = 0

	If Not $iExtensions Then
		; Get WGL extensions list
		; Try first with wglGetExtensionsStringARB
		Local $sExtensions = wglGetExtensionsStringARB( $hDeviceContext )
		If @error Then
			; Try then with wglGetExtensionsStringEXT
			$sExtensions = wglGetExtensionsStringEXT()
			If @error Then Return 0 ; No extensions list
		EndIf
		; The extensions are separated by blanks
		$aExtensions = StringSplit( $sExtensions, " " )
		$iExtensions = $aExtensions[0]
	EndIf

	; Check extension
	For $i = 1 To $iExtensions
		If $aExtensions[$i] = $sExtension Then Return 1
	Next
	Return 0
EndFunc

Func PrintOpenGLerror( $sFunction, $sOpenGLfunc, $iOpenGLerror )
	PrintError( "Function " & $sFunction & @CRLF & _
		"OpenGl ERROR in " & $sOpenGLfunc & ":" & @CRLF & _
		gluErrorString( $iOpenGLerror ) )
EndFunc

Func PrintError( $sErrStr, $hWnd = 0 )
	If $hWnd Then
		MsgBox( 16, ScriptName(), $sErrStr, 0, $hWnd )
	Else
		MsgBox( 16, ScriptName(), $sErrStr )
	EndIf
	ConsoleWrite( $sErrStr & @CRLF )
EndFunc
