; Windows Extension functions to OpenGL (WGL)
; And some functions from gdi32.dll and user32.dll used with the WGL-functions

; http://msdn.microsoft.com/en-us/library/windows/desktop/dd374278(v=vs.85).aspx

; Errors:
;   @error is set to 1 or 2
;   @extended is set to the error code returned by GetLastError where possible


#include-once

; ------------------------------------------------------------------------

Func ChoosePixelFormat( $hDC, $ppfd )
	Local $aRet = DllCall( $dllGDI32, "int", "ChoosePixelFormat", "hwnd", $hDC, "ptr", $ppfd )
	If @error Or Not $aRet[0] Then
		$aRet = DllCall( $dllKernel32, "dword", "GetLastError" )
		Return SetError( 1, $aRet[0], 0 )
	Else
		Return $aRet[0]
	EndIf
EndFunc

Func GetDC( $hWnd )
	Local $aRet = DllCall( $dllUser32, "hwnd", "GetDC", "hwnd", $hWnd )
	If @error Or Not $aRet[0] Then Return SetError( 1, 0, 0 )
	Return $aRet[0]
EndFunc

Func ReleaseDC( $hWnd, $hDC )
	Local $aRet = DllCall( $dllUser32, "int", "ReleaseDC", "hwnd", $hWnd, "hwnd", $hDC )
	If @error Or Not $aRet[0] Then Return SetError( 1, 0, 0 )
EndFunc

Func SetPixelFormat( $hDC, $iPixelFormat, $ppfd )
	Local $aRet = DllCall( $dllGDI32, "int", "SetPixelFormat", "hwnd", $hDC, "int", $iPixelFormat, "ptr", $ppfd )
	If @error Or Not $aRet[0] Then
		$aRet = DllCall( $dllKernel32, "dword", "GetLastError" )
		Return SetError( 1, $aRet[0], 0 )
	EndIf
EndFunc

Func SwapBuffers( $hDC )
	DllCall( $dllGDI32, "int", "SwapBuffers", "hwnd", $hDC )
EndFunc

Func wglCreateContext( $hDC )
	Local $aRet = DllCall( $dllOpenGL32, "hwnd", "wglCreateContext", "hwnd", $hDC )
	If @error Or Not $aRet[0] Then
		$aRet = DllCall( $dllKernel32, "dword", "GetLastError" )
		Return SetError( 1, $aRet[0], 0 )
	Else
		Return $aRet[0]
	EndIf
EndFunc

Func wglDeleteContext( $hglrc )
	Local $aRet = DllCall( $dllOpenGL32, "int", "wglDeleteContext", "hwnd", $hglrc )
	If @error Or Not $aRet[0] Then
		$aRet = DllCall( $dllKernel32, "dword", "GetLastError" )
		Return SetError( 1, $aRet[0], 0 )
	EndIf
EndFunc

Func wglGetCurrentDC()
	Local $aRet = DllCall( $dllOpenGL32, "hwnd", "wglGetCurrentDC" )
	If @error Or Not $aRet[0] Then Return SetError( 1, 0, 0 )
	Return $aRet[0]
EndFunc

Func wglGetProcAddress( $lpszProc )
	Local $aRet = DllCall( $dllOpenGL32, "ptr", "wglGetProcAddress", "str", $lpszProc )
	If @error Or Not $aRet[0] Then
		$aRet = DllCall( $dllKernel32, "dword", "GetLastError" )
		Return SetError( $aRet[0], 0, 0 )
	Else
		Return $aRet[0]
	EndIf
EndFunc

Func wglMakeCurrent( $hDC, $hglrc )
	Local $aRet = DllCall( $dllOpenGL32, "int", "wglMakeCurrent", "hwnd", $hDC, "hwnd", $hglrc )
	If @error Or Not $aRet[0] Then
		$aRet = DllCall( $dllKernel32, "dword", "GetLastError" )
		Return SetError( 1, $aRet[0], 0 )
	EndIf
EndFunc

Func wglUseFontBitmaps( $hDC, $first, $count, $listBase )
	Local $aRet = DllCall( $dllOpenGL32, "int", "wglUseFontBitmapsW", "hwnd", $hDC, "dword", $first, "dword", $count, "dword", $listBase )
	If @error Or Not $aRet[0] Then
		$aRet = DllCall( $dllKernel32, "dword", "GetLastError" )
		Return SetError( 1, $aRet[0], 0 )
	EndIf
EndFunc

Func wglUseFontOutlines( $hDC, $first, $count, $listBase, $deviation, $extrusion, $format, $lpgmf )
	Local $aRet = DllCall( $dllOpenGL32, "int", "wglUseFontOutlinesW", _
												 "hwnd", $hDC, "dword", $first, "dword", $count, "dword", $listBase, _
												 "float", $deviation, "float", $extrusion, "int", $format, "ptr", $lpgmf )
	If @error Or Not $aRet[0] Then
		$aRet = DllCall( $dllKernel32, "dword", "GetLastError" )
		Return SetError( 1, $aRet[0], 0 )
	EndIf
EndFunc

; ------------------------ WGL_ARB_create_context ------------------------

Func wglCreateContextAttribsARB( $hDC, $hShareContext, $attribList )
	Local $ptr = wglGetProcAddress( "wglCreateContextAttribsARB" )
	If @error Then Return SetError( 1, @error, 0 )
	Local $aRet = DllCallAddress( "hwnd", $ptr, "hwnd", $hDC, "hwnd", $hShareContext, "ptr", $attribList )
	If @error Or Not $aRet[0] Then
		$aRet = DllCall( $dllKernel32, "dword", "GetLastError" )
		Return SetError( 2, $aRet[0], 0 )
	Else
		Return $aRet[0]
	EndIf
EndFunc

; ---------------------- WGL_ARB_extensions_string -----------------------

Func wglGetExtensionsStringARB( $hdc )
	Local $ptr = wglGetProcAddress( "wglGetExtensionsStringARB" )
	If @error Then Return SetError( 1, @error, 0 )
	Local $aRet = DllCallAddress( "str", $ptr, "hwnd", $hdc )
	If @error Or Not $aRet[0] Then
		$aRet = DllCall( $dllKernel32, "dword", "GetLastError" )
		Return SetError( 2, $aRet[0], 0 )
	Else
		Return $aRet[0]
	EndIf
EndFunc

; ------------------------- WGL_ARB_pixel_format -------------------------

Func wglChoosePixelFormatARB( $hDC, $piAttribIList, $pfAttribFList, $nMaxFormats, ByRef $piFormats, ByRef $nNumFormats )
	Local $ptr = wglGetProcAddress( "wglChoosePixelFormatARB" )
	If @error Then Return SetError( 1, @error, 0 )
	Local $aRet = DllCallAddress( "int", $ptr, "hwnd", $hDC, "ptr", $piAttribIList, "ptr", $pfAttribFList, "uint", $nMaxFormats, "int*", $piFormats, "uint*", $nNumFormats )
	If @error Or Not $aRet[0] Then
		$aRet = DllCall( $dllKernel32, "dword", "GetLastError" )
		Return SetError( 2, $aRet[0], 0 )
	Else
		$piFormats = $aRet[5]
		$nNumFormats = $aRet[6]
	EndIf
EndFunc

; ---------------------- WGL_EXT_extensions_string -----------------------

Func wglGetExtensionsStringEXT()
	Local $ptr = wglGetProcAddress( "wglGetExtensionsStringEXT" )
	If @error Then Return SetError( 1, @error, 0 )
	Local $aRet = DllCallAddress( "str", $ptr )
	If @error Or Not $aRet[0] Then
		$aRet = DllCall( $dllKernel32, "dword", "GetLastError" )
		Return SetError( 2, $aRet[0], 0 )
	Else
		Return $aRet[0]
	EndIf
EndFunc
